Package: spatsoc
Title: Group Animal Relocation Data by Spatial and Temporal
        Relationship and Measure Intragroup Social Dynamics
Version: 0.2.13
Authors@R: c(
    person("Alec L.", "Robitaille", , "robit.alec@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-4706-1762")),
    person("Quinn", "Webber", role = "aut",
           comment = c(ORCID = "0000-0002-0434-9360")),
    person("Eric", "Vander Wal", role = "aut",
           comment = c(ORCID = "0000-0002-8534-4317"))
  )
Description: Detects spatial and temporal groups in GPS relocations
    (Robitaille et al. (2019) <doi:10.1111/2041-210X.13215>).  It can be
    used to convert GPS relocations to gambit-of-the-group format to build
    proximity-based social networks, and perform data-stream randomization 
    methods suitable for GPS data. Also provides measures of intragroup social 
    dynamics including distance and direction to leaders, centroids and nearest 
    neighbours. 
License: GPL-3 | file LICENSE
URL: https://docs.ropensci.org/spatsoc/,
        https://github.com/ropensci/spatsoc
BugReports: https://github.com/ropensci/spatsoc/issues
Depends: R (>= 3.4)
Imports: adehabitatHR (>= 0.4.21), data.table (>= 1.15.0), igraph, sf,
        lwgeom (>= 0.2.15), CircStats, stats, units (>= 0.8-6), rlang
Suggests: asnipe, knitr, markdown, move2, rmarkdown, s2, sftrack,
        testthat (>= 2.1.0)
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: GDAL (>= 2.0.1), GEOS (>= 3.4.0), PROJ (>= 4.8.0),
        sqlite3
NeedsCompilation: no
Packaged: 2026-04-27 12:47:46 UTC; alecr
Author: Alec L. Robitaille [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4706-1762>),
  Quinn Webber [aut] (ORCID: <https://orcid.org/0000-0002-0434-9360>),
  Eric Vander Wal [aut] (ORCID: <https://orcid.org/0000-0002-8534-4317>)
Maintainer: Alec L. Robitaille <robit.alec@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-27 13:20:03 UTC
Built: R 4.4.3; ; 2026-05-01 14:16:29 UTC; unix

# Generated by roxygen2: do not edit by hand

export(build_lines)
export(build_polys)
export(centroid_dyad)
export(centroid_fusion)
export(centroid_group)
export(direction_group)
export(direction_polarization)
export(direction_step)
export(direction_to_centroid)
export(direction_to_leader)
export(distance_to_centroid)
export(distance_to_leader)
export(dyad_id)
export(edge_alignment)
export(edge_delay)
export(edge_direction)
export(edge_dist)
export(edge_nn)
export(edge_zones)
export(fusion_id)
export(get_gbi)
export(get_geometry)
export(group_lines)
export(group_polys)
export(group_pts)
export(group_times)
export(leader_direction_group)
export(leader_edge_delay)
export(randomizations)
import(data.table)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  eval = FALSE,
  echo = TRUE,
  comment = "#>"
)

## ----eval = TRUE--------------------------------------------------------------
## Load packages
library(spatsoc)
library(data.table)

## ----echo = FALSE, eval = TRUE------------------------------------------------
data.table::setDTthreads(1)

## ----eval = TRUE--------------------------------------------------------------
## Read data as a data.table
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

## Cast datetime column to POSIXct
DT[, datetime := as.POSIXct(datetime)]

## ----eval = TRUE--------------------------------------------------------------
# Temporal groups
group_times(DT, datetime = 'datetime', threshold = '5 minutes')

# Edge-list generation
edges <- edge_dist(
  DT,
  threshold = 100,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  returnDist = TRUE,
  fillNA = TRUE
)

## ----eval = FALSE-------------------------------------------------------------
# # Temporal groups
# group_times(DT, datetime = 'datetime', threshold = '5 minutes')
# 
# # Edge-list generation
# edges <- edge_nn(
#   DT,
#   id = 'ID',
#   coords = c('X', 'Y'),
#   timegroup = 'timegroup'
# )
# 
# # Edge-list generation using maximum distance threshold
# edges <- edge_nn(
#   DT,
#   id = 'ID',
#   coords = c('X', 'Y'),
#   timegroup = 'timegroup',
#   threshold = 100
# )
# 
# # Edge-list generation using maximum distance threshold, returning distances
# edges <- edge_nn(
#   DT,
#   id = 'ID',
#   coords = c('X', 'Y'),
#   timegroup = 'timegroup',
#   threshold = 100,
#   returnDist = TRUE
# )
# 

## ----eval = TRUE--------------------------------------------------------------
# In this case, using the edges generated in 2. a) edge_dist
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

## ----eval = TRUE--------------------------------------------------------------
fusion_id(
  edges = edges,
  threshold = 100,
  n_min_length = 1,
  n_max_missing = 0,
  allow_split = FALSE
)

# Print first 10 fusion events
print(edges[fusionID <= 5])

## ----knitropts, include = FALSE-----------------------------------------------
knitr::opts_chunk$set(message = TRUE, 
                      warning = FALSE,
                      eval = FALSE, 
                      echo = FALSE)

## ----eval = TRUE, echo = TRUE-------------------------------------------------
# Load packages
library(spatsoc)
library(data.table)

## ----echo = FALSE, eval = TRUE------------------------------------------------
data.table::setDTthreads(1)

## ----eval = TRUE, echo = TRUE-------------------------------------------------
# Read in spatsoc's example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Use subset of individuals
DT <- DT[ID %in% c('H', 'I', 'J')]

# Cast character column 'datetime' as POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
DT <- DT[ID %chin% c('H', 'I', 'J')]

## ----eval = TRUE, echo = FALSE------------------------------------------------
knitr::kable(DT[, .SD[1:2], ID][order(ID)])

## ----groupmins, echo = TRUE---------------------------------------------------
# group_times(DT, datetime = 'datetime', threshold = '5 minutes')

## ----tableSetUp, eval = TRUE--------------------------------------------------
nRows <- 9

## ----tabgroupmins, eval = TRUE------------------------------------------------
knitr::kable(
  group_times(DT, threshold = '5 minutes', datetime = 'datetime')[, 
    .(ID, X, Y, datetime, minutes, timegroup)][
      order(datetime)][
        1:nRows])

## ----grouphours, echo = TRUE--------------------------------------------------
# group_times(DT, datetime = 'datetime', threshold = '2 hours')

## ----tabgrouphours, eval = TRUE-----------------------------------------------
knitr::kable(
  group_times(DT, threshold = '2 hours', datetime = 'datetime')[, 
    .(ID, X, Y, datetime, hours, timegroup)][
      order(datetime)][
        1:nRows])

## ----groupdays, echo = TRUE---------------------------------------------------
# group_times(DT, datetime = 'datetime', threshold = '5 days')

## ----tabgroupdays, eval = TRUE------------------------------------------------
knitr::kable(
  group_times(
    DT, 
    threshold = '5 days', 
    datetime = 'datetime')[,
      .SD[sample(.N, 3)], by = .(timegroup, block)][order(datetime)][
        1:nRows, .(ID, X, Y, datetime, block, timegroup)]
)

## ----grouppts, echo = TRUE----------------------------------------------------
# group_times(DT = DT, datetime = 'datetime', threshold = '15 minutes')
# group_pts(DT, threshold = 50, id = 'ID', coords = c('X', 'Y'), timegroup = 'timegroup')

## ----fakegrouppts, eval = TRUE------------------------------------------------
DT <- group_times(DT = DT, datetime = 'datetime', 
                     threshold = '15 minutes')
DT <- group_pts(
    DT = DT,
    threshold = 50, id = 'ID', coords = c('X', 'Y'),
    timegroup = 'timegroup')

knitr::kable(
  DT[
      between(group,  771, 774)][order(timegroup)][
        1:nRows, .(ID, X, Y, timegroup, group)]
)

## ----fakegrouplines, echo = TRUE----------------------------------------------
# utm <- 32736
# 
# group_times(DT = DT, datetime = 'datetime', threshold = '1 day')
# group_lines(DT, threshold = 50, crs = utm,
#             id = 'ID', coords = c('X', 'Y'),
#             timegroup = 'timegroup', sortBy = 'datetime')

## ----grouplines, eval = TRUE--------------------------------------------------
utm <- 32736

DT <- group_times(DT = DT, datetime = 'datetime', 
                threshold = '1 day')
DT <- group_lines(DT,
                  threshold = 50, crs = utm, 
                  id = 'ID', coords = c('X', 'Y'), 
                  sortBy = 'datetime', timegroup = 'timegroup')
knitr::kable(
  unique(DT[, .(ID, timegroup, group)])[, .SD[1:3], ID][order(timegroup)]
)

## ----fakegrouppolys, echo = TRUE----------------------------------------------
# utm <- 32736
# group_times(DT = DT, datetime = 'datetime', threshold = '8 days')
# group_polys(DT = DT, area = TRUE, hrType = 'mcp',
#            hrParams = list('percent' = 95),
#            crs = utm,
#            coords = c('X', 'Y'), id = 'ID')

## ----grouppolys, eval = TRUE--------------------------------------------------
utm <- 32736
DT <- group_times(DT = DT, datetime = 'datetime', threshold = '8 days')
knitr::kable(
  data.frame(group_polys(
    DT, 
    area = TRUE, hrType = 'mcp',
           hrParams = list('percent' = 95),
           crs = utm,
           coords = c('X', 'Y'), id = 'ID')[
             , .(ID1, ID2, area, proportion)])
)

## ----edgedist, echo = TRUE----------------------------------------------------
# group_times(DT = DT, datetime = 'datetime', threshold = '15 minutes')
# edge_dist(DT, threshold = 50, id = 'ID',
#           coords = c('X', 'Y'), timegroup = 'timegroup', fillNA = TRUE)

## ----fakeedgedist, eval = TRUE------------------------------------------------
DT <- group_times(DT = DT, datetime = 'datetime', 
                     threshold = '15 minutes')
edges <- edge_dist(DT, threshold = 50, id = 'ID', 
                   coords = c('X', 'Y'), timegroup = 'timegroup', fillNA = TRUE)

knitr::kable(
  edges[between(timegroup, 158, 160)]
)

## ----edgenn, echo = TRUE------------------------------------------------------
# group_times(DT = DT, datetime = 'datetime', threshold = '15 minutes')
# edge_nn(DT, id = 'ID', coords = c('X', 'Y'), timegroup = 'timegroup')

## ----fakeedgenn, eval = TRUE--------------------------------------------------
DT <- group_times(DT = DT, datetime = 'datetime', 
                     threshold = '15 minutes')
edges <- edge_nn(DT, id = 'ID', coords = c('X', 'Y'), timegroup = 'timegroup')

knitr::kable(
  edges[1:6]
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----echo = FALSE, eval = TRUE------------------------------------------------
data.table::setDTthreads(1)

## -----------------------------------------------------------------------------
# Load packages
library(spatsoc)
library(data.table)

# Load data
predator <- fread(system.file("extdata", "DT_predator.csv", package = "spatsoc"))
prey <- fread(system.file("extdata", "DT_prey.csv", package = "spatsoc"))

# Combine data
DT <- rbindlist(list(predator, prey))

# Set the datetime as a POSIxct
DT[, datetime := as.POSIXct(datetime)]


# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '10 minutes')

# Spatial grouping
group_pts(DT, threshold = 50, id = 'ID', 
          coords = c('X', 'Y'), timegroup = 'timegroup')


# Calculate the number of types within each group
DT[, n_type := uniqueN(type), by = group]
DT[, interact := n_type > 1]

# Prey's perspective
sub_prey <- DT[type == 'prey']
sub_prey[, mean(interact)]


# Plot --------------------------------------------------------------------
# If we subset only where there are interactions
sub_interact <- DT[(interact)]

plot(sub_prey$X, sub_prey$Y, col = 'grey', pch = 21)
points(sub_interact$X, sub_interact$Y, col = factor(sub_interact$type))

## ----knitropts, include = FALSE-----------------------------------------------
knitr::opts_chunk$set(message = TRUE, 
                      warning = FALSE,
                      eval = FALSE, 
                      echo = TRUE)

## -----------------------------------------------------------------------------
# ## Load packages
# library(spatsoc)
# library(data.table)
# library(asnipe)
# library(igraph)

## ----echo = FALSE, eval = TRUE------------------------------------------------
data.table::setDTthreads(1)

## -----------------------------------------------------------------------------
# ## Read data as a data.table
# DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
# 
# ## Cast datetime column to POSIXct
# DT[, datetime := as.POSIXct(datetime)]
# 
# ## Calculate the year of the relocation
# DT[, yr := year(datetime)]

## -----------------------------------------------------------------------------
# ## Temporal groups
# group_times(DT, datetime = 'datetime', threshold = '5 minutes')
# 
# ## Spatial groups
# group_pts(
#   DT,
#   threshold = 50,
#   id = 'ID',
#   coords = c('X', 'Y'),
#   timegroup = 'timegroup'
# )
# 

## ----eval = FALSE-------------------------------------------------------------
# # EPSG code for relocations
# utm <- 32736
# 
# ## Group relocations by julian day
# group_times(DT, datetime = 'datetime', threshold = '1 day')
# 
# ## Group lines for each individual and julian day
# group_lines(
#   DT,
#   threshold = 50,
#   crs = utm,
#   id = 'ID',
#   coords = c('X', 'Y'),
#   timegroup = 'timegroup',
#   sortBy = 'datetime'
# )

## ----eval = FALSE-------------------------------------------------------------
# # EPSG code for relocations
# utm <- 32736
# 
# ## Option 1: area = FALSE and home range intersection 'group' column added to DT
# group_polys(
#   DT,
#   area = FALSE,
#   hrType = 'mcp',
#   hrParams = list(percent = 95),
#   crs = utm,
#   id = 'ID',
#   coords = c('X', 'Y')
# )
# 
# ## Option 2: area = TRUE
# #  results must be assigned to a new variable
# #  data.table returned has ID1, ID2 and proportion and area overlap
# areaDT <- group_polys(
#   DT,
#   area = TRUE,
#   hrType = 'mcp',
#   hrParams = list(percent = 95),
#   crs = utm,
#   id = 'ID',
#   coords = c('X', 'Y')
# )
# 

## -----------------------------------------------------------------------------
# ## Subset DT to only year 2016
# subDT <- DT[yr == 2016]
# 
# ## Generate group by individual matrix
# # group column generated by spatsoc::group_pts
# gbiMtrx <- get_gbi(DT = subDT, group = 'group', id = 'ID')

## -----------------------------------------------------------------------------
# ## Generate observed network
# net <- get_network(gbiMtrx,
#                    data_format = "GBI",
#                    association_index = "SRI")

## -----------------------------------------------------------------------------
# # Calculate year column to ensure randomization only occurs within years
# #   since data spans multiple years
# DT[, yr := year(datetime)]
# 
# ## Step type randomizations
# #  providing 'timegroup' (from group_times) as datetime
# #  splitBy = 'yr' to force randomization only within year
# randStep <- randomizations(
#    DT,
#    type = 'step',
#    id = 'ID',
#    group = 'group',
#    coords = NULL,
#    datetime = 'timegroup',
#    iterations = 3,
#    splitBy = 'yr'
# )

## -----------------------------------------------------------------------------
# # Calculate year column to ensure randomization only occurs within years
# #   since data spans multiple years
# DT[, yr := year(datetime)]
# 
# ## Daily type randomizations
# # splitBy = 'yr' to force randomization only within year
# randDaily <- randomizations(
#    DT,
#    type = 'daily',
#    id = 'ID',
#    group = 'group',
#    coords = NULL,
#    datetime = 'datetime',
#    splitBy = 'yr',
#    iterations = 20
# )

## -----------------------------------------------------------------------------
# # Calculate year column to ensure randomization only occurs within years
# #   since data spans multiple years
# DT[, yr := year(datetime)]
# 
# ## Trajectory type randomization
# randTraj <- randomizations(
#    DT,
#    type = 'trajectory',
#    id = 'ID',
#    group = NULL,
#    coords = c('X', 'Y'),
#    datetime = 'datetime',
#    splitBy = 'yr',
#    iterations = 20
# )

## -----------------------------------------------------------------------------
# ## Create a data.table of unique combinations of iteration and year, excluding observed rows
# iterYearLs <- unique(randStep[!(observed), .(iteration, yr)])
# 
# ## Generate group by individual matrix
# # for each combination of iteration number and year
# # 'group' generated by spatsoc::group_pts
# # 'randomID' used instead of observed ID (type = 'step')
# gbiLs <- mapply(FUN = function(i, y) {
#   get_gbi(randStep[iteration == i & yr == y],
#           'group', 'randomID')
#   },
#   i = iterYearLs$iter,
#   y = iterYearLs$yr,
#   SIMPLIFY = FALSE
# )
# 
# ## Generate a list of random networks
# netLs <- lapply(gbiLs, FUN = get_network,
#                 data_format = "GBI", association_index = "SRI")
# 

## -----------------------------------------------------------------------------
# ## Generate fake population
# randDaily[, population := sample(1:2, .N, replace = TRUE)]
# 
# ## Create a data.table of unique combinations of iteration, year, and population,
# #    excluding observed rows
# iterYearLs <- unique(randStep[!(observed), .(iteration, yr, population)])
# 
# ## Generate group by individual matrix
# # for each combination of iteration number and year
# # 'group' generated by spatsoc::group_pts
# # 'randomID' used instead of observed ID (type = 'step')
# gbiLs <- mapply(FUN = function(i, y, p) {
#   get_gbi(randDaily[iteration == i &
#                       yr == y & population == p],
#           'group', 'randomID')
#   },
#   i = iterYearLs$iter,
#   y = iterYearLs$yr,
#   p = iterYearLs$population,
#   SIMPLIFY = FALSE
# )
# 
# ## Generate a list of random networks
# netLs <- lapply(gbiLs, FUN = get_network,
#                 data_format = "GBI", association_index = "SRI")
# 

## -----------------------------------------------------------------------------
# ## Randomized temporal groups
# # 'datetime' is the randomdatetime produced by randomizations(type = 'trajectory')
# group_times(randTraj, datetime = 'randomdatetime', threshold = '5 minutes')
# 
# ## Randomized spatial groups
# # 'iteration' used in splitBy to ensure only points within each iteration are grouped
# group_pts(randTraj, threshold = 50, id = 'ID', coords = c('X', 'Y'),
#           timegroup = 'timegroup', splitBy = 'iteration')
# 
# ## Create a data.table of unique combinations of iteration and year, excluding observed rows
# iterYearLs <- unique(randStep[!(observed), .(iteration, yr)])
# 
# ## Generate group by individual matrix
# # for each combination of iteration number and year
# # 'group' generated by spatsoc::group_pts
# # 'ID' used since datetimes were randomized within individuals
# gbiLs <- mapply(FUN = function(i, y) {
#   get_gbi(randTraj[iteration == i & yr == y],
#           'group', 'ID')
#   },
#   i = iterYearLs$iter,
#   y = iterYearLs$yr,
#   SIMPLIFY = FALSE
# )
# 
# ## Generate a list of random networks
# netLs <- lapply(gbiLs, FUN = get_network,
#                 data_format = "GBI", association_index = "SRI")
# 

## -----------------------------------------------------------------------------
# ## Generate graph
# g <- graph.adjacency(net, 'undirected',
#                      diag = FALSE, weighted = TRUE)
# 
# ## Metrics for all individuals
# observed <- data.table(
#   centrality = evcent(g)$vector,
#   strength = graph.strength(g),
#   degree = degree(g),
#   ID = names(degree(g)),
#   yr = subDT[, unique(yr)]
# )

## -----------------------------------------------------------------------------
# ## Generate graph and calculate network metrics
# mets <- lapply(seq_along(netLs), function(n) {
#   g <- graph.adjacency(netLs[[n]], 'undirected',
#                        diag = FALSE, weighted = TRUE)
# 
#   data.table(
#     centrality = evcent(g)$vector,
#     strength = graph.strength(g),
#     degree = degree(g),
#     ID = names(degree(g)),
#     iteration = iterYearLs$iter[[n]],
#     yr = iterYearLs$yr[[n]]
#     )
# })
# 
# ## Metrics for all individuals across all iterations and years
# random <- rbindlist(mets)
# 
# ## Mean values for each individual and year
# meanMets <- random[, lapply(.SD, mean), by = .(ID, yr),
#                 .SDcols = c('centrality', 'strength', 'degree')]

## -----------------------------------------------------------------------------
# ## Create a data.table of unique combinations of iteration and year, including observed and random rows
# iterYearLs <- unique(randStep[, .(iteration, yr)])
# 
# ## Generate group by individual matrix
# # for each combination of iteration and year
# # 'group' generated by spatsoc::group_pts
# # 'randomID' used instead of observed ID (type = 'step')
# gbiLs <- mapply(FUN = function(i, y) {
#   get_gbi(randStep[iteration == i & yr == y],
#           'group', 'randomID')
#   },
#   i = iterYearLs$iter,
#   y = iterYearLs$yr,
#   SIMPLIFY = FALSE
# )
# 
# ## Generate a list of random networks
# netLs <- lapply(gbiLs, FUN = get_network,
#                 data_format = "GBI", association_index = "SRI")
# 
# ## Generate graph and calculate network metrics
# mets <- lapply(seq_along(netLs), function(n) {
#   g <- graph.adjacency(netLs[[n]], 'undirected',
#                        diag = FALSE, weighted = TRUE)
# 
#   data.table(
#     centrality = evcent(g)$vector,
#     strength = graph.strength(g),
#     ID = names(degree(g)),
#     iteration = iterYearLs$iter[[n]],
#     yr = iterYearLs$yr[[n]]
#     )
# })
# 
# ## Observed and random for all individuals across all iterations and years
# out <- rbindlist(mets)
# 
# ## Split observed and random
# out[, observed := ifelse(iteration == 0, TRUE, FALSE)]
# 
# ## Mean values for each individual and year, by observed/random
# meanMets <- out[, lapply(.SD, mean), by = .(ID, yr, observed),
#                 .SDcols = c('centrality', 'strength')]
# 

#' Directional alignment based edge-lists
#'
#' `edge_alignment` returns edge-lists defined by directional alignment
#' (difference in movement direction) between individuals. The function expects
#' a `data.table` with relocation data and individual identifiers, a
#' direction column (generated by `direction_step`) and timegroup column
#' (generated by `group_times`).
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()].
#'
#' The `id`, `direction`, `timegroup`, and optional `group`
#' and `splitBy` arguments expect the names of a column in `DT` which
#' correspond to the individual identifier, direction (generated by
#' `direction_step`), timegroup (generated by `group_times`), group
#' (generated by `group_pts`) and additional grouping columns.
#'
#' There are two approaches to spatially restricting the calculation of
#' directional alignment. The `group` argument can be used to pass the
#' output group column from `group_pts` to calculate direction alignment
#' within spatiotemporal groups. Alternatively, the output of
#' `edge_alignment` can be merged with the output of `edge_dist` to
#' compare the difference in direction to the distance between individuals.
#'
#' The `splitBy` argument offers further control over the calculation of
#' directional alignment.  If within your `DT`, you have multiple
#' populations, subgroups or other distinct parts, you can provide the name of
#' the column which identifies them to `splitBy`. `edge_alignment`
#' will only consider rows within each `splitBy` subgroup.
#'
#' @inheritParams edge_dist
#' @inheritParams direction_polarization
#' @inheritParams diff_rad
#' @param timegroup character string of timegroup column name, default "timegroup"
#' @param group (optional) character string of group column name, used to restrict
#' the calculation of directional alignment to within spatiotemporal groups
#' @param splitBy (optional) vector of column names indicating subgroups within
#'   which the direction alignment will be calculated
#'
#' @return `edge_alignment` returns a `data.table` with columns ID1, ID2,
#'   timegroup, and a 'direction_diff' column indicating the difference in direction
#'   between ID1 and ID2 in radians, along with any columns provided in splitBy.
#'
#'  Note: unlike many other functions (eg. `group_pts`) in `spatsoc`,
#'  `edge_alignment` needs to be reassigned. See details in
#'  [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @references
#'
#' See examples of using directional alignment:
#'  * \doi{doi:10.1098/rsif.2013.0529}
#'  * \doi{doi:10.1016/j.beproc.2021.104473}
#'  * \doi{doi:10.1126/science.aaa5099}
#'
#' @export
#'
#' @family Edge-list generation
#' @family Direction functions
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#'
#' # Read example data
#' DT <- fread(system.file('extdata', 'DT.csv', package = 'spatsoc'))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Calculate direction
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Calculate directional alignment edge-list
#' align <- edge_alignment(
#'   DT,
#'   id = 'ID',
#'   signed = FALSE
#' )
#'
#' # Or, calculate directional alignment within spatiotemporal groups
#' group_pts(DT, threshold = 5, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' align_group <- edge_alignment(
#'   DT,
#'   id = 'ID',
#'   group = 'group',
#'   signed = FALSE
#' )
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' direction_step(DT, id = 'ID')
#' align_group <- edge_alignment(
#'   DT,
#'   id = 'ID',
#'   group = 'group',
#'   signed = FALSE
#' )
edge_alignment <- function(
    DT = NULL,
    id = NULL,
    direction = 'direction',
    timegroup = 'timegroup',
    group = NULL,
    splitBy = NULL,
    signed = FALSE) {
  # due to NSE notes in R CMD check
  ID1 <- ID2 <- N <- direction_diff <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)

  assert_not_null(id)
  assert_not_null(direction)
  assert_not_null(timegroup)

  check_colnames <- c(id, direction, timegroup, group, splitBy)
  assert_are_colnames(DT, check_colnames)

  assert_col_radians(DT, direction, ', did you use direction_step?')

  assert_inherits(signed, 'logical')

  if (any(unlist(lapply(DT[, .SD, .SDcols = timegroup], class)) %in%
    c('POSIXct', 'POSIXlt', 'Date', 'IDate', 'ITime', 'character'))) {
    warning(
      strwrap(
        prefix = ' ',
        initial = '',
        x = 'timegroup provided is a date/time
        or character type, did you use group_times?'
      )
    )
  }

  splitBy <- c(splitBy, timegroup)
  if (DT[, .N, by = c(id, splitBy, timegroup)][N > 1, sum(N)] != 0) {
    warning(
      strwrap(
        prefix = ' ',
        initial = '',
        x = 'found duplicate id in a
        timegroup and/or splitBy -
        does your group_times threshold match the fix rate?'
      )
    )
  }

  if ('splitBy' %in% colnames(DT)) {
    warning(
      strwrap(x = 'a column named "splitBy" was found in your data.table,
              renamed to "split_by" to avoid confusion with the argument
              "splitBy"')
    )
    data.table::setnames(DT, 'splitBy', 'split_by')
  }

  edges <- DT[, {
      m <- outer(direction, direction, FUN = diff_rad, signed = signed)

      data.table::data.table(
          ID1 = id[rep(seq_len(nrow(m)), ncol(m))],
          ID2 = id[rep(seq_len(ncol(m)), each = nrow(m))],
          direction_diff = c(m)
      )[ID1 != ID2]
    },
    by = splitBy,
    env = list(id = id, direction = direction)
  ]

  edges[, direction_diff := units::as_units(direction_diff, 'rad')][]

  return(edges)
}

# Test edge_alignment
context('test edge_alignment')

library(spatsoc)

id <- 'ID'
coords <- c('X', 'Y')
direction <- 'direction'
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736

DT <- fread('../testdata/DT.csv')
group_times(DT, 'datetime', '10 minutes')
direction_step(
  DT = DT,
  id = id,
  coords = coords,
  crs = utm
)

test_that('required arguments are provided else error', {
  expect_error(edge_alignment(DT = NULL),
               'DT')

  expect_error(edge_alignment(DT, id = NULL),
               'id')

  expect_error(edge_alignment(DT, id, direction = NULL),
               'direction')

  expect_error(edge_alignment(DT, id, direction, timegroup = NULL),
               'timegroup')
})

test_that('column names must exist in DT', {
  expect_error(edge_alignment(DT, id = 'potato', direction, timegroup),
               'field')

  expect_error(edge_alignment(DT, id, direction = 'potato', timegroup),
               'field')

  expect_error(edge_alignment(DT, id, direction, timegroup = 'potato'),
               'field')

  expect_error(edge_alignment(DT, id, direction, timegroup, group = 'potato'),
               'field')

  expect_error(edge_alignment(DT, id, direction, timegroup, splitBy = 'potato'),
               'field')
})

test_that('direction is units, timegroup is integer, signed is logical', {
  expect_error(edge_alignment(DT, id, direction = 'X', timegroup),
               'units')

  expect_warning(edge_alignment(DT, id, direction, timegroup = 'ID'),
                 'timegroup')

  expect_error(edge_alignment(DT, id, direction, timegroup, signed = 42),
               'signed')
})

test_that('duplicate IDs in a timegroup detected', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '8 hours')
  expect_warning(edge_alignment(copyDT, id, direction, timegroup),
                 'duplicate')
})

test_that('warns about splitBy column', {
  copyDT <- copy(DT)
  group_times(copyDT, 'datetime', '5 minutes')
  copyDT[, splitBy := as.IDate(datetime)]

  expect_warning(edge_alignment(copyDT, id, direction, timegroup),
                 'split_by')
})

test_that('returned IDs make sense', {
  eDT <- edge_alignment(DT, id, direction, timegroup)

  IDs <- DT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(eDT$ID2 %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)
})

test_that('returns a data.table', {
  expect_s3_class(edge_alignment(DT, id, direction, timegroup),
                  'data.table')
})

test_that('signed arg works', {
  expect_lt(
    mean(edge_alignment(DT, id, signed = TRUE)$direction_diff, na.rm = TRUE),
    mean(edge_alignment(DT, id, signed = FALSE)$direction_diff, na.rm = TRUE)
  )
})

#' Distance based edge-lists
#'
#' `edge_dist` returns edge-lists defined by a spatial distance within the user
#' defined threshold. The function expects a `data.table` with relocation data,
#' individual identifiers and a threshold argument. The threshold argument is
#' used to specify the criteria for distance between points which defines a
#' group. Relocation data should be in two columns representing the X and Y
#' coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()].
#'
#' The `id`, `timegroup` (and optional `splitBy`) arguments expect the names of
#' columns in `DT` which correspond to the individual identifier, and timegroup
#' (generated by `group_times`) and additional grouping columns.
#'
#' The `threshold` provided should match the units of the coordinates. The
#' `threshold` can be provided with units specified using the units package (eg.
#' `threshold = units::set_units(10, m)`) which will be checked against the
#' units of the coordinates using the `crs`. If units are not specified, the
#' `threshold` is assumed to be in the units of the coordinates.
#'
#' The `timegroup` argument is required to define the temporal groups within
#' which edges are calculated. The intended framework is to group rows
#' temporally with `group_times` then spatially with `edge_dist`. If you have
#' already calculated temporal groups without `group_times`, you can pass this
#' column to the `timegroup` argument. Note that the expectation is that each
#' individual will be observed only once per timegroup. Caution that
#' accidentally including huge numbers of rows within timegroups can overload
#' your machine since all pairwise distances are calculated within each
#' timegroup.
#'
#' The `splitBy` argument offers further control over grouping. If within your
#' `DT`, you have multiple populations, subgroups or other distinct parts, you
#' can provide the name of the column which identifies them to `splitBy`.
#' `edge_dist` will only consider rows within each `splitBy` subgroup.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Distance function" for details on underlying distance function used.
#'
#' @inheritSection distance_to_centroid Interface
#' @inheritSection distance_to_centroid Distance function
#'
#' @inheritParams group_pts
#' @inheritParams direction_step
#' @param returnDist logical indicating if the distance between individuals
#'   should be returned. If FALSE (default), only individual columns (and
#'   timegroup, splitBy columns if provided) are returned. If TRUE, a column
#'   "distance" is also returned indicating the distance between individuals in
#'   the units of the `crs`, or if `crs = NULL` no units are set.
#' @param fillNA logical indicating if NAs should be returned for individuals
#'   that were not within the threshold distance of any other. If TRUE, NAs are
#'   returned. If FALSE, only edges between individuals within the threshold
#'   distance are returned.
#'
#' @return `edge_dist` returns a `data.table` with columns ID1, ID2, timegroup
#'   (if supplied) and any columns provided in splitBy. If 'returnDist' is TRUE,
#'   column 'distance' is returned indicating the distance between ID1 and ID2.
#'
#'   The ID1 and ID2 columns represent the edges defined by the spatial (and
#'   temporal with `group_times`) thresholds.
#'
#'   Note: unlike many other functions (eg. `group_pts`) in `spatsoc`,
#'   `edge_dist` needs to be reassigned. See details in
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#'
#' @family Edge-list generation
#' @family Distance functions
#' @seealso [sf::st_distance()]
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge-list generation
#' edges <- edge_dist(
#'     DT,
#'     threshold = 100,
#'     id = 'ID',
#'     coords = c('X', 'Y'),
#'     timegroup = 'timegroup',
#'     crs = 32736,
#'     returnDist = TRUE,
#'     fillNA = TRUE
#'   )
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
edge_dist <- function(
    DT = NULL,
    threshold,
    id = NULL,
    coords = NULL,
    timegroup,
    crs = NULL,
    splitBy = NULL,
    geometry = 'geometry',
    returnDist = FALSE,
    fillNA = TRUE) {
  # due to NSE notes in R CMD check
  ID1 <- ID2 <- N <- geo <- x <- y <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_missing(threshold)
  assert_not_null(id)
  assert_not_missing(timegroup)
  assert_not_null(timegroup)

  check_cols <- c(timegroup, id, coords, splitBy)
  assert_are_colnames(DT, check_cols)

  if (any(unlist(lapply(DT[, .SD, .SDcols = timegroup], class)) %in%
          c('POSIXct', 'POSIXlt', 'Date', 'IDate', 'ITime', 'character'))) {
    warning(
      strwrap(
        prefix = " ",
        initial = "",
        x = 'timegroup provided is a date/time
        or character type, did you use group_times?'
      )
    )
  }

  splitBy <- c(splitBy, timegroup)
  if (DT[, .N, by = c(id, splitBy, timegroup)][N > 1, sum(N)] != 0) {
    warning(
      strwrap(
        prefix = " ",
        initial = "",
        x = 'found duplicate id in a
        timegroup and/or splitBy -
        does your group_times threshold match the fix rate?'
      )
    )
  }

  if ('splitBy' %in% colnames(DT)) {
    warning(
      strwrap(x = 'a column named "splitBy" was found in your data.table,
              renamed to "split_by" to avoid confusion with the argument
              "splitBy"')
    )
    data.table::setnames(DT, 'splitBy', 'split_by')
  }


  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    crs <- sf::st_crs(DT[[geometry]])

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    if (is.null(threshold)) {
      edges <- DT[, {
        distMatrix <- calc_distance(
          geometry_a = geo,
          use_dist = use_dist
        )
        diag(distMatrix) <- NA

        if (returnDist) {
          l <- data.table::data.table(
            ID1 = id[rep(seq_len(nrow(distMatrix)), ncol(distMatrix))],
            ID2 = id[rep(seq_len(ncol(distMatrix)), each = nrow(distMatrix))],
            distance = c(distMatrix)
          )[ID1 != ID2]
        } else {
          l <- data.table::data.table(
            ID1 = id[rep(seq_len(nrow(distMatrix)), ncol(distMatrix))],
            ID2 = id[rep(seq_len(ncol(distMatrix)), each = nrow(distMatrix))]
          )[ID1 != ID2]
        }
        l
      },
      by = splitBy,
      env = list(geo = geometry, id = id)]
    } else {
      assert_threshold(threshold, crs)

      if (!inherits(threshold, 'units') && !identical(crs, sf::NA_crs_) &&
          !use_dist) {
        threshold <- units::as_units(threshold, units(sf::st_crs(crs)$SemiMajor))
      }

      edges <- DT[, {
        distMatrix <- calc_distance(
          geometry_a = geo,
          use_dist = use_dist
        )
        diag(distMatrix) <- NA

        w <- which(distMatrix < threshold, arr.ind = TRUE)

        if (returnDist) {
          l <- list(ID1 = id[w[, 1]],
                    ID2 = id[w[, 2]],
                    distance = distMatrix[w])
        } else {
          l <- list(ID1 = id[w[, 1]],
                    ID2 = id[w[, 2]])
        }
        l
      },
      by = splitBy,
      env = list(geo = geometry, id = id)]
    }
  } else {
    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    if (is.null(threshold)) {
      edges <- DT[, {
        distMatrix <- calc_distance(
          x_a = x,
          y_a = y,
          crs = crs,
          use_dist = use_dist
        )
        diag(distMatrix) <- NA

        if (returnDist) {
          l <- data.table::data.table(
            ID1 = id[rep(seq_len(nrow(distMatrix)), ncol(distMatrix))],
            ID2 = id[rep(seq_len(ncol(distMatrix)), each = nrow(distMatrix))],
            distance = c(distMatrix)
          )[ID1 != ID2]
        } else {
          l <- data.table::data.table(
            ID1 = id[rep(seq_len(nrow(distMatrix)), ncol(distMatrix))],
            ID2 = id[rep(seq_len(ncol(distMatrix)), each = nrow(distMatrix))]
          )[ID1 != ID2]
        }
        l
      },
      by = splitBy,
      env = list(x = xcol, y = ycol, id = id)]
    } else {

      assert_threshold(threshold, crs)

      if (!inherits(threshold, 'units') && !identical(crs, sf::NA_crs_) &&
          !use_dist) {
        threshold <- units::as_units(threshold, 'm')
      }

      edges <- DT[, {
        distMatrix <- calc_distance(
          x_a = x,
          y_a = y,
          crs = crs,
          use_dist = use_dist
        )
        diag(distMatrix) <- NA

        w <- which(distMatrix < threshold, arr.ind = TRUE)

        if (returnDist) {
          l <- list(ID1 = id[w[, 1]],
                    ID2 = id[w[, 2]],
                    distance = distMatrix[w])
        } else {
          l <- list(ID1 = id[w[, 1]],
                    ID2 = id[w[, 2]])
        }
        l
      },
      by = splitBy,
      env = list(id = id, x = xcol, y = ycol)]
    }
  }

  if (fillNA) {
    merge(edges,
          unique(DT[, .SD, .SDcols = c(splitBy, id)]),
          by.x = c(splitBy, 'ID1'),
          by.y = c(splitBy, id),
          all = TRUE)
  } else {
    return(edges)
  }
}

# Test centroid_dyad
context('test centroid_dyad')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                   timegroup = timegroup, returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

clean_DT <- copy(DT)
clean_edges <- copy(edges)

# Note: since centroid_dyad and centroid_fusion share an internal function
#       that orchestrates the centroid calculation for both, testing is
#       overlapping. See test-centroid-fusion for full testing suite.

expected_DT <- copy(clean_DT)[timegroup < 3]
expected_DT[, X := timegroup * 10 + .I]
expected_DT[, Y := timegroup * 10]

expected_edges <- copy(clean_edges)[timegroup < 3]
expected_DT[ID %in% first(expected_edges$ID1), X := NA]

test_that('results are expected', {
  expect_equal(
    centroid_dyad(expected_edges, expected_DT, id = id, coords = coords)[
      timegroup == 1, unique(centroid_Y)],
    10
  )

  expect_equal(
    centroid_dyad(expected_edges, expected_DT, id = id, coords = coords)[
      timegroup == 2, .N],
    0
  )

  expect_gt(
    centroid_dyad(expected_edges, expected_DT, id = id, coords = coords)[
      timegroup == 1, unique(centroid_X)],
    10
  )

  expect_equal(
    expected_edges[, .N],
    centroid_dyad(expected_edges, expected_DT, id = id, coords = coords)[, .N]
  )

  expect_equal(
    expected_edges[, unique(ID1)],
    centroid_dyad(expected_edges, expected_DT,
                  id = id, coords = coords)[, unique(ID1)]
  )
})

test_that('NAs in dyadID result in NAs for centroid', {
  # Set fillNA = TRUE
  edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                     timegroup = timegroup, returnDist = TRUE,
                     fillNA = TRUE)
  dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

  centroids <- centroid_dyad(edges, DT, id, coords)
  expect_true(
    centroids[is.na(dyadID), all(is.na(centroid_X))]
  )
  expect_true(
    centroids[is.na(dyadID), all(is.na(centroid_Y))]
  )
})

# Test fusion_id
context("test fusion_id")

library(spatsoc)

DT <- fread('../testdata/DT.csv')

group_times(DT, datetime = 'datetime', threshold = '20 minutes')

edges <- edge_dist(
    DT,
    threshold = 100,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup',
    returnDist = TRUE,
    fillNA = TRUE
  )

dyad_id(edges, 'ID1', 'ID2')

test_that('edges is required', {
  expect_error(fusion_id(),
  'edges must be')
})


test_that('columns are required otherwise error detected', {
  expect_error(fusion_id(
    edges = edges[, .SD, .SDcols = -'dyadID']
  ))
  expect_error(fusion_id(
    edges = edges[, .SD, .SDcols = -'timegroup']
  ))
  expect_error(fusion_id(
    edges = edges[, .SD, .SDcols = -'distance']
  ))
})

test_that('arguments are correctly provided or error detected', {
  expect_error(fusion_id(
    edges = edges,
    threshold = NULL
  ))
  expect_error(fusion_id(
    edges = edges,
    threshold = -42
  ))
  expect_error(fusion_id(
    edges = edges,
    threshold = 'potato'
  ))

  expect_error(fusion_id(
    edges = edges,
    n_min_length = 'potato'
  ))

  expect_error(fusion_id(
    edges = edges,
    n_max_missing = 'potato'
  ))
  expect_error(fusion_id(
    edges = edges,
    allow_split = 'potato'
  ))
  expect_error(fusion_id(
    edges = edges,
    allow_split = 42
  ))
})

test_that('returns a data.table', {
  expect_s3_class(fusion_id(
    edges = edges,
    threshold = 50
  ), 'data.table')
})

test_that('returns a numeric fusionID column', {
  edges[, fusionID := NULL]
  expect_contains(colnames(fusion_id(
    edges = edges,
    threshold = 50
  )), 'fusionID')
  expect_type(edges$fusionID, 'integer')
})

test_that('message if fusionID column already present, overwritten', {
  fusionID_present <- copy(edges)[, fusionID := 42]
  expect_message(fusion_id(
    edges = fusionID_present,
    threshold = 50
  ), 'fusionID column will be overwritten by this function')
})



test_that('allow_split TRUE returns less unique fusionID', {
  # When splits are allowed, more fusion events are combined
  #  resulting in less unique fusion events
  expect_lt(
    fusion_id(edges, allow_split = TRUE, threshold = 50)[, uniqueN(fusionID)],
    fusion_id(edges, allow_split = FALSE, threshold = 50)[, uniqueN(fusionID)]
  )
})

test_that('larger n_max_missing returns less unique fusionID', {
  # When a larger n max missing is provided, more fusion events are combined
  #  resulting in less unique fusion events
  expect_lt(
    fusion_id(edges, n_max_missing = 3, threshold = 50)[, uniqueN(fusionID)],
    fusion_id(edges, n_max_missing = 0, threshold = 50)[, uniqueN(fusionID)]
  )
})

test_that('larger n_min_length returns less unique fusionID', {
  # When a larger n min length is provided, less fusion events are identified
  #  resulting in less unique fusion events
  expect_lt(
    fusion_id(edges, n_min_length = 3, threshold = 50)[, uniqueN(fusionID)],
    fusion_id(edges, n_min_length = 0, threshold = 50)[, uniqueN(fusionID)]
  )
})

edges_expected <- data.table(
  dyadID = rep('A-B', 11),
  timegroup = seq.int(12)[-5],
  distance = c(1, 50, 1, 1, 1, 50, 50, 1, 1, 50, 1)
)
threshold <- 25

test_that('n_min_length returns expected number of unique fusionIDs', {
  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 0,
      n_max_missing = 0,
      allow_split = FALSE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    5
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 0,
      allow_split = FALSE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    2
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 3,
      n_max_missing = 0,
      allow_split = FALSE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    0
  )
})


test_that('allow_split returns expected number of unique fusionIDs', {

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 0,
      allow_split = FALSE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    2
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 0,
      allow_split = FALSE
    )[!is.na(fusionID), .N, fusionID][, max(N)],
    2
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 0,
      allow_split = TRUE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    2
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 0,
      allow_split = TRUE
    )[!is.na(fusionID), .N, fusionID][, max(N)],
    4
  )
})


test_that('n_max_missing returns expected number of unique fusionIDs', {
  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 0,
      n_max_missing = 1,
      allow_split = FALSE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    4
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 1,
      allow_split = FALSE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    2
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 1,
      allow_split = FALSE
    )[!is.na(fusionID), .N, fusionID][, max(N)],
    3
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 1,
      allow_split = TRUE
    )[, uniqueN(fusionID,  na.rm = TRUE)],
    2
  )

  expect_equal(
    fusion_id(
      edges_expected,
      threshold = threshold,
      n_min_length = 2,
      n_max_missing = 1,
      allow_split = TRUE
    )[!is.na(fusionID), .N, fusionID][, max(N)],
    5
  )
})


test_that('lead edge cases identified as expected', {
  split_expected <- data.table(
    dyadID = rep('A-B', 7),
    timegroup = c(1,  2,   3, 10, 11, 12,  13),
    distance =  c(50, 1,  50, 1,  1,  50,   1)
  )
  threshold <- 25

  # If allow_split = TRUE,
  #  obs separate (tg 3) where not within 1 tg of lead+lag together should be NA
  #  obs separate (tg 12) where within lead+lag obs together should be same fusionID
  split_allow_true <- fusion_id(
    copy(split_expected),
    threshold = threshold,
    n_min_length = 0,
    n_max_missing = 0,
    allow_split = TRUE
  )
  expect_identical(split_allow_true[, fusionID[timegroup == 3]], NA_integer_)
  expect_false(identical(split_allow_true[, fusionID[timegroup == 3]],
                         split_allow_true[, fusionID[timegroup == 2]]))
  expect_false(identical(split_allow_true[, fusionID[timegroup == 3]],
                         split_allow_true[, fusionID[timegroup == 11]]))
  expect_identical(split_allow_true[, fusionID[timegroup == 11]],
                   split_allow_true[, fusionID[timegroup == 12]])

  # If allow_split = TRUE and n_max_missing = 7,
  #  obs separate (tg 3) where within lead+lag obs together should be
  #   same fusionID given n_max_missing = 7
  #  obs separate (tg 12) where within lead+lag obs together should be same fusionID
  split_allow_true_7_miss <- fusion_id(
    copy(split_expected),
    threshold = threshold,
    n_min_length = 0,
    n_max_missing = 7,
    allow_split = TRUE
  )
  expect_identical(split_allow_true_7_miss[, fusionID[timegroup == 3]],
                   split_allow_true_7_miss[, fusionID[timegroup == 2]])
  expect_identical(split_allow_true_7_miss[, fusionID[timegroup == 3]],
                   split_allow_true_7_miss[, fusionID[timegroup == 11]])
  expect_identical(split_allow_true_7_miss[, fusionID[timegroup == 11]],
                   split_allow_true_7_miss[, fusionID[timegroup == 12]])

  # Also check with n_min_length
  split_allow_true <- fusion_id(
    copy(split_expected),
    threshold = threshold,
    n_min_length = 1,
    n_max_missing = 0,
    allow_split = TRUE
  )
  expect_identical(split_allow_true[, fusionID[timegroup == 3]], NA_integer_)
  expect_false(identical(split_allow_true[, fusionID[timegroup == 3]],
                         split_allow_true[, fusionID[timegroup == 2]]))
  expect_false(identical(split_allow_true[, fusionID[timegroup == 3]],
                         split_allow_true[, fusionID[timegroup == 11]]))
  expect_identical(split_allow_true[, fusionID[timegroup == 11]],
                   split_allow_true[, fusionID[timegroup == 12]])


  lead_expected <- data.table(
    dyadID = rep('A-B', 5),
    timegroup = c(1,   3, 4, 5, 6),
    distance =  c(50,  1, 1, 1, 1)
  )
  threshold <- 25

  # If defaults,
  #  first obs together (tg 3) where lag timegroup difference is > 1, but
  #  lead timegroup difference is 1, fusionID should match tg 4-6s
  defaults_lead <- fusion_id(
    copy(lead_expected),
    threshold = threshold,
    n_min_length = 0,
    n_max_missing = 0,
    allow_split = FALSE
  )

  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 4]])
  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 5]])
  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 6]])

  # Also check with n_max_missing
  defaults_lead <- fusion_id(
    copy(lead_expected),
    threshold = threshold,
    n_min_length = 0,
    n_max_missing = 2,
    allow_split = FALSE
  )

  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 4]])
  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 5]])
  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 6]])

  # Also check with n_max_missing, n_min_length
  defaults_lead <- fusion_id(
    copy(lead_expected),
    threshold = threshold,
    n_min_length = 2,
    n_max_missing = 2,
    allow_split = FALSE
  )

  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 4]])
  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 5]])
  expect_identical(defaults_lead[, fusionID[timegroup == 3]],
                   defaults_lead[, fusionID[timegroup == 6]])

})

test_that('max timegroup difference is not greater than expected', {
  n_min_length <- 3
  n_max_missing <- 1
  allow_split <- TRUE

  fusions <- fusion_id(
    copy(edges),
    50,
    n_min_length = n_min_length,
    n_max_missing = n_max_missing,
    allow_split = allow_split
  )
  expect_all_false(
    fusions[, max(timegroup - shift(timegroup), na.rm = TRUE), by = fusionID][,
      V1 > (n_min_length + n_max_missing)]
  )
})

# Test edge_dist
context('test edge_dist')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
coords <- c('X', 'Y')
id <- 'ID'
utm <- 32736
threshold <- 10
timegroup <- 'timegroup'

group_times(DT, 'datetime', '10 minutes')

get_geometry(DT, coords = coords, crs = utm)
get_geometry(DT, coords = coords, crs = utm, output_crs = 4326,
             geometry_colname = 'geometry_longlat')

test_that('error/warn/msg if args are not provided as expected', {
  expect_error(
    edge_dist(
      DT = NULL
    ),
    'DT must be provided'
  )

  expect_error(
    edge_dist(
      DT = data.frame()
    ),
    'data.table'
  )

  expect_error(
    edge_dist(DT),
    'threshold'
  )

  expect_error(
    edge_dist(DT, threshold = threshold, id = NULL),
    'id'
  )

  expect_error(
    edge_dist(DT, threshold = threshold, id = id),
    'timegroup'
  )

  expect_error(
    edge_dist(DT, threshold = threshold, id = id, timegroup = NULL),
    'timegroup'
  )

  # geometry
  expect_message(
    edge_dist(DT, threshold = threshold, id = id, timegroup = timegroup, crs = utm),
    'crs argument is ignored'
  )
})


test_that('column names must exist in DT', {
  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = 'potato',
      coords = coords,
      timegroup = timegroup
    ),
    'not present in input'
  )

  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      coords = c('potatoX', 'potatoY'),
      timegroup = timegroup
    ),
    'not present in input'
  )

  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup,
      splitBy = 'potato'
    ),
    'not present in input'
  )

  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'potato'
    ),
    'not present in input'
  )
})

test_that('warns if timegroup is a datetime or character', {
  copyDT <- copy(DT)
  expect_warning(
    edge_dist(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'datetime'
    ),
    'timegroup provided is a',
    fixed = FALSE
  )

  copyDT <- copy(DT)
  copyDT[, posix := as.POSIXct(datetime)]
  expect_warning(
    edge_dist(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'posix'
    ),
    'timegroup provided is a',
    fixed = FALSE
  )

  copyDT <- copy(DT)
  copyDT[, idate := as.IDate(datetime)]
  expect_warning(
    edge_dist(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'idate'
    ),
    'timegroup provided is a',
    fixed = FALSE
  )
})

test_that('duplicate IDs in a timegroup detected', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '8 hours')
  expect_warning(
    edge_dist(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    ),
    'found duplicate id in a timegroup',
    fixed = FALSE
  )
})


test_that('warns about splitBy column', {
  copyDT <- copy(DT)

  group_times(copyDT, 'datetime', '5 minutes')
  copyDT[, splitBy := as.IDate(datetime)]

  expect_warning(
    edge_dist(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    ),
    'split_by'
  )
})

test_that('threshold correctly provided or error detected', {
  expect_error(
    edge_dist(DT,
      threshold = -10, timegroup = timegroup, id = id,
      coords = coords
    ),
    'threshold must be > 0'
  )

  expect_error(
    edge_dist(DT,
      threshold = 0, timegroup = timegroup, id = id,
      coords = coords
    ),
    'threshold must be > 0'
  )

  expect_error(
    edge_dist(DT,
      threshold = '0', timegroup = timegroup, id = id,
      coords = coords
    ),
    'threshold must be of class numeric'
  )

  expect_error(
    edge_dist(DT,
              threshold = -10, timegroup = timegroup, id = id,
              coords = coords, crs = utm
    ),
    'threshold must be > 0'
  )

  expect_error(
    edge_dist(DT,
              threshold = 0, timegroup = timegroup, id = id,
              coords = coords, crs = utm
    ),
    'threshold must be > 0'
  )

  expect_error(
    edge_dist(DT,
              threshold = '0', timegroup = timegroup, id = id,
              coords = coords, crs = utm
    ),
    'threshold must be of class numeric'
  )

  expect_error(
    edge_dist(DT,
              threshold = units::as_units(-1, 'm'),
              timegroup = timegroup, id = id,
              coords = coords, crs = utm
    ),
    'threshold must be > 0'
  )

  # geometry
  expect_error(
    edge_dist(DT,
              threshold = units::as_units(-1, 'm'),
              timegroup = timegroup, id = id
    ),
    'threshold must be > 0'
  )

  expect_error(
    edge_dist(DT,
              threshold = units::as_units(100, 'km'),
              timegroup = timegroup, id = id
    ),
    'units of threshold'
  )

})


test_that('error coords/geometry are not provided as expected', {
  # coords
  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      coords = c('X', NULL),
      timegroup = timegroup
    ),
    'coords must be length 2'
  )

  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      coords = c('X', 'ID'),
      timegroup = timegroup
    ),
    'coords must be of class numeric'
  )

  # geometry
  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      timegroup = timegroup,
      geometry = 'potato'
    ),
    'did you run'
  )

  expect_error(
    edge_dist(
      DT,
      threshold = threshold,
      id = id,
      timegroup = timegroup,
      geometry = 'X'
    ),
    'sfc_POINT'
  )
})


test_that('returned IDs make sense', {
  # coords
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')
  eDT <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup,
    fillNA = TRUE
  )

  IDs <- copyDT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(na.omit(eDT$ID2) %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)

  eDT <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup,
    fillNA = FALSE
  )

  IDs <- copyDT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(eDT$ID2 %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)

  # geometry
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')
  eDT <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    fillNA = TRUE
  )

  IDs <- copyDT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(na.omit(eDT$ID2) %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)

  eDT <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    fillNA = FALSE
  )

  IDs <- copyDT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(eDT$ID2 %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)

  # longlat
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')
  eDT <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    geometry = 'geometry_longlat',
    fillNA = TRUE
  )

  IDs <- copyDT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(na.omit(eDT$ID2) %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)

  eDT <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    geometry = 'geometry_longlat',
    fillNA = FALSE
  )

  IDs <- copyDT[, unique(ID)]
  expect_true(all(eDT$ID1 %in% IDs))
  expect_true(all(eDT$ID2 %in% IDs))
  expect_true(eDT[ID1 == ID2, .N] == 0)
})



test_that('returnDist works', {
  # coords
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')

  withDist <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup,
    returnDist = TRUE,
    fillNA = TRUE
  )

  woDist <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup,
    returnDist = FALSE,
    fillNA = TRUE
  )

  woThresh <- edge_dist(
    copyDT,
    threshold = NULL,
    id = id,
    coords = coords,
    timegroup = timegroup,
    returnDist = TRUE,
    fillNA = TRUE
  )

  expect_equal(
    withDist[, .(ID1, ID2, timegroup)],
    woDist[, .(ID1, ID2, timegroup)]
  )

  expect_equal(
    nrow(withDist),
    nrow(woDist)
  )

  expect_equal(
    withDist[is.na(ID2)],
    withDist[is.na(distance)]
  )

  expect_equal(
    withDist[!is.na(ID2)],
    withDist[!is.na(distance)]
  )

  expect_lt(
    withDist[, max(distance, na.rm = TRUE)],
    threshold
  )

  expect_gt(woThresh[, .N], withDist[, .N])
  expect_gt(woThresh[distance > threshold, .N], 0)

  withDistNoNA <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup,
    returnDist = TRUE,
    fillNA = FALSE
  )

  expect_true(withDistNoNA[is.na(distance), .N] == 0)
  expect_true(withDistNoNA[is.na(ID2), .N] == 0)
  expect_lt(
    withDistNoNA[, max(distance, na.rm = TRUE)],
    threshold
  )

  # geometry
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')

  withDist <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    returnDist = TRUE,
    fillNA = TRUE
  )

  woDist <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    returnDist = FALSE,
    fillNA = TRUE
  )

  woThresh <- edge_dist(
    copyDT,
    threshold = NULL,
    id = id,
    timegroup = timegroup,
    returnDist = TRUE,
    fillNA = TRUE
  )

  expect_equal(
    withDist[, .(ID1, ID2, timegroup)],
    woDist[, .(ID1, ID2, timegroup)]
  )

  expect_equal(
    nrow(withDist),
    nrow(woDist)
  )

  expect_equal(
    withDist[is.na(ID2)],
    withDist[is.na(distance)]
  )

  expect_equal(
    withDist[!is.na(ID2)],
    withDist[!is.na(distance)]
  )

  expect_lt(
    withDist[, max(distance, na.rm = TRUE)],
    threshold
  )

  expect_gt(woThresh[, .N], withDist[, .N])
  expect_gt(woThresh[distance > threshold, .N], 0)

  withDistNoNA <- edge_dist(
    copyDT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    returnDist = TRUE,
    fillNA = FALSE
  )

  expect_true(withDistNoNA[is.na(distance), .N] == 0)
  expect_true(withDistNoNA[is.na(ID2), .N] == 0)
  expect_lt(
    withDistNoNA[, max(distance, na.rm = TRUE)],
    threshold
  )
})


test_that('returns a data.table', {
  # coords
  expect_s3_class(edge_dist(
    DT,
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup
  ), 'data.table')

  # geometry
  expect_s3_class(edge_dist(
    DT,
    threshold = threshold,
    id = id,
    timegroup = timegroup
  ), 'data.table')

  expect_s3_class(edge_dist(
    DT,
    threshold = threshold,
    id = id,
    timegroup = timegroup,
    geometry = 'geometry_longlat'
  ), 'data.table')

})





test_that('handles NULL threshold', {
  # coords
  expect_equal(
    edge_dist(
      DT,
      threshold = NULL,
      id = id,
      coords = coords,
      timegroup = timegroup
    ),
    edge_dist(
      DT,
      threshold = Inf,
      id = id,
      coords = coords,
      timegroup = timegroup
    )
  )

  # geometry
  expect_equal(
    edge_dist(
      DT,
      threshold = NULL,
      id = id,
      timegroup = timegroup
    ),
    edge_dist(
      DT,
      threshold = Inf,
      id = id,
      timegroup = timegroup
    )
  )
  expect_equal(
    edge_dist(
      DT,
      threshold = NULL,
      id = id,
      timegroup = timegroup,
      geometry = 'geometry_longlat'
    ),
    edge_dist(
      DT,
      threshold = Inf,
      id = id,
      timegroup = timegroup,
      geometry = 'geometry_longlat'
    )
  )
})

# Test centroid_fusion
context('test centroid_fusion')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                   timegroup = timegroup, returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
fusion_id(edges, threshold = threshold)

clean_DT <- copy(DT)
clean_edges <- copy(edges)

test_that('edges, DT are required', {
  expect_error(centroid_fusion(edges, DT = NULL))
  expect_error(centroid_fusion(edges = NULL, DT))
})

test_that('arguments required, otherwise error detected', {
  expect_error(centroid_fusion(edges, DT, id = id, coords = 'X'),
               'coords must be length 2')
  expect_error(centroid_fusion(edges, DT, id = NULL),
               'id must be provided')
  expect_error(centroid_fusion(edges, DT, id = id,
                               coords = coords, timegroup = NULL),
               'timegroup must be provided')
})

test_that('column names must exist in DT', {
  expect_error(centroid_fusion(edges, DT, id = 'potato', coords = coords),
               'potato field')
  expect_error(centroid_fusion(edges, DT, id = id, coords = rep('potato', 2)),
               'potato field')
  expect_error(centroid_fusion(edges, DT, id = id,
                               coords = coords, timegroup = 'potato'),
               'potato field')
})

test_that('coords are correctly provided or error detected', {
  expect_error(centroid_fusion(edges, DT, id = id, coords = c('X', NULL)),
               'coords must be length 2')
  expect_error(centroid_fusion(edges, DT, id = id, coords = c('X', 'ID')),
               'coords must be of class numeric')
})

test_that('centroid column succesfully detected', {
  copyEdges <- copy(edges)[, centroid_X := 1]
  expect_message(
    centroid_fusion(copyEdges, DT, id = id, coords = coords),
    'centroid_X column will be overwritten'
  )
  copyEdges <- copy(edges)[, centroid_Y := 1]
  expect_message(
    centroid_fusion(copyEdges, DT, id = id, coords = coords),
    'centroid_Y column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  expect_equal(nrow(edges),
               nrow(centroid_fusion(edges, DT, id = id, coords = coords)))
})

test_that('two columns added to the result DT', {
  copyEdges <- copy(edges)

  expect_equal(ncol(copyEdges) + 2,
               ncol(centroid_fusion(edges, DT, id = id, coords = coords)))
})

test_that('two columns added to the result DT are doubles', {
  expect_type(centroid_fusion(edges, DT,
                              id = id, coords = coords)$centroid_X, 'double')
  expect_type(centroid_fusion(edges, DT,
                              id = id, coords = coords)$centroid_Y, 'double')
})

test_that('returns a data.table', {
  expect_s3_class(centroid_fusion(edges, DT,
                                  id = id, coords = coords), 'data.table')
})

expected_DT <- copy(clean_DT)[timegroup < 10]
expected_DT[, X := timegroup * 10 + .I]
expected_DT[, Y := timegroup * 10]

expected_edges <- copy(clean_edges)[timegroup < 10]
expected_DT[ID %in% first(expected_edges$ID1), X := NA]

test_that('xy results as expected', {
  expect_equal(
    centroid_fusion(expected_edges, expected_DT, id = id, coords = coords)[
      timegroup == 1, unique(centroid_Y)],
    10
  )

  expect_equal(
    centroid_fusion(expected_edges, expected_DT, id = id, coords = coords)[
      timegroup == 2, .N],
    0
  )

  expect_gt(
    centroid_fusion(expected_edges, expected_DT, id = id, coords = coords)[
                    timegroup == 1, unique(centroid_X)],
    10
  )

  expect_equal(
    expected_edges[, .N],
    centroid_fusion(expected_edges, expected_DT,
                    id = id, coords = coords)[, .N]
  )

  expect_equal(
    expected_edges[, unique(ID1)],
    centroid_fusion(expected_edges, expected_DT,
                    id = id, coords = coords)[, unique(ID1)]
  )

  expect_gte(
    centroid_fusion(expected_edges, expected_DT, id = id, coords = coords)[
      , uniqueN(centroid_X)],
    expected_edges[, uniqueN(fusionID)]
  )

  expect_gte(
    centroid_fusion(expected_edges, expected_DT, id = id, coords = coords)[
      , uniqueN(centroid_Y)],
    expected_edges[, uniqueN(fusionID)]
  )
})

expected_DT <- copy(clean_DT)[timegroup < 10]
expected_DT[, X := timegroup * 10 + .I]
expected_DT[, Y := timegroup * 10]

expected_edges <- copy(clean_edges)[timegroup < 10]
expected_DT[ID %in% first(expected_edges$ID1), X := NA]

get_geometry(expected_DT, coords = c('X', 'Y'), crs = 32736)

test_that('geometry results as expected', {
  expect_equal(
    centroid_fusion(expected_edges, expected_DT, id = id)[
      timegroup == 1, unique(centroid)],
    st_geometry(st_point(c(16, 10)))
  )

  expect_equal(
    centroid_fusion(expected_edges, expected_DT, id = id)[
      timegroup == 2, .N],
    0
  )

  expect_equal(
    expected_edges[, .N],
    centroid_fusion(expected_edges, expected_DT, id = id)[, .N]
  )

  expect_equal(
    expected_edges[, unique(ID1)],
    centroid_fusion(expected_edges, expected_DT, id = id)[, unique(ID1)]
  )

  expect_gte(
    centroid_fusion(expected_edges, expected_DT, id = id)[
      , length(unique(centroid))],
    expected_edges[, uniqueN(fusionID)]
  )

})


test_that('NAs in fusionID result in NAs for centroid', {
  # Set fillNA = TRUE
  edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                     timegroup = timegroup, returnDist = TRUE,
                     fillNA = TRUE)
  dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
  fusion_id(edges, threshold = threshold)

  centroids <- centroid_fusion(edges, DT, id, coords)
  expect_true(
    centroids[is.na(fusionID), all(is.na(centroid_X))]
  )
  expect_true(
    centroids[is.na(fusionID), all(is.na(centroid_Y))]
  )

})

# Test leader_edge_delay
context('test leader_edge_delay')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736
window <- 3


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
direction_step(DT, id, coords, utm)
edges <- edge_dist(DT, threshold = threshold, id = id,
                   coords = coords, timegroup = timegroup,
                   returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
fusion_id(edges, threshold = threshold)
delay <- edge_delay(edges, DT, id = id, window = window)
leader_delay <- leader_edge_delay(delay, threshold = 0.5)

# leader_edge_delay(edges = edges, threshold = 0.5, splitBy = 'population')

clean_delay <- copy(delay)

test_that('edges required', {
  expect_error(leader_edge_delay(edges = NULL))
})

test_that('threshold is numeric', {
  expect_error(leader_edge_delay(delay, threshold = 'potato'),
               'numeric')
})

test_that('column names must exist in DT', {
  missing_delay <- copy(clean_delay)[, direction_diff := NULL]
  expect_error(leader_edge_delay(missing_delay),
               'direction_diff')

  missing_delay <- copy(clean_delay)[, direction_delay := NULL]
  expect_error(leader_edge_delay(missing_delay),
               'direction_delay')

  missing_delay <- copy(clean_delay)[, ID1 := NULL]
  expect_error(leader_edge_delay(missing_delay),
               'ID1')

  missing_delay <- copy(clean_delay)[, ID2 := NULL]
  expect_error(leader_edge_delay(missing_delay),
               'ID2')

})

test_that('non-numeric cols passed as direction_diff and _delay error', {
  char_delay <- copy(delay)[, direction_diff := as.character(direction_diff)]
  expect_error(
    leader_edge_delay(char_delay, threshold = 0.5),
    "must be of class numeric"
  )
  char_delay <- copy(delay)[, direction_delay := as.character(direction_delay)]
  expect_error(
    leader_edge_delay(char_delay, threshold = 0.5),
    "must be of class integer"
  )
})

test_that('output length as expected', {
  # nrow(delay) is less than nrow(leader_delay) since leader_delay is aggregate
  expect_gt(nrow(delay), nrow(leader_delay))

  # nrow(delay) is less than when threshold is 1e-2 since some dyads will drop
  expect_gt(
    nrow(leader_edge_delay(delay)),
    nrow(leader_edge_delay(delay, threshold = 1e-2))
  )
  # nrow(delay) is the same when threshold is 1e3 since no rows will drop
  expect_equal(
    nrow(leader_edge_delay(delay)),
    nrow(leader_edge_delay(delay, threshold = 1e3))
  )

  # nrow(delay) with tiny threshold is 0 since all rows will drop
  expect_equal(nrow(leader_edge_delay(delay, threshold = 1e-3)), 0)
})

test_that('columns added to the result DT', {
  expected_cols <- c('ID1', 'ID2', 'dyadID',
                     'mean_direction_delay_dyad', 'mean_direction_delay')

  expect_setequal(expected_cols, colnames(leader_edge_delay(delay)))
  expect_length(expected_cols, ncol(leader_edge_delay(delay)))
})

test_that('column added to the result DT is double', {
  expect_type(leader_edge_delay(delay)$mean_direction_delay_dyad, 'double')
  expect_type(leader_edge_delay(delay)$mean_direction_delay, 'double')
})

test_that('returns a data.table', {
  expect_s3_class(leader_edge_delay(delay), 'data.table')
})


N_id <- 5
N_seq <- 10
seq_xy <- c(seq(0, 5, length.out = N_seq / 2),
            seq(5.1, 0, length.out = N_seq / 2))
DT_expect <- data.table(
  X = rep(seq_xy, each = N_id),
  Y = rep(seq_xy, each = N_id),
  ID = LETTERS[seq.int(N_id)]
)
DT_expect[, timegroup := seq.int(.GRP)[.GRP] + seq.int(.N), by = ID]
setorder(DT_expect, timegroup)
direction_step(DT_expect, id, coords, crs = 4326)

edge_expect <- edge_dist(DT_expect, threshold = 100, id, coords, timegroup,
                          returnDist = TRUE)
dyad_id(edge_expect, 'ID1', 'ID2')
fusion_id(edge_expect, threshold = 50)

window <- 5
delay_expect <- edge_delay(edge_expect, DT_expect, window = window, id = id)
leader_expect <- leader_edge_delay(delay_expect)

test_that('expected results are returned', {
  expect_lte(nrow(leader_expect), delay_expect[, uniqueN(dyadID) * 2])

  expect_equal(leader_expect[, uniqueN(abs(mean_direction_delay_dyad)),
                             by = dyadID],
               leader_expect[, 1, by = dyadID])

  expect_equal(leader_expect[, mean(mean_direction_delay_dyad)], 0,
               tolerance = 2)
  expect_equal(leader_expect[, mean(mean_direction_delay)], 0,
               tolerance = 2)

  expect_equal(
    leader_expect[mean_direction_delay == min(mean_direction_delay),
                  unique(ID1)],
    LETTERS[N_id]
  )
  expect_equal(
    leader_expect[mean_direction_delay == median(mean_direction_delay),
                  unique(ID1)],
    LETTERS[ceiling(N_id / 2)]
  )
  expect_equal(
    leader_expect[mean_direction_delay == max(mean_direction_delay),
                  unique(ID1)],
    LETTERS[1]
  )
})

# Test edge_direction
context('test edge_direction')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
utm <- 32736


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                   timegroup = timegroup, returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

clean_DT <- copy(DT)
clean_edges <- copy(edges)

test_that('edges, DT are required', {
  expect_error(edge_direction(edges, DT = NULL))
  expect_error(edge_direction(edges = NULL, DT))
})

test_that('arguments required, otherwise error detected', {
  expect_error(edge_direction(edges, DT, id = id, coords = 'X'),
               'coords must be length 2')
  expect_error(edge_direction(edges, DT, id = NULL), 'id must be provided')

  expect_error(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      timegroup = NULL
    ),
    'timegroup must be'
  )
  expect_error(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      timegroup = timegroup,
      crs = NULL
    ),
    'crs must be'
  )
})

test_that('column names must exist in DT', {
  expect_error(edge_direction(edges, DT, id = 'potato', coords = coords,
                              crs = utm),
               'potato field')
  expect_error(edge_direction(edges, DT, id = id, coords = rep('potato', 2),
                              crs = utm),
               'potato field')
  copyDT <- copy(DT)[, timegroup := NULL]
  expect_error(edge_direction(
    edges,
    copyDT,
    id = id,
    coords = coords,
    crs = utm,
    timegroup = 'timegroup'
  ),
  'timegroup field')

  copyEdges <- copy(edges)[, timegroup := NULL]
  expect_error(
    edge_direction(
      copyEdges,
      DT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    ),
    'timegroup field'
  )
})

test_that('coords are correctly provided or error detected', {
  expect_error(edge_direction(edges, DT, id = id, coords = c('X', NULL)),
               'coords must be length 2')
  expect_error(edge_direction(edges, DT, id = id, coords = c('X', 'ID')),
               'coords must be of class numeric')
})

test_that('direction_dyad column succesfully detected', {
  copyEdges <- copy(edges)[, direction_dyad := 1]
  expect_message(
    edge_direction(
      copyEdges,
      DT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    ),
    'direction_dyad column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  expect_equal(nrow(edges), nrow(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    )
  ))
})

test_that('one columns added to the result DT', {
  copyEdges <- copy(edges)

  expect_equal(ncol(copyEdges) + 1, ncol(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    )
  ))
})

test_that('column added to the result DT is unit', {
  expect_s3_class(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    )$direction_dyad,
    'units'
  )
})

test_that('returns a data.table', {
  expect_s3_class(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    ),
    'data.table'
  )
})

test_that('projection arg is deprecated', {
  expect_warning(
    edge_direction(
      edges,
      DT,
      id = id,
      coords = coords,
      timegroup = timegroup,
      projection = utm
    ),
    'projection argument is deprecated'
  )
})


n <- 4
DT_B <- data.table(
  X = c(0, 5, 5, 0),
  Y = c(0, 0, 5, 5),
  timegroup = rep(1, n),
  ID = LETTERS[seq.int(n)]
)
edges_B <- edge_dist(DT_B, id = id, coords = coords, timegroup = timegroup,
                     threshold = NULL, returnDist = TRUE)
dyad_id(edges_B, 'ID1', 'ID2')

dyad_dirs <- edge_direction(edges_B, DT_B, id = id,
                            coords = coords, crs = 4326)

test_that('East North West South dyads', {
  tolerance <- 0.01

  expect_equal(
    dyad_dirs[dyadID == 'A-B' & ID1 == 'A', direction_dyad],
    as_units(pi / 2, 'rad'),
    tolerance = tolerance
  )

  expect_equal(
    dyad_dirs[dyadID == 'A-C' & ID1 == 'A', direction_dyad],
    as_units(pi / 4, 'rad'),
    tolerance = tolerance
  )

  expect_equal(
    dyad_dirs[dyadID == 'A-D' & ID1 == 'A', direction_dyad],
    as_units(0, 'rad'),
    tolerance = tolerance
  )

  expect_equal(
    dyad_dirs[dyadID == 'A-B' & ID1 == 'B', direction_dyad],
    -1 * as_units(pi / 2, 'rad'),
    tolerance = tolerance
  )

  expect_equal(
    dyad_dirs[dyadID == 'B-C' & ID1 == 'C', direction_dyad],
    as_units(pi, 'rad'),
    tolerance = tolerance
  )

})

test_that('use_transform errors if crs not provided', {
  expect_error(
    edge_direction(edges, DT, id = id, coords = coords, crs = NA),
    'ensure crs is provided'
  )

  copyDT <- copy(DT)
  get_geometry(copyDT, coords = coords, crs = utm)
  st_crs(copyDT$geometry) <- NA

  expect_error(
    edge_direction(edges, copyDT, id = id),
    'ensure crs is provided'
  )
})


# geometry interface
test_that('if coords null, geometry required', {
  expect_error(edge_direction(edges, DT, id = id, coords = NULL),
               'get_geometry?')
})

test_that('crs provided with geometry gives message crs ignored', {
  get_geometry(DT, coords = coords, crs = utm)
  expect_message(edge_direction(edges, DT, id = id, crs = utm),
                 'ignored')
})

test_that('geometry correctly provided else error', {
  expect_error(edge_direction(edges, DT, id = id, geometry = 'potato'),
               'is not present')
  expect_error(edge_direction(edges, DT, id = id, geometry = 'X'),
               'must be of class')
})

test_that('geometry interface message before overwrite', {
  copyEdges <- copy(edges)[, direction_dyad := 42]
  expect_message(edge_direction(copyEdges, DT, id = id),
                 'overwritten')
})

test_that('geometry interface returns expected', {
  copyEdges <- copy(edges)
  get_geometry(DT, coords = coords, crs = utm)

  expect_equal(
    ncol(copyEdges) + 1,
    ncol(edge_direction(copyEdges, DT, id = id))
  )

  expect_equal(
    nrow(copyEdges),
    nrow(edge_direction(copyEdges, DT, id = id))
  )

  expect_true('direction_dyad' %in% colnames(edge_direction(copyEdges, DT, id = id)))

  direction_step(DT, id = id)
  expect_true('direction' %in% colnames(edge_direction(copyEdges, DT, id = id)))

  outEdges <- edge_direction(copyEdges, DT, id = id)
  expect_type(outEdges$direction_dyad, 'double')
  expect_s3_class(outEdges$direction_dyad, 'units')

  expect_equal(max(outEdges$direction_dyad, na.rm = TRUE), units::as_units(pi, 'rad'),
               tolerance = 0.01)
  expect_equal(min(outEdges$direction_dyad, na.rm = TRUE), units::as_units(-pi, 'rad'),
               tolerance = 0.01)


})

test_that('NAs in coordinates return NA', {
  copyDT <- copy(DT)

  full_edges <- edge_dist(DT, threshold = NULL, id = id, coords = coords,
                          timegroup = timegroup, returnDist = TRUE,
                          fillNA = TRUE)
  dyad_id(full_edges, id1 = 'ID1', id2 = 'ID2')

  copyDT[sample(.N, 100), X := NA]

  expect_lte(
    copyDT[is.na(X), .N],
    edge_direction(
      edges = full_edges,
      DT = copyDT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    )[is.na(direction_dyad), .N]
  )

  copyDT <- copy(DT)

  copyDT[sample(.N, 100), Y := NA]

  expect_lte(
    copyDT[is.na(Y), .N],
    edge_direction(
      edges = full_edges,
      DT = copyDT,
      id = id,
      coords = coords,
      crs = utm,
      timegroup = timegroup
    )[is.na(direction_dyad), .N]
  )
})


test_that('geometry interface can manage edges and DT both with geometry', {
  copy_DT <- copy(clean_DT)
  copy_edges <- copy(clean_edges)

  get_geometry(copy_DT, coords = coords, crs = utm)

  fusion_id(edges = copy_edges, threshold = 50)
  cent <- centroid_fusion(edges = copy_edges, DT = copy_DT, id = id)

  expect_true('direction_dyad' %in%
                colnames(edge_direction(edges = cent, DT = copy_DT, id = id)))

  # And can manage if not present
  cent[, geometry := NULL]
  expect_true('direction_dyad' %in%
                colnames(edge_direction(edges = cent, DT = copy_DT, id = id)))
})

# Test edge_zones
context('test edge_zones')

library(spatsoc)
library(units)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
utm <- 32736
timegroup <- 'timegroup'
group <- 'group'

zone_thresholds <- c(5, 20, 50)
zone_labels <- c('repulsion', 'orientation', 'attraction')
blind_volume <- 2

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                   timegroup = timegroup, returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

DT_blind <- copy(DT)
direction_step(DT_blind, id, coords, utm)
group_times(DT_blind, datetime = datetime, threshold = timethreshold)
edges <- edge_dist(DT, threshold = threshold, id = id, coords = coords,
                   timegroup = timegroup, returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
dyad_directions <- edge_direction(edges = edges, DT = DT_blind, id = id,
                                  coords = coords, crs = utm,
                                  timegroup = timegroup)

# edge_zones(
#   edges = edges,
#   zone_thresholds = zone_thresholds,
#   zone_labels = zone_labels
# )

# edge_zones(
#   edges = dyad_directions,
#   zone_thresholds = zone_thresholds,
#   zone_labels = zone_labels,
#   blind_volume = blind_volume
# )

test_that('arguments required, otherwise error detected', {
  expect_error(
    edge_zones(
      edges = NULL,
      zone_thresholds = zone_thresholds,
      zone_labels = zone_labels
    ),
    'edges'
  )

  expect_error(
    edge_zones(
      edges = edges,
      zone_thresholds = NULL,
      zone_labels = zone_labels
    ),
    'zone_thresholds'
  )

  expect_error(
    edge_zones(
      edges = edges,
      zone_thresholds = zone_thresholds,
      zone_labels = NULL
    ),
    'zone_labels'
  )
})

test_that('columns must exist in edges', {
  rm_col <- copy(edges)[, .SD, .SDcols = -'distance']
  expect_error(edge_zones(rm_col, zone_thresholds, zone_labels),
               'distance')

  rm_col <- copy(dyad_directions)[, .SD, .SDcols = -'direction']
  expect_error(edge_zones(rm_col, zone_thresholds, zone_labels, blind_volume),
               'direction')

  rm_col <- copy(dyad_directions)[, .SD, .SDcols = -'direction_dyad']
  expect_error(edge_zones(rm_col, zone_thresholds, zone_labels, blind_volume),
               'direction_dyad')

})

test_that('columns are correctly provided or error detected', {
  copyEdges <- copy(edges)[, distance := as.character(distance)]
  expect_error(edge_zones(copyEdges, zone_thresholds, zone_labels),
               'distance must be of class numeric')

  copyDyad <- copy(dyad_directions)[, direction := units::set_units(direction, 'degrees')]
  expect_error(edge_zones(copyDyad, zone_thresholds, zone_labels, blind_volume),
               'radians',
              )

  copyDyadDir <- copy(dyad_directions)[, direction_dyad := units::set_units(direction_dyad, 'degrees')]
  expect_error(edge_zones(copyDyadDir, zone_thresholds, zone_labels, blind_volume),
               'radians')
})

test_that('zone column succesfully detected', {
  zones_present <- copy(edges)[, zone := 1]
  expect_message(
    edge_zones(zones_present, zone_thresholds, zone_labels),
    'zone column will be overwritten'
  )
})

test_that('no rows are added to the result', {
  copyEdges <- copy(edges)

  expect_equal(nrow(copyEdges),
               nrow(edge_zones(copyEdges, zone_thresholds, zone_labels)))
})

test_that('one column added to the result', {
  copyEdges <- copy(edges)

  expect_equal(ncol(copyEdges) + 1,
               ncol(edge_zones(copyEdges, zone_thresholds, zone_labels)))

  expect_false('direction_dyad_relative' %in%
                 colnames(edge_zones(copyEdges, zone_thresholds, zone_labels)))
})

test_that('one column added to the result is a factor', {
  copyEdges <- copy(edges)

  expect_s3_class(edge_zones(copyEdges, zone_thresholds, zone_labels)$zone,
                  'factor')

  expect_equal(levels(edge_zones(copyEdges, zone_thresholds, zone_labels)$zone),
              zone_labels)

  copyDyadDirs <- copy(dyad_directions)

  expect_s3_class(edge_zones(copyDyadDirs, zone_thresholds, zone_labels,
                         blind_volume)$zone,
                  'factor')

  expect_equal(levels(edge_zones(copyDyadDirs, zone_thresholds, zone_labels,
                                 blind_volume)$zone),
               c(zone_labels, 'blind'))

})
test_that('returns a data.table', {
  expect_s3_class(edge_zones(edges, zone_thresholds, zone_labels),
                  'data.table')
})



test_that('results are expected', {
  expected <- merge(edge_zones(edges, zone_thresholds, zone_labels),
                    data.table(threshold = zone_thresholds, zone = zone_labels),
                    by = 'zone')

  expect_true(all(expected[, distance < threshold, by = zone]$V1))

  expected <- merge(edge_zones(dyad_directions, zone_thresholds,
                               zone_labels, blind_volume),
                    data.table(threshold = zone_thresholds, zone = zone_labels),
                    by = 'zone')

  expect_true(all(expected[, distance < threshold, by = zone]$V1))

})

# Test edge_delay
context('test edge_delay')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736
window <- 3


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
direction_step(DT, id, coords, utm)
edges <- edge_dist(DT, threshold = threshold, id = id,
                   coords = coords, timegroup = timegroup,
                   returnDist = TRUE, fillNA = FALSE)
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
fusion_id(edges, threshold = threshold)

clean_DT <- copy(DT)
clean_edges <- copy(edges)

# edge_delay(edges = edges, DT = DT, id = id, window = window)

test_that('edges, DT are required', {
  expect_error(edge_delay(edges, DT = NULL))
  expect_error(edge_delay(edges = NULL, DT))
})

test_that('arguments required, otherwise error detected', {
  expect_error(edge_delay(edges, DT, id = NULL),
               'id must be provided')
  expect_error(edge_delay(edges, DT, id = id, window = NULL),
               'window must be provided')
})

test_that('window is numeric, timegroup is integer', {
  expect_error(edge_delay(edges, DT, id = id, window = 'potato'),
               'numeric')
  copy_edges <- copy(clean_edges)
  copy_edges[, timegroup := as.character(timegroup)]
  expect_error(edge_delay(copy_edges, DT, id = id, window = 2),
               'integer')

  copy_DT <- copy(clean_DT)
  copy_DT[, timegroup := as.character(timegroup)]
  expect_error(edge_delay(edges, copy_DT, id = id, window = 2),
               'integer')
})

test_that('column names must exist in DT', {
  expect_error(edge_delay(edges, DT, id = 'potato'),
               'potato field')

  copy_edges <- copy(clean_edges)
  copy_edges[, timegroup := NULL]
  expect_error(edge_delay(copy_edges, DT, id = id, window = window),
               'timegroup field')

  copy_edges <- copy(clean_edges)
  copy_edges[, fusionID := NULL]
  expect_error(edge_delay(copy_edges, DT, id = id, window = window),
               'fusionID field')

  copy_edges <- copy(clean_edges)
  copy_edges[, dyadID := NULL]
  expect_error(edge_delay(copy_edges, DT, id = id, window = window),
               'dyadID field')

  expect_error(edge_delay(edges, DT, id = id, window = window,
                          direction = 'potato'),
               'potato field')

  copy_DT <- copy(clean_DT)
  copy_DT[, timegroup := NULL]
  expect_error(edge_delay(edges, copy_DT, id = id, window = window),
               'timegroup field')
})

test_that('no rows are added to the result edges', {
  expect_equal(nrow(edges),
               nrow(edge_delay(edges, DT, id = id, window = window)))
})

test_that('column added to the result DT', {
  copyEdges <- copy(edges)

  expected_cols <- c(colnames(copyEdges), 'direction_delay', 'direction_diff')

  expect_setequal(expected_cols,
                  colnames(edge_delay(edges, DT, id = id, window = window)))
  expect_length(expected_cols, ncol(edge_delay(edges, DT, id = id, window = window)))

})

test_that('column added to the result DT is integer', {
  expect_type(edge_delay(edges, DT, id = id,
                         window = window)$direction_delay, 'integer')
})

test_that('returns a data.table', {
  expect_s3_class(edge_delay(edges, DT, id = id, window = window), 'data.table')
})

test_that('direction colname can be different than default', {
  copyDT <- copy(clean_DT)
  setnames(copyDT, 'direction', 'potato')
  expect_s3_class(edge_delay(edges, copyDT, id = id, window = window,
                             direction = 'potato'),
                  'data.table')
})

test_that('window column in edge, DT does not influence results', {
  copyDT <- copy(clean_DT)
  copyDT[, window := 42]

  copyEdges <- copy(clean_edges)
  copyEdges[, window := 42]

  expect_equal(
    edge_delay(edges, copyDT, id = id, window = window),
    edge_delay(edges, DT, id = id, window = window)
  )

  expect_equal(
    edge_delay(copyEdges, DT, id = id,
               window = window)[, .SD, .SDcols = -'window'],
    edge_delay(edges, DT, id = id, window = window)
  )

  expect_equal(
    edge_delay(copyEdges, copyDT, id = id,
               window = window)[, .SD, .SDcols = -'window'],
    edge_delay(edges, DT, id = id, window = window)
  )

})

test_that('rows with NAs in ID1, ID2, dyadID, fusionID are dropped', {
  n_dropped <- 42
  na_edges <- copy(clean_edges)
  na_edges[seq.int(n_dropped), ID1 := NA]

  expect_equal(
    edge_delay(edges = na_edges, DT = DT, window, id)[, .N + n_dropped],
    edge_delay(edges = edges, DT = DT, window, id)[, .N]
  )

  na_edges <- copy(clean_edges)
  na_edges[seq.int(n_dropped), ID2 := NA]

  expect_equal(
    edge_delay(edges = na_edges, DT = DT, window, id)[, .N + n_dropped],
    edge_delay(edges = edges, DT = DT, window, id)[, .N]
  )

  na_edges <- copy(clean_edges)
  na_edges[seq.int(n_dropped), dyadID := NA]

  expect_equal(
    edge_delay(edges = na_edges, DT = DT, window, id)[, .N + n_dropped],
    edge_delay(edges = edges, DT = DT, window, id)[, .N]
  )
  na_edges <- copy(clean_edges)
  na_edges[seq.int(n_dropped), fusionID := NA]

  expect_equal(
    edge_delay(edges = na_edges, DT = DT, window, id)[, .N + n_dropped],
    edge_delay(edges = edges, DT = DT, window, id)[, .N]
  )

})

test_that('setorder doesnt impact results', {
  reorder_edges <- copy(edges)
  setorder(reorder_edges, ID1)
  expect_equal(
    edge_delay(edges = edges, DT, window, id)[, sort(direction_delay)],
    edge_delay(edges = reorder_edges, DT, window, id)[, sort(direction_delay)]
  )

  reorder_DT <- copy(DT)
  setorder(reorder_DT, ID)
  expect_equal(
    edge_delay(edges = edges, DT, window, id)[, sort(direction_delay)],
    edge_delay(edges = edges, reorder_DT, window, id)[, sort(direction_delay)]
  )
})

test_that('forward/reverse are equal', {
  swap_edges <- edge_dist(DT,
                          threshold = threshold, id = id,
                          coords = coords, timegroup = timegroup,
                          returnDist = TRUE, fillNA = FALSE
  )
  setnames(swap_edges, c('ID1', 'ID2'), c('ID2', 'ID1'))
  dyad_id(swap_edges, id1 = 'ID1', id2 = 'ID2')
  fusion_id(swap_edges, threshold = threshold)

  swap_delay <- edge_delay(swap_edges, DT, window, id)
  delay <- edge_delay(edges, DT, window, id)

  swap_leader <- leader_edge_delay(swap_delay)
  leader <- leader_edge_delay(delay)

  expect_equal(
    swap_leader[order(dyadID, ID1)],
    leader[order(dyadID, ID1)]
  )
})



N_id <- 5
N_seq <- 10
seq_xy <- c(seq(0, 5, length.out = N_seq / 2),
            seq(5.1, 0, length.out = N_seq / 2))
DT_expect <- data.table(
  X = rep(seq_xy, each = N_id),
  Y = rep(seq_xy, each = N_id),
  ID = LETTERS[seq.int(N_id)]
)
DT_expect[, timegroup := seq.int(.GRP)[.GRP] + seq.int(.N), by = ID]
setorder(DT_expect, timegroup)
direction_step(DT_expect, id, coords, crs = 4326)

edge_expect <- edge_dist(DT_expect, threshold = 100, id, coords, timegroup,
                         returnDist = TRUE)
dyad_id(edge_expect, 'ID1', 'ID2')
fusion_id(edge_expect, threshold = 100)

window <- 5
delay_expect <- edge_delay(edge_expect, DT_expect, window = window, id = id)

leader_expect <- leader_edge_delay(delay_expect)

test_that('expected results are returned', {
  expect_lte(nrow(delay_expect), nrow(edge_expect))
  expect_lte(nrow(DT_expect), nrow(delay_expect))

  mean_delays <- delay_expect[, mean(direction_delay, na.rm = TRUE), by = ID1]

  expect_equal(mean_delays[V1 == min(V1), ID1], LETTERS[N_id])
  expect_equal(mean_delays[V1 == median(V1), ID1], LETTERS[ceiling(N_id / 2)])
  expect_equal(mean_delays[V1 == max(V1), ID1], LETTERS[1])

  mean_delays_wrt_A <- delay_expect[
    ID1 == 'A', mean(direction_delay, na.rm = TRUE), by = ID2]
  expect_equal(mean_delays_wrt_A[V1 == max(V1), ID2], LETTERS[N_id])
})

test_that('exaggerated window size returns the same',  {
  expect_equal(
    edge_delay(edge_expect, DT_expect, window = window, id = id),
    edge_delay(edge_expect, DT_expect, window = window + 10, id = id)
  )

  expect_equal(
    edge_delay(edge_expect, DT_expect, window = window, id = id),
    edge_delay(edge_expect, DT_expect, window = window + 1000, id = id)
  )
})

#' Dyad centroid
#'
#' `centroid_dyad` calculates the centroid (mean location) of a dyad in each
#' observation identified by `edge_nn` or `edge_dist`. The function expects an
#' edge-list generated by `edge_nn` or `edge_dist` and a `data.table` with
#' relocation data appended with a `timegroup` column from `group_times`.
#' Relocation data should be in two columns representing the X and Y
#' coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `edges` and `DT` must be `data.table`s. If your data is a `data.frame`,
#' you can convert it by reference using [data.table::setDT()] or by reassigning
#' using [data.table::data.table()].
#'
#' The `edges` and `DT` are internally merged in this function using the columns
#' `id`, `dyadID` and `timegroup`. This function expects a `dyadID` present,
#' generated with the `dyad_id` function. The `id` and `timegroup` arguments
#' expect the names of a column in `DT` which correspond to the id and timegroup
#' columns.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Centroid function" for details on the underlying centroid function used.
#'
#' @inheritSection centroid_group Interface
#' @inheritSection centroid_group Centroid function
#'
#' @param edges edge-list generated generated by `edge_dist` or
#'  `edge_nn`, with dyad ID column generated by `dyad_id`
#' @param DT input data.table with timegroup column generated with
#'   `group_times` matching the input data.table used to generate the edge
#'   list with `edge_nn` or `edge_dist`
#' @inheritParams edge_alignment
#' @inheritParams group_pts
#' @inheritParams centroid_group
#'
#' @return `centroid_dyad` returns the input `edges` appended with centroid
#'   column(s) for each timestep and dyad id.
#'
#'  If `coords` are provided, the centroid columns will be named by prefixing
#'  the coordinate column names with "centroid_" (eg. "X" = "centroid_X"). If
#'  `geometry` is used, the centroid column will be named "centroid".
#'
#'  Note: due to the merge required within this function, the output needs to be
#'  reassigned unlike some other `spatsoc` functions like `dyad_id` and
#'  `group_pts`. See details in
#'  [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'  A message is returned when the centroid column(s) already exist in the input
#'  because they will be overwritten.
#'
#' @export
#' @family Centroid functions
#' @seealso [dyad_id] [edge_dist] [edge_nn] [group_pts]
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge-list generation
#' edges <- edge_dist(
#'     DT,
#'     threshold = 100,
#'     id = 'ID',
#'     coords = c('X', 'Y'),
#'     timegroup = 'timegroup',
#'     returnDist = TRUE,
#'     fillNA = FALSE
#'   )
#'
#' # Generate dyad id
#' dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
#'
#' # Calculate dyad centroid
#' centroids <- centroid_dyad(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   timegroup = 'timegroup'
#' )
#'
#' print(centroids)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup')
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
#' centroids <- centroid_dyad(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   timegroup = 'timegroup'
#' )
#' print(centroids)
centroid_dyad <- function(
    edges = NULL,
    DT = NULL,
    id = NULL,
    coords = NULL,
    crs = NULL,
    timegroup = 'timegroup',
    geometry = 'geometry') {

  centroid_(
    'dyadID',
    edges = edges,
    DT = DT,
    id = id,
    coords = coords,
    crs = crs,
    timegroup = timegroup,
    geometry = geometry
  )
}

#' Group lines
#'
#' `group_lines` groups rows into spatial groups by generating LINESTRINGs and
#' grouping based on spatial intersection. The function expects a `data.table`
#' with relocation data, individual identifiers and a distance threshold. The
#' relocation data is transformed into sf LINESTRINGs using [build_lines] and
#' intersecting LINESTRINGs are grouped. The threshold argument is used to
#' specify the distance criteria for grouping. Relocation data should be in two
#' columns representing the X and Y coordinates.
#'
#' ## R-spatial evolution
#'
#' Please note, spatsoc has followed updates from R spatial, GDAL and PROJ for
#' handling coordinate reference systems, see more at
#' <https://r-spatial.org/r/2020/03/17/wkt.html>.
#'
#' In addition, `group_lines` (and [build_lines]) previously used
#' [sp::SpatialLines], rgeos::gIntersects, rgeos::gBuffer but have been
#' updated to use [sf::st_as_sf], [sf::st_linestring], [sf::st_intersects], and
#' [sf::st_buffer] according to the R-spatial evolution, see more at
#' <https://r-spatial.org/r/2022/04/12/evolution.html>.
#'
#' ## Notes on arguments
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using [data.table::setDT].
#'
#' The `id`, `coords`, `sortBy` (and optional `timegroup`
#' and `splitBy`) arguments expect the names of respective columns in
#' `DT` which correspond to the individual identifier, X and Y coordinates,
#' sorting, timegroup (generated by [group_times]) and additional grouping
#' columns.
#'
#' The `crs` argument expects a numeric or character defining the coordinate
#' reference system. For example, for UTM zone 36N (EPSG 32736), the crs
#' argument is either `crs = 'EPSG:32736'` or `crs = 32736`. See details in
#' [`sf::st_crs()`] and <https://spatialreference.org> for a list of EPSG codes.
#'
#' The `sortBy` argument is used to order the input `DT` when creating sf
#' LINESTRINGs. It must a column in the input `DT` of type POSIXct to ensure the
#' rows are sorted by date time.
#'
#' The `threshold` must be provided in the units of the coordinates. The
#' `threshold` can be equal to 0 if strict overlap is intended, otherwise it
#' should be some value greater than 0, in the units of the coordinates. Eg. in
#' the case of UTM, a `threshold = 50` would indicate a 50m distance threshold.
#'
#' The `timegroup` argument is optional, but recommended to pair with
#' [group_times]. The intended framework is to group rows temporally with
#' [group_times] then spatially with [group_lines] (or [group_pts],
#' [group_polys]). With [group_lines], pick a relevant [group_times] threshold
#' such as `'1 day'` or `'7 days'` which is informed by your study species,
#' system or question.
#'
#' The `splitBy` argument offers further control building LINESTRINGs. If in
#' your input `DT`, you have multiple temporal groups (e.g.: years) for example,
#' you can provide the name of the column which identifies them and build
#' LINESTRINGs for each individual in each year. The grouping performed by
#' [group_lines] will only consider rows within each `splitBy` subgroup.
#'
#' @return `group_lines` returns the input `DT` appended with a "group"
#'   column.
#'
#'   This column represents the spatial (and if `timegroup` was provided -
#'   spatiotemporal) group calculated by intersecting lines. As with the other
#'   grouping functions,  the actual value of group is arbitrary and represents
#'   the identity of a given group where 1 or more individuals are assigned to a
#'   group. If the data was reordered, the group may change, but the contents of
#'   each group would not.
#'
#'   A message is returned when a column named "group" already exists in the
#'   input `DT`, because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @inheritParams group_pts
#' @inheritParams build_polys
#' @inheritParams build_lines
#' @param threshold The width of the buffer around the lines in the units of the
#'   crs. Use `threshold = 0` to compare intersection without buffering.
#' @param sfLines Alternatively to providing a DT, provide a simple feature
#'   LINESTRING object generated with the sf package. The id argument is
#'   required to provide the identifier matching each LINESTRING.
#'   If an sfLines object is provided, groups cannot be calculated by timegroup
#'   or splitBy.
#' @param sortBy Character string of date time column(s) to sort rows by. Must
#'   be a POSIXct.
#'
#' @export
#'
#' @family Spatial grouping
#' @seealso [build_lines] [group_times]
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Subset only individuals A, B, and C
#' DT <- DT[ID %in% c('A', 'B', 'C')]
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # EPSG code for example data
#' utm <- 32736
#'
#' group_lines(DT, threshold = 50, crs = utm, sortBy = 'datetime',
#'             id = 'ID', coords = c('X', 'Y'))
#'
#' ## Daily movement tracks
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '1 day')
#'
#' # Subset only first 50 days
#' DT <- DT[timegroup < 25]
#'
#' # Spatial grouping
#' group_lines(DT, threshold = 50, crs = utm,
#'             id = 'ID', coords = c('X', 'Y'),
#'             timegroup = 'timegroup', sortBy = 'datetime')
#'
#' ## Daily movement tracks by population
#' group_lines(DT, threshold = 50, crs = utm,
#'             id = 'ID', coords = c('X', 'Y'),
#'             timegroup = 'timegroup', sortBy = 'datetime',
#'             splitBy = 'population')
group_lines <- function(
    DT = NULL,
    threshold = NULL,
    crs = NULL,
    id = NULL,
    coords = NULL,
    timegroup = NULL,
    sortBy = NULL,
    splitBy = NULL,
    sfLines = NULL,
    projection = NULL) {

  # due to NSE notes in R CMD check
  group <- ..coords <- ..id <- ..sortBy <- withinGroup <- NULL

  if (!is.null(projection)) {
    warning('projection argument is deprecated, setting crs = projection')
    crs <- projection
  }

  if (is.null(threshold)) {
    message('threshold missing, using 0 by default')
    threshold <- 0
  } else if (!is.numeric(threshold)) {
    assert_inherits(threshold, 'numeric')
  } else if (threshold < 0) {
    assert_relation(threshold, `>`, 0)
  }

  if (!is.null(sfLines) && !is.null(DT)) {
    stop('cannot provide both DT and sfLines')
  } else if (is.null(sfLines) && is.null(DT)) {
    stop('must provide either DT or sfLines')
  } else if (!is.null(sfLines) && is.null(DT)) {
    if (!inherits(sfLines, 'sf') ||
        !'LINESTRING' %in% sf::st_geometry_type(sfLines)) {
      stop('sfLines provided must be a sf object with LINESTRINGs')
    }
    assert_not_null(id)

    if (data.table::uniqueN(sfLines[[id]]) != nrow(sfLines)) {
      stop('number of unique values in sfLines does not match nrow(sfLines)')
    }

    if (threshold == 0) {
      inter <- sf::st_intersects(sfLines, sfLines, sparse = FALSE)
    } else {
      buffered <- sf::st_buffer(sfLines, dist = threshold)
      inter <- sf::st_intersects(sfLines, buffered, sparse = FALSE)
    }
    dimnames(inter) <- list(sfLines[[id]], sfLines[[id]])
    g <- igraph::graph_from_adjacency_matrix(inter)
    ovr <- igraph::components(g)$membership
    out <- data.table::data.table(names(ovr),
                                  unlist(ovr))
    data.table::setnames(out, c('ID', 'group'))
    return(out[])
  } else if (is.null(sfLines) && !is.null(DT)) {
    assert_not_null(crs)
    assert_not_null(coords)
    assert_not_null(id)
    assert_not_null(sortBy)

    assert_are_colnames(DT, c(id, coords, timegroup, sortBy))

    if ('group' %in% colnames(DT)) {
      message('group column will be overwritten by this function')
      data.table::set(DT, j = 'group', value = NULL)
    }
  }

  if (is.null(timegroup)) {
    withCallingHandlers({
      lns <- build_lines(
        DT = DT,
        crs = crs,
        coords = coords,
        id = id,
        sortBy = sortBy,
        splitBy = splitBy
      )},
      warning = function(w){
        if (startsWith(conditionMessage(w), 'some rows dropped')) {
          invokeRestart('muffleWarning')
        }
      }
    )
    if (nrow(lns) != 0) {
      if (threshold == 0) {
        inter <- sf::st_intersects(lns, lns, sparse = FALSE)
      } else {
        buffered <- sf::st_buffer(lns, dist = threshold)
        inter <- sf::st_intersects(lns, buffered, sparse = FALSE)
      }
      dimnames(inter) <- list(lns[[id]], lns[[id]])
      g <- igraph::graph_from_adjacency_matrix(inter)
      ovr <- igraph::components(g)$membership
      ovrDT <- data.table::data.table(ID = names(ovr),
                                      group = unlist(ovr))
    } else {
      ovrDT <- data.table::data.table(ID = DT[[id]], group = NA_integer_)
    }

    data.table::setnames(ovrDT, c(id, 'group'))
    DT[ovrDT, group := group, on = id]
    if (DT[is.na(group), .N] > 0) {
      warning(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'some rows were dropped,
          cannot build a line with < 2 points.
          in this case, group set to NA.'
        )
      )
    }
    return(DT[])
  } else {

    if (is.null(splitBy)) {
      splitBy <- timegroup
    }
    else {
      splitBy <- c(splitBy, timegroup)
    }
    ovrDT <-
      DT[, {
        withCallingHandlers({
          lns <- build_lines(
            DT = .SD,
            crs = crs,
            coords = ..coords,
            id = ..id,
            sortBy = ..sortBy
          )},
          warning = function(w){
            if (startsWith(conditionMessage(w), 'some rows dropped')) {
              invokeRestart('muffleWarning')
            }
          }
        )
        if (!is.null(lns)) {
          if (threshold == 0) {
            inter <- sf::st_intersects(lns, lns, sparse = FALSE)
          } else {
            buffered <- sf::st_buffer(lns, dist = threshold)
            inter <- sf::st_intersects(lns, buffered, sparse = FALSE)
          }
          dimnames(inter) <- list(lns[[id]], lns[[id]])
          g <- igraph::graph_from_adjacency_matrix(inter)
          ovr <- igraph::components(g)$membership
          out <- data.table::data.table(names(ovr),
                                        unlist(ovr))
          data.table::setnames(out, c(..id, 'withinGroup'))
        } else {
          out <- data.table(get(..id), withinGroup = NA_real_)
          data.table::setnames(out, c(..id, 'withinGroup'))

        }
      }, by = c(splitBy), .SDcols = c(coords, id, sortBy)]

    DT[ovrDT, withinGroup := withinGroup, on = c(id, splitBy)]
    DT[, group := ifelse(is.na(withinGroup), NA_integer_, .GRP),
       by = c(splitBy, 'withinGroup')]
    data.table::set(DT, j = 'withinGroup', value = NULL)
    if (DT[is.na(group), .N] > 0) {
      warning(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'some rows were dropped,
          cannot build a line with < 2 points.
          in this case, group set to NA.'
        )
      )
    }
    return(DT[])
  }
}

# Test group_lines
context('test group_lines')

library(spatsoc)

DT <- fread('../testdata/DT.csv')

utm <- 32736

DT[, datetime := as.POSIXct(datetime)]
DT[, jul := data.table::yday(datetime)]
DT[, family := sample(c(1, 2, 3, 4), .N, replace = TRUE)]

test_that('one of DT or sfLines is required, not both or neither', {
  expect_error(
    group_lines(
      DT = NULL,
      threshold = 10,
      sfLines = NULL
    ),
    'must provide either DT or sfLines'
  )

  expect_error(
    group_lines(
      DT = DT,
      threshold = 10,
      sfLines = build_lines(
        DT,
        crs = utm,
        coords = c('X', 'Y'),
        id = 'ID',
        sortBy = 'datetime'
      )
    ),
    'cannot provide both DT and sfLines'
  )
})


test_that('coords, id, crs provided and proper format', {
  copyDT <- copy(DT)
  group_times(copyDT, datetime = 'datetime', threshold = '1 day')
  expect_error(
    group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = 'timegroup',
      id = NULL,
      coords = c('X', 'Y'),
      crs = utm
    ),
    'id must be provided'
  )

  expect_error(
    group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = NULL
    ),
    'crs must be provided',
    fixed = FALSE
  )

  expect_error(
    group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = NULL,
      crs = utm
    ),
    'coords must be provided'
  )

  expect_error(
    group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = NULL
    ),
    'sortBy must be provided'
  )

})


test_that('column names must exist in DT', {
  expect_error(
    group_lines(
      DT = DT,
      threshold = 10,
      timegroup = 'timegroup',
      id = 'potatoID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    group_lines(
      DT = DT,
      threshold = 10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('potatoX', 'potatoY'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'not present in input',
    fixed = FALSE
  )
})



test_that('timegroup is correctly provided but is not required', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, timegroup)]
  expect_error(
    group_lines(
      DT = copyDT[N > 2],
      threshold = 10,
      timegroup = 'potato',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'potato field', fixed = FALSE
  )

  expect_true(inherits(group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = NULL,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ), 'data.table')
  )
})



test_that('threshold is correctly provided, or error', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, timegroup)]
  expect_message(
    group_lines(
      DT = copyDT[N > 2],
      threshold = NULL,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'threshold missing, using 0 by default'
  )

  expect_error(
    group_lines(
      DT = DT,
      threshold = -10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'threshold must be > 0'
  )

  expect_error(
    group_lines(
      DT = DT,
      threshold = '0',
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'threshold must be of class numeric'
  )
})


test_that('group lines returns a single warning for <2 locs', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )

  copyDT[1, ID := 'Z']
  expect_warning(
    group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'some rows were dropped, cannot build a line with',
    fixed = FALSE
  )

  copyDT <- copy(DT)
  copyDT[1, ID := 'Z']
  expect_warning(
    group_lines(
      DT = copyDT,
      threshold = 10,
      timegroup = NULL,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'some rows were dropped, cannot build a line with', fixed = FALSE
  )

})

test_that('group column is added to result or NA if < 2 locs', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, timegroup)]

  expect_true('group' %in%
                colnames(
                  group_lines(
                    DT = copyDT[N > 2],
                    threshold = 10,
                    timegroup = 'timegroup',
                    id = 'ID',
                    coords = c('X', 'Y'),
                    crs = utm,
                    sortBy = 'datetime'
                  )
                ))

  copyDT <- copy(DT)
  copyDT[1, ID := 'Z']
  expect_true(suppressWarnings(
    group_lines(
      DT = copyDT,
      threshold = 10,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    )[is.na(group), .N] != 0)
  )

  copyDT <- DT[, .SD[1], by = ID]
  expect_true(suppressWarnings(
    group_lines(
      DT = copyDT,
      threshold = 10,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    )[is.na(group), .N] != 0)
  )
})

test_that('only one column added to the result DT', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, timegroup)]

  expect_equal(ncol(copyDT[N > 2]) + 1,
               ncol(
                 group_lines(
                   DT = copyDT[N > 2],
                   threshold = 10,
                   timegroup = 'timegroup',
                   id = 'ID',
                   coords = c('X', 'Y'),
                   crs = utm,
                   sortBy = 'datetime'
                 )
               ))
})

test_that('no rows are added to the result DT', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, timegroup)]
  copyDT <- copyDT[N > 2]
  expect_equal(nrow(copyDT),
               nrow(
                 group_lines(
                   DT = copyDT,
                   threshold = 10,
                   timegroup = 'timegroup',
                   id = 'ID',
                   coords = c('X', 'Y'),
                   crs = utm,
                   sortBy = 'datetime'
                 )
               ))
})


test_that('withinGroup is not returned to the user', {
  copyDT <- copy(DT)
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, timegroup)]
  expect_false('withinGroup' %in% colnames(
    group_lines(
      DT = copyDT[N > 2],
      threshold = 10,
      timegroup = 'timegroup',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    )
  ))
})

test_that('only 1 unique timegroup * splitBy', {
  copyDT <- DT[, mnth := month(datetime)][, yr := year(datetime)]
  copyDT[, nBy := .N, by = .(yr, mnth, ID)]
  copyDT <- copyDT[nBy > 2]
  group_lines(
    DT = copyDT,
    threshold = 100,
    timegroup = 'mnth',
    id = 'ID',
    coords = c('X', 'Y'),
    crs = utm,
    splitBy = 'yr',
    sortBy = 'datetime'
  )
  expect_equal(copyDT[, .(uniqueMonths = uniqueN(mnth)),
                      by = .(group)][, max(uniqueMonths)],
               1)

  expect_equal(copyDT[, .(uniqueYears = uniqueN(yr)),
                      by = .(group)][, max(uniqueYears)],
               1)

})
# or uniquen(.SD, by = c(id, splitBy))


test_that('group column succesfully detected', {
  copyDT <- copy(DT)[, group := 1]
  copyDT[, mnth := month(datetime)][, yr := year(datetime)]
  copyDT[, nBy := .N, by = .(yr, mnth, ID)]
  copyDT <- copyDT[nBy > 2]

  expect_message(
    group_lines(
      DT = copyDT,
      threshold = 0,
      timegroup = 'mnth',
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      splitBy = 'yr',
      sortBy = 'datetime'
    ),
    'group column will be overwritten'
  )

  expect_message(
    group_lines(
      DT = copyDT,
      threshold = 0,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      splitBy = 'yr',
      sortBy = 'datetime'
    ),
    'group column will be overwritten'
  )
})


# group_lines(
#   DT = DT,
#   threshold = 10,
#   datetime = 'timegroup',
#   splitBy = 'family',
#   id = 'ID',
#   coords = c('X', 'Y'),
#   crs = utm
# )


test_that('sfLines provided must be an sf LINESTRING', {
  expect_error(
    group_lines(sfLines = DT, threshold = 10, id = 'ID'),
    'sfLines provided must be a sf object'
  )

  # TODO:  test linestring
})

test_that('sfLines id provided, uniqueN(id) match n rows', {
  sfLines <- build_lines(
    DT = DT,
    id = 'ID',
    coords = c('X', 'Y'),
    crs = utm,
    sortBy = 'datetime'
  )

  expect_error(
    group_lines(sfLines = sfLines, threshold = 10, id = NULL),
    'id must be provided'
  )

  sfLines_dup <- rbind(sfLines, sfLines)

  expect_error(
    group_lines(sfLines = sfLines_dup, threshold = 10, id = 'ID'),
    'number of unique values in sfLines'
  )
})

test_that('sfLines provided returns data.table', {
  sfLines <- build_lines(
    DT = DT,
    id = 'ID',
    coords = c('X', 'Y'),
    crs = utm,
    sortBy = 'datetime'
  )

  expect_true(
    inherits(group_lines(sfLines = sfLines, threshold = 10, id = 'ID'), 'data.table')
  )

  expect_true(
    inherits(group_lines(sfLines = sfLines, threshold = 0, id = 'ID'), 'data.table')
  )

  expect_equal(
    nrow(group_lines(sfLines = sfLines, threshold = 10, id = 'ID')),
    nrow(sfLines)
  )

})


test_that('splitBy argument doesnt use splitBy column', {
  copyDT <- copy(DT)

  copyDT[, splitBy := as.IDate(datetime)]
  group_times(copyDT, 'datetime', '1 day')

  utm <- 'EPSG:32736'

  expect_true(
    group_lines(
      DT = copyDT,
      threshold = 0,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      timegroup = 'timegroup',
      splitBy = 'splitBy',
      sortBy = 'datetime'
    )[, uniqueN(splitBy), group][, all(V1 == 1)]
  )
})



test_that('group_lines can accomodate NULL timegroup and non null splitBy', {
  copyDT <- copy(DT)
  copyDT[, yr := year(datetime)]

  expect_contains(
    group_lines(
      DT = copyDT,
      threshold = 10,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      timegroup = NULL,
      splitBy = 'yr',
      sortBy = 'datetime'
    ) |> colnames(),
    'group'
  )

})


test_that('projection arg is deprecated', {
  expect_warning(
    group_lines(
      DT = DT,
      threshold = 0,
      id = 'ID',
      coords = c('X', 'Y'),
      splitBy = 'yr',
      sortBy = 'datetime',
      projection = utm
    ),
    'projection argument is deprecated'
  )
})

#' Dyad id
#'
#' Generate a dyad ID for edge-list generated by `edge_nn` or
#' `edge_dist`.
#'
#' An undirected edge identifier between, for example individuals A and B will
#' be A-B (and reverse B and A will be A-B). Internally sorts and pastes id
#' columns.
#'
#' More details in the edge and dyad vignette (in progress).
#'
#' @param id1 ID1 column name generated by `edge_dist` or `edge_nn`
#' @param id2 ID2 column name generated by `edge_dist` or `edge_nn`
#' @param DT input data.table with columns id1 and id2, as generated by
#'   `edge_dist` or `edge_nn`
#'
#' @return `dyad_id` returns the input `data.table` with appended "dyadID"
#'   column.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge-list generation
#' edges <- edge_dist(
#'     DT,
#'     threshold = 100,
#'     id = 'ID',
#'     coords = c('X', 'Y'),
#'     timegroup = 'timegroup',
#'     returnDist = TRUE,
#'     fillNA = TRUE
#'   )
#'
#' # Generate dyad IDs
#' dyad_id(edges, 'ID1', 'ID2')
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup')
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
dyad_id <- function(DT = NULL, id1 = NULL, id2 = NULL) {
  # due to NSE notes in R CMD check
  ID1 <- ID2 <- dyadID <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(id1)
  assert_not_null(id2)
  assert_are_colnames(DT, c(id1, id2))

  if ('dyadID' %in% colnames(DT)) {
    message('dyadID column will be overwritten by this function')
    data.table::set(DT, j = 'dyadID', value = NULL)
  }

  ids <- unique(stats::na.omit(c(DT[[id1]], DT[[id2]])))
  dyads <- data.table::CJ(ID1 = ids, ID2 = ids)[ID1 != ID2]
  dyads[, dyadID :=
          apply(
            X = .SD,
            MARGIN = 1,
            FUN = function(x)
              paste(sort(x), collapse = '-')
          )]

  data.table::setnames(dyads, c('ID1', 'ID2'), c(id1, id2))

  return(DT[dyads, dyadID := dyadID, on = c(id1, id2)][])
}

# Test dyad_id
context("test dyad_id")

library(spatsoc)

DT <- fread('../testdata/DT.csv')
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = 'datetime', threshold = '20 minutes')
edges <- edge_nn(DT,
                 id = 'ID',
                 coords = c('X', 'Y'),
                 timegroup = 'timegroup')


test_that('DT is required', {
  expect_error(dyad_id(DT = NULL),
  'DT must be provided')
})

test_that('id1 and id2 are required', {
  expect_error(dyad_id(DT = edges, id1 = NULL),
               'id1 must be')

  expect_error(dyad_id(DT = edges, id1 = 'ID', id2 = NULL),
               'id2 must be')
})

test_that('columns must be in DT', {
  expect_error(dyad_id(DT = edges, id1 = 'potato', id2 = 'ID2'),
               'potato field provided is not present in input', fixed = FALSE)

  expect_error(dyad_id(DT = edges, id1 = 'ID', id2 = 'potato'),
               'potato field provided is not present in input', fixed = FALSE)
})


test_that('dyadID column succesfully detected', {
  edges[, dyadID := 1]
  expect_message(
    dyad_id(DT = edges, id1 = 'ID', id2 = 'NN'),
    'dyadID column will be overwritten'
  )
})



test_that('dyadID handles NAs', {
  naedges <- copy(edges)[, NN := NA]
  expect_equal(dyad_id(DT = naedges,
                       id1 = 'ID',
                       id2 = 'NN')[is.na(dyadID), .N],
               naedges[is.na(NN), .N])

  naedges <- copy(edges)[, ID := NA]
  expect_equal(dyad_id(DT = naedges,
                       id1 = 'ID',
                       id2 = 'NN')[is.na(dyadID), .N],
               naedges[is.na(ID), .N])
})


test_that('dyadID handles numeric ids', {
  nDyads <- dyad_id(copy(edges), 'ID', 'NN')[, uniqueN(dyadID)]
  numids <- copy(edges)
  numids[, ID := .GRP, ID]
  numids[, NN := .GRP, NN]

  expect_equal(dyad_id(DT = numids,
                       id1 = 'ID',
                       id2 = 'NN')[, uniqueN(dyadID)],
               nDyads)

})

#' Group polygons
#'
#' `group_polys` groups rows into spatial groups by overlapping polygons (home
#' ranges). The function expects a `data.table` with relocation data, individual
#' identifiers and an `area` argument.  The relocation data is transformed into
#' home range POLYGONs using [build_polys()] with [adehabitatHR::mcp] or
#' [adehabitatHR::kernelUD]. If the `area` argument is `FALSE`, `group_polys`
#' returns grouping calculated by spatial overlap. If the `area` argument is
#' `TRUE`, `group_polys` returns the area area and proportion of overlap.
#' Relocation data should be in two columns representing the X and Y
#' coordinates.
#'
#' ## R-spatial evolution
#'
#' Please note, spatsoc has followed updates from R spatial, GDAL and PROJ for
#' handling coordinate reference systems, see more below and  details at
#' <https://r-spatial.org/r/2020/03/17/wkt.html>.
#'
#' In addition, `group_polys` previously used rgeos::gIntersection,
#' rgeos::gIntersects and rgeos::gArea but has been
#' updated to use [sf::st_intersects], [sf::st_intersection] and [sf::st_area]
#' according to the R-spatial evolution, see more
#' at <https://r-spatial.org/r/2022/04/12/evolution.html>.
#'
#' ## Notes on arguments
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()].
#'
#' The `id`, `coords` (and optional `splitBy`) arguments expect
#' the names of respective columns in `DT` which correspond to the
#' individual identifier, X and Y coordinates, and additional grouping columns.
#'
#' The `crs` argument expects a character string or numeric defining the
#' coordinate reference system to be passed to [sf::st_crs]. For example, for
#' UTM zone 36S (EPSG 32736), the crs argument is `crs = "EPSG:32736"` or `crs =
#' 32736`. See <https://spatialreference.org> for a list of EPSG codes.
#'
#' The `hrType` must be either one of "kernel" or "mcp". The
#' `hrParams` must be a named list of arguments matching those of
#' [adehabitatHR::kernelUD()] or [adehabitatHR::mcp()].
#'
#' The `splitBy` argument offers further control over grouping. If within
#' your `DT`, you have multiple populations, subgroups or other distinct
#' parts, you can provide the name of the column which identifies them to
#' `splitBy`. The grouping performed by `group_polys` will only
#' consider rows within each `splitBy` subgroup.
#'
#' @return When `area` is `FALSE`, `group_polys` returns the input `DT` appended
#'   with a `group` column. As with the other grouping functions,  the actual
#'   value of `group` is arbitrary and represents the identity of a given group
#'   where 1 or more individuals are assigned to a group. If the data was
#'   reordered, the `group` may change, but the contents of each group would
#'   not. When `area` is `TRUE`, `group_polys` returns a proportional area
#'   overlap `data.table`. In this case, ID refers to the focal individual of
#'   which the total area is compared against the overlapping area of ID2.
#'
#'   If `area` is `FALSE`, a message is returned when a column named `group`
#'   already exists in the input `DT`, because it will be overwritten.
#'
#'   Along with changes to follow the R-spatial evolution, `group_polys` also
#'   now returns area and proportion of overlap with units explicitly specified
#'   through the `units` package.
#'
#'  Note: if area is TRUE, the output of `group_polys` needs to be
#'  reassigned. See details in
#'  [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'
#' @inheritParams group_pts
#' @inheritParams group_lines
#' @inheritParams build_polys
#' @param area logical indicating either overlap group (when `FALSE`) or
#'   area and proportion of overlap (when `TRUE`)
#' @param hrType type of HR estimation, either 'mcp' or 'kernel'
#' @param hrParams a named list of parameters for `adehabitatHR` functions
#' @param sfPolys Alternatively, provide solely a simple features object with
#' POLYGONs or MULTIPOLYGONs. If sfPolys are provided, id is required and
#' splitBy cannot be used.
#'
#' @export
#'
#' @family Spatial grouping
#' @seealso [build_polys()] [group_times()]
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # EPSG code for example data
#' utm <- 32736
#'
#' group_polys(DT, area = FALSE, hrType = 'mcp',
#'             hrParams = list(percent = 95), crs = utm,
#'             id = 'ID', coords = c('X', 'Y'))
#'
#' areaDT <- group_polys(DT, area = TRUE, hrType = 'mcp',
#'                       hrParams = list(percent = 95), crs = utm,
#'                       id = 'ID', coords = c('X', 'Y'))
#' print(areaDT)
group_polys <- function(
    DT = NULL,
    area = NULL,
    hrType = NULL,
    hrParams = NULL,
    crs = NULL,
    id = NULL,
    coords = NULL,
    splitBy = NULL,
    sfPolys = NULL,
    projection = NULL) {
  # due to NSE notes in R CMD check
  nBy <- ..coords <- ..id <- withinGroup <- group <- outGroup <- NULL

  if (!is.null(projection)) {
    warning('projection argument is deprecated, setting crs = projection')
    crs <- projection
  }

  assert_not_null(area)
  assert_inherits(area, 'logical')

  if (is.null(DT) && is.null(sfPolys)) {
    stop('must provide either DT or sfPolys')
  } else if (!is.null(DT) && !is.null(sfPolys)) {
    stop('cannot provide both DT and sfPolys')
  }

  if (is.null(splitBy)) {
    if (is.null(DT) && !is.null(sfPolys)) {
      input <- 'sfPolys'
    } else if (!is.null(DT) && is.null(sfPolys)) {
      sfPolys <-
        build_polys(
          DT = DT,
          crs = crs,
          hrType = hrType,
          hrParams = hrParams,
          coords = coords,
          id = id,
          splitBy = NULL,
          spPts = NULL
        )
      input <- 'DT'
    }

    assert_not_null(id)

    if (!area) {
      if ('group' %in% colnames(DT)) {
        message('group column will be overwritten by this function')
        data.table::set(DT, j = 'group', value = NULL)
      }
      inter <- sf::st_intersects(sfPolys, sfPolys, sparse = FALSE)
      dimnames(inter) <- list(sfPolys[[id]], sfPolys[[id]])
      g <- igraph::graph_from_adjacency_matrix(inter)
      ovr <- igraph::components(g)$membership
      out <- data.table::data.table(names(ovr),
                                    as.integer(unlist(ovr)))
      data.table::setnames(out, c(id, 'group'))
      if (input == 'DT') {
        DT[out, group := group, on = c(id)]
        return(DT)
      } else if (input == 'sfPolys'){
        return(out)
      }
    } else if (area) {
      if (any(grepl(' ', sfPolys[[id]]))) {
        stop('please ensure IDs do not contain spaces')
      }
      if (! 'area' %in% colnames(sfPolys)) {
        stop('please ensure column "area" present in input DT or sfPolys')
      }
      sf::st_agr(sfPolys) <- 'constant'
      inter <- sf::st_intersection(sfPolys, sfPolys)
      areas <- sf::st_area(inter)
      out_inter <- data.table::data.table(
        ID1 = inter[[id]],
        ID2 = inter[[paste0(id, '.1')]],
        area = areas,
        area_ID1 = units::as_units(inter[['area']], units(areas),
                                   set_units_mode = 'standard')
      )
      data.table::set(out_inter, j = 'proportion',
          value = units::set_units(
            out_inter[['area']] / out_inter[['area_ID1']],
            'percent')
      )
      data.table::set(out_inter, j = 'area_ID1',  value = NULL)

      disjointed <- data.frame(sf::st_disjoint(sfPolys))
      out_disjointed <- data.frame(
        ID1 = sfPolys[[id]][disjointed$row.id],
        ID2 = sfPolys[[id]][disjointed$col.id],
        area = rep(units::as_units(0, units(out_inter$area)),
                   nrow(disjointed)),
        proportion = rep(units::set_units(0, 'percent'), nrow(disjointed))
      )
      out <- rbind(out_inter, out_disjointed)

      data.table::setcolorder(out, c('ID1', 'ID2', 'area', 'proportion'))
      return(out[])
    }
  } else if (!is.null(splitBy)) {
    if (!is.null(sfPolys)) {
      stop('cannot provide sfPolys if providing splitBy')
    }
    assert_are_colnames(DT, c(id, splitBy))

    DT[, nBy := .N, c(splitBy, id)]

    if (!area) {
      if ('group' %in% colnames(DT)) {
        message('group column will be overwritten by this function')
        data.table::set(DT, j = 'group', value = NULL)
      }
      ovrDT <-
        DT[nBy > 5, {
          try(
            sfPolys <-
              build_polys(
                DT = .SD,
                crs = crs,
                hrType = hrType,
                hrParams = hrParams,
                coords = ..coords,
                id = ..id,
                splitBy = NULL,
                spPts = NULL
              ),
            silent = TRUE
          )
          if (!is.null(sfPolys)) {
            inter <- sf::st_intersects(sfPolys, sfPolys, sparse = FALSE)
            dimnames(inter) <- list(sfPolys[[..id]], sfPolys[[..id]])
            g <- igraph::graph_from_adjacency_matrix(inter)
            ovr <- igraph::components(g)$membership
            out <- data.table::data.table(names(ovr),
                                          as.integer(unlist(ovr)))
            data.table::setnames(out, c(id, 'withinGroup'))
            out
          } else {
            data.table(ID = get(..id),
                       withinGroup = NA_integer_)
          }
        }, by = c(splitBy), .SDcols = c(coords, id)]
      DT[ovrDT, withinGroup := withinGroup, on = c(id, splitBy)]
      DT[, group := ifelse(is.na(withinGroup), NA_integer_, .GRP),
         by = c(splitBy, 'withinGroup')]
      data.table::set(DT, j = c('withinGroup', 'nBy'), value = NULL)
      if (DT[is.na(group), .N] > 0) {
        warning(
          strwrap(
            prefix = " ",
            initial = "",
            x = 'build_polys failed for some rows,
            check `DT[is.na(group)]` and choice of hrParams'
          )
        )
      }
      return(DT[])
    } else if (area) {
      if (any(DT[, grepl(' ', .SD[[1]]), .SDcols = id])) {
        stop('please ensure IDs do not contain spaces')
      }
      outDT <-
        DT[nBy > 5, {
          try(
            sfPolys <-
              build_polys(
                DT = .SD,
                crs = crs,
                hrType = hrType,
                hrParams = hrParams,
                id = ..id,
                coords = ..coords,
                splitBy = NULL,
                spPts = NULL
              ),
            silent = TRUE
          )
          if (!is.null(sfPolys)) {
            sf::st_agr(sfPolys) <- 'constant'
            inter <- sf::st_intersection(sfPolys, sfPolys)
            areas <- sf::st_area(inter)
            out_inter <- data.table::data.table(
              ID1 = inter[[id]],
              ID2 = inter[[paste0(id, '.1')]],
              area = areas,
              area_ID1 = units::as_units(inter[['area']], units(areas),
                                         set_units_mode = 'standard')
            )
            data.table::set(out_inter, j = 'proportion',
                value = units::set_units(
                  out_inter[['area']] / out_inter[['area_ID1']],
                  'percent')
            )
            data.table::set(out_inter, j = 'area_ID1',  value = NULL)
            data.table::set(out_inter, j = 'proportion',
                i = which(unclass(out_inter$proportion) > 100),
                value = units::set_units(100, 'percent'))

            disjointed <- data.frame(sf::st_disjoint(sfPolys))
            out_disjointed <- data.frame(
              ID1 = sfPolys[[id]][disjointed$row.id],
              ID2 = sfPolys[[id]][disjointed$col.id],
              area = rep(
                units::as_units(0, units(out_inter$area)),
                nrow(disjointed)
              ),
              proportion = rep(
                units::set_units(0, 'percent'),
                nrow(disjointed)
              )
            )
            out <- rbind(out_inter, out_disjointed)

            data.table::setcolorder(out,
                                    c('ID1', 'ID2', 'area', 'proportion'))
            out
          } else {
            out <- data.table(ID = get(..id),
                              ID2 = NA_character_,
                              NA_real_,
                              NA_real_)
            data.table::setnames(out, c(..id, paste0(..id, '2'),
                            'area', 'proportion'))
            out
          }
        }, by = c(splitBy), .SDcols = c(coords, id)]
      dropped <-
        unique(DT[nBy <= 5, .SD, .SDcols = c(splitBy, id)])
      out <- data.table::rbindlist(list(dropped, outDT), fill = TRUE)
      if (out[is.na(area), .N] > 0) {
        warning(
          strwrap(
            prefix = " ",
            initial = "",
            x = 'build_polys failed for some rows,
            check `DT[is.na(group)]` and choice of hrParams'
          )
        )
      }
      return(out)
    }
  }
}

# Test group_polys
context('test group_polys')

library(spatsoc)

DT <- fread('../testdata/DT.csv')

DT[, datetime := as.POSIXct(datetime)]
DT[, family := sample(c(1, 2, 3, 4), .N, replace = TRUE)]
DT[, jul := data.table::yday(datetime)]

utm <- 32736

test_that('DT or spPts are required but not both', {
  expect_error(group_polys(
    DT = NULL,
    sfPolys = NULL,
    area = FALSE
  ),
  'must provide either DT or sfPolys')

  expect_error(group_polys(
    DT = DT,
    sfPolys = 10,
    area = FALSE
  ),
  'cannot provide both DT and sfPolys')
})

test_that('id required', {
  expect_error(group_polys(DT, area = TRUE, id = NULL), 'id')
})

test_that('area provided and logical, or error', {
  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = NULL,
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'area must be provided',
    fixed = TRUE
  )

  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = 'potato',
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'area must be of class logical',
    fixed = TRUE
  )

  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = NULL,
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'area must be provided',
    fixed = TRUE
  )
})

test_that('id provided', {
  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = TRUE,
      coords = c('X', 'Y'),
      id = NULL
    ),
    'id must be provided',
    fixed = TRUE
  )
})

test_that('crs provided or error', {
  expect_error(
    group_polys(
      DT = DT,
      crs = NULL,
      hrType = 'mcp',
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'crs must be provided'
  )
})

test_that('mising hrParams warns default used', {
  copyDT <- copy(DT)
  missingpromise <- evaluate_promise(
    group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID'
    )
  )
  expect_match(
    missingpromise$messages,
    'hrParams is not provided, using defaults'
  )
})

test_that('missing hrType fails', {
  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = NULL,
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'hrType must be provided'
  )
})

test_that('column names must exist in DT', {
  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = FALSE,
      coords = c('potatoX', 'potatoY'),
      id = 'ID'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    group_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )
})

test_that('splitBy and sfPolys are not both provided', {
  expect_error(
    group_polys(
      splitBy = 'yr',
      sfPolys = 10,
      area = TRUE
    ),
    'cannot provide sfPolys if providing splitBy'
  )
})


test_that('ID field does not have spaces', {
  copyDT <- copy(DT)[ID == unique(ID)[1], ID := paste(ID, ID)]
  idpromise <- evaluate_promise(
    expect_error(group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID'
    ))
  )
  expect_match(
    as.character(idpromise$result),
    'please ensure IDs do not contain spaces'
  )

  copyDT[, population := 1]
  expect_error(
    group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'population'
    ),
    'please ensure IDs do not contain spaces'
  )
})

test_that('column and row lengths returned make sense', {
  copyDT <- copy(DT)
  group_polys_mcp <- group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID'
  )
  expect_lte(nrow(group_polys_mcp),
             nrow(expand.grid(DT[, unique(ID)], DT[, unique(ID)])))

  copyDT <- copy(DT)[, yr := year(datetime)]

  group_polys_mcp_split <- group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
    )

  expect_equal(nrow(group_polys_mcp_split), nrow(copyDT))

  group_sf_polys <- group_polys(
      id = 'ID',
      area = FALSE,
      sfPolys = build_polys(
        DT = DT,
        crs = utm,
        hrType = 'mcp',
        hrParams = list(percent = 95),
        coords = c('X', 'Y'),
        id = 'ID'
    )
  )
  group_sf_polys_area <- group_polys(
    id = 'ID',
    area = TRUE,
    sfPolys = build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      coords = c('X', 'Y'),
      id = 'ID'
    )
  )
  expect_length(unique(DT$ID), nrow(group_sf_polys))
  expect_equal(nrow(group_sf_polys_area), length(unique(DT$ID)) ^ 2)
})


test_that('withinGroup is not returned to the user', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '14 days')
  )
  copyDT[, N := .N, by = .(ID, block)]
  within_group <- group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID'
  )
  expect_false('withinGroup' %in% colnames(within_group))
})


test_that('group column succesfully detected', {
  copyDT <- copy(DT)[, group := 1][, mnth := month(datetime)]
  copyDT <- copyDT[, nBy := .N, by = .(mnth, ID)][nBy > 30]
  groupdelpromise <- evaluate_promise(
    group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'mnth'
    )
  )
  expect_match(groupdelpromise$messages, 'group column will be overwritten')

  copyDT <- copy(DT)[, group := 1][, mnth := month(datetime)]
  copyDT <- copyDT[, nBy := .N, by = .(mnth, ID)][nBy > 30]
  overwritepromise <- evaluate_promise(
    group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID'
    )
  )
  expect_match(
    overwritepromise$messages,
    'group column will be overwritten'
  )
})


test_that('area provided with splitBy does not return errors', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  copyDT[, yr := year(datetime)]
  within_split <- group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
  )
  expect_false('withinGroup' %in% colnames(within_split))

  area <- group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
  )
  expect_true('area' %in% colnames(area))

  prop <- group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
  )
  expect_true('proportion' %in% colnames(prop))
})


test_that('less than 5 locs returns NAs and warning', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  copyDT[, yr := year(datetime)]
  copyDT[1, yr := 2020]
  expect_warning(
    group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
    ),
    'build_polys failed for some rows', fixed = FALSE
  )

  expect_warning(
    group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
    ),
    'build_polys failed for some rows', fixed = FALSE
  )

  expect_true(
    suppressWarnings(group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
    )[is.na(group), .N != 0]
  ))

  expect_true(
    suppressWarnings(group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'yr'
    )[is.na(area), .N != 0]
  ))


})



test_that('splitBy argument doesnt use splitBy column', {
  copyDT <- copy(DT)

  copyDT[, splitBy := family]

  utm <- 32736

  expect_true(
    suppressWarnings(group_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = FALSE,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'jul'
    ))[, uniqueN(jul), group][, all(V1 == 1)]
  )

  expect_false('splitBy' %in%
      suppressWarnings(group_polys(
        DT = copyDT,
        crs = utm,
        hrType = 'mcp',
        hrParams = list(percent = 95),
        area = TRUE,
        coords = c('X', 'Y'),
        id = 'ID',
        splitBy = 'jul'
      )))

})


test_that('proportion within 0-100, area > 0', {

  out_mcp <- suppressWarnings(group_polys(
    DT = DT,
    crs = utm,
    hrType = 'mcp',
    hrParams = list(percent = 95),
    area = TRUE,
    coords = c('X', 'Y'),
    id = 'ID'
  ))

  expect_gte(min(out_mcp$proportion),
             units::as_units(0, 'percent'))
  expect_lte(max(out_mcp$proportion),
             units::as_units(100.01, 'percent'))
  expect_gte(min(out_mcp$area),
             units::as_units(0, 'm^2'))


  out_kernel <- suppressWarnings(group_polys(
    DT = DT,
    crs = utm,
    hrType = 'kernel',
    hrParams = list(percent = 95),
    area = TRUE,
    coords = c('X', 'Y'),
    id = 'ID'
  ))

  expect_gte(min(out_kernel$proportion),
             units::as_units(0, 'percent'))
  expect_lte(max(out_kernel$proportion),
             units::as_units(100.01, 'percent'))
  expect_gte(min(out_kernel$area),
             units::as_units(0, 'm^2'))

})


test_that('sfPolys has area column', {
  sfPolys <- build_polys(DT, crs = utm, hrType = 'kernel',
                         hrParams = list(grid = 60, percent = 95),
                         id = 'ID', coords = c('X', 'Y'))

  sfPolys$area <- NULL

  expect_error(
    group_polys(sfPolys = sfPolys,
                area = TRUE,
                id = 'id'),
    'please ensure column "area"',
    fixed = TRUE
  )
})


test_that('projection arg is deprecated', {
  expect_warning(
    group_polys(
      DT = DT,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      area = TRUE,
      coords = c('X', 'Y'),
      id = 'ID',
      projection = utm
    ),
    'projection argument is deprecated'
  )
})

#' Build polygons
#'
#' `build_polys` generates a simple feature collection with POLYGONs from a
#' `data.table`. The function expects a `data.table` with
#' relocation data, individual identifiers, a crs,
#' home range type and parameters. The relocation
#' data is transformed into POLYGONs using either [adehabitatHR::mcp] or
#' [adehabitatHR::kernelUD] for each individual and, optionally,
#' combination of columns listed in `splitBy`. Relocation data should be in two
#' columns representing the X and Y coordinates.
#'
#' [group_polys] uses `build_polys` for grouping overlapping
#' polygons created from relocations.
#'
#' ## R-spatial evolution
#'
#' Please note, spatsoc has followed updates from R spatial, GDAL and PROJ for
#' handling coordinate reference systems, see more below and  details at
#' <https://r-spatial.org/r/2020/03/17/wkt.html>.
#'
#' In addition, `build_polys` previously used [sp::SpatialPoints] but has been
#' updated to use [sf::st_as_sf] according to the R-spatial evolution, see more
#' at <https://r-spatial.org/r/2022/04/12/evolution.html>.
#'
#' ## Notes on arguments
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT].
#'
#' The `id`, `coords` (and optional `splitBy`) arguments
#' expect the names of respective columns in `DT` which correspond
#' to the individual identifier, X and Y coordinates, and additional
#' grouping columns.
#'
#' The `crs` argument expects a character string or numeric
#' defining the coordinate reference system to be passed to [sf::st_crs].
#' For example, for UTM zone 36S (EPSG 32736), the crs
#' argument is `crs = "EPSG:32736"` or `crs = 32736`.
#' See <https://spatialreference.org>
#' for a list of EPSG codes.
#'
#' The `hrType` must be either one of "kernel" or "mcp". The
#' `hrParams` must be a named list of arguments matching those
#' of [adehabitatHR::kernelUD] and [adehabitatHR::getverticeshr]
#' or [adehabitatHR::mcp].
#'
#' The `splitBy` argument offers further control building
#' POLYGONs. If in your `DT`, you have multiple
#' temporal groups (e.g.: years) for example, you can provide the
#' name of the column which identifies them and build POLYGONs
#' for each individual in each year.
#'
#'
#' @return `build_polys` returns a simple feature collection with POLYGONs
#' for each individual (and optionally `splitBy` combination).
#'
#' An error is returned when `hrParams` do not match the arguments
#' of the respective `hrType` `adehabitatHR` function.
#'
#'
#' @inheritParams build_lines
#' @inheritParams group_polys
#' @param spPts alternatively, provide solely a SpatialPointsDataFrame with one
#' column representing the ID of each point, as specified by [adehabitatHR::mcp]
#' or [adehabitatHR::kernelUD]
#' @param crs numeric or character defining the coordinate reference
#'   system to be passed to [sf::st_crs]. For example, either
#'   `crs = "EPSG:32736"` or `crs = 32736`.
#' @export
#'
#' @family Build functions
#' @seealso [group_polys]
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # EPSG code for example data
#' utm <- 32736
#'
#' # Build polygons for each individual using kernelUD and getverticeshr
#' build_polys(DT, crs = utm, hrType = 'kernel',
#'             hrParams = list(grid = 60, percent = 95),
#'             id = 'ID', coords = c('X', 'Y'))
#'
#' # Build polygons for each individual by year
#' DT[, yr := year(datetime)]
#' build_polys(DT, crs = utm, hrType = 'mcp',
#'             hrParams = list(percent = 95),
#'             id = 'ID', coords = c('X', 'Y'), splitBy = 'yr')
build_polys <- function(
    DT = NULL,
    crs = NULL,
    hrType = NULL,
    hrParams = NULL,
    id = NULL,
    coords = NULL,
    splitBy = NULL,
    spPts = NULL,
    projection = NULL) {
  # due to NSE notes in R CMD check
  . <- NULL

  if (!is.null(projection)) {
    warning('projection argument is deprecated, setting crs = projection')
    crs <- projection
  }

  if (is.null(DT) && is.null(spPts)) {
    stop('input DT or spPts required')
  }

  if (!is.null(DT) && !is.null(spPts)) {
    stop('cannot provide both DT and spPts')
  }

  if (!is.null(DT) && is.null(spPts)) {
    assert_not_null(coords)
    assert_not_null(id)
    assert_not_null(crs)
    assert_length(coords, 2)

    if (is.null(splitBy)) {
      splitBy <- id
    } else {
      splitBy <- c(id, splitBy)
    }

    assert_are_colnames(DT, c(splitBy, coords))
  }

  assert_col_inherits(DT, coords, 'numeric')
  assert_col_inherits(DT, splitBy, c('numeric', 'character', 'integer'))
  assert_not_null(hrType)

  if (is.null(hrParams)) {
    message('hrParams is not provided, using defaults')
  }

  if (is.null(spPts)) {
    ade_id <- DT[, do.call(function(...) paste(..., sep = '-'), .SD),
                           .SDcols = c(splitBy)]

    spPts <- sf::as_Spatial(
      sf::st_as_sf(
        DT[, cbind(.SD, ade_id), .SDcols = c(coords)],
        coords = coords,
        crs = sf::st_crs(crs)
      ),
      IDs = ade_id
    )
  }

  hrParams$xy <- spPts

  if (hrType == 'mcp') {
    functionParams <- formals(adehabitatHR::mcp)
    if (all(names(hrParams) %in% names(functionParams))) {
      if (!('unout' %in% names(hrParams))) {
        hrParams$unout <- 'm2'
      }
      out <- do.call(adehabitatHR::mcp, hrParams)
    } else {
      stop(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'hrParams provided do not match function parameters,
               see ?adehabitatHR::mcp'
        )
      )
    }
  } else if (hrType == 'kernel') {
    kernelParam <- formals(adehabitatHR::kernelUD)
    verticesParam <- formals(adehabitatHR::getverticeshr.estUD)

    if (all(names(hrParams) %in% c(names(kernelParam), names(verticesParam)))) {
      if (!('unout' %in% names(hrParams))) {
        hrParams$unout <- 'm2'
      }
      kern <- do.call(adehabitatHR::kernelUD,
                      hrParams[intersect(names(hrParams), names(kernelParam))])
      out <- do.call(adehabitatHR::getverticeshr,
                     c(x = list(kern),
                       hrParams[intersect(names(hrParams),
                                          names(verticesParam))]))
    } else {
      stop(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'hrParams provided do not match
          function parameters, see ?adehabitatHR::kernelUD
          and ?adehabitatHR::getverticeshr'
        )
      )
    }
  } else {
    stop('hrType not one of "kernel" or "mcp"')
  }

  out_sf <- sf::st_as_sf(out)
  colnames(out_sf) <- gsub('id', id, colnames(out_sf))
  return(out_sf)
}

# Test build_polys
context('test build_polys')

DT <- fread('../testdata/DT.csv')

utm <- 'EPSG:32736'


test_that('DT or spPts are required but not both', {
  expect_error(build_polys(DT = NULL, spPts = NULL),
               'input DT or spPts required')

  expect_error(build_polys(DT = DT, spPts = 10),
               'cannot provide both DT and spPts')
})



test_that('coords, id, crs, (splitBy) provided and proper format', {
  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      coords = c('X', 'Y'),
      id = NULL
    ),
    'id must be provided'
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = NULL,
      hrType = 'mcp',
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'crs must be provided'
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      coords = NULL,
      id = 'ID'
    ),
    'coords must be provided'
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      coords = 'X',
      id = 'ID'
    ),
    'coords must be length 2',
    fixed = FALSE
  )

  copyDT <- copy(DT)
  copyDT[, X := as.character(X)]
  expect_error(
    build_polys(
      DT = copyDT,
      crs = utm,
      hrType = 'mcp',
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'coords must be of class numeric'
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = NULL,
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'hrType must be provided'
  )

  copyDT <- copy(DT)
  copyDT[, s := factor(1)]
  expect_error(
    build_polys(
      DT = copyDT,
      crs = utm,
      hrType = NULL,
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 's'
    ),
    'splitBy must be', fixed = FALSE
  )
})


test_that('column names must exist in DT', {
  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      coords = c('X', 'Y'),
      id = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      coords = c('potatoX', 'potatoY'),
      id = 'ID'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      coords = c('X', 'Y'),
      id = 'ID',
      splitBy = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )
})

test_that('hrType either mcp or kernel', {
  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'potato',
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'hrType not one of',
    fixed = FALSE
  )
})


test_that('hrParams returns error if params do not match function params', {
  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95, potato = TRUE),
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'hrParams provided do not match function parameters',
    fixed = FALSE
  )

  expect_error(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'kernel',
      hrParams = list(grid = 60, potato = TRUE),
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'hrParams provided do not match function parameters',
    fixed = FALSE
  )

})

test_that('if hrParams NULL, warns', {
  paramspromise <- evaluate_promise(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      hrParams = NULL,
      coords = c('X', 'Y'),
      id = 'ID'
    )
  )

  expect_match(
    paramspromise$messages,
    'hrParams is not provided, using defaults'
  )

})


test_that('build_polys returns sf, POLYGONs', {
  expect_s3_class(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'sf'
  )

  expect_s3_class(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'kernel',
      hrParams = list(grid = 60),
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'sf'
  )

  expect_true(
    any(c('POLYGON', 'MULTIPOLYGON') %in%
    sf::st_geometry_type(
      build_polys(
        DT = DT,
        crs = utm,
        hrType = 'mcp',
        hrParams = list(percent = 95),
        coords = c('X', 'Y'),
        id = 'ID'
      ),
      by_geometry = FALSE)
    )
  )


  expect_true(
    any(c('POLYGON', 'MULTIPOLYGON') %in%
    sf::st_geometry_type(
        build_polys(
          DT = DT,
          crs = utm,
          hrType = 'kernel',
          hrParams = list(grid = 60),
          coords = c('X', 'Y'),
          id = 'ID'),
        by_geometry = FALSE
      )
    )
  )
})


test_that('hrParams can have both vertices and kernel args', {
  expect_s3_class(
    build_polys(
      DT = DT,
      crs = utm,
      hrType = 'kernel',
      hrParams = list(percent = 95, grid = 60),
      coords = c('X', 'Y'),
      id = 'ID'
    ),
    'sf'
  )

  expect_true(
    any(c('POLYGON', 'MULTIPOLYGON') %in%
      sf::st_geometry_type(
        build_polys(
          DT = DT,
          crs = utm,
          hrType = 'kernel',
          hrParams = list(percent = 95, grid = 60),
          coords = c('X', 'Y'),
          id = 'ID'),
        by_geometry = FALSE
      )
    )
  )
})


test_that('id provided matches id retured', {
  expect_in(
    'ID',
    colnames(build_polys(
      DT = DT,
      crs = utm,
      hrType = 'kernel',
      hrParams = list(percent = 95, grid = 60),
      coords = c('X', 'Y'),
      id = 'ID'
    ))
  )
})


test_that('projection arg is deprecated', {
  expect_warning(
    build_polys(
      DT = DT,
      hrType = 'mcp',
      hrParams = list(percent = 95),
      coords = c('X', 'Y'),
      id = 'ID',
      projection = utm
    ),
    'projection argument is deprecated'
  )
})

#' Group points
#'
#' `group_pts` groups rows into spatial groups. The function expects a
#' `data.table` with relocation data, individual identifiers and a
#' threshold argument. The threshold argument is used to specify the criteria
#' for distance between points which defines a group. Relocation data should be
#' in two columns representing the X and Y coordinates, or in a
#' geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `id`, `timegroup`  (and optional `splitBy`)
#' arguments expect the names of columns in `DT` which correspond to the
#' individual identifier, and timegroup (generated by
#' `group_times`) and additional grouping columns.
#'
#' The `threshold` provided should match the units of the coordinates. The
#' `threshold` can be provided with units specified using the units package (eg.
#' `threshold = units::set_units(10, m)`) which will be checked against the
#' units of the coordinates using the `crs`. If units are not specified, the
#' `threshold` is assumed to be in the units of the coordinates.
#'
#' The `timegroup` argument is required to define the temporal groups
#' within which spatial groups are calculated. The intended framework is to
#' group rows temporally with `group_times` then spatially with
#' `group_pts` (or `group_lines`, `group_polys`).
#' If you have already calculated temporal groups without
#' `group_times`, you can pass this column to the `timegroup`
#' argument. Note that the expectation is that each individual will be observed
#' only once per timegroup. Caution that accidentally including huge numbers of
#' rows within timegroups can overload your machine since all pairwise distances
#' are calculated within each timegroup.
#'
#' The `splitBy` argument offers further control over grouping. If within
#' your `DT`, you have multiple populations, subgroups or other distinct
#' parts, you can provide the name of the column which identifies them to
#' `splitBy`. The grouping performed by `group_pts` will only consider
#' rows within each `splitBy` subgroup.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Distance function" for details on underlying distance function used.
#'
#' @inheritSection distance_to_centroid Interface
#' @inheritSection distance_to_centroid Distance function
#'
#' @return `group_pts` returns the input `DT` appended with a
#'   `group` column.
#'
#'   This column represents the spatialtemporal group. As with the other
#'   grouping functions,  the actual value of `group` is arbitrary and
#'   represents the identity of a given group where 1 or more individuals are
#'   assigned to a group. If the data was reordered, the `group` may
#'   change, but the contents of each group would not.
#'
#'   A message is returned when a column named `group` already exists in
#'   the input `DT`, because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @param DT input data.table
#' @param threshold distance for grouping points, either numeric or units, in
#'   the units of the crs / the coordinates or geometry
#' @param id character string of ID column name
#' @param coords character vector of X coordinate and Y coordinate column names.
#' Note: the order is assumed X followed by Y column names
#' @param timegroup timegroup field in the DT within which the output will be
#'   calculated
#' @param crs numeric or character defining the coordinate reference
#'   system to be passed to [sf::st_crs]. For example, either
#'   `crs = "EPSG:32736"` or `crs = 32736`. Used only if coords are provided,
#'   see details under Interface
#' @param splitBy (optional) character string or vector of grouping column
#'   name(s) upon which the output will be calculated
#' @param geometry simple feature geometry list column name, generated by
#'   [get_geometry()]. Default 'geometry', see details under Interface
#'
#' @export
#'
#' @family Spatial grouping
#' @seealso `group_times`
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Select only individuals A, B, C for this example
#' DT <- DT[ID %in% c('A', 'B', 'C')]
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 5, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Spatial grouping with timegroup and splitBy on population
#' group_pts(DT, threshold = 5, id = 'ID', coords = c('X', 'Y'),
#'          timegroup = 'timegroup', splitBy = 'population')
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 50, id = 'ID', timegroup = 'timegroup')
group_pts <- function(
    DT = NULL,
    threshold = NULL,
    id = NULL,
    coords = NULL,
    timegroup,
    crs = NULL,
    splitBy = NULL,
    geometry = 'geometry') {
  # due to NSE notes in R CMD check
  N <- withinGroup <- group <- geo <- x <- y <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(id)
  assert_not_null(threshold)
  assert_not_missing(timegroup)
  check_colnames <- c(timegroup, id, splitBy)
  assert_are_colnames(DT, check_colnames)

  if (!is.null(timegroup)) {
    if (any(unlist(lapply(DT[, .SD, .SDcols = timegroup], class)) %in%
            c('POSIXct', 'POSIXlt', 'Date', 'IDate', 'ITime', 'character'))) {
      warning(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'timegroup provided is a date/time
          or character type, did you use group_times?'
        )
      )
    }
  }

  if (DT[, .N, by = c(id, splitBy, timegroup)][N > 1, sum(N)] != 0) {
    warning(
      strwrap(
        prefix = " ",
        initial = "",
        x = 'found duplicate id in a
          timegroup and/or splitBy -
          does your group_times threshold match the fix rate?'
      )
    )
  }

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    crs <- sf::st_crs(DT[[geometry]])
    assert_threshold(threshold, crs)

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    if (!inherits(threshold, 'units') && !identical(crs, sf::NA_crs_) &&
        !use_dist) {
      threshold <- units::as_units(threshold, 'm')
    }

    DT[!sf::st_is_empty(geo), withinGroup := {
      distMatrix <- calc_distance(
        geometry_a = geo,
        use_dist = use_dist
      )
      graphAdj <-
        igraph::graph_from_adjacency_matrix(distMatrix <= threshold)
      igraph::components(graphAdj)$membership
    },
    by = c(splitBy, timegroup),
    env = list(geo = geometry)]

    if ('group' %in% colnames(DT)) {
      message('group column will be overwritten by this function')
      data.table::set(DT, j = 'group', value = NULL)
    }

    DT[!sf::st_is_empty(geo), group := .GRP,
      by = c(splitBy, timegroup, 'withinGroup'),
      env = list(geo = geometry)]
  } else {
    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    assert_threshold(threshold, crs)
    assert_length(coords, 2)
    assert_are_colnames(DT, coords)
    assert_col_inherits(DT, coords, 'numeric')

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)

    if ('group' %in% colnames(DT)) {
      message('group column will be overwritten by this function')
      data.table::set(DT, j = 'group', value = NULL)
    }

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    if (!inherits(threshold, 'units') && !identical(crs, sf::NA_crs_) &&
        !use_dist) {
      threshold <- units::as_units(threshold, 'm')
    }

    DT[!is.na(x) & !is.na(y),
      withinGroup := {
        distMatrix <- calc_distance(
          x_a = x, y_a = y, crs = crs,
          use_dist = use_dist
        )
        graphAdj <-
          igraph::graph_from_adjacency_matrix(distMatrix <= threshold)
        igraph::components(graphAdj)$membership
      },
      by = c(splitBy, timegroup),
      env = list(x = xcol, y = ycol)
    ]

    DT[!is.na(x) & !is.na(y),
      group := .GRP,
      by = c(splitBy, timegroup, 'withinGroup'),
      env = list(x = xcol, y = ycol)
    ]
  }

  data.table::set(DT, j = 'withinGroup', value = NULL)

  return(DT[])
}

# Test direction_to_centroid
context('test direction_to_centroid')

library(spatsoc)
library(units)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, timethreshold)
group_pts(DT, threshold = threshold, id = id,
          coords = coords, timegroup = timegroup)

DT_without_centroid <- copy(DT)

centroid_group(DT, coords = coords, group = group)

clean_DT <- copy(DT)

test_that('DT is required', {
  expect_error(direction_to_centroid(DT = NULL))
})

test_that('column names must exist in DT', {
  expect_error(direction_to_centroid(DT, coords = rep('potato', 2),
                                     crs = utm),
               'potato field')
  copy_DT <- copy(clean_DT)
  setnames(copy_DT, 'centroid_X', 'potato_X')
  expect_error(direction_to_centroid(copy_DT, coords = coords,
                                     crs = utm),
               'did you run centroid_group')
})

test_that('coords are correctly provided or error detected', {
  expect_error(direction_to_centroid(DT, coords = c('X', NULL), crs = utm),
               'coords must be length 2')
  expect_error(direction_to_centroid(DT, coords = c('X', 'ID'), crs = utm),
               'coords must be of class numeric')
  copy_DT <- copy(clean_DT)[, X := as.character(X)]
  expect_error(direction_to_centroid(copy_DT, coords = coords, crs = utm),
               'coords must be of class numeric')
  copy_DT <- copy(clean_DT)[, centroid_X := as.character(centroid_X)]
  expect_error(direction_to_centroid(copy_DT, coords = coords, crs = utm),
               'coords_centroid must be of class numeric')
})

test_that('direction_centroid column succesfully detected', {
  copy_DT <- copy(clean_DT)[, direction_centroid := 1]
  expect_message(
    direction_to_centroid(copy_DT, coords = coords, crs = utm),
    'direction_centroid column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  copy_DT <- copy(clean_DT)

  expect_equal(nrow(copy_DT),
               nrow(direction_to_centroid(copy_DT, coords = coords, crs = utm)))
})

test_that('one column added to the result DT', {
  copy_DT <- copy(clean_DT)

  expect_equal(ncol(copy_DT) + 1,
               ncol(direction_to_centroid(copy_DT, coords = coords, crs = utm)))
})

test_that('column added to the result DT is a double with units rad', {
  expect_type(direction_to_centroid(DT, coords = coords, crs = utm)$direction_centroid,
              'double')
  expect_equal(
    units(
      direction_to_centroid(DT, coords = coords, crs = utm)$direction_centroid)$numerator,
    'rad')
})

test_that('returns a data.table', {
  expect_s3_class(direction_to_centroid(DT, coords = coords, crs = utm), 'data.table')
})

test_that('use_transform errors if crs not provided', {
  expect_error(
    direction_to_centroid(DT, coords = coords, crs = NA),
    'ensure crs is provided'
  )

  copyDT <- copy(DT)
  get_geometry(copyDT, coords = coords, crs = utm)
  st_crs(copyDT$geometry) <- NA
  get_geometry(copyDT, coords = paste0('centroid_', coords), crs = utm,
               geometry_colname = 'centroid')

  expect_error(
    direction_to_centroid(
      DT = copyDT
    ),
    'ensure crs is provided'
  )
})


# geometry interface
test_that('if coords null, geometry required', {
  expect_error(direction_to_centroid(DT, crs = utm),
               'get_geometry?')

  get_geometry(DT, coords = coords, crs = utm)
  expect_error(direction_to_centroid(DT, crs = utm),
               'centroid_group?')

  copy_DT <- copy(DT)
  copy_DT[, centroid := 42]
  expect_error(direction_to_centroid(copy_DT, crs = utm),
               'sfc_POINT')

})

test_that('message if overwrite', {
  copy_DT <- copy(clean_DT)[, direction_centroid := 1]
  get_geometry(copy_DT, coords = coords, crs = utm)
  centroid_group(copy_DT)
  expect_message(
    direction_to_centroid(copy_DT),
    'direction_centroid column will be overwritten'
  )
})

test_that('NAs in coordinates return NA', {
  copyDT <- copy(DT)
  copyDT[sample(.N, 100), X := NA]

  expect_equal(
    copyDT[is.na(X), .N],
    direction_to_centroid(copyDT, coords = coords,
                          crs = utm)[is.na(X)][is.na(direction_centroid), .N]
  )

  copyDT <- copy(DT)
  copyDT[sample(.N, 100), Y := NA]

  expect_equal(
    copyDT[is.na(Y), .N],
    direction_to_centroid(copyDT, coords = coords,
                          crs = utm)[is.na(Y)][is.na(direction_centroid), .N]
  )

  copyDT <- copy(DT)
  copyDT[sample(.N, 100), X := NA]
  copyDT[sample(.N, 100), centroid_X := NA]
  get_geometry(copyDT, coords, crs = utm)
  get_geometry(copyDT, c('centroid_X', 'centroid_Y'), crs = utm,
               geometry_colname = 'centroid')

  expect_equal(
    copyDT[is.na(Y), .N],
    direction_to_centroid(copyDT)[is.na(Y)][is.na(direction_centroid), .N]
  )

  copyDT <- copy(DT)
  copyDT[sample(.N, 100), Y := NA]
  copyDT[sample(.N, 100), centroid_Y := NA]
  get_geometry(copyDT, coords, crs = utm)
  get_geometry(copyDT, c('centroid_X', 'centroid_Y'), crs = utm,
               geometry_colname = 'centroid')

  expect_equal(
    copyDT[is.na(Y), .N],
    direction_to_centroid(copyDT)[is.na(Y)][is.na(direction_centroid), .N]
  )

})

# Test get_gbi
context('test get_gbi')

library(spatsoc)

DT <- fread('../testdata/DT.csv')

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime)]
DT[, yr := year(datetime)]

utm <- 'EPSG:32736'

group_times(DT, 'datetime', '5 minutes')
group_pts(DT, 50, timegroup = 'timegroup',
          id = 'ID', coords = c('X', 'Y'), splitBy = 'yr')

test_that('DT is required', {
  expect_error(get_gbi(
    DT = NULL,
    group = 'group',
    id = 'ID'
  ),
  'DT must be provided')
})


test_that('ID and group column names provided', {
  expect_error(get_gbi(
    DT = DT,
    group = NULL,
    id = 'ID'
  ),
  'group must be')

  expect_error(get_gbi(
    DT = DT,
    group = 'group',
    id = NULL
  ),
  'id must be')

})


test_that('columns in DT', {
  expect_error(get_gbi(
    DT = DT,
    group = 'potato',
    id = 'ID'
  ),
  'potato field provided')

  expect_error(get_gbi(
    DT = DT,
    group = 'group',
    id = 'potato'
  ),
  'potato field provided')

})



test_that('matrix returned and type integer', {
  expect_type(
    get_gbi(
      DT = DT,
      group = 'group',
      id = 'ID'
    ),
    'integer'
  )

  expect_is(
    get_gbi(
      DT = DT,
      group = 'group',
      id = 'ID'
    ),
    'matrix'
  )

})


test_that('gbi length returned makes sense', {
  expect_equal(sum(get_gbi(
    DT = DT,
    group = 'group',
    id = 'ID'
  )),
  unique(DT[, .(ID, group)])[, .N])
})

test_that('NAs detected', {
  copyDT <- copy(DT)[1, group := NA]

  expect_warning(get_gbi(
    DT = copyDT,
    group = 'group',
    id = 'ID'
  ),
  'DT contains NA', fixed = FALSE)
})

# Test distance_to_leader
context('test distance_to_leader')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, timethreshold)
group_pts(DT, threshold = threshold, id = id,
          coords = coords, timegroup = timegroup)
centroid_group(DT, coords = coords, group = group)
direction_step(DT = DT, id = id, coords = coords, crs = utm)
direction_group(DT)
leader_direction_group(DT, coords = coords, group = group, crs = utm,
                       return_rank = TRUE)

get_geometry(DT, coords = coords, crs = utm)

# Removing group with missing leader
DT_with_missing <- copy(DT)
DT <- copy(DT)[
  !group %in%
    DT[, any(rank_position_group_direction == 1, na.rm = TRUE), by = group][
      !(V1), group]][group != 868]

clean_DT <- copy(DT)

test_that('arguments required, otherwise error detected', {
  expect_error(distance_to_leader(DT = NULL))

  expect_error(distance_to_leader(DT, coords = coords, group = NULL),
               'group must be provided')
})

test_that('column names must exist in DT', {
  expect_error(distance_to_leader(DT, coords = rep('potato', 2), group = group),
               'potato field')
  expect_error(distance_to_leader(DT, coords = coords, group = 'potato'),
               'potato field')
  copy_DT <- copy(DT)
  setnames(copy_DT, 'rank_position_group_direction', 'potato')
  expect_error(distance_to_leader(copy_DT, coords = coords, group = group),
               'did you run leader?')

  copy_DT <- copy(DT)[, rank_position_group_direction := NULL]
  expect_error(distance_to_leader(copy_DT, coords = coords,
                                  group = group))
})

test_that('coords/geometry are correctly provided or error detected', {
  expect_error(distance_to_leader(DT, coords = c('X', NULL), group = group),
               'coords must be length 2')
  copy_DT <- copy(DT)[, X := as.character(X)]
  expect_error(distance_to_leader(copy_DT, coords = coords, group = group),
               'coords must be of class numeric')
  copy_DT <- copy(DT)[, X := as.character(X)]
  expect_error(distance_to_leader(copy_DT, coords = coords,
                                  group = group),
               'coords must be of class numeric')

  # geometry
  expect_error(distance_to_leader(DT, geometry = 'potato'),
               'get_geometry')
  expect_message(distance_to_leader(DT, crs = utm),
                 'crs argument is ignored')
  expect_error(distance_to_leader(DT, geometry = 'X'))
})

test_that('leader is correctly provided or error detected', {
  copy_DT <- copy(DT)[, rank_position_group_direction :=
                        as.character(rank_position_group_direction)]
  expect_error(distance_to_leader(copy_DT, coords = coords, group = group),
               'must be of class numeric')
})

test_that('warns if group does not have a leader', {
  # coords
  expect_warning(
    distance_to_leader(
      DT = DT_with_missing,
      coords = coords,
      group = 'group'
    ),
    'missing leader'
  )

  # geometry
  expect_warning(
    distance_to_leader(
      DT = DT_with_missing,
      group = 'group'
    ),
    'missing leader'
  )
})

test_that('message when distance_leader column overwritten', {
  # coords
  copyDT <- copy(clean_DT)[, distance_leader := 1]
  expect_message(
    distance_to_leader(copyDT, coords = coords, group = group),
    'distance_leader column will be overwritten'
  )

  # geometry
  copyDT <- copy(clean_DT)[, distance_leader := 1]
  expect_message(
    distance_to_leader(copyDT, group = group),
    'distance_leader column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  # coords
  copyDT <- copy(clean_DT)
  expect_equal(nrow(copyDT),
               nrow(distance_to_leader(copyDT, coords = coords, group = group)))

  # geometry
  copyDT <- copy(clean_DT)
  expect_equal(nrow(copyDT),
               nrow(distance_to_leader(copyDT, group = group)))
})

test_that('one column added to the result DT', {
  # coords
  copyDT <- copy(clean_DT)
  expect_equal(ncol(clean_DT) + 1,
               ncol(distance_to_leader(copyDT, coords = coords, group = group)))

  # And check modifies by reference
  copyDT <- copy(clean_DT)
  distance_to_leader(copyDT, coords = coords, group = group)
  expect_equal(ncol(clean_DT) + 1, ncol(copyDT))

  # geometry
  copyDT <- copy(clean_DT)
  expect_equal(ncol(clean_DT) + 1,
               ncol(distance_to_leader(copyDT, group = group)))

  # And check modifies by reference
  copyDT <- copy(clean_DT)
  distance_to_leader(copyDT, group = group)
  expect_equal(ncol(clean_DT) + 1, ncol(copyDT))
})

test_that('column added to the result DT is a double', {
  # coords
  expect_type(
    distance_to_leader(DT, coords = coords, group = group)$distance_leader,
    'double'
  )

  # geometry
  expect_type(
    distance_to_leader(DT, group = group)$distance_leader,
    'double'
  )
})

test_that('zzz columns not added to the result', {
  # coords
  zzz_cols <- c('has_leader', 'zzz_N_by_group')
  expect_false(
    any(zzz_cols %in% colnames(direction_to_leader(DT, coords = coords, crs = utm)))
  )

  # geometry
  zzz_cols <- c('has_leader', 'zzz_N_by_group', 'zzz_geometry_leader')
  expect_false(
    any(zzz_cols %in% colnames(direction_to_leader(DT, crs = utm)))
  )
})

test_that('returns a data.table', {
  # coords
  expect_s3_class(distance_to_leader(DT, coords = coords, group = group),
                  'data.table')

  # geometry
  expect_s3_class(distance_to_leader(DT, group = group),
                  'data.table')
})

test_that('NA for leader returns NA for distance', {
  # coords
  out <- suppressWarnings(distance_to_leader(
    DT = DT_with_missing,
    coords = coords,
    group = 'group'
  ))
  leaderless <- out[, !any(rank_position_group_direction == 1, na.rm = TRUE),
                    by = group]
  expect_all_true(
    out[group %in% leaderless[(V1)]$group, is.na(distance_leader)]
  )

  # geometry
  out <- suppressWarnings(distance_to_leader(
    DT = DT_with_missing,
    group = 'group'
  ))
  leaderless <- out[, !any(rank_position_group_direction == 1, na.rm = TRUE),
                    by = group]
  expect_all_true(
    out[group %in% leaderless[(V1)]$group, is.na(distance_leader)]
  )
})


expect_DT <- data.table(
  ID = c('A', 'B'),
  X = c(0, 10),
  Y = c(0, 0),
  group_direction = rep(units::as_units(0, 'rad'), 2),
  group = c(1, 1)
)
centroid_group(expect_DT, coords = coords)
leader_direction_group(expect_DT, coords = coords, crs = utm,
                       return_rank = TRUE, group = group)
distance_to_leader(expect_DT, coords = c('X', 'Y'))

test_that('expected results for simple case', {
  expect_lte(
    expect_DT[, max(distance_leader)],
    10
  )

  expect_gte(
    expect_DT[, min(distance_leader)],
    0
  )
  expect_equal(
    expect_DT[distance_leader == min(distance_leader), ID],
    'B'
  )
})

# geometry
expect_DT <- data.table(
  ID = c('A', 'B'),
  X = c(0, 10),
  Y = c(0, 0),
  group_direction = rep(units::as_units(0, 'rad'), 2),
  group = c(1, 1)
)
centroid_group(expect_DT, coords = coords)
leader_direction_group(expect_DT, coords = coords, crs = utm,
                       return_rank = TRUE, group = group)
get_geometry(expect_DT, coords = coords, crs = NA)
distance_to_leader(expect_DT)

test_that('expected results for simple case', {
  expect_lte(
    expect_DT[, max(distance_leader)],
    10
  )

  expect_gte(
    expect_DT[, min(distance_leader)],
    0
  )
  expect_equal(
    expect_DT[distance_leader == min(distance_leader), ID],
    'B'
  )
})

# Test distance_to_centroid
context('test distance_to_centroid')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, timethreshold)
group_pts(DT, threshold = threshold, id = id,
          coords = coords, timegroup = timegroup)
centroid_group(DT, coords = coords, group = group)

get_geometry(DT, coords = coords, crs = utm)
centroid_group(DT, group = group)

clean_DT <- copy(DT)

test_that('arguments as expected, otherwise error/message/warn', {
  expect_error(distance_to_centroid(DT = NULL))

  expect_error(distance_to_centroid(DT, coords = coords, group = NULL,
                                    return_rank = TRUE),
               'group must be provided')
  expect_error(distance_to_centroid(DT, coords = coords, group = group,
                                    return_rank = NULL),
               'return_rank')
})

test_that('column names must exist in DT', {
  expect_error(distance_to_centroid(DT, coords = rep('potato', 2)),
               'potato field')
  expect_error(distance_to_centroid(DT, coords = coords, group = 'potato',
                                    return_rank = TRUE),
               'potato field')
})

test_that('coords/geometry are correctly provided or error detected', {
  expect_error(distance_to_centroid(DT, coords = c('X', NULL)),
               'coords must be length 2')
  copy_DT <- copy(DT)[, X := as.character(X)]
  expect_error(distance_to_centroid(copy_DT, coords = coords),
               'coords must be of class numeric')
  copy_DT <- copy(DT)[, centroid_X := as.character(centroid_X)]
  expect_error(distance_to_centroid(copy_DT, coords = coords),
               'coords_centroid must be of class numeric')

  # geometry
  expect_error(distance_to_centroid(DT, geometry = 'potato'),
               'get_geometry')
  expect_message(distance_to_centroid(DT, crs = utm),
                 'crs argument is ignored')
  expect_error(distance_to_centroid(DT, geometry = 'X'))
})

test_that('distance_centroid column succesfully detected', {
  # coords
  copyDT <- copy(clean_DT)[, distance_centroid := 1]
  expect_message(
    distance_to_centroid(copyDT, coords = coords),
    'distance_centroid column will be overwritten'
  )

  # geometry
  copyDT <- copy(clean_DT)[, distance_centroid := 1]
  expect_message(
    distance_to_centroid(copyDT),
    'distance_centroid column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  # coords
  copyDT <- copy(clean_DT)
  expect_equal(nrow(copyDT),
               nrow(distance_to_centroid(copyDT, coords = coords)))

  # geometry
  copyDT <- copy(clean_DT)
  expect_equal(nrow(copyDT),
               nrow(distance_to_centroid(copyDT)))
})

test_that('1/2 columns added to the result DT (depending on return_rank)', {
  # coords
  copyDT <- copy(clean_DT)
  expect_equal(ncol(copyDT) + 2,
               ncol(distance_to_centroid(copyDT, coords = coords,
                                         return_rank = TRUE)))

  copyDT <- copy(clean_DT)
  expect_equal(ncol(copyDT) + 1,
               ncol(distance_to_centroid(copyDT, coords = coords,
                                         return_rank = FALSE)))

  # geometry
  copyDT <- copy(clean_DT)
  expect_equal(ncol(copyDT) + 2,
               ncol(distance_to_centroid(copyDT, return_rank = TRUE)))

  copyDT <- copy(clean_DT)
  expect_equal(ncol(copyDT) + 1,
               ncol(distance_to_centroid(copyDT, return_rank = FALSE)))
})

test_that('column added to the result DT is a double', {
  # coords
  expect_type(distance_to_centroid(DT, coords = coords)$distance_centroid,
              'double')

  expect_type(distance_to_centroid(DT, coords = coords, crs = utm)$distance_centroid,
              'double')

  # geometry
  copyDT <- copy(clean_DT)
  get_geometry(copyDT, coords = coords, crs = utm, output_crs = 4326,
               geometry_colname = 'geometry_longlat')
  centroid_group(copyDT, group = group, geometry = 'geometry_longlat')
  expect_type(distance_to_centroid(copyDT)$distance_centroid, 'double')
})

test_that('returns a data.table', {
  # coords
  expect_s3_class(distance_to_centroid(DT, coords = coords), 'data.table')

  # geometry
  expect_s3_class(distance_to_centroid(DT), 'data.table')
})

test_that('message if overwritting rank_distance_centroid', {
  # coords
  copy_DT <- copy(clean_DT)
  copy_DT[, rank_distance_centroid := 'potato']

  expect_message(distance_to_centroid(copy_DT, coords = coords, group = group,
                                      return_rank = TRUE),
                 'rank_distance_centroid')

  # geometry
  copy_DT <- copy(clean_DT)
  copy_DT[, rank_distance_centroid := 'potato']

  expect_message(distance_to_centroid(copy_DT, group = group,
                                      return_rank = TRUE),
                 'rank_distance_centroid')
})

test_that('NA for centroid returns NA for distance', {
  # coords
  copy_DT <- copy(DT)[sample(.N, 1e3)]
  copy_DT[sample(.N, 1e2), X := NA]
  expect_all_true(
    distance_to_centroid(copy_DT, coords = coords, crs = utm)[
      is.na(X), is.na(distance_centroid)
    ]
  )

  copy_DT <- copy(DT)[sample(.N, 1e3)]
  copy_DT[sample(.N, 1e2), Y := NA]
  expect_all_true(
    distance_to_centroid(copy_DT, coords = coords, crs = utm)[
      is.na(Y), is.na(distance_centroid)
    ]
  )

  copy_DT <- copy(DT)[sample(.N, 1e3)]
  copy_DT[sample(.N, 1e2), centroid_X := NA]
  expect_all_true(
    distance_to_centroid(copy_DT, coords = coords, crs = utm)[
      is.na(centroid_X), is.na(distance_centroid)
    ]
  )

  copy_DT <- copy(DT)[sample(.N, 1e3)]
  copy_DT[sample(.N, 1e2), centroid_Y := NA]
  expect_all_true(
    distance_to_centroid(copy_DT, coords = coords, crs = utm)[
      is.na(centroid_Y), is.na(distance_centroid)
    ]
  )

  # geometry
  copy_DT <- copy(DT)[sample(.N, 1e3)]
  copy_DT[sample(.N, 1e2), geometry := st_sfc(st_point())]
  expect_all_true(
    distance_to_centroid(copy_DT)[
      sf::st_is_empty(geometry), is.na(distance_centroid)
    ]
  )

  copy_DT <- copy(DT)[sample(.N, 1e3)]
  copy_DT[sample(.N, 1e2), centroid := st_sfc(st_point())]
  expect_all_true(
    distance_to_centroid(copy_DT)[
      sf::st_is_empty(centroid), is.na(distance_centroid)
    ]
  )

  copy_DT <- copy(DT)[sample(.N, 1e3)]
  get_geometry(copy_DT, coords = coords, crs = utm, output_crs = 4326,
               geometry_colname = 'geometry_longlat')
  centroid_group(copy_DT, geometry = 'geometry_longlat')
  copy_DT[sample(.N, 1e2), geometry_longlat := st_sfc(st_point())]
  expect_all_true(
    distance_to_centroid(copy_DT, geometry = 'geometry_longlat')[
      sf::st_is_empty(geometry_longlat), is.na(distance_centroid)
    ]
  )

  copy_DT <- copy(DT)[sample(.N, 1e3)]
  get_geometry(copy_DT, coords = coords, crs = utm, output_crs = 4326,
               geometry_colname = 'geometry_longlat')
  centroid_group(copy_DT, geometry = 'geometry_longlat')
  copy_DT[sample(.N, 1e2), centroid := st_sfc(st_point())]
  expect_all_true(
    distance_to_centroid(copy_DT)[
      sf::st_is_empty(centroid), is.na(distance_centroid)
    ]
  )
})

# Test centroid_group
context('test centroid_group')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
group_pts(DT, threshold = threshold, id = id,
          coords = coords, timegroup = timegroup)

clean_DT <- copy(DT)

test_that('DT is required', {
  expect_error(centroid_group(DT = NULL))
})

test_that('arguments required, otherwise error detected', {
  expect_error(centroid_group(DT, group = group, coords = 'X'),
               'coords must be length 2')
  expect_error(centroid_group(DT, coords = coords, group = NULL),
               'group must be provided')
})

test_that('column names must exist in DT', {
  expect_error(centroid_group(DT, coords = rep('potato', 2)),
               'potato field')
  expect_error(centroid_group(DT, coords = coords, group = 'potato'),
               'potato field')
})

test_that('coords are correctly provided or error detected', {
  expect_error(centroid_group(DT, coords = c('X', NULL)),
               'coords must be length 2')
  expect_error(centroid_group(DT, coords = c('X', 'ID')),
               'coords must be of class numeric')
})

test_that('centroid column succesfully detected', {
  copyDT <- copy(clean_DT)[, centroid_X := 1]
  expect_message(
    centroid_group(copyDT, coords = coords),
    'centroid_X column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  copyDT <- copy(clean_DT)

  expect_equal(nrow(copyDT),
               nrow(centroid_group(copyDT, coords = coords)))
})

test_that('two columns added to the result DT', {
  copyDT <- copy(clean_DT)

  expect_equal(ncol(copyDT) + 2,
               ncol(centroid_group(DT, coords = coords)))
})

test_that('two columns added to the result DT are doubles', {
  expect_type(centroid_group(DT, coords = coords)$centroid_X, 'double')
  expect_type(centroid_group(DT, coords = coords)$centroid_Y, 'double')
})

test_that('returns a data.table', {
  expect_s3_class(centroid_group(DT, coords = coords), 'data.table')
})


DT <- data.table(
  group = c(1, 1, 2, 2),
  X = c(10, 20, 10, NA),
  Y = c(10, 20, 10, 20)
)

test_that('xy results are expected', {
  expect_equal(
    centroid_group(copy(DT), coords)[
      group == 1, unique(centroid_X)],
    15
  )

  expect_equal(
    centroid_group(copy(DT), coords)[
      group == 1, unique(centroid_X)],
    15
  )

  expect_equal(
    centroid_group(copy(DT), coords)[
      group == 2, unique(centroid_X)],
    10
  )
})

DT <- data.table(
  group = c(1, 1, 2, 2),
  X = c(10, 20, 10, NA),
  Y = c(10, 20, 10, 20)
)

get_geometry(DT, coords = c('X', 'Y'), crs = 32736)

test_that('geometry results are expected', {
  expect_equal(
    centroid_group(copy(DT))[
      group == 1, sf::st_coordinates(unique(centroid))[,1]],
      setNames(15, 'X')
  )

  expect_equal(
    centroid_group(copy(DT))[
      group == 1, sf::st_coordinates(unique(centroid))[,2]],
    setNames(15, 'Y')
  )

  expect_equal(
    centroid_group(copy(DT))[
      group == 2, sf::st_coordinates(unique(centroid))[,1]],
    setNames(10, 'X')
  )

})


test_that('geometry interface', {
  expect_message(centroid_group(copy(DT), crs = utm), 'crs')
  copyDT <- copy(DT)
  copyDT[, centroid := 42]
  expect_message(centroid_group(copyDT), 'centroid')
})

# Test group_pts
context('test group_pts')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
coords <- c('X', 'Y')
threshold <- 10
timegroup <- 'timegroup'
utm <- 32736


DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = 'datetime', threshold = '20 minutes')
get_geometry(DT, coords, utm)
get_geometry(DT, coords, utm, 4326, 'geometry_longlat')

test_that('args provided as expected else conditions', {
  expect_error(
    group_pts(
      DT = NULL,
      threshold = threshold,
      id = id
    ),
    'DT must be provided'
  )
  expect_error(
    group_pts(DT, threshold = threshold, id = NULL),
    'id must be'
  )

  expect_error(
    group_pts(DT, threshold = NULL, id = id),
    'threshold must be'
  )

  expect_error(
    group_pts(DT,
      threshold = threshold, id = id,
      coords = coords
    ),
    'timegroup must be'
  )
})

test_that('column names must exist in DT', {
  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = 'potato',
      coords = coords,
      timegroup = timegroup
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      coords = c('potatoX', 'potatoY'),
      timegroup = timegroup
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup,
      splitBy = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )
})


test_that('threshold correctly provided or error detected', {
  expect_error(
    group_pts(DT,
            threshold = -10, timegroup = timegroup, id = id,
            coords = coords
    ),
    'threshold must be > 0'
  )

  expect_error(
    group_pts(DT,
            threshold = 0, timegroup = timegroup, id = id,
            coords = coords
    ),
    'threshold must be > 0'
  )

  expect_error(
    group_pts(DT,
            threshold = '0', timegroup = timegroup, id = id,
            coords = coords
    ),
    'threshold must be of class numeric'
  )

  expect_error(
    group_pts(DT,
            threshold = -10, timegroup = timegroup, id = id,
            coords = coords, crs = utm
    ),
    'threshold must be > 0'
  )

  expect_error(
    group_pts(DT,
            threshold = 0, timegroup = timegroup, id = id,
            coords = coords, crs = utm
    ),
    'threshold must be > 0'
  )

  expect_error(
    group_pts(DT,
            threshold = '0', timegroup = timegroup, id = id,
            coords = coords, crs = utm
    ),
    'threshold must be of class numeric'
  )

  expect_error(
    group_pts(DT,
            threshold = units::as_units(-1, 'm'),
            timegroup = timegroup, id = id,
            coords = coords, crs = utm
    ),
    'threshold must be > 0'
  )

  # geometry
  expect_error(
    group_pts(DT,
            threshold = units::as_units(-1, 'm'),
            timegroup = timegroup, id = id
    ),
    'threshold must be > 0'
  )

  expect_error(
    group_pts(DT,
            threshold = units::as_units(100, 'km'),
            timegroup = timegroup, id = id
    ),
    'units of threshold'
  )
})


test_that('coords/geometry are correctly provided or error detected', {
  # coords
  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      coords = c('X', NULL),
      timegroup = timegroup
    ),
    'coords must be length 2'
  )

  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      coords = c('X', 'ID'),
      timegroup = timegroup
    ),
    'coords must be of class numeric'
  )

  # geometry
  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      timegroup = timegroup,
      geometry = 'potato'
    ),
    'did you run'
  )

  expect_error(
    group_pts(
      DT,
      threshold = threshold,
      id = id,
      timegroup = timegroup,
      geometry = 'X'
    ),
    'sfc_POINT'
  )
})

test_that('warns if timegroup is a datetime or character', {
  copyDT <- copy(DT)
  expect_warning(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'datetime'
    ),
    'timegroup provided is a',
    fixed = FALSE
  )

  copyDT <- copy(DT)
  copyDT[, posix := as.POSIXct(datetime)]
  expect_warning(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'posix'
    ),
    'timegroup provided is a',
    fixed = FALSE
  )

  copyDT <- copy(DT)
  copyDT[, idate := as.IDate(datetime)]
  expect_warning(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = 'idate'
    ),
    'timegroup provided is a',
    fixed = FALSE
  )
})


test_that('group column succesfully detected', {
  # coords
  copyDT <- copy(DT)[, group := 1]
  expect_message(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    ),
    'group column will be overwritten'
  )

  # geometry
  copyDT <- copy(DT)[, group := 1]
  expect_message(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      timegroup = timegroup
    ),
    'group column will be overwritten'
  )
})


test_that('withinGroup is not returned to the user', {
  # coords
  expect_false('withinGroup' %in% colnames(
    group_pts(
      copy(DT),
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    )
  ))

  # geometry
  expect_false('withinGroup' %in% colnames(
    group_pts(
      copy(DT),
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    )
  ))
})

test_that('no rows are added to the result DT', {
  # coords
  expect_equal(
    nrow(copy(DT)),
    nrow(
      group_pts(
        copy(DT),
        threshold = threshold,
        id = id,
        coords = coords,
        timegroup = timegroup
      )
    )
  )

  # geometry
  expect_equal(
    nrow(copy(DT)),
    nrow(
      group_pts(
        copy(DT),
        threshold = threshold,
        id = id,
        timegroup = timegroup
      )
    )
  )
})

test_that('group column added to the result DT', {
  # coords
  expect_equal(
    ncol(copy(DT)) + 1,
    ncol(
      group_pts(
        copy(DT),
        threshold = threshold,
        id = id,
        coords = coords,
        timegroup = timegroup
      )
    )
  )

  expect_true('group' %in%
    colnames(
      group_pts(
        copy(DT),
        threshold = threshold,
        id = id,
        coords = coords,
        timegroup = timegroup
      )
    ))

  # geometry
  expect_equal(
    ncol(copy(DT)) + 1,
    ncol(
      group_pts(
        copy(DT),
        threshold = threshold,
        id = id,
        timegroup = timegroup
      )
    )
  )

  expect_true('group' %in%
    colnames(
      group_pts(
        copy(DT),
        threshold = threshold,
        id = id,
        timegroup = timegroup
      )
    ))
})

test_that('duplicate IDs in a timegroup detected', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '8 hours')
  expect_warning(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    ),
    'found duplicate id in a timegroup',
    fixed = FALSE
  )
})

test_that('returns a data.table', {
  # coords
  expect_s3_class(group_pts(
    copy(DT),
    threshold = threshold,
    id = id,
    coords = coords,
    timegroup = timegroup
  ), 'data.table')

  # geometry
  expect_s3_class(group_pts(
    copy(DT),
    threshold = threshold,
    id = id,
    timegroup = timegroup
  ), 'data.table')
})


test_that('splitBy argument doesnt use splitBy column', {
  copyDT <- copy(DT)

  copyDT[, splitBy := sample(seq.int(5), .N, TRUE)]

  expect_true(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    )[, uniqueN(splitBy), group][V1 > 1, .N != 0]
  )
})


test_that('group_pts returns NA for group when X/Y are NA', {
  # coords
  copyDT <- copy(DT)
  n <- 10
  copyDT[sample(.N, n), X := NA]

  expect_equal(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    )[is.na(group), .N],
    n
  )

  copyDT <- copy(DT)
  n <- 10
  copyDT[sample(.N, n), Y := NA]

  expect_equal(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    )[is.na(group), .N],
    n
  )

  copyDT <- copy(DT)
  n <- 10
  copyDT[sample(.N, n), c('X', 'Y') := NA]

  expect_equal(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      coords = coords,
      timegroup = timegroup
    )[is.na(group), .N],
    n
  )

  # geometry
  copyDT <- copy(DT)
  n <- 10
  copyDT[sample(.N, n), geometry := st_sfc(st_point())]

  expect_equal(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      timegroup = timegroup
    )[is.na(group), .N],
    n
  )

  # longlat
  copyDT <- copy(DT)
  n <- 10
  copyDT[sample(.N, n), geometry_longlat := st_sfc(st_point())]

  expect_equal(
    group_pts(
      copyDT,
      threshold = threshold,
      id = id,
      timegroup = timegroup,
      geometry = 'geometry_longlat'
    )[is.na(group), .N],
    n
  )
})

# Test direction_polarization
context('test direction_polarization')

library(spatsoc)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, threshold = timethreshold)
group_pts(DT, threshold = threshold, id = id,
          coords = coords, timegroup = timegroup)
direction_step(DT, id = id, coords = coords, crs = utm)

clean_DT <- copy(DT)

test_that('DT is required', {
  expect_error(direction_polarization(DT = NULL))
})

test_that('arguments required, otherwise error detected', {
  expect_error(direction_polarization(DT, group = NULL),
               'group must be provided')
  expect_error(direction_polarization(DT, direction = NULL),
               'direction must be provided')
})

test_that('column names must exist in DT', {
  expect_error(direction_polarization(DT, direction = 'potato'),
               'potato field')
  expect_error(direction_polarization(DT, group = 'potato'),
               'potato field')
})

test_that('radians expected else error', {
  copyDT <- copy(clean_DT)[, deg := units::as_units(1.1, 'degree')]
  expect_error(direction_group(copyDT, direction = 'deg'),
               'direction must be of units radians')
})


test_that('polarization column succesfully detected', {
  copyDT <- copy(clean_DT)[, polarization := 1]
  expect_message(
    direction_polarization(copyDT),
    'polarization column will be overwritten'
  )
})

test_that('no rows are added to the result DT', {
  copyDT <- copy(clean_DT)

  expect_equal(nrow(copyDT),
               nrow(direction_polarization(copyDT)))
})

test_that('one column added to the result DT', {
  copyDT <- copy(clean_DT)

  expect_equal(ncol(copyDT) + 1,
               ncol(direction_polarization(DT)))
})

test_that('column added to the result DT is numeric, between 0-1', {
  expect_type(direction_polarization(DT)$polarization, 'double')
  expect_gte(min(direction_polarization(DT)$polarization, na.rm = TRUE), 0)
  expect_lte(max(direction_polarization(DT)$polarization, na.rm = TRUE), 1)
})

test_that('returns a data.table', {
  expect_s3_class(direction_polarization(DT), 'data.table')
})

# Test randomizations
context('test randomizations')

library(spatsoc)

DT <- fread('../testdata/DT.csv')

DT[, datetime := as.POSIXct(datetime)]
DT[, yr := year(datetime)]

group_times(DT, datetime = 'datetime', threshold = '1 hour')
group_pts(
  DT,
  threshold = 100,
  id = 'ID',
  timegroup = 'timegroup',
  coords = c('X', 'Y')
)

test_that('DT, type, id, datetime, (group) are required', {
  expect_error(randomizations(DT = NULL),
               'DT must be provided')

  expect_error(randomizations(DT = DT,
                              type = NULL),
               'type must be', fixed = FALSE)

  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = NULL),
               'id must be provided')

  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = 'ID',
                              datetime = NULL),
               'datetime must be')

  # Group required for daily and step
  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = 'ID',
                              iterations = 1,
                              group = NULL,
                              datetime = 'timegroup'),
               'group must be provided if type is "step" or "daily"')

  expect_error(randomizations(DT = DT,
                              type = 'daily',
                              id = 'ID',
                              group = NULL,
                              iterations = 1,
                              datetime = 'datetime'),
               'group must be provided if type is "step" or "daily"')

})

test_that('type must be one of options', {
  expect_error(randomizations(DT = DT,
                              type = 'potato'),
               'type of randomization must be one of', fixed = FALSE)
})

test_that('fields provided must be in DT', {
  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = 'potato',
                              datetime = 'datetime'),
               'potato field', fixed = TRUE)

  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = 'ID',
                              datetime = 'potato'),
               'potato field', fixed = TRUE)

  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = 'ID',
                              datetime = 'datetime',
                              splitBy = 'potato'),
               'potato field', fixed = TRUE)
})

test_that('iterations is NULL or correctly provided', {
  copyDT <- copy(DT)
  expect_warning(randomizations(DT = copyDT,
                              type = 'step',
                              id = 'ID',
                              group = 'group',
                              datetime = 'datetime',
                              iterations = NULL),
               'iterations is not provided', fixed = FALSE)

  expect_error(randomizations(DT = DT,
                              type = 'step',
                              id = 'ID',
                              datetime = 'datetime',
                              iterations = 'potato'),
               'iterations must be of class', fixed = FALSE)
})

test_that('dateFormatted or not depending on randomization type', {
  expect_warning(randomizations(DT = DT,
                                type = 'step',
                                id = 'ID',
                                group = 'group',
                                datetime = 'datetime',
                                iterations = 1),
                 'datetime provided is POSIXct', fixed = FALSE)

  DT[, numDate := 1]
  expect_error(randomizations(DT = DT,
                              type = 'daily',
                              id = 'ID',
                              group = 'group',
                              datetime = 'numDate',
                              iterations = 3),
                 'datetime must be POSIXct', fixed = FALSE)

  expect_error(randomizations(DT = DT,
                              type = 'trajectory',
                              id = 'ID',
                              coords = c('X', 'Y'),
                              datetime = 'numDate',
                              iterations = 3),
                 'datetime must be POSIXct', fixed = FALSE)
})


test_that('step randomization returns as expected', {
  # N unique randomIDs same as N unique IDs in each timegroup
  expect_equal(
    randomizations(
      DT = DT,
      type = 'step',
      id = 'ID',
      group = 'group',
      iterations = 1,
      datetime = 'timegroup'
    )[, uniqueN(randomID), by = timegroup],
    DT[, uniqueN(ID), by = timegroup])

  # N rows in output is (nrows * iterations + 1)
  expect_equal(nrow(
    randomizations(
      DT = DT,
      type = 'step',
      id = 'ID',
      group = 'group',
      iterations = 3,
      datetime = 'timegroup'
    )
  ),
  nrow(DT) * (3 + 1))

  copyDT <- copy(DT)[, population := 1][1:50, population := 2]
  expect_equal(
    randomizations(
      DT = copyDT,
      type = 'step',
      id = 'ID',
      group = 'group',
      iterations = 1,
      datetime = 'timegroup',
      splitBy = 'population'
    )[, uniqueN(randomID), by = timegroup],
    copyDT[, uniqueN(ID), by = timegroup])
})


test_that('daily randomization returns as expected', {
  expect_equal(
    randomizations(
      DT = DT,
      type = 'daily',
      id = 'ID',
      group = 'group',
      iterations = 1,
      datetime = 'datetime'
    )[, .(N = uniqueN(randomID)),
      by = .(jul, ID, iteration)][, max(N)],
    1)
})

test_that('trajectory randomization returns as expected', {
  expect_equal(
    randomizations(
      DT = DT,
      type = 'trajectory',
      id = 'ID',
      coords = c('X', 'Y'),
      iterations = 1,
      datetime = 'datetime'
    )[, uniqueN(randomJul), by = .(ID, jul, iteration)][, max(V1)],
    1)


  expect_equal(
    nrow(randomizations(
      DT = DT,
      type = 'trajectory',
      coords = c('X', 'Y'),
      id = 'ID',
      iterations = 3,
      datetime = 'datetime'
    )),
    nrow(DT) * (3 + 1))

  expect_true(all(c('X', 'Y', 'randomdatetime') %in%
                    colnames(
                      randomizations(
                        DT = DT,
                        type = 'trajectory',
                        coords = c('X', 'Y'),
                        id = 'ID',
                        iterations = 3,
                        datetime = 'datetime'
                      )
                    )))

})

test_that('non uniques are found', {
  copyDT <- copy(DT)
  copyDT[2, (colnames(DT)) := copyDT[1, .SD]]

  expect_warning(randomizations(DT = copyDT,
                                type = 'daily',
                                id = 'ID',
                                group = 'group',
                                datetime = 'datetime',
                                iterations = 2),
                 'found non-unique rows of id, datetime',
                 fixed = FALSE)
})

test_that('randomization returns expected columns', {
  expect_true(all(c('observed', 'iteration')
                  %in% colnames(
    randomizations(
      DT = DT,
      type = 'daily',
      id = 'ID',
      group = 'group',
      datetime = 'datetime',
      iterations = 2
    )
  )))

  # if step, randomID
  expect_true('randomID' %in% colnames(
    randomizations(
      DT = DT,
      type = 'step',
      id = 'ID',
      group = 'group',
      datetime = 'timegroup',
      iterations = 2
    )
  ))

  # if daily, randomID and jul
  expect_true(all(c('randomID', 'jul') %in% colnames(
    randomizations(
      DT = DT,
      type = 'daily',
      id = 'ID',
      group = 'group',
      datetime = 'datetime',
      iterations = 2
    )
  )))

  # if trajectory, randomJul and randomdatetime and jul
  expect_true('randomJul' %in% colnames(
    randomizations(
      DT = DT,
      type = 'trajectory',
      id = 'ID',
      coords = c('X', 'Y'),
      datetime = 'datetime',
      iterations = 2
    )
  ))

  expect_true('randomdatetime' %in% colnames(
    randomizations(
      DT = DT,
      type = 'trajectory',
      id = 'ID',
      coords = c('X', 'Y'),
      datetime = 'datetime',
      iterations = 2
    )
  ))

  expect_error(
    randomizations(
      DT = DT,
      type = 'trajectory',
      id = 'ID',
      coords = NULL,
      datetime = 'datetime',
      iterations = 2),
      'coords must be provided if type is "trajectory"'
    )
})





test_that('randomization is silent when an individual only has one row', {
  ## Fixed here: 6092ad8d055ff269d39dd481a80f27069c790630

  # ** Fake an individual with 1 relocation**
  copyDT <- copy(DT)[1, ID := 'Z']

  # Temporal grouping
  group_times(copyDT, datetime = 'datetime', threshold = '5 minutes')

  # Spatial grouping with timegroup
  group_pts(
    copyDT,
    threshold = 5,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup'
  )

  expect_s3_class(randomizations(
    copyDT,
    type = 'step',
    id = 'ID',
    group = 'group',
    datetime = 'timegroup',
    splitBy = 'yr',
    iterations = 2
  ), 'data.table')


  expect_s3_class(  randomizations(
    copyDT,
    type = 'daily',
    id = 'ID',
    group = 'group',
    datetime = 'datetime',
    splitBy = 'yr',
    iterations = 2
  ), 'data.table')


  expect_s3_class(  randomizations(
    copyDT,
    type = 'trajectory',
    id = 'ID',
    group = NULL,
    coords = c('X', 'Y'),
    datetime = 'datetime',
    splitBy = 'yr',
    iterations = 2
  ), 'data.table')


})



test_that('n individuals consistent across years', {
  DT[1, ID := 'Z']

  ## Step
  randStep <- randomizations(
    DT,
    type = 'step',
    id = 'ID',
    group = 'group',
    splitBy = 'yr',
    datetime = 'timegroup',
    iterations = 20
  )

  uID <- randStep[, .(nID = .N), by = .(randomID, yr, iteration)]

  expect_equal(
    uID[, sd(nID), by = .(randomID, yr)]$V1,
    rep(0, randStep[, uniqueN(randomID)])
  )

  ## Daily
  randDaily <- randomizations(
    DT,
    type = 'daily',
    id = 'ID',
    group = 'group',
    splitBy = 'yr',
    datetime = 'datetime',
    iterations = 20
  )
  uJul <- randDaily[, .(nDay = uniqueN(jul)), by = .(randomID, yr, iteration)]

  expect_equal(
    uJul[, sd(nDay), by = .(randomID, yr)]$V1,
    rep(0, nrow(randDaily[, .N, .(randomID, yr)]))
  )

})

# Test leader_direction_group
context('test leader_direction_group')

library(spatsoc)
library(units)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
datetime <- 'datetime'
timethreshold <- '20 minutes'
threshold <- 50
coords <- c('X', 'Y')
timegroup <- 'timegroup'
group <- 'group'
utm <- 32736

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
group_times(DT, datetime = datetime, timethreshold)
group_pts(DT, threshold = threshold, id = id,
          coords = coords, timegroup = timegroup)
centroid_group(DT, coords = coords, group = group)
direction_step(DT = DT, id = id, coords = coords, crs = utm)
direction_group(DT)

DT_geo <- copy(DT)
get_geometry(DT_geo, coords = coords, crs = utm)
get_geometry(DT_geo, coords = coords, crs = utm, output_crs = 4326,
             geometry_colname = 'geometry_longlat')
centroid_group(DT_geo)
direction_step(DT_geo, id = id)
direction_group(DT_geo)

clean_DT <- copy(DT)
clean_DT_geo <- copy(DT_geo)

test_that('DT is required', {
  expect_error(leader_direction_group(DT = NULL))
})

test_that('arguments required, otherwise error detected', {
  expect_error(leader_direction_group(DT, coords = coords, return_rank = NULL,
                                      crs = utm),
               'return_rank must be')
  expect_error(leader_direction_group(DT, coords = coords, group = NULL,
                                      crs = utm),
               'group must be')

  expect_error(leader_direction_group(DT, coords = coords, group = NULL,
                                      crs = NULL),
               'crs must be')

  # geo
  copy_DT <- copy(clean_DT_geo)[, geometry := NULL]
  expect_error(leader_direction_group(copy_DT), 'geometry')

  expect_message(leader_direction_group(DT_geo, crs = 4326),
                 'ignored')
})

test_that('column names must exist in DT', {
  expect_error(leader_direction_group(DT, coords = rep('potato', 2)),
               'potato field')
  copy_DT <- copy(clean_DT)
  setnames(copy_DT, 'centroid_X', 'potato_X')
  expect_error(leader_direction_group(copy_DT, coords = coords, crs = utm),
               'centroid_X field')
  expect_error(leader_direction_group(DT, coords = coords, crs = utm,
                                      group = 'potato'),
               'group_pts')

  # geo
  expect_error(leader_direction_group(DT_geo, geometry = 'potato'),
               'potato')

})


test_that('coords are correctly provided or error detected', {
  expect_error(leader_direction_group(DT, coords = c('X', NULL), crs = utm),
               'coords must be length 2')
  copy_DT <- copy(clean_DT)[, X := as.character(X)]
  expect_error(leader_direction_group(copy_DT, coords = coords, crs = utm),
               'coords must be of class numeric')
  copy_DT <- copy(clean_DT)[, centroid_X := as.character(centroid_X)]
  expect_error(leader_direction_group(copy_DT, coords = coords, crs = utm),
               'coords_centroid must be of class numeric')
  copy_DT <- copy(clean_DT)[, centroid_Y := as.character(centroid_X)]
  expect_error(leader_direction_group(copy_DT, coords = coords, crs = utm),
               'coords_centroid must be of class numeric')

  # geo
  expect_error(leader_direction_group(DT_geo, geometry = 'X'),
               'geometry')
})

test_that('radians expected else error', {
  copyDT <- copy(DT)[, group_direction := units::set_units(group_direction, 'degree')]
  expect_error(leader_direction_group(copyDT, coords = coords, crs = utm),
               'direction_group')
})

test_that('position_group_direction column succesfully detected', {
  copy_DT <- copy(clean_DT)[, position_group_direction := 1]
  expect_message(
    leader_direction_group(copy_DT, coords = coords, crs = utm),
    'position_group_direction column will be overwritten'
  )

  # geo
  copy_DT <- copy(clean_DT_geo)[, position_group_direction := 1]
  expect_message(leader_direction_group(copy_DT),
                 'position_group_direction')
})

test_that('no rows are added to the result DT', {
  copy_DT <- copy(clean_DT)
  expect_equal(nrow(copy_DT),
               nrow(leader_direction_group(copy_DT, coords = coords, crs = utm)))

  # geo
  copy_DT <- copy(clean_DT_geo)
  expect_equal(nrow(copy_DT),
               nrow(leader_direction_group(copy_DT)))
})

test_that('1 or 2 (return_rank = TRUE) column(s) added to the result DT', {
  copy_DT <- copy(clean_DT)
  expect_equal(ncol(copy_DT) + 1,
               ncol(leader_direction_group(copy_DT, coords = coords, crs = utm,
                                           return_rank = FALSE)))

  copy_DT <- copy(clean_DT)
  expect_equal(ncol(copy_DT) + 2,
               ncol(leader_direction_group(copy_DT, coords = coords, crs = utm,
                                           return_rank = TRUE,
                                           group = 'group')))

  # geo
  copy_DT <- copy(clean_DT_geo)
  expect_equal(ncol(copy_DT) + 1,
               ncol(leader_direction_group(copy_DT, return_rank = FALSE)))

  copy_DT <- copy(clean_DT_geo)
  expect_equal(ncol(copy_DT) + 2,
               ncol(leader_direction_group(copy_DT,
                                           return_rank = TRUE,
                                           group = 'group')))
})

test_that('column(s) added to the result DT are expected type', {
  expect_type(
    leader_direction_group(DT, coords = coords, crs = utm)$position_group_direction,
    'double'
  )
  expect_type(
    leader_direction_group(
      DT,
      coords = coords,
      crs = utm,
      group = 'group',
      return_rank = TRUE)$rank_position_group_direction,
    'double'
  )

  # geo
  copy_DT <- copy(clean_DT_geo)
  expect_type(
    leader_direction_group(copy_DT)$position_group_direction,
    'double'
  )
  expect_type(
    leader_direction_group(
      copy_DT)$rank_position_group_direction,
    'double'
  )

})

test_that('column(s) added to the result DT are expected range', {
  expect_gt(
    leader_direction_group(DT, coords = coords, crs = utm)[
      position_group_direction < 0, .N],
    0
  )

  expect_gt(
    leader_direction_group(DT, coords = coords, crs = utm)[
      position_group_direction > 0, .N],
    0
  )

  expect_equal(
    leader_direction_group(DT, coords = coords, crs = utm,
                           group = 'group')[
      rank_position_group_direction < 0, .N],
    0
  )

  expect_gt(
    leader_direction_group(DT, coords = coords, crs = utm,
                           group = 'group')[
      position_group_direction > 0, .N],
    0
  )

  # geo
  copy_DT <- copy(clean_DT_geo)
  expect_gt(
    leader_direction_group(copy_DT)[
      position_group_direction < 0, .N],
    0
  )

  expect_gt(
    leader_direction_group(copy_DT)[
      position_group_direction > 0, .N],
    0
  )

  expect_equal(
    leader_direction_group(copy_DT)[
      rank_position_group_direction < 0, .N],
    0
  )

  expect_gt(
    leader_direction_group(copy_DT)[
      position_group_direction > 0, .N],
    0
  )
})

test_that('returns a data.table', {
  expect_s3_class(leader_direction_group(DT, coords = coords, crs = utm), 'data.table')
  expect_s3_class(leader_direction_group(DT_geo), 'data.table')
})



expect_DT <- data.table(
  ID = c('A', 'B'),
  X = c(0, 10),
  Y = c(0, 0),
  group_direction = rep(as_units(0, 'rad'), 2),
  group = c(1, 1)
)
centroid_group(expect_DT, coords = coords)
leader_direction_group(expect_DT, coords = coords,
                       return_rank = TRUE, group = group,
                       crs = utm)

test_that('expected results for simple case', {
  expect_lte(
    expect_DT[, max(rank_position_group_direction)],
    2
  )

  expect_lte(
    expect_DT[, max(position_group_direction)],
    5
  )
  expect_gte(
    expect_DT[, min(position_group_direction)],
    -5
  )

  expect_equal(
    expect_DT[rank_position_group_direction == 1, ID],
    'B'
  )
})

# geo
expect_DT_geo <- data.table(
  ID = c('A', 'B'),
  X = c(0, 10),
  Y = c(0, 0),
  group_direction = rep(as_units(0, 'rad'), 2),
  group = c(1, 1)
)
get_geometry(expect_DT_geo, coords = coords, crs = utm)
centroid_group(expect_DT_geo)
leader_direction_group(expect_DT_geo)

test_that('expected results for simple case', {
  expect_lte(
    expect_DT_geo[, max(rank_position_group_direction)],
    2
  )

  expect_lte(
    expect_DT_geo[, max(position_group_direction)],
    5
  )
  expect_gte(
    expect_DT_geo[, min(position_group_direction)],
    -5
  )

  expect_equal(
    expect_DT_geo[rank_position_group_direction == 1, ID],
    'B'
  )
})


test_that('ud_units must be m', {
  expect_error(leader_direction_group(DT, coords = coords, group = 'group',
                                      crs = 4326),
               'planar')

  expect_error(leader_direction_group(DT_geo, geometry = 'geometry_longlat'),
               'planar')
})

#' Fusion centroid
#'
#' `centroid_fusion` calculates the centroid of each timestep in fusion events.
#' The function expects an edge-list of fusion events identified by
#' [fusion_id()] from edge-lists generated with [edge_dist()] and a `data.table`
#' with relocation data appended with a `timegroup` column from [group_times()].
#' Relocation data should be in two columns representing the X and Y
#' coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `edges` and `DT` must be `data.table`s. If your data is a `data.frame`,
#' you can convert it by reference using [data.table::setDT()] or by reassigning
#' using [data.table::data.table()].
#'
#' The `edges` and `DT` are internally merged in this function using the columns
#' `timegroup` (from [group_times()]) and `ID1` and `ID2` (in `edges`, from
#' `dyad_id`) and `id` (in `DT`). This function expects a `fusionID` present,
#' generated with the [fusion_id()] function. The `timegroup` argument expects
#' the names of a column in `edges` which correspond to the timegroup column.
#' The `id` and `timegroup` arguments expect the names of columns in `DT` which
#' correspond to the id, and timegroup columns.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Centroid function" for details on the underlying centroid function used.
#'
#' @inheritSection centroid_group Interface
#' @inheritSection centroid_group Centroid function
#'
#' @param edges edge-list generated generated by [edge_dist()] or
#'  [edge_nn()], with fusionID column generated by [fusion_id()]
#' @param DT input data.table with timegroup column generated with
#'   [group_times()] matching the input data.table used to generate the edge
#'   list with [edge_nn()] or [edge_dist()]
#' @inheritParams group_pts
#' @inheritParams centroid_group
#'
#' @return `centroid_fusion` returns the input `edges` appended with
#'  centroid column(s) for each timestep and fusion id.
#'
#'  If `coords` are provided, the centroid columns will be named by prefixing
#'  the coordinate column names with "centroid_" (eg. "X" = "centroid_X"). If
#'  `geometry` is used, the centroid column will be named "centroid".
#'
#'  Note: due to the merge required within this function, the output needs to be
#'  reassigned unlike some other `spatsoc` functions like `dyad_id` and
#'  `group_pts`. See details in
#'  [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'  A message is returned when the centroid column(s) already exist in the input
#'  because they will be overwritten.
#'
#' @export
#' @family Centroid functions
#' @seealso [fusion_id] [edge_dist] [group_pts]
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge-list generation
#' edges <- edge_dist(
#'     DT,
#'     threshold = 100,
#'     id = 'ID',
#'     coords = c('X', 'Y'),
#'     timegroup = 'timegroup',
#'     returnDist = TRUE,
#'     fillNA = FALSE
#'   )
#'
#' # Generate dyad id
#' dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
#'
#' # Generate fusion id
#' fusion_id(edges, threshold = 100)
#'
#' # Calculate fusion centroid
#' centroids <- centroid_fusion(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   timegroup = 'timegroup'
#' )
#'
#' print(centroids)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
#' fusion_id(edges, threshold = 100)
#' centroids <- centroid_fusion(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   timegroup = 'timegroup'
#' )
#' print(centroids)
centroid_fusion <- function(
    edges = NULL,
    DT = NULL,
    id = NULL,
    coords = NULL,
    crs = NULL,
    timegroup = 'timegroup',
    geometry = 'geometry') {

  centroid_(
    centroid_groupings = 'fusionID',
    edges = edges,
    DT = DT,
    id = id,
    coords = coords,
    crs = crs,
    timegroup = timegroup,
    geometry = geometry
  )
}

# Internal function reused for centroid_dyad and centroid_fusion
centroid_ <- function(
    centroid_groupings = NULL,
    edges = NULL,
    DT = NULL,
    id = NULL,
    coords = NULL,
    crs = NULL,
    timegroup = 'timegroup',
    geometry = 'geometry') {

  # Due to NSE notes in R CMD check
  centroid <- geo <- . <- x <- y <- NULL

  assert_not_null(centroid_groupings)

  assert_not_null(DT)
  assert_is_data_table(DT)

  assert_not_null(edges)
  assert_not_null(id)
  assert_not_null(timegroup)

  check_cols_edges <- c(centroid_groupings, 'ID1', 'ID2', timegroup)
  assert_are_colnames(edges, check_cols_edges)
  check_cols_DT <- c(id, timegroup)
  assert_are_colnames(DT, check_cols_DT)

  if (is.null(coords)) {
    m <- merge(edges,
               DT[, .SD, .SDcols = c(geometry, id, 'timegroup')],
               by.x = c('ID1', timegroup),
               by.y = c(id, timegroup),
               all.x = TRUE,
               sort = FALSE)

    use_mean <- crs_use_mean(sf::st_crs(DT[[geometry]]))

    m[!is.na(centroid_groupings),
      centroid := calc_centroid(geometry = geo, use_mean = use_mean),
      env = list(geo = geometry,
                 centroid_groupings = centroid_groupings,
                 timegroup = timegroup),
      by = .(centroid_groupings, timegroup)]
    m[, centroid := sf::st_sfc(centroid, recompute_bbox = TRUE)]

  } else {
    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)

    out_xcol <- paste0('centroid_', gsub(' ', '', xcol))
    out_ycol <- paste0('centroid_', gsub(' ', '', ycol))

    m <- merge(edges,
               DT[, .SD, .SDcols = c(coords, id, 'timegroup')],
               by.x = c('ID1', timegroup),
               by.y = c(id, timegroup),
               all.x = TRUE,
               sort = FALSE)

    if (out_xcol %in% colnames(m)) {
      message(paste(out_xcol, 'column will be overwritten by this function'))
      data.table::set(m, j = out_xcol, value = NULL)
    }

    if (out_ycol %in% colnames(m)) {
      message(paste(out_ycol, 'column will be overwritten by this function'))
      data.table::set(m, j = out_ycol, value = NULL)
    }

    use_mean <- crs_use_mean(crs)

    m[!is.na(centroid_groupings),
      (c(out_xcol, out_ycol)) :=
        calc_centroid(, x, y, crs = crs, use_mean = use_mean),
      env = list(x = xcol, y = ycol,
                 centroid_groupings = centroid_groupings,
                 timegroup = timegroup),
      by = .(centroid_groupings, timegroup)]

    data.table::set(m, j = coords, value = NULL)
  }

  data.table::setcolorder(m, colnames(edges))
  return(m[])
}

#' Nearest neighbour based edge-lists
#'
#' `edge_nn` returns edge-lists defined by the nearest neighbour. The function
#' expects a `data.table` with relocation data, individual identifiers and a
#' threshold argument. The threshold argument is used to specify the criteria
#' for distance between points which defines a group. Relocation data should be
#' in two columns representing the X and Y coordinates, or in a geometry column
#' prepared by the helper function [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()].
#'
#' The `id`, `timegroup` (and optional `splitBy`) arguments expect the names of
#' columns in `DT` which correspond to the individual identifier, and timegroup
#' (generated by `group_times`) and additional grouping columns.
#'
#' If a `threshold` is provided, it should match the units of the coordinates.
#' The `threshold` can be provided with units specified using the units package
#' (eg. `threshold = units::set_units(10, m)`) which will be checked against the
#' units of the coordinates using the `crs`. If units are not specified, the
#' `threshold` is assumed to be in the units of the coordinates.
#'
#' The `timegroup` argument is required to define the temporal groups within
#' which edge nearest neighbours are calculated. The intended framework is to
#' group rows temporally with `group_times` then spatially with `edge_nn`. If
#' you have already calculated temporal groups without `group_times`, you can
#' pass this column to the `timegroup` argument. Note that the expectation is
#' that each individual will be observed only once per timegroup. Caution that
#' accidentally including huge numbers of rows within timegroups can overload
#' your machine since all pairwise distances are calculated within each
#' timegroup.
#'
#' The `splitBy` argument offers further control over grouping. If within your
#' `DT`, you have multiple populations, subgroups or other distinct parts, you
#' can provide the name of the column which identifies them to `splitBy`.
#' `edge_nn` will only consider rows within each `splitBy` subgroup.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Distance function" for details on underlying distance function used.
#'
#' @inheritSection distance_to_centroid Interface
#' @inheritSection distance_to_centroid Distance function
#'
#' @param threshold (optional) spatial distance threshold to set maximum
#'   distance between an individual and their neighbour.
#' @inheritParams group_pts
#' @inheritParams edge_dist
#'
#' @return `edge_nn` returns a `data.table`  with three columns: timegroup, ID
#'   and NN. If 'returnDist' is TRUE, a column 'distance' is returned indicating
#'   the distance between ID and NN.  The ID and NN columns represent the edges
#'   defined by the nearest neighbours (and temporal thresholds with
#'   `group_times`).
#'
#'   If an individual was alone in a timegroup or splitBy, or did not have any
#'   neighbours within the threshold distance, they are assigned NA for nearest
#'   neighbour.
#'
#'   Note: unlike many other functions (eg. `group_pts`) in `spatsoc`, `edge_nn`
#'   needs to be reassigned. See details in
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'
#' @export
#'
#' @family Edge-list generation
#' @family Distance functions
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Select only individuals A, B, C for this example
#' DT <- DT[ID %in% c('A', 'B', 'C')]
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge-list generation
#' edges <- edge_nn(DT, id = 'ID', coords = c('X', 'Y'),
#'         timegroup = 'timegroup')
#'
#' # Edge-list generation using maximum distance threshold
#' edges <- edge_nn(DT, id = 'ID', coords = c('X', 'Y'),
#'         timegroup = 'timegroup', threshold = 100)
#'
#' # Edge-list generation, returning distance between nearest neighbours
#' edge_nn(DT, id = 'ID', coords = c('X', 'Y'),
#'         timegroup = 'timegroup', threshold = 100,
#'         returnDist = TRUE)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' edge_nn(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
edge_nn <- function(
    DT = NULL,
    id = NULL,
    coords = NULL,
    timegroup,
    crs = NULL,
    splitBy = NULL,
    threshold = NULL,
    geometry = 'geometry',
    returnDist = FALSE) {
  # NSE
  N <- geo <- x <- y <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(id)
  assert_not_missing(timegroup)
  assert_not_null(timegroup)

  check_cols <- c(timegroup, id, splitBy)
  assert_are_colnames(DT, check_cols)

  if (any(unlist(lapply(DT[, .SD, .SDcols = timegroup], class)) %in%
          c('POSIXct', 'POSIXlt', 'Date', 'IDate', 'ITime', 'character'))) {
    warning(
      strwrap(
        prefix = " ",
        initial = "",
        x = 'timegroup provided is a date/time
        or character type, did you use group_times?'
      )
    )
  }

  if ('splitBy' %in% colnames(DT)) {
    warning(
      strwrap(x = 'a column named "splitBy" was found in your data.table,
              renamed to "split_by" to avoid confusion with the argument
              "splitBy"')
    )
    data.table::setnames(DT, 'splitBy', 'split_by')
  }

  splitBy <- c(splitBy, timegroup)
  if (DT[, .N, by = c(id, splitBy)][N > 1, sum(N)] != 0) {
    warning(
      strwrap(
        prefix = " ",
        initial = "",
        x = 'found duplicate id in a
        timegroup and/or splitBy -
        does your group_times threshold match the fix rate?'
      )
    )
  }

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    crs <- sf::st_crs(DT[[geometry]])

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    if (!is.null(threshold)) {
      assert_threshold(threshold, crs)

      if (!inherits(threshold, 'units') && !identical(crs, sf::NA_crs_) &&
          !use_dist) {
        threshold <- units::as_units(threshold, 'm')
      }
    }

    DT[, {

      distMatrix <- calc_distance(
        geometry_a = geo,
        use_dist = use_dist
      )
      diag(distMatrix) <- NA

      if (!is.null(threshold)) {
        distMatrix[distMatrix > threshold] <- NA
      }
      wm <- as.numeric(apply(distMatrix, MARGIN = 2, which.min))

      out_id <- id
      if (all(is.na(wm))) {
        out_nn <- NA_character_
        if (returnDist) {
          out_dist <- NA_real_
        }
      } else {
        out_nn <- id[wm]
        if (returnDist) {
          w <- wm + (length(wm) * (seq_along(wm) - 1))
          out_dist <- distMatrix[w]
        }
      }

      l <- list(ID = out_id,
                NN = out_nn)

      if (returnDist) {
        l <- c(l, list(distance = out_dist))
      }

      l
    },
    by = c(splitBy),
    env = list(geo = geometry, id = id)]
  } else {
    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    assert_length(coords, 2)
    assert_are_colnames(DT, coords)
    assert_col_inherits(DT, coords, 'numeric')

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    if (!is.null(threshold)) {
      assert_threshold(threshold, crs)

      if (!inherits(threshold, 'units') && !identical(crs, sf::NA_crs_) &&
          !use_dist) {
        threshold <- units::as_units(threshold, 'm')
      }
    }

    DT[, {

      distMatrix <- calc_distance(
        x_a = x,
        y_a = y,
        crs = crs,
        use_dist = use_dist
      )
      diag(distMatrix) <- NA

      if (!is.null(threshold)) {
        distMatrix[distMatrix > threshold] <- NA
      }
      wm <- as.numeric(apply(distMatrix, MARGIN = 2, which.min))

      out_id <- id
      if (all(is.na(wm))) {
        out_nn <- NA_character_
        if (returnDist) {
          out_dist <- NA_real_
        }
      } else {
        out_nn <- id[wm]
        if (returnDist) {
          w <- wm + (length(wm) * (seq_along(wm) - 1))
          out_dist <- distMatrix[w]
        }
      }

      l <- list(ID = out_id,
                NN = out_nn)

      if (returnDist) {
        l <- c(l, list(distance = out_dist))
      }

      l
    },
    by = c(splitBy),
    env = list(x = xcol, y = ycol, id = id)]
  }


}

#' Movement of 10 "Newfoundland Bog Cows"
#'
#' A dataset containing the GPS relocations of 10 individuals in winter
#' 2016-2017.
#'
#' @format A data.table with 14297 rows and 5 variables: \describe{
#'   \item{ID}{individual identifier} \item{X}{X coordinate of the relocation
#'   (UTM 36N)} \item{Y}{Y coordinate of the relocation (UTM 36N)}
#'   \item{datetime}{character string representing the date time}
#'   \item{population}{sub population within the individuals} }
#'
#' @name DT
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
NULL

#' Leadership in directional correlation delay
#'
#' Given the directional correlation delay, `leader_edge_delay`
#' calculates the mean directional correlation delay for individuals in a
#' group to identify leadership patterns.
#'
#' The function expects an edge-list from `edge_delay` with columns
#' 'direction_delay' indicating the directional correlation delay between
#' individuals and 'direction_diff' indicating the unsigned difference in
#' movement directions at the temporal delay, columns 'ID1' and 'ID2' indicating
#' individuals and column 'dyadID' indicating the dyad.
#'
#' The `edge` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' @return `leader_edge_delay` returns the input `edges` aggregated
#'   with a `mean_direction_delay_dyad` column indicating the mean
#'   directional correlation delay between ID1 and ID2 and a
#'   `mean_direction_delay` column indicating the mean directional
#'   correlation delay for each individual in 'ID1' column.
#'
#'  Note: since `leader_edge_delay` returns an aggregation of the input
#'  `edges`, the output needs to be reassigned unlike some other
#'  `spatsoc` functions like `dyad_id`. See details in
#'  [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'
#' @inheritParams edge_delay
#' @param threshold (optional) threshold difference in direction used to subset
#'   rows included in calculation of mean directional delay. eg.
#'   `threshold = 0.5` corresponds to only rows where direction_diff is
#'   less than 0.5. Expects that unit is radians, see `edge_delay`.
#' @param splitBy (optional) character string or vector of grouping column
#'   name(s) upon which the mean directional correlation delay will be
#'   calculated
#'
#' @export
#' @seealso `edge_delay`
#' @family Leadership functions
#' @family Direction functions
#'
#' @references
#' See examples of measuring leadership using the directional correlation
#' delay:
#'  * \doi{doi:10.1016/j.anbehav.2013.07.005}
#'  * \doi{doi:10.1073/pnas.1305552110}
#'  * \doi{doi:10.1126/science.aap7781}
#'  * \doi{doi:10.1111/jfb.15315}
#'  * \doi{doi:10.1371/journal.pcbi.1003446}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Select only individuals A, B, C for this example
#' DT <- DT[ID %in% c('A', 'B', 'C')]
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Calculate direction
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Distance based edge-list generation
#' edges <- edge_dist(
#'   DT,
#'   threshold = 100,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   timegroup = 'timegroup',
#'   returnDist = TRUE,
#'   fillNA = FALSE
#' )
#'
#' # Generate dyad id
#' dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
#'
#' # Generate fusion id
#' fusion_id(edges, threshold = 100)
#'
#' # Directional correlation delay
#' delay <- edge_delay(
#'   edges = edges,
#'   DT = DT,
#'   window = 3,
#'   id = 'ID'
#' )
#'
#' # Leadership from directional correlation delay
#' leadership <- leader_edge_delay(
#'   delay,
#'   threshold = 0.5
#' )
#' print(leadership)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' direction_step(DT, id = 'ID')
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
#' fusion_id(edges, threshold = 100)
#' delay <- edge_delay(
#'   edges = edges,
#'   DT = DT,
#'   window = 3,
#'   id = 'ID'
#' )
#' leadership <- leader_edge_delay(
#'   delay,
#'   threshold = 0.5
#' )
#' print(leadership)
leader_edge_delay <- function(
    edges = NULL,
    threshold = NULL,
    splitBy = NULL) {
  # Due to NSE notes
  . <- direction_diff <- direction_delay <- mean_direction_delay <-
    mean_direction_delay_dyad <- NULL
  direction_delay <- 'direction_delay'
  direction_diff <- 'direction_diff'

  assert_not_null(edges)
  assert_is_data_table(edges)

  check_cols <- c(direction_delay, direction_diff, 'ID1', 'ID2', splitBy)
  assert_are_colnames(edges, check_cols, ', did you use edge_delay?')
  assert_col_inherits(edges, direction_delay, 'integer',
                      ', did you use edge_delay?')
  assert_col_inherits(edges, direction_diff, 'numeric',
                      ', did you use edge_delay?')

  if (is.null(threshold)) {
    threshold <- Inf
  } else {
    assert_inherits(threshold, 'numeric')
  }

  subset_threshold <- edges[direction_diff < threshold]

  out <- subset_threshold[,
    .(mean_direction_delay_dyad = mean(direction_delay, na.rm = TRUE)),
    by = c('ID1', 'ID2', 'dyadID', splitBy)]

  out[, mean_direction_delay := mean(mean_direction_delay_dyad, na.rm = TRUE),
      by = c('ID1', splitBy)]

  return(out[])
}

#' Direction based edge-lists
#'
#' [edge_direction()] returns edge lists defined by the direction between
#' individuals. The function expects a distance based edge-list generated by
#' `edge_nn` or [edge_dist()] and a `data.table` with relocation data appended
#' with a `timegroup` column from [group_times()]. It is required to use the
#' argument `fillNA = FALSE` for [edge_dist()] to ensure there are no NAs in the
#' coordinate columns. Relocation data should be in two columns representing the
#' X and Y coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `edges` and `DT` must be `data.table`s. If your data is a `data.frame`,
#' you can convert it by reference using [data.table::setDT()] or by reassigning
#' using [data.table::data.table()].
#'
#' The `edges` and `DT` are internally merged in this function using the columns
#' `id` / `ID1` and `ID2`, `timegroup`. The `id`, and `timegroup` arguments
#' expect the names of columns which correspond to the ID, and timegroup
#' columns in `DT`.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Direction function" for details on the underlying direction function used.
#'
#' @inheritSection direction_step Interface
#' @inheritSection direction_step Direction function
#'
#' @inheritParams centroid_dyad
#' @inheritParams direction_step
#' @inheritParams edge_alignment
#'
#' @references
#'
#' See examples of measuring the direction between individuals:
#'  * \doi{doi:10.1098/rsif.2013.0529}
#'  * \doi{doi:10.1098/rstb.2019.0380}
#'  * \doi{doi:10.1111/jfb.15315}
#'
#' @return [edge_direction()] returns the input `edges` appended with a
#'   "direction_dyad" column representing the direction between ID1 and ID2 in
#'   radians. A value of NaN is returned when the coordinates of ID1 equal the
#'   coordinates of ID2.
#'
#'   If the "direction" column is found in input DT, it will be retained for ID1
#'   in the output for use in downstream functions (eg. [edge_zones()]).
#'
#'   Missing values in coordinates / geometry are ignored and NA is returned.
#'
#'   Note: due to the merge required within this function, the output needs to
#'   be reassigned unlike some other `spatsoc` functions like [dyad_id()] and
#'   [group_pts()]. See details in
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @family Edge-list generation
#' @family Direction functions
#' @seealso [dyad_id], [edge_dist], [edge_nn], [group_times],
#'   [lwgeom::st_geod_azimuth()]
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge list generation
#' edges <- edge_dist(
#'     DT,
#'     threshold = 100,
#'     id = 'ID',
#'     coords = c('X', 'Y'),
#'     timegroup = 'timegroup',
#'     returnDist = TRUE,
#'     fillNA = FALSE
#'   )
#'
#' # Direction based edge-lists
#' dyad_directions <- edge_direction(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736,
#'   timegroup = 'timegroup'
#' )
#'
#' print(dyad_directions)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' dyad_directions <- edge_direction(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   timegroup = 'timegroup'
#' )
edge_direction <- function(
    edges = NULL,
    DT = NULL,
    id = NULL,
    coords = NULL,
    crs = NULL,
    timegroup = 'timegroup',
    geometry = 'geometry',
    projection = NULL) {

  # Due to NSE notes in R CMD check
  geo_id1 <- geo_id2 <- x_id1 <- y_id1 <- x_id2 <- y_id2 <- NULL

  if (!is.null(projection)) {
    warning('projection argument is deprecated, setting crs = projection')
    crs <- projection
  }

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(edges)
  assert_not_null(id)
  assert_not_null(timegroup)
  assert_are_colnames(DT, c(id, timegroup))
  assert_are_colnames(edges, timegroup)

  out_col <- 'direction_dyad'

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    geometry_id1 <- paste0(geometry, '_id1')
    geometry_id2 <- paste0(geometry, '_id2')

    ID1_cols <- if ('direction' %in% colnames(DT)) {
      c('direction', geometry, id, timegroup)
    } else {
      c(geometry, id, timegroup)
    }

    if (geometry %in% colnames(edges)) {
      edges <- data.table::copy(edges)[, .SD, .SDcols = -c(geometry)]
    }

    m <- merge(edges,
               DT[, .SD, .SDcols = ID1_cols],
               by.x = c('ID1', timegroup),
               by.y = c(id, timegroup),
               all.x = TRUE,
               sort = FALSE)
    data.table::setnames(m, geometry, geometry_id1)
    m <- merge(m,
               DT[, .SD, .SDcols = c(geometry, id, timegroup)],
               by.x = c('ID2', timegroup),
               by.y = c(id, timegroup),
               all.x = TRUE,
               sort = FALSE)
    data.table::setnames(m, geometry, geometry_id2)

    if (out_col %in% colnames(m)) {
      message(paste(out_col, 'column will be overwritten by this function'))
      data.table::set(m, j = out_col, value = NULL)
    }

    crs <- sf::st_crs(DT[[geometry]])
    use_transform <- !sf::st_is_longlat(crs)

    if (is.na(use_transform)) {
      rlang::abort(paste0('sf::st_is_longlat(crs) is ', use_transform,
                          ', ensure crs is provided for direction functions'))
    }

    m[!sf::st_is_empty(geo_id1) & !sf::st_is_empty(geo_id2),
      (out_col) := calc_direction(
        geometry_a = geo_id1,
        geometry_b = geo_id2,
        use_transform = use_transform
      ),
      env = list(
        geo_id1 = geometry_id1,
        geo_id2 = geometry_id2
      )
    ]

    data.table::set(m, j = c(geometry_id1, geometry_id2), value = NULL)
    data.table::setcolorder(m, c(timegroup, 'ID1', 'ID2'))

  } else {
    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')
    assert_not_null(crs)

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)
    coords_id1 <- paste0('id1_', coords)
    coords_id2 <- paste0('id2_', coords)

    ID1_cols <- if ('direction' %in% colnames(DT)) {
      c('direction', coords, id, timegroup)
    } else {
      c(coords, id, timegroup)
    }

    m <- merge(edges,
               DT[, .SD, .SDcols = ID1_cols],
               by.x = c('ID1', timegroup),
               by.y = c(id, timegroup),
               all.x = TRUE,
               sort = FALSE)
    data.table::setnames(m, coords, coords_id1)
    m <- merge(m,
               DT[, .SD, .SDcols = c(coords, id, timegroup)],
               by.x = c('ID2', timegroup),
               by.y = c(id, timegroup),
               all.x = TRUE,
               sort = FALSE)
    data.table::setnames(m, coords, coords_id2)

    if (out_col %in% colnames(m)) {
      message(paste(out_col, 'column will be overwritten by this function'))
      data.table::set(m, j = out_col, value = NULL)
    }

    use_transform <- !sf::st_is_longlat(crs)

    if (is.na(use_transform)) {
      rlang::abort(paste0('sf::st_is_longlat(crs) is ', use_transform,
                          ', ensure crs is provided for direction functions'))
    }

    m[!is.na(x_id1) & !is.na(y_id1) & !is.na(x_id2) & !is.na(y_id2),
      (out_col) := calc_direction(
        x_a = x_id1,
        y_a = y_id1,
        x_b = x_id2,
        y_b = y_id2,
        crs = crs,
        use_transform = use_transform
      ),
      env = list(
        x_id1 = data.table::first(coords_id1),
        y_id1 = data.table::last(coords_id1),
        x_id2 = data.table::first(coords_id2),
        y_id2 = data.table::last(coords_id2)
      )
    ]

    data.table::set(m, j = c(coords_id1, coords_id2), value = NULL)
    data.table::setcolorder(m, c(timegroup, 'ID1', 'ID2'))
  }



  return(m[])
}

#' Group by individual matrix
#'
#' `get_gbi` generates a group by individual matrix. The function expects a
#' `data.table` with individual identifiers and a group column. The group
#' by individual matrix can then be used to build a network using
#' [asnipe::get_network()].
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()].
#'
#' The `group` argument expects the name of a column which corresponds to
#' an integer group identifier (generated by `spatsoc`'s grouping
#' functions).
#'
#' The `id` argument expects the name of a column which corresponds to the
#' individual identifier.
#'
#' @return `get_gbi` returns a group by individual matrix (columns
#'   represent individuals and rows represent groups).
#'
#'   Note that `get_gbi` is identical in function for turning the outputs
#'   of `spatsoc` into social networks as
#'   [asnipe::get_group_by_individual()]
#'   but is more efficient thanks to
#'   [`data.table::dcast()`][data.table::dcast.data.table].
#'
#' @inheritParams group_pts
#' @param group Character string of group column (generated from one of
#'   spatsoc's spatial grouping functions)
#' @export
#'
#' @seealso `group_pts` `group_lines`
#'   `group_polys`
#' @family Social network tools
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#' DT[, yr := year(datetime)]
#'
#' # EPSG code for example data
#' utm <- 'EPSG:32736'
#'
#' group_polys(DT, area = FALSE, hrType = 'mcp',
#'             hrParams = list(percent = 95),
#'             crs = utm, id = 'ID', coords = c('X', 'Y'),
#'             splitBy = 'yr')
#'
#' gbiMtrx <- get_gbi(DT = DT, group = 'group', id = 'ID')
#'
get_gbi <- function(DT = NULL,
    group = 'group',
    id = NULL) {

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(group)
  assert_not_null(id)
  assert_are_colnames(DT, c(group, id))

  if (anyNA(DT[[group]])) {
    warning('DT contains NA(s) in group column, these rows will be dropped')
  }

  uDT <-
    stats::na.omit(unique(DT[, .SD, .SDcols = c(group, id)]),
            cols = group)

  cDT <-
    data.table::dcast(
      uDT,
      formula = stats::reformulate(id, group),
      fun.aggregate = length,
      value.var = group
    )

  ids <- colnames(cDT)[!grepl(group, colnames(cDT))]

  m <- as.matrix(cDT[, .SD, .SDcols = ids])

  rownames(m) <- cDT[[group]]
  return(m)
}

#' Group centroid
#'
#' `centroid_group` calculates the centroid of all individuals in each
#' spatiotemporal group identified by `group_pts`. The function expects a
#' `data.table` with relocation data appended with a `group` column from
#' `group_pts`. Relocation data should be in two columns representing the X and
#' Y coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `group` argument expects the name of a column in
#' `DT` which correspond to the group column.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Centroid function" for details on the underlying centroid function used.
#'
#' @section Interface:
#'  Two interfaces are available for providing coordinates:
#'
#'  1. Provide `coords` and optionally `crs`. The `coords` argument expects the
#'  names of the X and Y coordinate columns. The `crs` argument expects a
#'  character string or numeric defining the coordinate reference system to be
#'  passed to [sf::st_crs]. For example, for UTM zone 36S (EPSG 32736), the crs
#'  argument is `crs = "EPSG:32736"` or `crs = 32736`. See
#'  <https://spatialreference.org> for a list of EPSG codes. For centroid
#'  calculations, if `crs` is NULL, it will be internally set to `NA_crs_`.
#'  2. (New!) Provide `geometry`. The `geometry` argument allows the user to
#'  supply a `geometry` column that represents the coordinates as a simple
#'  feature geometry list column. This interface expects the user to prepare
#'  their input DT with [get_geometry()]. To use this interface, leave the
#'  `coords` and `crs` arguments `NULL`, and the default argument for `geometry`
#'  ('geometry') will be used directly.
#'
#' @section Centroid function:
#'
#' The underlying centroid function used depends on the crs of the coordinates
#'  or geometry provided.
#'
#'  - If the crs is longlat degrees (as determined by
#'  [sf::st_is_longlat()]) and [sf::sf_use_s2()] is TRUE, the distance function
#'  is [sf::st_centroid()] which passes to [s2::s2_centroid()].
#'  - If the crs is longlat degrees but [sf::sf_use_s2()] is FALSE, the centroid
#'  calculated will be incorrect. See [sf::st_centroid()].
#'  - If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
#'  centroid function used is mean.
#'
#'  Note: if the input is length 1, the input is returned.
#'
#' @param DT input data.table with group column generated with `group_pts`
#' @inheritParams group_pts
#' @param group Character string of group column
#' @param geometry simple feature geometry list column name, generated by
#'   [get_geometry()]. Default 'geometry', see details under Interface
#' @param crs numeric or character defining the coordinate reference
#'   system to be passed to [sf::st_crs]. For example, either
#'   `crs = "EPSG:32736"` or `crs = 32736`. Used only if coords are provided,
#'   see details under Interface
#'
#' @return `centroid_group` returns the input `DT` appended with
#'  centroid column(s) for each group.
#'
#'  If the `crs` for `coords` or `st_crs(geometry)` for `geometry` is long lat
#'  (see [sf::st_is_longlat()]), centroids will be calculated using
#'  [s2::s2_centroid()] through [sf::st_centroid()]. If the `crs` for `coords`
#'  or `st_crs(geometry)` for `geometry` is projected or NA, the centroids will
#'  be calculated using a mean on the coordinates.
#'
#'  If `coords` are provided, the centroid columns will be named by prefixing
#'  the coordinate column names with "centroid_" (eg. "X" = "centroid_X"). If
#'  `geometry` is used, the centroid column will be named "centroid".
#'
#'  A message is returned when the centroid column(s) already exist in the
#'  input because they will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @seealso `group_pts`
#' @family Centroid functions
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 5, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate group centroid
#' centroid_group(DT, coords = c('X', 'Y'), group = 'group')
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' centroid_group(DT)
centroid_group <- function(
    DT = NULL,
    coords = NULL,
    crs = NULL,
    group = 'group',
    geometry = 'geometry') {

  # Due to NSE notes in R CMD check
  centroid <- geo <- cent <- . <- x <- y <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)

  assert_not_null(group)
  assert_are_colnames(DT, group)

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    out <- 'centroid'
    if (out %in% colnames(DT)) {
      message(out, ' column will be overwritten by this function')
      data.table::set(DT, j = out, value = NULL)
    }

    use_mean <- crs_use_mean(sf::st_crs(DT[[geometry]]))

    DT[, (out) := calc_centroid(geo, use_mean = use_mean),
       env = list(geo = geometry),
       by = c(group)]
    DT[, (out) := sf::st_sfc(cent, recompute_bbox = TRUE),
       env = list(cent = out)]

  } else {
    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)

    out_xcol <- paste0('centroid_', gsub(' ', '', xcol))
    out_ycol <- paste0('centroid_', gsub(' ', '', ycol))

    if (out_xcol %in% colnames(DT)) {
      message(paste(out_xcol, 'column will be overwritten by this function'))
      data.table::set(DT, j = out_xcol, value = NULL)
    }

    if (out_ycol %in% colnames(DT)) {
      message(paste(out_ycol, 'column will be overwritten by this function'))
      data.table::set(DT, j = out_ycol, value = NULL)
    }
    use_mean <- crs_use_mean(crs)
    DT[, (c(out_xcol, out_ycol)) :=
         calc_centroid(, x, y, crs = crs, use_mean = use_mean),
       env = list(x = xcol, y = ycol),
       by = c(group)]
  }

  return(DT[])
}

#' Polarization
#'
#' `direction_polarization` calculates the polarization of individual
#' directions in each spatiotemporal group identified by `group_pts`. The
#' function expects a `data.table` with relocation data appended with a
#' `direction` column from `direction_step` and a `group` column
#' from `group_pts`.
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `direction` and `group` arguments expect the names of columns
#' in `DT` which correspond to the direction and group columns. The
#' direction column is expected in units of radians and the polarization is
#' calculated with [CircStats::r.test()].
#'
#' @inheritParams direction_group
#'
#' @return `direction_polarization` returns the input `DT` appended
#'   with a `polarization` column representing the direction polarization
#'   of all individuals in each spatiotemporal group.
#'
#'   The direction polarization is calculated using [CircStats::r.test()]
#'   which expects units of radians.
#'
#'   A message is returned when the `polarization` columns already
#'   exists in the input `DT`, because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @seealso `direction_step`, `group_pts`,
#'   [CircStats::r.test()]
#' @family Direction functions
#'
#' @references
#' See examples of using polarization:
#'  * \doi{doi:10.1016/j.cub.2017.08.004}
#'  * \doi{doi:10.1371/journal.pcbi.1009437}
#'  * \doi{doi:10.7554/eLife.19505}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 50, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate direction at each step
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Calculate polarization
#' direction_polarization(DT)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' direction_step(DT, id = 'ID')
#' direction_polarization(DT)
direction_polarization <- function(
    DT,
    direction = 'direction',
    group = 'group') {

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(direction)
  assert_not_null(group)
  assert_are_colnames(DT, c(direction, group))

  assert_col_inherits(DT, direction, 'units', ', did you use direction_step?')
  assert_col_radians(DT, direction, ', did you use direction_step?')

  out <- 'polarization'

  if (out %in% colnames(DT)) {
    message(paste(out, 'column will be overwritten by this function'))
    data.table::set(DT, j = out, value = NULL)
  }

  DT[, out :=  CircStats::r.test(as.numeric(dir), degree = FALSE)$r.bar,
     by = c(group),
     env = list(dir = direction, out = out)]

  return(DT[])
}

#' Behavioural zones
#'
#' `edge_zones` returns edge-lists defined by behavioural zones (Couzin
#' 2002). The function expects a distance based edge-list generated by
#' `edge_dist` (optionally with directions measured by
#' `direction_step` and `edge_direction`), zone thresholds, labels
#' and (optionally) a blind volume.
#'
#' `edge_zones` uses interindividual distances, and optionally directions,
#' to assign neighboring individuals to a focal individual's behavioural zones.
#' The user provides zone thresholds (eg. 25 m, 100 m, 250 m) along with zone
#' labels (eg. zone of repulsion, zone of orientation, zone of attraction),
#' according to their objectives, study species and system. The optional blind
#' volume can be provided to define a range of interindividual directions that
#' correspond to the limits of the focal individual's perception.
#'
#' Two workflows for this function exist, depending on if the blind volume
#' argument is used:
#'
#' a) If the blind volume is not provided, simply provide your distance based
#' edge-lists from `edge_dist` with zone thresholds and labels.
#'
#' b) If the blind volume is provided, the following order of functions is
#' expected to ensure the relevant direction columns are available:
#'    1. direction_step(DT)
#'    1. edges <- edge_dist(DT)
#'    1. dyad_id(edges)
#'    1. dyad_directions <- edge_direction(edges, DT)
#'    1. edge_zones(dyad_directions)
#'
#'  Interindividual distances are converted into behavioural zones using
#'  `cut`. The thresholds provided are used as cut points for a series of
#'  intervals that are open on the left and closed on the right, starting at 0.
#'  See details in [base::cut()].
#'
#'  The (optional) blind volume defines the range of interindividual directions
#'  between the focal individual (ID1) and the neighbour (ID2) that is outside
#'  of the focal individual's perception. The interindividual direction (column
#'  "direction_dyad" from `edge_direction`) is made relative to the focal
#'  individual's movement direction (column "direction" from
#'  `direction_step`). The argument blind_volume expects a single value to
#'  define a symmetrical window behind the focal individual's movement direction
#'  eg. where blind_volume = 2, the symmetrical window is from (-2, 2).
#'
#' The `edges` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' @param edges distance based edge-list generated by `edge_dist`,
#'   optionally with direction columns from `direction_step` and
#'   `edge_direction` (see Details)
#' @param zone_thresholds upper thresholds to define behavioural zones, eg.
#'   c(10, 20, 30) defines behavioural zones (0-10], (10-20], (20-30]
#' @param zone_labels labels for zones defined by zone_thresholds, must match
#'   zone_thresholds in length
#' @param blind_volume (optional) interindividual direction to define
#'   symmetrical window outside of focal individual's perception, eg. 2 becomes
#'   (-2, 2), see Details
#'
#' @return `edge_zones` returns the input `edges` appended with a zone
#'   column indicating the behavioural zone, using the zone label provided.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'   A message is returned when a zone column already exists in the input
#'   `edges`, because it will be overwritten.
#'
#' @export
#' @references
#'
#' The behavioural zones metric is defined in Couzin et al. 2002
#' (\doi{doi:10.1006/jtbi.2002.3065}).
#'
#' See examples of measuring behavioural zones:
#'  * \doi{doi:10.1073/pnas.1001763107}
#'  * \doi{doi:10.1371/journal.pcbi.1008832}
#'  * \doi{doi:10.3389/fphy.2021.715996}
#'
#' @family Edge-list generating functions
#' @family Distance functions
#' @family Direction functions
#' @seealso [edge_dist] [direction_step] [edge_direction]
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge list generation
#' edges <- edge_dist(
#'   DT,
#'   threshold = 100,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   timegroup = 'timegroup',
#'   returnDist = TRUE,
#'   fillNA = FALSE
#' )
#'
#' # Calculate behavioural zones
#' edge_zones(
#'   edges,
#'   zone_thresholds = c(25, 50, 75),
#'   zone_labels = c('repulsion', 'orientation', 'attraction')
#' )
#'
#' # Alternatively, if a user wants to specify a blind volume,
#' #   we need to measure directions
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Edge list generation
#' edges <- edge_dist(
#'   DT,
#'   threshold = 100,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   timegroup = 'timegroup',
#'   returnDist = TRUE,
#'   fillNA = FALSE
#' )
#'
#' # Generate dyad id
#' dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
#'
#' # Interindividual directions
#' dyad_directions <- edge_direction(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736,
#'   timegroup = 'timegroup'
#' )
#'
#' # Calculate behavioural zones
#' edge_zones(
#'   dyad_directions,
#'   zone_thresholds = c(25, 50, 75),
#'   zone_labels = c('repulsion', 'orientation', 'attraction'),
#'   blind_volume = 2
#' )
#' print(dyad_directions[, .SD[1:3], by = zone])
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' direction_step(DT, id = 'ID')
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
#' dyad_directions <- edge_direction(
#'   edges,
#'   DT,
#'   id = 'ID',
#'   timegroup = 'timegroup'
#' )
#' edge_zones(
#'   dyad_directions,
#'   zone_thresholds = c(25, 50, 75),
#'   zone_labels = c('repulsion', 'orientation', 'attraction'),
#'   blind_volume = 2
#' )
#' print(dyad_directions[, .SD[1:3], by = zone])
edge_zones <- function(
    edges = NULL,
    zone_thresholds = NULL,
    zone_labels = NULL,
    blind_volume = NULL) {
  # due to NSE notes in R CMD check
  distance <- direction_dyad <- direction <- direction_dyad_relative <- NULL
  direction <- 'direction'
  direction_dyad <- 'direction_dyad'
  distance <- 'distance'

  assert_not_null(edges)
  assert_is_data_table(edges)
  assert_not_null(zone_thresholds)
  assert_not_null(zone_labels)
  assert_are_colnames(edges, distance, ', did you run edge_dist?')
  assert_col_inherits(edges, distance, 'numeric')

  out_col <- 'zone'
  if (out_col %in% colnames(edges)) {
    message(paste(out_col, 'column will be overwritten by this function'))
    data.table::set(edges, j = out_col, value = NULL)
  }

  edges[, (out_col) := cut(distance, breaks = c(0, zone_thresholds),
                           labels = c(zone_labels))]

  if (!is.null(blind_volume)) {
    assert_are_colnames(edges, direction_dyad, ', did you use edge_direction?')
    assert_are_colnames(edges, direction, ', did you use direction_step?')
    assert_col_inherits(edges, direction_dyad, 'units', ', did you use direction_step?')
    assert_col_inherits(edges, direction, 'units', ', did you use direction_step?')
    assert_col_radians(edges, direction_dyad, ', did you use edge_direction?')
    assert_col_radians(edges, direction, ', did you use edge_direction?')

    edges[, direction_dyad_relative :=
            diff_rad(direction, direction_dyad, signed = TRUE)]

    edges[abs(direction_dyad_relative) > blind_volume & !is.na(get(out_col)),
          (out_col) := 'blind']

    data.table::set(edges, j = 'direction_dyad_relative', value = NULL)

  }

  return(edges)

}

assert_are_colnames <- function(x, nms, ...) {
  if (length(nms)) {
    for (nm in nms) {
      if(!nm %in% colnames(x)) {
        rlang::abort(paste0(nm, ' field provided is not present in input',
                            ...),
                     call = rlang::caller_env())
      } else {
        invisible(NULL)
      }
    }
  }
}

assert_col_inherits <- function(x, cols, classes, ...) {
  if (length(cols)) {
    for (col in cols) {
      if(!inherits(x[[col]], classes)) {
        rlang::abort(paste0(rlang::caller_arg(cols), ' must be of class ',
                            paste0(classes, collapse = '/'),
                            ...),
                     call = rlang::caller_env())
      } else {
        invisible(NULL)
      }
    }
  }
}

assert_col_radians <- function(x, col, ...) {
  if(!identical(units(x[[col]])$numerator, 'rad')) {
    rlang::abort(paste0(rlang::caller_arg(col),
                        ' must be of units radians', ...),
                 call = rlang::caller_env())
  } else {
    invisible(NULL)
  }
}

assert_col_typeof <- function(x, cols, type, ...) {
  if (length(cols)) {
    for (col in cols) {
      if(!identical(typeof(x[[col]]), type)) {
        rlang::abort(paste0(rlang::caller_arg(cols), ' must be of type ',
                            paste0(type, collapse = '/'),
                            ...),
                     call = rlang::caller_env())
      } else {
        invisible(NULL)
      }
    }
  }
}

assert_inherits <- function(x, classes, ...) {
  if(!inherits(x, classes)) {
    rlang::abort(paste0(rlang::caller_arg(x), ' must be of class ',
                        paste0(classes, collapse = '/'),
                        ...),
                 call = rlang::caller_env())
  } else {
    invisible(NULL)
  }
}

assert_is_data_table <- function(x) {
  if (!data.table::is.data.table(x)) {
    rlang::abort(paste0('input ', rlang::caller_arg(x),
                        ' must be a data.table'),
                 call = rlang::caller_env())
  }
  if (nrow(x) == 0L) {
    rlang::abort(paste0('input ', rlang::caller_arg(x), ' has zero rows'),
                 call = rlang::caller_env())
  }
  return(invisible(NULL))
}

assert_length <- function(x, len) {
  if (length(x) != len) {
    rlang::abort(paste0(rlang::caller_arg(x), ' must be length ', len),
                 call = rlang::caller_env())
  } else {
    return(invisible(NULL))
  }
}

assert_not_missing <- function(x) {
  if (missing(x)) {
    rlang::abort(paste0(rlang::caller_arg(x), ' must be provided'),
                 call = rlang::caller_env())
  }
  return(invisible(NULL))
}

assert_not_null <- function(x, ...) {
  if (is.null(x)) {
    rlang::abort(paste0(rlang::caller_arg(x), ' must be provided', ...),
                 call = rlang::caller_env())
  }
  return(invisible(NULL))
}

assert_relation <- function(x, fun, y, ..., n = 1) {
  if (!(fun(x, y))) {
    rlang::abort(
      paste0(
        rlang::caller_arg(x),
        ' must be ',
        rlang::caller_arg(fun),
        ' ',
        y,
        ...
      ),
      call = rlang::caller_env(n = n)
    )
  }
  return(invisible(NULL))
}

assert_threshold <- function(threshold = NULL, crs = NULL) {
  if (inherits(threshold, 'units')) {
    if (any(is.null(crs), is.na(crs))) {
      assert_relation(threshold, `>`, units::as_units(0, units(threshold)))
    } else {
      assert_units_match(threshold, sf::st_crs(crs)$ud_unit, n = 2)
      assert_relation(threshold, `>`, units::as_units(0, units(threshold)))
    }
  } else if (inherits(threshold, 'numeric')){
    if (any(is.null(crs), is.na(crs))) {
      assert_relation(threshold, `>`, 0)
    } else {
      threshold <- units::as_units(threshold,
                                   units(sf::st_crs(crs)$ud_unit))
      assert_relation(threshold, `>`,
        units::as_units(0, units(sf::st_crs(crs)$ud_unit)),
        n = 2
      )
    }
    return(invisible(NULL))
  } else if (is.null(threshold)) {
    return(invisible(NULL))
  } else {
    rlang::abort('threshold must be of class numeric, units, or NULL')
  }
}

assert_units_match <- function(x, y, n = 1) {
  if (isFALSE(identical(units(x), units(y)))) {
    rlang::abort(
      paste0(
        'units of ',
        rlang::caller_arg(x),
        ' (', units(x), ')',
        ' do not match units of ',
        rlang::caller_arg(y),
        ' (', units(y), ')'
      ),
      call = rlang::caller_env(n = n)
    )
  }
  return(invisible(NULL))
}

#' Calculate centroid
#'
#' **Internal function** - not developed to be used outside of spatsoc functions
#'
#' Calculate centroid for one of:
#' - geometry
#' - the points in x, y
#' - the pairwise points in geometry_a and geometry_b
#' - the pairwise points in x_a, y_a and x_b, y_b
#'
#' @param geometry sfc (simple feature geometry list column) from
#' [get_geometry()]
#' @param x X coordinate column, numeric
#' @param y Y coordinate column, numeric
#' @param crs crs for x, y coordinates, ignored for geometry argument
#' @param use_mean boolean predetermine if centroid calculated via mean
#'
#' @returns
#'
#' The underlying centroid function used depends on the crs of the coordinates
#' or geometry provided.
#'
#'  - If the crs is longlat degrees (as determined by
#' [sf::st_is_longlat()]) and [sf::sf_use_s2()] is TRUE, the distance function
#' is [sf::st_centroid()] which passes to [s2::s2_centroid()].
#'  - If the crs is longlat degrees but [sf::sf_use_s2()] is FALSE, the centroid
#' calculated will be incorrect. See [sf::st_centroid()].
#'  - If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
#' centroid function used is mean.
#'
#' Note: if the input is length 1, the input is returned.
#'
#' @keywords internal
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{
#' data.table::setDTthreads(1)
#' }
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' DT[, spatsoc:::calc_centroid(x = X, y = Y, crs = 32736)]
#'
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#'
#' # Calculating centroids requires recomputing the bbox when:
#' #  1- by = , 2- providing geometry
#' DT[, centroid := spatsoc:::calc_centroid(geometry), by = ID]
#' DT[, centroid := sf::st_sfc(centroid, recompute_bbox = TRUE)]
#' plot(DT$centroid)
calc_centroid <- function(geometry, x, y, crs, use_mean = FALSE) {
  if (isFALSE(use_mean)) {
    if (!missing(geometry) && missing(x) && missing(y) && missing(crs)) {
      if (identical(length(geometry), 1L)) {
        return(sf::st_as_sf(geometry))
      } else {
        sf::st_as_sf(sf::st_centroid(sf::st_combine(geometry)))
      }
    } else if (missing(geometry) && !missing(x) && !missing(y) &&
               !missing(crs)) {
      if (identical(length(x), 1L) & identical(length(y), 1L)) {
        return(data.frame(x, y))
      } else {
        data.frame(sf::st_coordinates(sf::st_centroid(sf::st_combine(
          sf::st_as_sf(
            data.frame(x, y),
            crs = crs,
            coords = seq.int(2),
            na.fail = FALSE
          )
        ))))
      }
    } else {
      rlang::abort(c(
        'arguments incorrectly provided, use one of the following combinations:',
        '1. geometry',
        '2. x, y, crs'
      ))
    }
  } else {
    if (!missing(geometry) && missing(x) && missing(y) && missing(crs)) {
      if (identical(length(geometry), 1L)) {
        return(sf::st_as_sf(geometry))
      } else {
        sf::st_as_sf(
          data.frame(apply(sf::st_coordinates(geometry), 2, mean, na.rm = TRUE,
                           simplify = FALSE)), coords = seq.int(2), crs = crs)
      }
    } else if (missing(geometry) && !missing(x) && !missing(y) && !missing(crs)) {
      if (identical(length(x), 1L) & identical(length(y), 1L)) {
        return(data.frame(x, y))
      } else {
        data.frame(mean(x, na.rm = TRUE), mean(y, na.rm = TRUE))
      }
    } else {
      rlang::abort(c(
        'arguments incorrectly provided, use one of the following combinations:',
        '1. geometry',
        '2. x, y, crs'
      ))
    }
  }
}

#' Calculate direction
#'
#' **Internal function** - not developed to be used outside of spatsoc functions
#'
#' Calculate direction using [lwgeom::st_geod_azimuth()] between one of:
#' - the sequence of points in geometry_a
#' - the sequence of points in x_a, y_a
#' - the pairwise points in geometry_a and geometry_b
#' - the pairwise points in x_a, y_a and x_b, y_b
#'
#' Requirements:
#' - matching length between a and b objects if b provided
#' - crs is provided. If crs is not longlat (as determined by
#' [sf::st_is_longlat()]), the coordinates or geometry will be transformed to
#'  `sf::st_crs(4326)`
#' - no missing values in coordinates
#'
#' @param geometry_a,geometry_b sfc (simple feature geometry list column) from
#' [get_geometry()]
#' @param x_a,x_b X coordinate column, numeric
#' @param y_a,y_b Y coordinate column, numeric
#' @param crs crs for x_a, y_a (and if provided, x_b, y_b) coordinates,
#' ignored for geometry_a and geometry_b arguments
#' @param use_transform boolean predetermine if coordinates/geometry need transform
#'
#' @returns
#'
#' Direction in units of radians, in range of pi, -pi where North = 0.
#'
#' @keywords internal
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Example result for East, North, West, South directions
#' example <- data.table(
#'   X = c(0, 5, 5, 0, 0),
#'   Y = c(0, 0, 5, 5, 0)
#' )
#' # E, N, W, S
#' example[, spatsoc:::calc_direction(x_a = X, y_a = Y, crs = 4326, use_transform = FALSE)]
calc_direction <- function(
    geometry_a, geometry_b,
    x_a, y_a,
    x_b, y_b,
    crs,
    use_transform) {
  crs_longlat <- 4326

  if (!missing(geometry_a) && missing(x_a) && missing(y_a)
      && missing(x_b) && missing(y_b)) {
    if(any(rowSums(is.na(sf::st_coordinates(geometry_a))) == 2)) {
      rlang::abort('missing values in coordinates')
    }
    if (!missing(geometry_b)) {
      if(any(rowSums(is.na(sf::st_coordinates(geometry_b))) == 2)) {
        rlang::abort('missing values in coordinates')
      }
      if (use_transform) {
        lwgeom::st_geod_azimuth(
          x = sf::st_transform(
            geometry_a,
            crs = crs_longlat
          ),
          y = sf::st_transform(
            geometry_b,
            crs = crs_longlat
          )
        )
      } else {
        lwgeom::st_geod_azimuth(geometry_a, geometry_b)
      }
    } else {
      if (use_transform) {
        lwgeom::st_geod_azimuth(
          x = sf::st_transform(
            geometry_a,
            crs = crs_longlat
          )
        )
      } else {
        lwgeom::st_geod_azimuth(geometry_a)
      }
    }
  } else if (missing(geometry_a) && !missing(x_a) && !missing(y_a)) {
    if (!missing(x_b) && !missing(y_b)) {
      if (use_transform) {
        lwgeom::st_geod_azimuth(
          x = sf::st_transform(
            sf::st_as_sf(data.frame(x_a, y_a), crs = crs, coords = seq.int(2),
                         na.fail = TRUE),
            crs = crs_longlat
          ),
          y = sf::st_transform(
            sf::st_as_sf(data.frame(x_b, y_b), crs = crs, coords = seq.int(2),
                         na.fail = TRUE),
            crs = crs_longlat
          )
        )
      } else {
        lwgeom::st_geod_azimuth(
          x = sf::st_as_sf(data.frame(x_a, y_a), crs = crs, coords = seq.int(2),
                           na.fail = TRUE),
          y = sf::st_as_sf(data.frame(x_b, y_b), crs = crs, coords = seq.int(2),
                           na.fail = TRUE)
        )
      }
    } else {
      if (use_transform) {
        lwgeom::st_geod_azimuth(
          x = sf::st_transform(
            sf::st_as_sf(data.frame(x_a, y_a), crs = crs, coords = seq.int(2),
                         na.fail = TRUE),
            crs = crs_longlat
          )
        )
      } else {
        lwgeom::st_geod_azimuth(
          x = sf::st_as_sf(data.frame(x_a, y_a), crs = crs, coords = seq.int(2),
                           na.fail = TRUE)
        )
      }
    }
  } else {
    rlang::abort(c(
      'arguments incorrectly provided, use one of the following combinations:',
      '1. geometry_a',
      '2. geometry_a and geometry_b',
      '3. x_a, y_a, and crs',
      '4. x_a, y_a, x_b, y_b, and crs'
    ))
  }
}

#' Calculate distance
#'
#' **Internal function** - not developed to be used outside of spatsoc functions
#'
#' Calculate distance for one of the following combinations:
#' - the distance matrix of points in geometry_a
#' - the distance matrix of points in x_a, y_a
#' - the pairwise distance between points in geometry_a and geometry_b
#' - the pairwise distance between points in x_a, y_a and x_b, y_b
#'
#' Requirements:
#' - matching length between a and b objects if b provided
#'
#' @param geometry_a,geometry_b sfc (simple feature geometry list column)
#' from [get_geometry()]
#' @param x_a,x_b X coordinate column, numeric
#' @param y_a,y_b Y coordinate column, numeric
#' @param crs crs for x_a, y_a (and if provided, x_b, y_b) coordinates,
#' ignored for geometry_a and geometry_b arguments
#' @param use_dist boolean predetermine if distance calculated via dist
#'
#' @returns
#'
#'  The underlying distance function used depends on the crs of the coordinates
#'  or geometry provided.
#'
#'  - If the crs is longlat degrees (as determined by
#'  [sf::st_is_longlat()]), the distance function is [sf::st_distance()] which
#'  passes to [s2::s2_distance()] if [sf::sf_use_s2()] is TRUE and
#'  [lwgeom::st_geod_distance()] if [sf::sf_use_s2()] is FALSE. The distance
#'  returned has units set according to the crs.
#'
#'  - If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
#'  distance function used is [stats::dist()] (or Euclidean distance for
#'  pairwise distances), maintaining expected behaviour from previous versions.
#'  The distance returned does not have units set.
#'
#'
#'  Note: in both cases, if the coordinates are NA then the result will be NA.
#'
#' @keywords internal
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Example points
#' example <- data.table(
#'   X = c(0, 5, 5, 0, 0, NA_real_, 0,        NA_real_),
#'   Y = c(0, 0, 5, 5, 0, 0,        NA_real_, NA_real_)
#' )
#' # E, N, W, S
#' example[, spatsoc:::calc_distance(x_a = X, y_a = Y, crs = 4326, use_dist = FALSE)]
calc_distance <- function(
    geometry_a, geometry_b,
    x_a, y_a,
    x_b, y_b,
    crs,
    use_dist) {
  if (!missing(geometry_a) && missing(x_a) && missing(y_a) &&
    missing(x_b) && missing(y_b)) {
    if (!missing(geometry_b)) {
      # Pairwise
      if (use_dist) {
        calc_distance_pairwise(geometry_a, geometry_b)
      } else {
        sf::st_distance(geometry_a, geometry_b, by_element = TRUE)
      }
    } else {
      # Matrix
      if (use_dist) {
        coords <- sf::st_coordinates(geometry_a)
        m <- as.matrix(stats::dist(coords))
        m[is.na(coords[, 1]),] <- NA_real_
        m[, is.na(coords[, 2])] <- NA_real_
        m
      } else {
        sf::st_distance(geometry_a, by_element = FALSE)
      }
    }
  } else if (missing(geometry_a) && !missing(x_a) && !missing(y_a)) {
    if (!missing(x_b) && !missing(y_b)) {
      # Pairwise
      if (use_dist) {
        calc_distance_pairwise(x_a = x_a, y_a = y_a, x_b = x_b, y_b = y_b)
      } else {
        sf::st_distance(
          x = sf::st_as_sf(data.frame(x_a, y_a),
                           crs = crs, coords = seq.int(2),
                           na.fail = FALSE
          ),
          y = sf::st_as_sf(data.frame(x_b, y_b),
                           crs = crs, coords = seq.int(2),
                           na.fail = FALSE
          ),
          by_element = TRUE
        )
      }
    } else {
      # Matrix
      if (use_dist) {
        m <- as.matrix(stats::dist(cbind(x_a, y_a)))
        m[is.na(x_a),] <- NA_real_
        m[, is.na(y_a)] <- NA_real_
        m
      } else {
        sf::st_distance(
          x = sf::st_as_sf(data.frame(x_a, y_a),
                           crs = crs, coords = seq.int(2),
                           na.fail = FALSE
          ),
          by_element = FALSE
        )
      }
    }
  } else {
    rlang::abort(c(
      'arguments incorrectly provided, use one of the following combinations:',
      '1. geometry_a',
      '2. geometry_a and geometry_b',
      '3. x_a, y_a, and crs',
      '4. x_a, y_a, x_b, y_b, and crs'
    ))
  }
}

# Internal calc pairwise distance used in internal calc_distance
calc_distance_pairwise <- function(geometry_a, geometry_b,
                          x_a, y_a,
                          x_b, y_b) {
  if (!missing(geometry_a) && missing(x_a) && missing(y_a) &&
      !missing(geometry_b)) {
    a <- sf::st_coordinates(geometry_a)
    b <- sf::st_coordinates(geometry_b)
    sqrt((a[, 1] - b[, 1]) ^ 2 + (a[, 2] - b[, 2]) ^ 2)
  } else if (missing(geometry_a) && !missing(x_a) && !missing(y_a) &&
             !missing(x_b) && !missing(y_b)) {
    sqrt((x_a - x_b) ^ 2 + (y_a - y_b) ^ 2)
  }
}

# Internal decider for calc_centroid
crs_use_mean <- function(crs) {
  if (isTRUE(sf::st_is_longlat(crs))) {
    if (sf::sf_use_s2()) {
      use_mean <- FALSE
    } else {
      warning('st_centroid does not give correct centroids for longlat',
              '\nsee ?sf::st_centroid')
      use_mean <- TRUE
    }
  } else {
    use_mean <- TRUE
  }
  return(use_mean)
}


#' Difference of two angles measured in radians
#'
#' **Internal function** - not developed to be used outside of spatsoc functions
#'
#' @param x angle in radians
#' @param y angle in radians
#' @param signed logical if signed difference should be returned, default FALSE
#' @param return_units return difference with units = 'rad'
#'
#' @return Difference between x and y in radians. If signed is TRUE, the signed
#'   difference is returned. If signed is FALSE, the absolute difference is
#'   returned. Note: The difference is the smallest difference, eg. the
#'   difference between 2 rad and -2.5 rad is 1.78.
#' @references adapted from https://stackoverflow.com/a/7869457
#' @keywords internal
#' @examples
#' # Load data.table, units
#' library(data.table)
#' library(units)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Set order using data.table::setorder
#' setorder(DT, datetime)
#'
#' # Calculate direction
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Differences
#' spatsoc:::diff_rad(DT[1, direction], DT[2, direction])
#'
#' # Note smallest difference returned
#' spatsoc:::diff_rad(as_units(2, 'rad'), as_units(-2.5, 'rad'))
diff_rad <- function(x, y, signed = FALSE, return_units = FALSE) {
  if (!inherits(x, 'units') || units(x)$numerator != 'rad') {
    stop('units(x) is not radians')
  }
  if (!inherits(y, 'units') || units(y)$numerator != 'rad') {
    stop('units(y) is not radians')
  }

  d <- units::drop_units(y) - units::drop_units(x)
  d <- ((d + pi) %% (2 * pi)) - pi

  if (signed) {
    out <- d
  } else {
    out <- abs(d)
  }

  if (return_units) {
    return(units::as_units(out, 'rad'))
  } else {
    return(out)
  }
}

# Test calc_direction
context('test calc_direction')

library(sf)
sf_use_s2(TRUE)

DT <- fread('../testdata/DT.csv')
id <- 'ID'
coords <- c('X', 'Y')
crs <- 32736
crs_longlat <- 4326

get_geometry(DT, coords = coords, crs = crs)
get_geometry(DT, coords = coords, crs = crs, output_crs = 4326,
             geometry_colname = 'geometry_longlat')

DT[, dest_geometry_longlat := sf::st_centroid(sf::st_union(geometry_longlat))]

coords_longlat <- paste0(coords, '_longlat')
DT[, (coords_longlat) := as.data.table(sf::st_coordinates(geometry_longlat))]

dest_coords <- paste0('dest_', coords)
DT[, (dest_coords) := as.data.table(sf::st_coordinates(dest_geometry_longlat))]

test_that('arguments provided correctly else error', {
  expect_error(
    DT[, calc_direction(x_a = X, geometry_a = geometry)]
  )

  expect_error(
    DT[, calc_direction(x_b = X, geometry_b = geometry)]
  )

  expect_error(
    DT[, calc_direction(y_a = X, geometry_a = geometry)]
  )

  expect_error(
    DT[, calc_direction(x_b = X, geometry_a = geometry)]
  )
})

test_that('units are returned', {
  expect_s3_class(DT[, calc_direction(geometry_longlat, use_transform = FALSE)],
                  'units')
  expect_identical(DT[, units(
    calc_direction(geometry_longlat, use_transform = FALSE))$numerator],
    'rad'
  )

  expect_s3_class(
    DT[, calc_direction(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat,
                        use_transform = FALSE)],
    'units'
  )
  expect_identical(
    DT[, units(
      calc_direction(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat,
                     use_transform = FALSE)
    )$numerator],
    "rad"
  )
})

test_that('expected dims returned', {
  # N - 1 since missing start/dest for direction
  expect_length(DT[, calc_direction(geometry_longlat, use_transform = FALSE)],
                DT[, .N - 1])
  expect_length(
    DT[, calc_direction(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat,
                        use_transform = FALSE)],
    DT[, .N - 1]
  )

  expect_length(DT[, calc_direction(geometry_longlat, geometry_longlat,
                                    use_transform = FALSE)], DT[, .N])
  expect_length(
    DT[, calc_direction(x_a = X_longlat, y_a = Y_longlat,
                        x_b = X_longlat, y_b = Y_longlat, crs = crs_longlat,
                        use_transform = FALSE)],
    DT[, .N]
  )
})

test_that('expected range returned', {
  y <- units::set_units(pi, 'rad')
  expect_equal(DT[, min(calc_direction(geometry_longlat, use_transform = FALSE))], -y, tolerance = 0.1)
  expect_equal(DT[, max(calc_direction(geometry_longlat, use_transform = FALSE))], y, tolerance = 0.1)

  expect_equal(
    DT[, min(calc_direction(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat, use_transform = FALSE))],
    -y,
    tolerance = 0.1
  )
  expect_equal(
    DT[, max(calc_direction(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat, use_transform = FALSE))],
    y,
    tolerance = 0.1
  )
})

test_that('NAs returned as expected', {
  X_NA <- copy(DT)[seq.int(100)][sample(.N, 10), X_longlat := NA]
  expect_error(X_NA[, calc_direction(x_a = X_longlat, y_a = Y_longlat,
                                     crs = crs_longlat,
                                     use_transform = FALSE)],
               'missing values in coordinates')

  Y_NA <- copy(DT)[seq.int(100)][sample(.N, 10), Y_longlat := NA]
  expect_error(Y_NA[, calc_direction(x_a = X_longlat, y_a = Y_longlat,
                                     crs = crs_longlat,
                                     use_transform = FALSE)],
               'missing values in coordinates')

  XY_NA <- copy(DT)[seq.int(100)][sample(.N, 10), (coords_longlat) := NA]
  expect_error(XY_NA[, calc_direction(x_a = X_longlat, y_a = Y_longlat,
                                      crs = crs_longlat,
                                      use_transform = FALSE)],
               'missing values in coordinates')

  get_geometry(XY_NA, coords_longlat, crs_longlat)
  expect_error(XY_NA[, calc_direction(geometry,
                                      use_transform = FALSE)],
               'missing values in coordinates')
  expect_error(XY_NA[, calc_direction(geometry, geometry_b = geometry,
                                      use_transform = FALSE)],
               'missing values in coordinates')

  geometry_NA <- copy(DT)[seq.int(100)]
  geometry_NA[, geometry_longlat_miss := geometry]
  geometry_NA[sample(.N, 10), geometry_longlat_miss := st_sfc(st_point())]
  expect_error(geometry_NA[, calc_direction(geometry_longlat_miss)],
               'missing values in coordinates')
  expect_error(geometry_NA[, calc_direction(geometry_longlat, geometry_longlat_miss)],
               'missing values in coordinates')
})

# Test diff_rad
context('test diff_rad')

library(units)

pt_01 <- as_units(0.1, 'rad')
pt_02 <- as_units(0.2, 'rad')
pi_rad <- as_units(pi, 'rad')

# Adapted from: https://gist.github.com/bradphelan/7fe21ad8ebfcb43696b8
expect_equal(diff_rad(pt_01, pt_02, TRUE, TRUE), pt_01)
expect_equal(diff_rad(pt_01, pt_02 + 2 * pi_rad, TRUE, TRUE), pt_01)
expect_equal(diff_rad(pt_01, pt_02 - 2 * pi_rad, TRUE, TRUE), pt_01)
expect_equal(diff_rad(pt_01 + 2 * pi_rad, pt_02, TRUE, TRUE), pt_01)
expect_equal(diff_rad(pt_01 - 2 * pi_rad, pt_02, TRUE, TRUE), pt_01)
expect_equal(diff_rad(pt_02, pt_01, TRUE, TRUE), -pt_01)
expect_equal(diff_rad(pt_02 + 2 * pi_rad, pt_01, TRUE, TRUE), -pt_01)
expect_equal(diff_rad(pt_02 - 2 * pi_rad, pt_01, TRUE, TRUE), -pt_01)
expect_equal(diff_rad(pt_02, pt_01 + 2 * pi_rad, TRUE, TRUE), -pt_01)
expect_equal(diff_rad(pt_02, pt_01 - 2 * pi_rad, TRUE, TRUE), -pt_01)

expect_equal(diff_rad(pt_01, pt_02, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_01, pt_02 + 2 * pi_rad, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_01, pt_02 - 2 * pi_rad, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_01 + 2 * pi_rad, pt_02, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_01 - 2 * pi_rad, pt_02, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_02, pt_01, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_02 + 2 * pi_rad, pt_01, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_02 - 2 * pi_rad, pt_01, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_02, pt_01 + 2 * pi_rad, FALSE, TRUE), pt_01)
expect_equal(diff_rad(pt_02, pt_01 - 2 * pi_rad, FALSE, TRUE), pt_01)

expect_error(diff_rad(pt_01, 0.2), 'rad')
expect_error(diff_rad(0.1, pt_02), 'rad')

# Test calc_centroid
context('test calc_centroid')

library(sf)
sf_use_s2(TRUE)

DT <- fread('../testdata/DT.csv')
coords <- c('X', 'Y')
crs <- 32736
crs_longlat <- 4326

get_geometry(DT, coords = coords, crs = crs)
get_geometry(DT, coords = coords, crs = crs, output_crs = 4326,
             geometry_colname = 'geometry_longlat')

DT[, dest_geometry := st_centroid(st_union(geometry))]

coords_longlat <- paste0(coords, '_longlat')
DT[, (coords_longlat) := as.data.table(st_coordinates(geometry_longlat))]

dest_coords <- paste0('dest_', coords)
DT[, (dest_coords) := as.data.table(st_coordinates(dest_geometry))]

test_that('arguments provided correctly else error', {
  expect_error(
    DT[, calc_centroid(x = X, geometry = geometry, use_mean = FALSE)],
    'arguments incorrectly provided'
  )

  expect_error(
    DT[, calc_centroid(y = Y, geometry = geometry, use_mean = TRUE)],
    'arguments incorrectly provided'
  )
})

test_that('data.table/numeric returned when coords provided, sf when geo', {
  expect_s3_class(DT[, calc_centroid(geometry)], 'sf')
  expect_s3_class(DT[, .(centroid = calc_centroid(geometry))][[1]], 'sfc')

  expect_s3_class(
    DT[, calc_centroid(x = X_longlat, y = Y_longlat, crs = crs_longlat)],
    'data.frame'
  )
  expect_s3_class(
    DT[, .(centroid = calc_centroid(x = X, y = Y, crs = crs))][1],
    'data.table'
  )

  expect_s3_class(
    DT[, calc_centroid(x = X, y = Y, crs = crs), by = ID],
    'data.table'
  )

  expect_type(
    DT[, calc_centroid(x = X, y = Y, crs = crs), by = ID]$X,
    'double'
  )
})

test_that('expected dims returned', {
  expect_length(DT[, calc_centroid(geometry)], 1L)
  expect_length(DT[, calc_centroid(x = X_longlat, y = Y_longlat, crs = crs_longlat)],
                2L)
  expect_equal(ncol(DT[, calc_centroid(x = X_longlat, y = Y_longlat, crs = crs_longlat)]),
               2L)
})

test_that('calc_centroid equals st_centroid for mean and length 1 inputs', {
  new_nms <- c('X', 'Y')

  i <- DT[, sample(.I, 1)]

  expect_equal(
    data.frame(setnames(DT[i, calc_centroid(x = X, y = Y, crs = crs, use_mean = TRUE)],
                        new = new_nms)),
    data.frame(DT[i, .(X, Y)])
  )

  expect_equal(
    data.frame(setnames(DT[i, calc_centroid(x = X, y = Y, crs = crs, use_mean = FALSE)],
                        new = new_nms)),
    data.frame(DT[i, .(X, Y)])
  )

  expect_equal(
    DT[i, calc_centroid(geometry = geometry, use_mean = FALSE)]$x,
    DT[i, sf::st_sf(geometry)]$geometry
  )

  expect_equal(
    DT[i, calc_centroid(geometry = geometry, use_mean = TRUE)]$x,
    DT[i, sf::st_sf(geometry)]$geometry
  )

  i_seq <- DT[, sample(.I, 100)]

  expect_equal(
    setnames(DT[i_seq, calc_centroid(x = X, y = Y, crs = crs, use_mean = TRUE)], new = new_nms),
    data.frame(st_coordinates(
      st_centroid(st_combine(st_as_sf(DT[i_seq, .(X, Y)], coords = seq.int(2), crs = crs)))
    ))
  )

  expect_equal(
    setnames(DT[i_seq, calc_centroid(x = X, y = Y, crs = NA_crs_, use_mean = TRUE)], new = new_nms),
    data.frame(st_coordinates(
      st_centroid(st_combine(st_as_sf(DT[i_seq, .(X, Y)], coords = seq.int(2), crs = NA_crs_)))
    ))
  )
})


test_that('crs_use_mean decides as expected', {
  expect_true(crs_use_mean(NA))
  expect_true(crs_use_mean(NULL))
  expect_true(crs_use_mean(32736))
  expect_false({sf::sf_use_s2(TRUE); crs_use_mean(4326)})
  expect_true(suppressWarnings({sf::sf_use_s2(FALSE); crs_use_mean(4326)}))
  expect_warning({sf::sf_use_s2(FALSE); crs_use_mean(4326)})
})


# These specific test results require recent commits to sf. Save for later.
# test_that('NAs returned as expected', {
#   X_NA <- copy(DT)[seq.int(100)][sample(.N, 10), X := NA]
#   res <- X_NA[, calc_centroid(x = X, y = Y, crs = crs)]
#   expect_length(res, 1L)
#   expect_false(any(is.na(st_coordinates(res))))
#
#   Y_NA <- copy(DT)[seq.int(100)][sample(.N, 10), Y := NA]
#   res <- Y_NA[, calc_centroid(x = X, y = Y, crs = crs)]
#   expect_length(res, 1L)
#   expect_false(any(is.na(st_coordinates(res))))
#
#   XY_NA <- copy(DT)[seq.int(100)][sample(.N, 10), (coords_longlat) := NA]
#   res <- XY_NA[, calc_centroid(x = X_longlat, y = Y_longlat, crs = 4326)]
#   expect_length(res, 1L)
#   expect_false(any(is.na(st_coordinates(res))))
#
#   get_geometry(XY_NA, coords_longlat, crs_longlat)
#   res <- XY_NA[, calc_centroid(geometry)]
#   expect_length(res, 1L)
#   expect_false(any(is.na(st_coordinates(res))))
# })

# Test calc_distance
context('test calc_distance')

library(sf)

DT <- fread('../testdata/DT.csv')
coords <- c('X', 'Y')
crs <- 32736
crs_longlat <- 4326

# Reduce test data size
DT <- DT[ID %in% c('A', 'B', 'C')]

get_geometry(DT, coords = coords, crs = crs)
get_geometry(DT, coords = coords, crs = crs, output_crs = 4326,
             geometry_colname = 'geometry_longlat')

DT[, dest_geometry := sf::st_centroid(sf::st_union(geometry))]
DT[, dest_geometry_longlat := sf::st_centroid(sf::st_union(geometry_longlat))]

coords_longlat <- paste0(coords, '_longlat')
DT[, (coords_longlat) := as.data.table(sf::st_coordinates(geometry_longlat))]

dest_coords <- paste0('dest_', coords)
DT[, (dest_coords) := as.data.table(sf::st_coordinates(dest_geometry))]

dest_coords_longlat <- paste0('dest_', coords, '_longlat')
DT[, (dest_coords_longlat) := as.data.table(sf::st_coordinates(dest_geometry_longlat))]

test_that('arguments provided correctly else error', {
  expect_error(
    DT[, calc_distance(x_a = X, geometry_a = geometry)]
  )

  expect_error(
    DT[, calc_distance(x_b = X, geometry_b = geometry)]
  )

  expect_error(
    DT[, calc_distance(y_a = X, geometry_a = geometry)]
  )

  expect_error(
    DT[, calc_distance(x_b = X, geometry_a = geometry)]
  )
})

test_that('units are returned when use_dist is FALSE else numeric', {
  expect_s3_class(DT[, calc_distance(geometry_longlat, use_dist = FALSE)],
                  'units')
  expect_s3_class(DT[, calc_distance(geometry_longlat,
                                     st_shift_longitude(geometry_longlat),
                                     use_dist = FALSE)],
                  'units')
  expect_s3_class(
    DT[, calc_distance(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat,
                       use_dist = FALSE)],
    'units'
  )

  expect_type(
    DT[, calc_distance(x_a = X, y_a = Y, crs = crs, use_dist = TRUE)],
    'double'
  )
  expect_type(
    DT[, calc_distance(x_a = X, y_a = Y, x_b = X + 100, y_b = Y + 100,
                       crs = crs, use_dist = TRUE)],
    'double'
  )
})

test_that('expected dims returned', {
  N <- 100
  expect_length(DT[seq.int(N), calc_distance(geometry_longlat, use_dist = FALSE)],
                N * N)
  expect_length(DT[seq.int(N), calc_distance(geometry_longlat, dest_geometry_longlat,
                                             use_dist = FALSE)],
                N)

  expect_length(
    DT[seq.int(N), calc_distance(x_a = X_longlat, y_a = Y_longlat,
                                 crs = crs_longlat, use_dist = FALSE)],
    N * N
  )
  expect_length(
    DT[seq.int(N), calc_distance(x_a = X_longlat, y_a = Y_longlat,
                                 x_b = dest_X_longlat, y_b = dest_Y_longlat,
                                 crs = crs_longlat,
                                 use_dist = FALSE)],
    N
  )

  expect_length(DT[seq.int(N), calc_distance(geometry, use_dist = TRUE)],
                N * N)
  expect_length(DT[seq.int(N), calc_distance(geometry, geometry[1],
                                             use_dist = TRUE)],
                N)

  expect_length(
    DT[seq.int(N), calc_distance(x_a = X_longlat, y_a = Y_longlat, crs = crs_longlat,
                                 use_dist = TRUE)],
    N * N
  )
  expect_length(
    DT[seq.int(N), calc_distance(x_a = X_longlat, y_a = Y_longlat,
                                 x_b = dest_X_longlat, y_b = dest_Y_longlat,
                                 crs = crs_longlat,
                                 use_dist = TRUE)],
    N
  )
})

test_that('expected range returned', {
  expect_gte(DT[, min(calc_distance(geometry, use_dist = TRUE))],
             0)
  expect_gte(DT[, min(calc_distance(x_a = X, y_a = Y, crs = crs,
                                    use_dist = TRUE))],
             0)
  expect_gte(DT[, min(calc_distance(geometry, geometry_b = dest_geometry,
                                    use_dist = FALSE))],
             units::set_units(0, 'm'))
  expect_gte(DT[, min(calc_distance(x_a = X_longlat, y_a = Y_longlat,
                                    x_b = dest_X_longlat, y_b = dest_Y_longlat,
                                    crs = crs_longlat,
                                    use_dist = FALSE))],
             units::set_units(0, 'm'))
})

test_that('NAs returned as expected', {
  # NAs returned if use_dist = FALSE, in recent updates to sf
  # NAs are not returned if use_dist = TRUE, unless X and Y are both NA
  # This is handled in group_pts by subsetting where X and Y are not NA

  X_NA <- copy(DT)[seq.int(100)][sample(.N, 10), X := NA]
  res <- X_NA[, calc_distance(x_a = X, y_a = Y, crs = crs, use_dist = TRUE)]
  expect_length(res, nrow(X_NA) * nrow(X_NA))
  expect_true(any(is.na(X_NA$X)))
  # expect_true(any(is.na(res)))

  Y_NA <- copy(DT)[seq.int(100)][sample(.N, 10), Y := NA]
  res <- Y_NA[, calc_distance(x_a = X, y_a = Y, crs = crs, use_dist = TRUE)]
  expect_length(res, nrow(Y_NA) * nrow(Y_NA))
  expect_true(any(is.na(Y_NA$Y)))
  # expect_true(any(is.na(res)))

  X_NA <- copy(DT)[seq.int(100)][sample(.N, 10), X_longlat := NA]
  res <- X_NA[, calc_distance(x_a = X_longlat, y_a = Y_longlat,
                              crs = crs_longlat, use_dist = FALSE)]
  expect_length(res, nrow(X_NA) * nrow(X_NA))
  expect_true(any(is.na(X_NA$X_longlat)))
  # expect_true(any(is.na(res)))

  Y_NA <- copy(DT)[seq.int(100)][sample(.N, 10), Y := NA]
  res <- Y_NA[, calc_distance(x_a = X, y_a = Y, crs = crs, use_dist = FALSE)]
  expect_length(res, nrow(Y_NA) * nrow(Y_NA))
  expect_true(any(is.na(Y_NA$Y)))
  # expect_true(any(is.na(res)))

  XY_NA <- copy(DT)[seq.int(100)][sample(.N, 10), (coords_longlat) := NA]
  res <- XY_NA[, calc_distance(x_a = X_longlat, y_a = Y_longlat, crs = 4326,
                               use_dist = FALSE)]
  expect_length(res, nrow(XY_NA) * nrow(XY_NA))
  expect_true(any(is.na(c(XY_NA$X_longlat, XY_NA$Y_longlat))))
  expect_true(any(is.na(res)))

  get_geometry(XY_NA, coords_longlat, crs_longlat)
  res <- XY_NA[, calc_distance(geometry, use_dist = FALSE)]
  expect_length(res, nrow(XY_NA) * nrow(XY_NA))
  expect_true(any(sf::st_is_empty(XY_NA$geometry)))
  # expect_true(any(is.na(res)))
})

# Test assert family
context('test assert_*')

test <- data.table(X = LETTERS, Y = LETTERS,
                   ANGLE = runif(length(LETTERS), 0, 180))
units(test$ANGLE) <- 'degrees'


test_that('assert_not_null', {
  expect_error(assert_not_null(NULL))
})

test_that('assert_is_data_table', {
  expect_error(assert_is_data_table(data.frame()))
  expect_error(assert_is_data_table(data.table()))
})

test_that('assert_are_colnames', {
  expect_error(assert_are_colnames(test, 'Z'))
  expect_error(assert_are_colnames(test, c('X', 'Z')))
})

test_that('assert_col_inherits', {
  expect_error(assert_col_inherits(test, 'X', 'numeric'))
  expect_error(assert_col_inherits(test, c('X', 'Y'), 'numeric'))
})

test_that('assert_col_typeof', {
  expect_error(assert_col_typeof(test, 'X', 'units'))
  expect_null(assert_col_typeof(test, 'X', 'character'))
})

test_that('assert_inherits', {
  expect_error(assert_inherits(42, 'character'))
  expect_error(assert_inherits(TRUE, 'numeric'))
})


test_that('assert_col_radians', {
  expect_error(assert_col_radians(test, 'ANGLE',
                                  ', did you use direction_step?'))
})
test_that('assert_length', {
  expect_error(assert_length(1:2, 3L))
})
test_that('assert_relation', {
  expect_error(assert_relation(1, '>', 3))
})

test_that('assert_not_missing', {
  expect_error((function(x) assert_not_missing(x))())
})

test_that('assert_threshold', {
  expect_error(assert_threshold(0), '> 0')
  expect_error(assert_threshold(-1), '> 0')
  expect_silent(assert_threshold(1))
  expect_error(assert_threshold(units::as_units(-1, 'm')), '> 0')
  expect_silent(assert_threshold(units::as_units(1, 'degree'), 4326))
  expect_error(assert_threshold(units::as_units(1, 'm'), 4326),
               'do not match')
  expect_silent(assert_threshold(units::as_units(1, 'm'), 32649))
  expect_silent(assert_threshold(NULL))
  expect_silent(assert_threshold(100, 4326))
  expect_error(assert_threshold(units::as_units(-1, 'm'), 32736))
})

test_that('assert_units_match', {
  expect_error(assert_units_match(units::as_units(1, 'm'),
                                  units::as_units(1, 'km')))
})

#' Group times
#'
#' `group_times` groups rows into time groups. The function expects date
#' time formatted data and a threshold argument. The threshold argument is used
#' to specify a time window within which rows are grouped.
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()].
#'
#' The `datetime` argument expects the name of a column in `DT` which
#' is of type `POSIXct` or the name of two columns in `DT` which are
#' of type `IDate` and `ITime`.
#'
#' `threshold` must be provided in units of minutes, hours or days. The
#' character string should start with an integer followed by a unit, separated
#' by a space. It is interpreted in terms of 24 hours which poses the following
#' limitations:
#'
#' \itemize{ \item minutes, hours and days cannot be fractional \item minutes
#' must divide evenly into 60 \item minutes must not exceed 60 \item minutes,
#' hours which are nearer to the next day, are grouped as such \item hours must
#' divide evenly into 24 \item multi-day blocks should divide into the range of
#' days, else the blocks may not be the same length }
#'
#' In addition, the `threshold` is considered a fixed window throughout the
#' time series and the rows are grouped to the nearest interval.
#'
#' If `threshold` is NULL, rows are grouped using the `datetime`
#' column directly.
#'
#' @return `group_times` returns the input `DT` appended with a
#'   `timegroup` column and additional temporal grouping columns to help
#'   investigate, troubleshoot and interpret the timegroup.
#'
#'   The actual value of `timegroup` is arbitrary and represents the
#'   identity of a given `timegroup` which 1 or more individuals are
#'   assigned to. If the data was reordered, the group may change, but the
#'   contents of each group would not.
#'
#'   The temporal grouping columns added depend on the `threshold`
#'   provided:
#'
#'   \itemize{ \item `threshold` with unit minutes: "minutes" column added
#'   identifying the nearest minute group for each row. \item `threshold`
#'   with unit hours: "hours" column added identifying the nearest hour group
#'   for each row. \item `threshold` with unit days: "block" columns added
#'   identifying the multiday block for each row. }
#'
#'   A message is returned when any of these columns already exist in the input
#'   `DT`, because they will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#'
#' @inheritParams group_pts
#' @param datetime name of date time column(s). either 1 POSIXct or 2 IDate and
#'   ITime. e.g.: 'datetime' or c('idate', 'itime')
#' @param threshold threshold for grouping times. e.g.: '2 hours', '10 minutes',
#'   etc. if not provided, times will be matched exactly. Note that provided
#'   threshold must be in the expected format: '## unit'
#'
#' @export
#'
#' @family Temporal grouping
#' @seealso `group_pts` `group_lines`
#'   `group_polys`
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' group_times(DT, datetime = 'datetime', threshold = '5 minutes')
#'
#' group_times(DT, datetime = 'datetime', threshold = '2 hours')
#'
#' group_times(DT, datetime = 'datetime', threshold = '10 days')
#'
group_times <- function(
    DT = NULL,
    datetime = NULL,
    threshold = NULL) {
  # due to NSE notes in R CMD check
  minutes <- block <- hours <- itime <- . <- idate <- timegroup <- NULL
  minday <- maxday <- rangeday <- adjIDate <- adjHour <-  NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(datetime)
  assert_are_colnames(DT, datetime)

  checkCols <- c('hours', 'minutes', 'block', 'timegroup')

  if (any(checkCols %in% colnames(DT))) {
    message(paste0(
      paste(as.character(intersect(
        colnames(DT), checkCols
      )), collapse = ', '),
      ' columns found in input DT and will be overwritten by this function'
    ))
    data.table::set(DT, j = intersect(colnames(DT), checkCols), value = NULL)
  }

  if (is.null(threshold)) {
    message('no threshold provided, using the time field directly to group')
    DT[, timegroup := .GRP, by = c(datetime)]
    return(DT[])
  } else {
    if (length(datetime) == 1 &&
        'POSIXct' %in% unlist(lapply(DT[, .SD, .SDcols = datetime], class))) {
      dtm <-
        DT[, cbind(.SD[[1]], data.table::IDateTime(.SD[[1]])),
           .SDcols = datetime]
      data.table::setnames(dtm, c(datetime, 'idate', 'itime'))
    } else if (length(datetime) == 2 &&
               all(c('IDate', 'ITime') %in%
                   unlist(
                     lapply(DT[, .SD, .SDcols = datetime], class)))) {
      dtm <- DT[, .SD, .SDcols = datetime]
      data.table::setnames(dtm, c('idate', 'itime'))
    } else {
      stop(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'time field provided must be
          either 1 column: POSIXct or 2 columns: IDate and ITime'
        )
      )
    }

    if (grepl('hour', threshold) &&
        data.table::tstrsplit(threshold, ' ')[[1]] == 1L) {
      threshold <- '60 minutes'
    }

    if (grepl('hour', threshold)) {
      nHours <- data.table::tstrsplit(threshold, ' ',
                                      type.convert = TRUE)[[1]]
      if (!is.integer(nHours)) {
        if (nHours %% 1 != 0) {
          warning(
            strwrap(
              prefix = " ",
              initial = "",
              x = 'number of hours provided
              cannot be a fractional - threshold will be rounded'
            )
          )
        }
        nHours <- as.integer(nHours)
      }

      if (24 %% nHours != 0) {
        stop(
          strwrap(
            prefix = " ",
            initial = "",
            x = 'number of hours provided
                 does not evenly divide into 24'
          )
        )
      }

      dtm[data.table::hour(itime) %% nHours < (nHours / 2),
          hours := nHours * (data.table::hour(itime) %/% nHours)]
      dtm[data.table::hour(itime) %% nHours >= (nHours / 2),
          hours := nHours * ((data.table::hour(itime) %/% nHours) + 1L)]

      dtm[, adjIDate := idate]
      dtm[hours == 24, c('adjIDate', 'hours') := .(idate + 1, 0)]

      dtm[, timegroup := .GRP, by = .(hours, adjIDate)]
      data.table::set(dtm, j = 'adjIDate', value = NULL)
      data.table::set(dtm, j = c('idate', 'itime'), value = NULL)
      DT[, (colnames(dtm)) := dtm]

      return(DT[])
    } else if (grepl('minute', threshold)) {
      nMins <- data.table::tstrsplit(threshold, ' ',
                                     type.convert = TRUE)[[1]]
      if (!is.integer(nMins)) {
        if (nMins %% 1 != 0) {
          warning(
            strwrap(
              prefix = " ",
              initial = "",
              x = 'number of minutes provided
              cannot be a fractional - threshold will be rounded'
            )
          )
        }
        nMins <- as.integer(nMins)
      }

      assert_relation(nMins, `<=`, 60, ... = ' minutes')

      if (60 %% nMins != 0) {
        stop('threshold not evenly divisible by 60')
      }

      dtm[data.table::minute(itime) %% nMins < (nMins / 2) ,
          minutes := nMins * (data.table::minute(itime) %/% nMins)]
      dtm[data.table::minute(itime) %% nMins >= (nMins / 2),
          minutes := nMins * ((data.table::minute(itime) %/% nMins) + 1L)]

      dtm[, c('adjMinute', 'adjHour', 'adjDate') :=
            .(minutes, data.table::hour(itime), idate)]

      dtm[minutes == 60L,
          c('adjMinute', 'adjHour') :=
            .(0L, adjHour + 1L)]

      dtm[adjHour == 24L,
      c('adjMinute', 'adjHour', 'adjDate') :=
        .(0L, 0L, idate + 1)]

      dtm[, timegroup := .GRP,
          by = c('adjMinute', 'adjHour', 'adjDate')]

      data.table::set(dtm, j = c('adjMinute', 'adjHour', 'adjDate'), value = NULL)
      data.table::set(dtm, j = c('idate', 'itime'), value = NULL)

      DT[, (colnames(dtm)) := dtm]
      return(DT[])
    } else if (grepl('day', threshold)) {
      nDays <- data.table::tstrsplit(threshold, ' ')[[1]]
      if (!is.integer(nDays)) {
        nDays <- as.integer(nDays)
      }
      if (nDays == 1) {
        dtm[, timegroup := .GRP,
            by = .(data.table::year(idate), data.table::yday(idate))]

        data.table::set(dtm, j = c('idate', 'itime'), value = NULL)
        DT[, (colnames(dtm)) := dtm]
        return(DT[])
      } else {

        minday <- dtm[, min(data.table::yday(idate))]
        maxday <- dtm[, max(data.table::yday(idate))]
        rangeday <- dtm[, maxday - minday]

        if (!(rangeday %% nDays == 0)) {
          warning(
            strwrap(
              prefix = " ",
              initial = "",
              x = paste0('the minimum and maximum days in
              DT are not evenly divisible by the provided block length'
              ))
          )
        }

        dtm[, block := cut(
          data.table::yday(idate),
          breaks = seq.int(minday, maxday + nDays, by = nDays),
          right = FALSE,
          labels = FALSE
        )]

        dtm[, timegroup := .GRP, .(year(idate), block)]
        data.table::set(dtm, j = c('idate', 'itime'),
            value = NULL)
        DT[, (colnames(dtm)) := dtm]
        return(DT[])
      }
    } else {
      stop("must provide threshold in units of hour, day, or minute")
    }
  }
}

# Test group_times
context('test group_times')

library(spatsoc)

DT <- fread('../testdata/DT.csv', tz = 'UTC')

test_that('DT is required', {
  expect_error(group_times(DT = NULL,
                           datetime = NULL, threshold = '10 minutes'),
               'DT must be provided')
})


test_that('time field correctly provided or error detected', {
  expect_error(group_times(DT, datetime = NULL, threshold = '10 minutes'),
               'datetime must be')

  expect_error(group_times(DT,
                           datetime = 'potato',
                           threshold = '10 minutes'),
               'potato field provided')
})

test_that('threshold properly provided', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_message(
    group_times(copyDT, datetime = 'datetime',
                threshold = NULL),
    'no threshold provided',
    fixed = FALSE
  )

  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_error(
    group_times(copyDT, datetime = 'datetime',
                threshold = '50 potato'),
    'must provide threshold in units of hour, day, or minute'
  )
})


test_that('time fields are already present', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')
  expect_message(group_times(copyDT, datetime = 'datetime',
                             threshold = '10 minutes'),
                 'columns found in input', fixed = FALSE)
})

test_that('time field is appropriate format', {
  # where character is provided
  copyDT <- copy(DT)
  expect_error(
    group_times(copyDT, datetime = 'ID', threshold = '60 minutes'),
    'time field provided must be either',
    fixed = FALSE
  )

  # where numeric is provided
  copyDT <- copy(DT)
  copyDT[, datetimenumeric := 1]
  expect_error(group_times(copyDT, datetime = 'datetimenumeric',
                          threshold = '60 minutes'),
               'time field provided must be either', fixed = FALSE)


})

test_that('threshold with minutes fails with > 60', {
  copyDT <- copy(DT)[, c('idate', 'itime') := IDateTime(datetime)]
  expect_error(group_times(copyDT, datetime = c('idate', 'itime'),
                           threshold = '70 minutes'),
               '<= 60', fixed = FALSE)
})

test_that('threshold with minutes fails if not divisible by 60', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_error(
    group_times(copyDT, datetime = 'datetime', threshold = '13 minutes'),
    'threshold not evenly',
    fixed = FALSE
  )
})

test_that('threshold provided must be in units of hours, minutes, days', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_error(group_times(copyDT,
                           datetime = 'datetime',
                           threshold = '13 potatoes'),
               'must provide threshold in units', fixed = FALSE)
})

test_that('check that 60 minutes and 1 hour are the same result', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]

  expect_equal({
    copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
    group_times(copyDT, datetime = 'datetime',
               threshold = '1 hour')
  },
  {
    copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
    group_times(copyDT, datetime = 'datetime',
               threshold = '60 minutes')
  })
})

test_that('warns if block is not even', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]

  if (copyDT[, max(data.table::yday(datetime)) -
             min(data.table::yday(datetime))] %% 13 == 0) {
    blockLength <- '17 days'
  } else {
    blockLength <- '13 days'
  }

  expect_warning(group_times(copyDT, datetime = 'datetime',
                            threshold = blockLength),
                 'the minimum and maximum days', fixed = FALSE)

})


test_that('timegroup column + time fields are added to result', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_true('timegroup' %in%
                colnames(group_times(
                  copyDT, datetime = 'datetime',
                  threshold = '1 day'
                )))

  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  # to avoid block length warning
  expect_true(all(c('timegroup', 'block') %in%
                    colnames(suppressWarnings(
                      group_times(copyDT,
                                  datetime = 'datetime',
                                  threshold = '2 days')
                    ))))
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  copyDT <- copyDT[year(datetime) == unique(year(datetime))[1]]
  expect_true(all(c('timegroup', 'block') %in%
                    colnames(
                      group_times(copyDT,
                                  datetime = 'datetime',
                                  threshold = '2 days')
                    )))


  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_true(all(c('timegroup', 'hours') %in%
                    colnames(
                      group_times(copyDT, datetime = 'datetime',
                                 threshold = '2 hours')
                    )))

  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_true(all(c('timegroup', 'minutes') %in%
                    colnames(
                      group_times(copyDT, datetime = 'datetime',
                                 threshold = '10 minutes')
                    )))
})

test_that('timegroup column and fields are detected if already present', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  # to avoid block length warning
  suppressWarnings(
    group_times(copyDT, datetime = 'datetime', threshold = '2 days'))

  expect_message(
    group_times(copyDT, datetime = 'datetime', threshold = '1 day'),
    'block, timegroup ',
    fixed = FALSE
  )

  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '10 minutes')

  expect_message(
    group_times(copyDT, datetime = 'datetime', threshold = '10 minutes'),
    'minutes, timegroup ',
    fixed = FALSE
  )

  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  group_times(copyDT, datetime = 'datetime', threshold = '2 hours')

  expect_message(
    group_times(copyDT, datetime = 'datetime', threshold = '2 hours'),
    'hours, timegroup ',
    fixed = FALSE
  )
})

test_that('warns if no threshold provided', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]

  expect_message(
    group_times(copyDT, datetime = 'datetime'),
    'no threshold provided',
    fixed = FALSE
  )
})

test_that('warns if threshold is fractional', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]

  expect_warning(
    group_times(copyDT, datetime = 'datetime', threshold = '2.5 hours'),
    'number of hours provided cannot be a fractional',
    fixed = FALSE
  )

  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_warning(
    group_times(copyDT, datetime = 'datetime', threshold = '2.5 minutes'),
    'number of minutes provided cannot be a fractional',
    fixed = FALSE
  )

})

test_that('error if threshold not divisible by 24 hours', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]

  expect_error(
    group_times(copyDT, datetime = 'datetime', threshold = '5 hours'),
    'does not evenly divide into 24',
    fixed = FALSE
  )

})

test_that('multiyear blocks are well handled', {
  copyDT <- copy(DT)
  copyDT[, isoDate := as.POSIXct(
    ISOdatetime(2006, 10, 10, 10, 10, 10)
  )]
  expect_equal(group_times(copyDT, datetime = 'isoDate',
                           threshold = '1 day')[, .N,
                                                by = timegroup]$N,
               copyDT[, .N,
                by = .(data.table::yday(isoDate))]$N)

  copyDT[1, isoDate := as.POSIXct(
    ISOdatetime(2010, 10, 10, 10, 10, 10)
  )]
  copyDT[, timegroup := NULL]
  expect_equal(group_times(copyDT, datetime = 'isoDate',
                           threshold = '1 day')[, .N,
                                                by = timegroup]$N,
               copyDT[, .N, by = .(data.table::yday(isoDate),
                                   data.table::year(isoDate))]$N)

})


test_that('group_times nearest hour with minutes threshold', {
  DT <- data.table(
    idate = as.IDate('2018-12-12'),
    itime = as.ITime(
      c('01:00', '01:02', '00:58',
        '03:00', '03:02', '02:59',
        '05:00', '05:01', '04:58')
    )
  )

  expect_equal(
    group_times(DT, c('idate', 'itime'),
                threshold = '5 minutes')[, uniqueN(timegroup)],
               3)

  DT <- data.table(
    idate = as.IDate(
      c('2018-12-12','2018-12-12', '2018-12-12',
        '2018-12-12','2018-12-12', '2018-12-13',
        '2018-12-13','2018-12-13', '2018-12-13')),
    itime = as.ITime(
      c('22:00', '22:02', '21:58',
        '23:58', '23:59', '0:01',
        '2:00', '2:01', '1:58'))
  )

  expect_equal(
    group_times(DT, c('idate', 'itime'),
                threshold = '5 minutes')[, uniqueN(timegroup)],
    3)

})


test_that(
  'timegroups are based off years but blocks are consistent across all years', {
  copyDT <- copy(DT)
  copyDT[, datetime := as.POSIXct(datetime)]
  suppressWarnings(group_times(copyDT, 'datetime', '10 days'))
  blocks <- copyDT[, as.list(range(yday(datetime))),
                   .(block, year(datetime))]
  setkey(blocks, V1, V2)
  expect_true(
    nrow(foverlaps(blocks, blocks)) == nrow(blocks)
  )

  # need multi year data for same jul days to test if
  # n years by block > 1, but n years by timegroup == 1
})


test_that('datetime argument doesnt use datetime column', {
  copyDT <- copy(DT)
  # copyDT[, datetime := as.POSIXct(datetime)]
  copyDT[, date := as.IDate(datetime)]
  group_times(copyDT, 'date')
  expect_equal(
    copyDT[, uniqueN(date)],
    copyDT[, uniqueN(timegroup)]
  )

})

#' Data-stream randomizations
#'
#' `randomizations` performs data-stream social network randomization. The
#' function expects a `data.table` with relocation data, individual
#' identifiers and a randomization `type`. The `data.table` is
#' randomized either using `step` or `daily` between-individual
#' methods, or within-individual daily `trajectory` method described by
#' Spiegel et al. (2016).
#'
#' The `DT` must be a `data.table`. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()].
#'
#' Three randomization `type`s are provided: \enumerate{ \item step -
#' randomizes identities of relocations between individuals within each time
#' step. \item daily - randomizes identities of relocations between individuals
#' within each day. \item trajectory - randomizes daily trajectories within
#' individuals (Spiegel et al. 2016). }
#'
#' Depending on the `type`, the `datetime` must be a certain format:
#'
#' \itemize{ \item step - datetime is integer group created by
#' `group_times` \item daily - datetime is `POSIXct` format \item
#' trajectory - datetime is `POSIXct` format }
#'
#' The `id`, `datetime`,  (and optional `splitBy`) arguments
#' expect the names of respective columns in `DT` which correspond to the
#' individual identifier, date time, and additional grouping columns. The
#' `coords` argument is only required when the `type` is "trajectory",
#' since the coordinates are required for recalculating spatial groups with
#' `group_pts`, `group_lines` or `group_polys`.
#'
#' Please note that if the data extends over multiple years, a column indicating
#' the year should be provided to the `splitBy` argument. This will ensure
#' randomizations only occur within each year.
#'
#' The `group` argument is expected only when `type` is 'step' or
#' 'daily'.
#'
#' For example, using [`data.table::year()`][data.table::IDateTime]:
#'
#' \preformatted{ DT[, yr := year(datetime)] randomizations(DT, type = 'step',
#' id = 'ID', datetime = 'timegroup', splitBy = 'yr') }
#'
#' `iterations` is set to 1 if not provided. Take caution with a large
#' value for `iterations` with large input `DT`.
#'
#' @return `randomizations` returns the random date time or random id along
#'   with the original `DT`, depending on the randomization `type`.
#'   The length of the returned `data.table` is the original number of rows
#'   multiplied by the number of iterations + 1. For example, 3 iterations will
#'   return 4x - one observed and three randomized.
#'
#'   Two columns are always returned: \itemize{ \item observed - if the rows
#'   represent the observed (TRUE/FALSE) \item iteration - iteration of rows
#'   (where 0 is the observed) }
#'
#'   In addition, depending on the randomization type, random ID or random date
#'   time columns are returned:
#'
#'   \itemize{ \item step - `randomID` each time step \item daily -
#'   `randomID` for each day and `jul` indicating julian day \item
#'   trajectory - a random date time ("random" prefixed to `datetime`
#'   argument), observed `jul` and `randomJul` indicating the random
#'   day relocations are swapped to. }
#'
#'
#' @param type one of 'daily', 'step' or 'trajectory' - see details
#' @param datetime field used for providing date time or time group - see
#'   details
#' @param group generated from spatial grouping functions - see details
#' @inheritParams group_pts
#' @param splitBy List of fields in DT to split the randomization process by
#' @param iterations The number of iterations to randomize
#'
#' @references
#' <doi:10.1111/2041-210X.12553>
#' @export
#'
#' @family Social network tools
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Select only individuals A, B, C for this example
#' DT <- DT[ID %in% c('A', 'B', 'C')]
#'
#' # Date time columns
#' DT[, datetime := as.POSIXct(datetime)]
#' DT[, yr := year(datetime)]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '5 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 5, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Randomization: step
#' randStep <- randomizations(
#'     DT,
#'     type = 'step',
#'     id = 'ID',
#'     group = 'group',
#'     datetime = 'timegroup',
#'     splitBy = 'yr',
#'     iterations = 2
#' )
#'
#' # Randomization: daily
#' randDaily <- randomizations(
#'     DT,
#'     type = 'daily',
#'     id = 'ID',
#'     group = 'group',
#'     datetime = 'datetime',
#'     splitBy = 'yr',
#'     iterations = 2
#' )
#'
#' # Randomization: trajectory
#' randTraj <- randomizations(
#'     DT,
#'     type = 'trajectory',
#'     id = 'ID',
#'     group = NULL,
#'     coords = c('X', 'Y'),
#'     datetime = 'datetime',
#'     splitBy = 'yr',
#'     iterations = 2
#' )
#'
randomizations <- function(
    DT = NULL,
    type = NULL,
    id = NULL,
    group = NULL,
    coords = NULL,
    datetime = NULL,
    splitBy = NULL,
    iterations = NULL) {
  # due to NSE notes in R CMD check
  randomID <- jul <- randomJul <- rowID <- iteration <- observed <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(type)

  if (!(type %in% c('step', 'daily', 'trajectory'))) {
    stop('type of randomization must be one of: step, daily or trajectory')
  }

  assert_not_null(id)
  assert_not_null(datetime)

  assert_are_colnames(DT, c(id, datetime, splitBy))

  if (is.null(iterations)) {
    warning('iterations is not provided therefore iterations set to 1')
    iterations <- 1L
  }

  assert_inherits(iterations, c('numeric', 'integer'))

  if (length(datetime) == 1 &&
      any(class(DT[[datetime]]) %in% c('POSIXct', 'POSIXt'))) {
    dateFormatted <- TRUE
  } else {
    dateFormatted <- FALSE
  }

  if (type == 'step') {
    if (dateFormatted) {
      warning(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'datetime provided is POSIXct,
          step randomization will only be
          performed within each datetime -
          consider using group_times first and providing timegroup'
        )
      )
    }
  } else if (type == 'daily' || type == 'trajectory') {
    if (!dateFormatted) {
      stop(
        strwrap(
          prefix = " ",
          initial = "",
          x = 'datetime must be POSIXct for daily and trajectory randomization'
        )
      )
    }
  }
  if (type %in% c('step', 'daily')) {
    assert_not_null(group, ' if type is "step" or "daily"')
    assert_are_colnames(DT, group)
    selCols <- c(splitBy, id, datetime, group)
  } else if (type == 'trajectory') {
    assert_not_null(coords, ' if type is "trajectory"')
    assert_are_colnames(DT, coords)
    selCols <- c(splitBy, id, coords, datetime)
  }

  repDT <- DT[, .SD, .SDcols = selCols][rep(1:.N, iterations + 1)]
  repDT[, rowID := .GRP, by = c(selCols)]

  if (any(repDT[, .N , by = rowID]$N != iterations + 1)) {
    warning('found non-unique rows of id, datetime (and splitBy)')
  }

  repDT[, iteration := seq(0, .N - 1, 1), by = rowID]
  repDT[iteration == 0, observed := TRUE]
  repDT[iteration != 0, observed := FALSE]

  data.table::set(repDT, j = 'rowID', value = NULL)

  if (type == 'step') {
    if (is.null(splitBy)) {
      splitBy <- c(datetime, 'iteration')
    } else {
      splitBy <- c(datetime, 'iteration', splitBy)
    }

    repDT[!(observed), randomID := .SD[sample(.N, size = .N)],
          by = splitBy, .SDcols = id]

    repDT[(observed), randomID := .SD, .SDcols = id]

    return(repDT)
  }

  repDT[, jul := data.table::yday(.SD[[1]]), .SDcols = datetime]
  idDays <- unique(
    repDT[, .SD, .SDcols = c(splitBy, id, 'jul', 'iteration', 'observed')]
  )

  if (type == 'daily') {
    if (is.null(splitBy)) {
      splitBy <- c('jul', 'iteration', 'observed')
    } else {
      splitBy <- c('jul', 'iteration', 'observed', splitBy)
    }

    idDays[, randomID := .SD[sample(.N, size = .N)], by = c(splitBy),
           .SDcols = id]
    idDays[(observed), randomID := .SD[[1]], .SDcols = id]

    return(merge(repDT, idDays, by = c(splitBy, id), all = TRUE))

  } else if (type == 'trajectory') {
    if (is.null(splitBy)) {
      splitBy <- c(id, 'iteration')
    } else {
      splitBy <- c(id, 'iteration', splitBy)
    }

    idDays[, randomJul := sample(jul, size = .N), by = c(splitBy)]

    merged <- merge(
      x = repDT,
      y = idDays,
      by = c('jul', splitBy),
      all = TRUE
    )

    randomDateCol <- paste0('random', datetime)
    merged[, (randomDateCol) :=
             as.POSIXct(.SD[[1]] + (86400 * (randomJul - jul))),
           .SDcols = datetime]

    merged[(observed),
           c(randomDateCol, 'randomJul') := .SD,
           .SDcols = c(datetime, 'jul')]

    return(merged)
  }
}

#' Leadership along group direction
#'
#' Given the mean direction of a group of individuals, `leader_direction_group`
#' shifts the coordinate system to a new origin at the group centroid and
#' rotates the coordinate system by the mean direction to return each
#' individual's position along the mean direction, representing leadership in
#' terms of the front-back position in each group's mean direction. Relocation
#' data should be in two columns representing the X and Y coordinates, or in a
#' geometry column prepared by the helper function [get_geometry()].
#'
#' The function expects a `data.table` with relocation data appended with a
#' `group_direction` column from [direction_group()] and group centroid columns
#' from [centroid_group()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `group_direction` argument expects the names of columns in `DT` which
#' correspond to the mean group direction generated by [direction_group()]. The
#' mean group direction column is expected in units of radians. The
#' `return_rank` argument controls if the rank of each individual's distance to
#' the group centroid is also returned. If `return_rank` is TRUE, the `group`
#' argument is required to specify the group column generated by [group_pts()].
#' The `ties.method` argument is passed to [data.table::frank()], see details at
#' [`?data.table::frank()`][data.table::frank].
#'
#' See below under "Interface" for details on providing coordinates.
#'
#' @inheritSection direction_step Interface
#'
#' @return `leader_direction_group` returns the input `DT` appended with a
#'   `position_group_direction` column indicating the position along the group
#'   direction in the units of the crs and, optionally when `return_rank =
#'   TRUE`, a `rank_position_group_direction` column indicating the ranked
#'   position along the group direction.
#'
#'   A message is returned when `position_group_direction` or
#'   `rank_position_group_direction` columns already exist in the input `DT`,
#'   because they will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @param DT input data.table with group direction columns generated by
#'   `direction_group` and centroid columns generated by
#'   `centroid_group`
#' @param group_direction group_direction column name generated using
#'   `direction_group`, default 'group_direction'
#' @inheritParams direction_step
#' @inheritParams distance_to_centroid
#' @inheritParams group_pts
#'
#' @export
#' @seealso `direction_group`, `centroid_group`
#' @family Leadership functions
#' @family Direction functions
#'
#' @references
#' See examples of measuring leadership along group direction (also called
#' forefront index):
#'  * \doi{doi:10.1371/journal.pone.0036567}
#'  * \doi{doi:10.1111/jfb.15315}
#'  * \doi{doi:10.1098/rspb.2021.0839}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 50, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate direction at each step
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Calculate group centroid
#' centroid_group(DT, coords = c('X', 'Y'))
#'
#' # Calculate group direction
#' direction_group(DT)
#'
#' # Calculate leader in terms of position along group direction
#' leader_direction_group(DT, coords = c('X', 'Y'), crs = 32736)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' direction_step(DT = DT, id = 'ID')
#' centroid_group(DT)
#' direction_group(DT)
#' leader_direction_group(DT)
leader_direction_group <- function(
    DT = NULL,
    group_direction = 'group_direction',
    coords = NULL,
    group = 'group',
    crs = NULL,
    geometry = 'geometry',
    return_rank = TRUE,
    ties.method = NULL) {
  # Due to NSE notes
  geo <- cent <- group_dir <- x <- x_centroid <- y <- y_centroid <- pos <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(return_rank)

  assert_are_colnames(DT, group_direction)
  assert_col_radians(DT, group_direction, ', did you use direction_group?')

  pos_col <- 'position_group_direction'

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    centroid <- 'centroid'
    assert_are_colnames(DT, centroid, ', did you run get_geometry()?')
    assert_col_inherits(DT, centroid, 'sfc_POINT')

    crs_ud_unit <- sf::st_crs(DT[[geometry]])$ud_unit

    if (!identical(crs_ud_unit, units::as_units('m'))) {
      rlang::abort(
        paste0(
          'sf::st_crs(crs)$ud_unit is ',
          ifelse(is.null(crs_ud_unit), NULL, units(crs_ud_unit)),
          ', leader_direction_group requires planar coordinates in m'
        )
      )
    }

    if (pos_col %in% colnames(DT)) {
      message(
        pos_col, ' column will be overwritten by this function'
      )
      data.table::set(DT, j = pos_col, value = NULL)
    }

    DT[, c(pos_col) := {
      coords_geo <- sf::st_coordinates(geo)
      coords_cent <- sf::st_coordinates(cent)

      cos(units::drop_units(group_dir)) * (coords_geo[, 1] - coords_cent[, 1]) +
        sin(units::drop_units(group_dir)) * (coords_geo[, 2] - coords_cent[, 2])
    },
    env = list(
      group_dir = group_direction,
      geo = geometry,
      cent = centroid
    )
    ]

  } else {
    assert_not_null(coords)
    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')
    assert_not_null(crs)

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)
    pre <- 'centroid_'
    centroid_xcol <- paste0(pre, xcol)
    centroid_ycol <- paste0(pre, ycol)
    coords_centroid  <- c(centroid_xcol, centroid_ycol)

    assert_are_colnames(DT, coords_centroid, ', did you run centroid_group?')
    assert_col_inherits(DT, coords_centroid, 'numeric')

    crs_ud_unit <- sf::st_crs(crs)$ud_unit

    if (!identical(crs_ud_unit, units::as_units('m'))) {
      rlang::abort(
        paste0(
          'sf::st_crs(crs)$ud_unit is ',
          ifelse(is.null(crs_ud_unit), NULL, units(crs_ud_unit)),
          ', leader_direction_group requires planar coordinates in m'
        )
      )
    }

    if (pos_col %in% colnames(DT)) {
      message(
        pos_col, ' column will be overwritten by this function'
      )
      data.table::set(DT, j = pos_col, value = NULL)
    }

    DT[, c(pos_col) :=
      cos(units::drop_units(group_dir)) * (x - x_centroid) +
      sin(units::drop_units(group_dir)) * (y - y_centroid),
    env = list(
      group_dir = group_direction,
      x = xcol, y = ycol,
      x_centroid = centroid_xcol, y_centroid = centroid_ycol
    )
    ]
  }

  if (return_rank) {
    rank_col <- 'rank_position_group_direction'
    if (rank_col %in% colnames(DT)) {
      message(
        paste0(rank_col, ' column will be overwritten by this function')
      )
      data.table::set(DT, j = rank_col, value = NULL)
    }

    assert_not_null(group)
    assert_are_colnames(DT, group, ', did you run group_pts?')

    DT[, c(rank_col) :=
         data.table::frank(-pos, ties.method = ties.method,
                           na.last = 'keep'),
       by = c(group),
       env = list(pos = pos_col)]
  }

  return(DT[])
}

#' Group mean direction
#'
#' `direction_group` calculates the mean direction of all individuals in each
#' spatiotemporal group identified by [group_pts()]. The function expects a
#' `data.table` with relocation data appended with a `direction` column from
#' [direction_step()] and a `group` column from [group_pts()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `direction` and `group` arguments expect the names of columns in `DT`
#' which correspond to the direction and group columns. The direction column is
#' expected in units of radians and the mean calculated with
#' [CircStats::circ.mean()].
#'
#' @param DT input data.table with direction column generated by
#'   [direction_step()] and group column generated with [group_pts()]
#' @param direction character string of direction column name, default
#'   "direction", expects that the unit of the direction column is radians.
#' @param group character string of group column name, default "group"
#'
#' @return `direction_group` returns the input `DT` appended with a
#'   `group_direction` column representing the mean direction of all individuals
#'   in each spatiotemporal group.
#'
#'   The mean direction is calculated using [CircStats::circ.mean()] which
#'   expects units of radians.
#'
#'   A message is returned when the `group_direction` columns already exists in
#'   the input `DT`, because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @seealso [direction_step()], [group_pts()],
#'   [CircStats::circ.mean()]
#' @family Direction functions
#'
#' @references
#' See examples of using mean group direction:
#'  * \doi{doi:10.1098/rsos.170148}
#'  * \doi{doi:10.1098/rsos.201128}
#'  * \doi{doi:10.1016/j.beproc.2018.01.013}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 50, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate direction at each step
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Calculate group direction
#' direction_group(DT)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' direction_step(DT, id = 'ID')
#' direction_group(DT)
direction_group <- function(
    DT,
    direction = 'direction',
    group = 'group') {
  # due to NSE notes in R CMD check
  dir <- out <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)

  assert_not_null(direction)
  assert_not_null(group)
  assert_are_colnames(DT, c(direction, group))

  assert_col_inherits(DT, direction, 'units')
  assert_col_radians(DT, direction, ', did you use direction_step?')

  out_mean <- 'group_direction'
  if (out_mean %in% colnames(DT)) {
    message(paste(out_mean, 'column will be overwritten by this function'))
    data.table::set(DT, j = out_mean, value = NULL)
  }

  DT[, out := CircStats::circ.mean(as.numeric(dir)),
    by = c(group),
    env = list(dir = direction, out = out_mean)]

  DT[, out := units::as_units(out, 'rad'), env = list(out = out_mean)]

  return(DT[])
}

#' Distance to group centroid
#'
#' `distance_to_centroid` calculates the distance of each relocation to the
#' centroid of the spatiotemporal group identified by `group_pts`. The function
#' expects a `data.table` with relocation data appended with a `group` column
#' from `group_pts` and centroid columns from `centroid_group`. Relocation data
#' should be provided in two columns representing the X and Y coordinates, or in
#' a geometry column prepared by the helper function [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' This function expects a `group` column present generated with the `group_pts`
#' function and centroid coordinate column(s) generated with the
#' `centroid_group` function. The `group` arguments expect the names of columns
#' in `DT` which correspond to the group column. The `return_rank` argument
#' controls if the rank of each individual's distance to the group centroid is
#' also returned. The `ties.method` argument is passed to `data.table::frank`,
#' see details at [`?data.table::frank()`][data.table::frank].
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Distance function" for details on underlying distance function used.
#'
#' @section Interface:
#'  Two interfaces are available for providing coordinates:
#'
#'  1. Provide `coords` and `crs`. The `coords` argument expects the names of
#'  the X and Y coordinate columns. The `crs` argument expects a character
#'  string or numeric defining the coordinate reference system to be passed to
#'  [sf::st_crs]. For example, for UTM zone 36S (EPSG 32736), the crs argument
#'  is `crs = "EPSG:32736"` or `crs = 32736`. See <https://spatialreference.org>
#'  for a list of EPSG codes.
#'  2. (New!) Provide `geometry`. The `geometry` argument allows the user to
#'  supply a `geometry` column that represents the coordinates as a simple
#'  feature geometry list column. This interface expects the user to prepare
#'  their input DT with [get_geometry()]. To use this interface, leave the
#'  `coords` and `crs` arguments `NULL`, and the default argument for `geometry`
#'  ('geometry') will be used directly.
#'
#' @section Distance function:
#'
#'  The underlying distance function used depends on the crs of the coordinates
#'  or geometry provided.
#'
#'  - If the crs is longlat degrees (as determined by
#'  [sf::st_is_longlat()]), the distance function is [sf::st_distance()] which
#'  passes to [s2::s2_distance()] if [sf::sf_use_s2()] is TRUE and
#'  [lwgeom::st_geod_distance()] if [sf::sf_use_s2()] is FALSE. The distance
#'  returned has units set according to the crs.
#'
#'  - If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
#'  distance function used is [stats::dist()], maintaining expected behaviour
#'  from previous versions. The distance returned does not have units set.
#'
#'  Note: in both cases, if the coordinates are NA then the result will be NA.
#'
#' @param DT input data.table with centroid columns generated by eg.
#'   `centroid_group`
#' @param group group column name, generated by `group_pts`, default
#'   'group'
#' @param return_rank logical if rank distance should also be returned, default
#'   TRUE
#' @param ties.method see [`?data.table::frank()`][data.table::frank]
#' @inheritParams group_pts
#' @inheritParams direction_step
#'
#' @return `distance_to_centroid` returns the input `DT` appended with a
#'   `distance_centroid` column indicating the distance to the group centroid
#'   and, optionally, a `rank_distance_centroid` column indicating the within
#'   group rank distance to the group centroid (if `return_rank = TRUE`).
#'
#'   A message is returned when `distance_centroid` and optional
#'   `rank_distance_centroid` columns already exist in the input `DT`, because
#'   they will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @family Distance functions
#' @family Centroid functions
#' @seealso [centroid_group], [group_pts], [sf::st_distance()]
#' @references
#' See examples of using distance to group centroid:
#'  * \doi{doi:10.1016/j.anbehav.2021.08.004}
#'  * \doi{doi:10.1111/eth.12336}
#'  * \doi{doi:10.1007/s13364-018-0400-2}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 5, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate group centroid
#' centroid_group(DT, coords = c('X', 'Y'), group = 'group')
#'
#' # Calculate distance to group centroid
#' distance_to_centroid(
#'   DT,
#'   coords = c('X', 'Y'),
#'   group = 'group',
#' )
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' centroid_group(DT)
#' direction_to_centroid(DT)
distance_to_centroid <- function(
    DT = NULL,
    coords = NULL,
    group = 'group',
    crs = NULL,
    return_rank = TRUE,
    ties.method = NULL,
    geometry = 'geometry') {

  # Due to NSE notes in R CMD check
  geo <- cent <- x <- y <- x_centroid <- y_centroid <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)

  assert_not_null(return_rank)

  out <- 'distance_centroid'

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')
    centroid_col <- 'centroid'
    assert_are_colnames(DT, centroid_col, ', did you run centroid_group?')
    assert_col_inherits(DT, centroid_col, 'sfc_POINT')

    if (out %in% colnames(DT)) {
      message(out, ' column will be overwritten by this function')
      data.table::set(DT, j = out, value = NULL)
    }
    crs <- sf::st_crs(DT[[geometry]])
    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    DT[, c(out) := calc_distance(
      geometry_a = geo,
      geometry_b = cent,
      use_dist = use_dist
    ),
    env = list(geo = geometry, cent = centroid_col)
    ]

  } else {
    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)
    pre <- 'centroid_'
    xcol_centroid <- paste0(pre, xcol)
    ycol_centroid <- paste0(pre, ycol)
    coords_centroid  <- c(xcol_centroid, ycol_centroid)

    assert_are_colnames(DT, coords_centroid, ', did you run centroid_group?')
    assert_col_inherits(DT, coords_centroid, 'numeric')

    if (out %in% colnames(DT)) {
      message(out, ' column will be overwritten by this function')
      data.table::set(DT, j = out, value = NULL)
    }

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    DT[,
      c(out) := calc_distance(
        x_a = x,
        y_a = y,
        x_b = x_centroid,
        y_b = y_centroid,
        crs = crs,
        use_dist = use_dist
      ),
      env = list(
        x = xcol, y = ycol,
        x_centroid = xcol_centroid, y_centroid = ycol_centroid
      )
    ]

  }

  if (return_rank) {
    assert_not_null(group)
    assert_are_colnames(DT, group, ', did you run group_pts?')

    out_rank <- 'rank_distance_centroid'
    if (out_rank %in% colnames(DT)) {
      message(
        out_rank, ' column will be overwritten by this function'
      )
      data.table::set(DT, j = out_rank, value = NULL)
    }

    DT[, c(out_rank) :=
         data.table::frank(out,  ties.method = ties.method, na.last = 'keep'),
       by = group,
       env = list(out = out, group = group)]
  }

  return(DT[])
}

#' Direction step
#'
#' `direction_step` calculates the direction of movement steps in radians. The
#' function expects a `data.table` with relocation data and individual
#' identifiers. Relocation data should be in two columns representing the X and
#' Y coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()]. Note the order of rows is not modified by this function and
#' therefore users must be cautious to set it explicitly. See example for one
#' approach to setting order of rows using a datetime field.
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `id`, and optional `splitBy` arguments expect the names of a column in
#' `DT` which correspond to the individual identifier and additional grouping
#' columns.
#'
#' The `splitBy` argument offers further control over grouping. If within your
#' `DT`, you have distinct sampling periods for each individual, you can provide
#' the column name(s) which identify them to `splitBy`. The direction
#' calculation by [direction_step()] will only consider rows within each `id`
#' and `splitBy` subgroup.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Direction function" for details on the underlying direction function used.
#'
#' @section Interface:
#'  Two interfaces are available for providing coordinates:
#'
#'  1. Provide `coords` and `crs`. The `coords` argument expects the names of
#'  the X and Y coordinate columns. The `crs` argument expects a character
#'  string or numeric defining the coordinate reference system to be passed to
#'  [sf::st_crs]. For example, for UTM zone 36S (EPSG 32736), the crs argument
#'  is `crs = "EPSG:32736"` or `crs = 32736`. See <https://spatialreference.org>
#'  for a list of EPSG codes.
#'  2. (New!) Provide `geometry`. The `geometry` argument allows the user to
#'  supply a `geometry` column that represents the coordinates as a simple
#'  feature geometry list column. This interface expects the user to prepare
#'  their input DT with [get_geometry()]. To use this interface, leave the
#'  `coords` and `crs` arguments `NULL`, and the default argument for `geometry`
#'  ('geometry') will be used directly.
#'
#' @section Direction function:
#'
#'  The underlying distance function used depends on the crs of the coordinates
#'  or geometry provided.
#'
#'  - If the crs is provided and longlat degrees (as determined by
#'  [sf::st_is_longlat()]), the distance function is
#'  [lwgeom::st_geod_azimuth()].
#'  - If the crs is provided and not longlat degrees (eg. a projected UTM),
#'  the coordinates or geometry are transformed to `sf::st_crs(4326)` before the
#'  distance is measured using [lwgeom::st_geod_azimuth()].
#'  - If the crs is NULL or NA_crs_, the distance function cannot be used
#'  and an error is returned.
#'
#' @return `direction_step` returns the input `DT` appended with a `direction`
#'   column with units set to radians using the `units` package.
#'
#'   This column represents the azimuth between the sequence of points for each
#'   individual computed using [lwgeom::st_geod_azimuth()]. Note, the order of
#'   points is not modified by this function and therefore it is crucial the
#'   user sets the order of rows to their specific question before using
#'   `direction_step`. In addition, the direction column will include an `NA`
#'   value for the last point in each sequence of points since there is no
#'   future point to calculate a direction to.
#'
#'   A message is returned when a direction column are already exists in the
#'   input `DT`, because it will be overwritten.
#'
#'   An error is returned if there are any missing values in coordinates /
#'   geometry as the underlying direction function ([lwgeom::st_geod_azimuth()])
#'   does not accept missing values.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @inheritParams group_pts
#' @inheritParams build_polys
#' @param geometry simple feature geometry list column name, generated by
#'   [get_geometry()]. Default 'geometry', see details under Interface
#'
#' @family Direction functions
#' @seealso [lwgeom::st_geod_azimuth()], [amt::direction_abs()],
#' [geosphere::bearing()]
#' @export
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Set order using data.table::setorder
#' setorder(DT, datetime)
#'
#' # Calculate direction
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Or: geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' direction_step(DT, id = 'ID')
#'
#' # Example result for East, North, West, South steps
#' example <- data.table(
#'   X = c(0, 5, 5, 0, 0),
#'   Y = c(0, 0, 5, 5, 0),
#'   step = c('E', 'N', 'W', 'S', NA),
#'   ID = 'A'
#' )
#'
#' direction_step(example, 'ID', c('X', 'Y'), crs = 4326)
#' example[, .(step, direction, units::set_units(direction, 'degree'))]
direction_step <- function(
    DT = NULL,
    id = NULL,
    coords = NULL,
    crs = NULL,
    splitBy = NULL,
    geometry = 'geometry',
    projection = NULL) {

  # due to NSE notes in R CMD check
  geo <- x <- y <- NULL

  if (!is.null(projection)) {
    warning('projection argument is deprecated, setting crs = projection')
    crs <- projection
  }

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(id)
  assert_are_colnames(DT, c(id, splitBy))

  out_colname <- 'direction'

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    if ('direction' %in% colnames(DT)) {
      message('direction column will be overwritten by this function')
      data.table::set(DT, j = 'direction', value = NULL)
    }

    crs <- sf::st_crs(DT[[geometry]])
    use_transform <- !sf::st_is_longlat(crs)

    if (is.na(use_transform)) {
      rlang::abort(paste0('sf::st_is_longlat(crs) is ', use_transform,
                          ', ensure crs is provided for direction functions'))
    }

    DT[, c(out_colname) := c(calc_direction(
      geometry_a = geo,
      use_transform = use_transform
    ), units::set_units(NA, 'rad')),
    by = c(id, splitBy),
    env = list(geo = geometry)
    ]

  } else {
    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')
    assert_not_null(crs)

    if (out_colname %in% colnames(DT)) {
      message(out_colname, ' column will be overwritten by this function')
      data.table::set(DT, j = 'direction', value = NULL)
    }

    use_transform <- !sf::st_is_longlat(crs)

    if (is.na(use_transform)) {
      rlang::abort(paste0('sf::st_is_longlat(crs) is ', use_transform,
                          ', ensure crs is provided for direction functions'))
    }

    DT[, c(out_colname) := c(calc_direction(
      x_a = x,
      y_a = y,
      crs = crs,
      use_transform = use_transform
    ), units::set_units(NA, 'rad')),
    by = c(id, splitBy),
    env = list(x = data.table::first(coords), y = data.table::last(coords))
    ]
  }

}

# Test direction_step
context('test direction_step')

library(spatsoc)
library(units)

DT <- fread('../testdata/DT.csv')

DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
setorder(DT, datetime)

id <- 'ID'
coords <- c('X', 'Y')
utm <- 32736

clean_DT <- copy(DT)

test_that('DT is required', {
  expect_error(direction_step(DT = NULL), 'DT must be provided')
})

test_that('args required else error', {
  expect_error(direction_step(DT,  id = NULL), 'id must be')

  expect_error(direction_step(DT,  id = id, coords = coords, crs = NULL),
               'crs must')

  expect_error(direction_step(DT, id = id, coords = 'X', crs = utm),
               'coords must be length 2')
})

test_that('column names must exist in DT', {
  expect_error(direction_step(DT, id = 'potato', coords = coords,
                              crs = utm),
               'not present')

  expect_error(direction_step(DT, id = id, coords = c('potato', 'potato'),
                              crs = utm),
               'not present')

  expect_error(direction_step(DT, id = id, coords = coords,
                              crs = utm, splitBy = 'potato'),
               'not present')
})

test_that('coords are correctly provided or error detected', {
  expect_error(direction_step(DT, id = id, coords = c('X', NULL),
                              crs = utm),
               'coords must be')

  expect_error(direction_step(DT, id = id, coords = c('X', 'ID'),
                              crs = utm),
               'numeric')
})

test_that('dimensions returned expected', {

  expect_equal(
    ncol(clean_DT) + 1,
    ncol(direction_step(copy(clean_DT), id = id,
                        coords = coords, crs = utm))
  )

  expect_equal(
    nrow(clean_DT),
    nrow(direction_step(copy(clean_DT), id = id,
                        coords = coords, crs = utm))
  )

  expect_true('direction' %in% colnames(
    direction_step(
      copy(clean_DT),
      id = id,
      coords = coords,
      crs = utm
    )
  ))

})


test_that('direction column succesfully detected', {
  copyDT <- copy(clean_DT)[, direction := 1]
  expect_message(
    direction_step(
      copyDT,
      id = id,
      coords = coords,
      crs = utm
    ),
    'direction column will be overwritten'
  )
})


test_that('returns a data.table', {
  expect_s3_class(direction_step(DT, id = id, coords = coords,
                                 crs = utm),
                  'data.table')
})


test_that('splitBy returns expected', {
  DT[, split := sample(seq.int(5), .N, replace = TRUE)]
  expect_gte(
    direction_step(DT, id = id, coords = coords,
                   crs = utm,
                   splitBy = 'split')[is.na(direction), .N],
    direction_step(DT, id = id, coords = coords,
                   crs = utm)[is.na(direction), .N]

  )
})


DT_A <- data.table(
  X = c(-5, -5, 0, 14, 10, 0),
  Y = c(5, 3, 1, 1, 11, 11),
  ID =  'A'
)[, timegroup := seq.int(.N)]

# Related to: PR 92
test_that('longlat NA radian returned', {
  expect_equal(
    class(direction_step(DT_A, id = id, coords = coords,
                         crs = 4326)[]$direction),
    'units'
  )
  expect_gte(
    sum(is.na(direction_step(DT_A, id = id, coords = coords,
                             crs = 4326)$direction)),
    1
  )
})


test_that('use_transform errors if crs not provided', {
  expect_error(
    direction_step(DT, id = id, coords = coords, crs = NA),
    'ensure crs is provided'
  )

  copyDT <- copy(DT)
  get_geometry(copyDT, coords = coords, crs = utm)
  st_crs(copyDT$geometry) <- NA
  expect_error(
    direction_step(
      DT = copyDT,
      id = id
    ),
    'ensure crs is provided'
  )
})


DT_B <- data.table(
  X = c(0, 5, 5, 0, 0),
  Y = c(0, 0, 5, 5, 0),
  step = c('E', 'N', 'W', 'S', NA),
  timegroup = seq.int(5),
  ID = 'B'
)
direction_step(DT_B, id, coords, crs = 4326)

test_that('East North West South steps', {
  tolerance <- 0.01

  expect_equal(
    DT_B[step == 'E', direction],
    as_units(pi / 2, 'rad'),
    tolerance = tolerance
  )

  expect_equal(
    DT_B[step == 'N', direction],
    as_units(0),
    tolerance = tolerance
  )

  expect_equal(
    DT_B[step == 'W', direction],
    as_units(- pi / 2),
    tolerance = tolerance
  )

  expect_equal(
    DT_B[step == 'S', direction],
    as_units(pi),
    tolerance = tolerance
  )
})


test_that('projection arg is deprecated', {
  expect_warning(
    direction_step(
      DT,
      id = id,
      coords = coords,
      projection = utm
    ),
    'projection argument is deprecated'
  )
})

# geometry interface
test_that('if coords null, geometry required', {
  expect_error(direction_step(DT, id = id, coords = NULL, crs = utm),
               'get_geometry?')
})

get_geometry(DT, coords = coords, crs = utm)
get_geometry(DT, coords = coords, crs = utm, output_crs = 4326,
             geometry_colname = 'geometry_longlat')

test_that('crs provided with geometry gives message crs ignored', {
  expect_message(direction_step(DT, id = id, crs = utm), 'ignored')
})

test_that('geometry correctly provided else error', {
  expect_error(direction_step(DT, id = id, geometry = 'potato'),
               'is not present')
  expect_error(direction_step(DT, id = id, geometry = 'X'),
               'must be of class')
})

test_that('geometry interface message before overwrite', {
  copyDT <- copy(DT)[, direction := 42]
  expect_message(direction_step(copyDT, id = id),
                 'overwritten')
})

test_that('geometry interface returns expected', {
  copyDT <- get_geometry(copy(clean_DT), coords = coords, crs = utm,
                         output_crs = 4326)

  expect_equal(
    ncol(copyDT) + 1,
    ncol(direction_step(copyDT, id = id))
  )

  expect_equal(
    nrow(copyDT),
    nrow(direction_step(copy(copyDT), id = id))
  )

  expect_true('direction' %in% colnames(direction_step(copy(copyDT), id = id)))

  expect_type(copyDT$direction, 'double')
  expect_s3_class(copyDT$direction, 'units')

  expect_equal(min(copyDT$direction, na.rm = TRUE), units::as_units(-pi, 'rad'),
               tolerance = 0.01)
  expect_equal(max(copyDT$direction, na.rm = TRUE), units::as_units(pi, 'rad'),
               tolerance = 0.01)


  copyDT <- get_geometry(copy(clean_DT), coords = coords, crs = utm)

  expect_equal(
    ncol(copyDT) + 1,
    ncol(direction_step(copyDT, id = id))
  )

  expect_equal(
    nrow(copyDT),
    nrow(direction_step(copy(copyDT), id = id))
  )

  expect_true('direction' %in% colnames(direction_step(copy(copyDT), id = id)))

  expect_type(copyDT$direction, 'double')
  expect_s3_class(copyDT$direction, 'units')

  expect_equal(min(copyDT$direction, na.rm = TRUE), units::as_units(-pi, 'rad'),
               tolerance = 0.01)
  expect_equal(max(copyDT$direction, na.rm = TRUE), units::as_units(pi, 'rad'),
               tolerance = 0.01)
})

#' Get geometry
#'
#' `get_geometry` sets up an input `DT` with a 'geometry' column for spatsoc's
#' geometry interface. The function expects a `data.table` with relocation data and
#' a coordinate reference system.
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' The `coords` argument expects the names of columns in `DT` which correspond
#' to the X and Y coordinates.
#'
#' The `output_crs` argument allows the user to set an output `crs` for their
#' geometry column. Note: some functions in spatsoc (eg. those that measure
#' directions like `edge_direction` and `direction_to_leader`) require
#' geographic coordinates and it is therefore simpler to leave the default
#' `output_crs = 4326`.
#'
#' @return `get_geometry` returns the input `DT` appended with a
#'   `geometry` column which represents the input coordinates
#'   as a `sfc` (simple feature geometry list column). If the `output_crs`
#'   was provided, the `geometry` will be transformed to the `output_crs`.
#'
#'   A message is returned when a column named `geometry` already exists in
#'   the input `DT`, because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @inheritParams group_pts
#' @param crs numeric or character defining the coordinate reference system to
#'   be passed to [sf::st_crs]. For example, `crs = "EPSG:32736"` or
#'   `crs = 32736`.
#' @param output_crs default NULL, the output crs to transform the input
#'   coordinates to with [sf::st_transform]. If output_crs is NULL or
#'   matching the crs argument, the coordinates will not be transformed
#' @param geometry_colname default "geometry", to optionally set output name
#'   of simple feature geometry list column
#'
#' @export
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file('extdata', 'DT.csv', package = 'spatsoc'))
#'
#' # Get geometry
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#'
#' # Print
#' print(DT)
get_geometry <- function(
    DT = NULL,
    coords = NULL,
    crs = NULL,
    output_crs = NULL,
    geometry_colname = 'geometry') {

  assert_not_null(DT)
  assert_is_data_table(DT)

  assert_are_colnames(DT, coords)
  assert_length(coords, 2)
  assert_col_inherits(DT, coords, 'numeric')

  assert_not_null(crs)

  if (geometry_colname %in% colnames(DT)) {
    message(paste0(geometry_colname,
                   ' column will be overwritten by this function'))
    data.table::set(DT, j = geometry_colname, value = NULL)
  }

  DT[, (geometry_colname) := {
    x <- sf::st_as_sf(data.frame(.SD),
                      coords = coords,
                      crs = crs,
                      na.fail = FALSE
    )
    if (!is.null(output_crs) && !identical(crs, sf::st_crs(output_crs))) {
      x <- sf::st_transform(x, output_crs)
    }
    x
  }, .SDcols = coords]
  return(DT)
}

# Test get_geometry
context('test get_geometry')

library(sf)

DT <- fread('../testdata/DT.csv')

coords <- c('X', 'Y')
crs <- 32736

test_that('DT is required', {
  expect_error(get_geometry(
    DT = NULL
  ),
  'DT must be provided')
})

test_that('coords provided correctly, else error', {
  expect_error(
    get_geometry(
      DT,
      coords = 'X'
    ),
    'coords must be length 2'
  )

  expect_error(
    get_geometry(
      DT,
      coords = c('potatoX', 'Y')
    ),
    'not present in input'
  )

  expect_error(
    get_geometry(
      DT,
      coords = c('X', 'potatoY')
    ),
    'not present in input'
  )

  expect_error(
    get_geometry(
      DT,
      coords = c('ID', 'ID')
    ),
    'coords must be of class numeric'
  )

})

test_that('crs provided correctly, else error', {
  expect_error(
    get_geometry(
      DT,
      coords = coords,
      crs = NULL
    ),
    'crs must be provided'
  )
})

test_that('group column succesfully detected', {
  copyDT <- copy(DT)[, geometry := 1]
  expect_message(
    get_geometry(
      copyDT,
      coords = coords,
      crs = crs
    ),
    'geometry column will be overwritten'
  )
})

test_that('geometry column returned is sfc, as expected', {
  copyDT <- copy(DT)
  get_geometry(copyDT, coords = coords, crs = crs)

  expect_s3_class(copyDT$geometry, 'sfc_POINT')
  expect_s3_class(copyDT$geometry, 'sfc')

  copyDT <- copy(DT)
  get_geometry(copyDT, coords = coords, crs = crs)

  expect_equal(st_coordinates(copyDT$geometry),
               copyDT[, as.matrix(.SD), .SDcols = coords])

  copyDT <- copy(DT)
  get_geometry(copyDT, coords = coords, crs = crs)

  expect_equal(
    mean(st_coordinates(copyDT$geometry) -
      copyDT[, as.matrix(.SD), .SDcols = coords]),
    0
  )

  copyDT <- copy(DT)[sample(seq.int(.N), 10), c(coords) := NA]
  get_geometry(copyDT, coords = coords, crs = crs)
  expect_equal(copyDT[is.na(X), .N], copyDT[st_is_empty(geometry), .N])
})

#' Build lines
#'
#' `build_lines` generates a simple feature collection with LINESTRINGs from a
#' `data.table`. The function expects a `data.table` with relocation data,
#' individual identifiers, a sorting column and a `crs`. The relocation
#' data is transformed into LINESTRINGs for each individual and, optionally,
#' combination of columns listed in `splitBy`. Relocation data should be in two
#' columns representing the X and Y coordinates.
#'
#' ## R-spatial evolution
#'
#'  Please note, spatsoc has followed updates from R spatial, GDAL and PROJ
#'  for handling coordinate reference systems, see more at
#'  <https://r-spatial.org/r/2020/03/17/wkt.html>.
#'
#' In addition, `build_lines` previously used [sp::SpatialLines] but has been
#' updated to use [sf::st_as_sf] and [sf::st_linestring] according to the
#' R-spatial evolution, see more at
#' <https://r-spatial.org/r/2022/04/12/evolution.html>.
#'
#' ## Notes on arguments
#'
#' The `crs` argument expects a numeric or character defining the
#' coordinate reference system.
#' For example, for UTM zone 36N (EPSG 32736), the crs argument is either
#'  `crs = 'EPSG:32736'` or `crs = 32736`.
#'  See details in [`sf::st_crs()`] and <https://spatialreference.org>
#'  for a list of EPSG codes.
#'
#' The `sortBy` argument is used to order the input `DT` when creating
#' sf LINESTRINGs. It must a column in the input `DT` of type
#' POSIXct to ensure the rows are sorted by date time.
#'
#' The `splitBy` argument offers further control building LINESTRINGs.
#' If in your input `DT`, you have multiple temporal groups (e.g.: years) for
#' example, you can provide the name of the column which identifies them and
#' build LINESTRINGs for each individual in each year.
#'
#' `build_lines` is used by `group_lines` for grouping overlapping
#' lines generated from relocations.
#'
#' @return `build_lines` returns an sf LINESTRING object with a line
#' for each individual (and optionally `splitBy` combination).
#'
#' Individuals (or combinations of individuals and `splitBy`) with less than two
#'  relocations are dropped since it requires at least two relocations to
#'  build a line.
#'
#' @inheritParams group_lines
#' @inheritParams build_polys
#' @param projection (deprecated) use crs argument instead
#'
#' @export
#'
#' @family Build functions
#' @seealso `group_lines`
#'
#' @import data.table
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # EPSG code for example data
#' utm <- 32736
#'
#' # Build lines for each individual
#' lines <- build_lines(DT, crs = utm, id = 'ID', coords = c('X', 'Y'),
#'             sortBy = 'datetime')
#'
#' # Build lines for each individual by year
#' DT[, yr := year(datetime)]
#' lines <- build_lines(DT, crs = utm, id = 'ID', coords = c('X', 'Y'),
#'             sortBy = 'datetime', splitBy = 'yr')
build_lines <- function(
    DT = NULL,
    crs = NULL,
    id = NULL,
    coords = NULL,
    sortBy = NULL,
    splitBy = NULL,
    projection = NULL) {

  # due to NSE notes in R CMD check
  dropped <- . <- NULL

  if (!is.null(projection)) {
    warning('projection argument is deprecated, setting crs = projection')
    crs <- projection
  }

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(id)
  assert_not_null(crs)
  assert_not_null(sortBy)
  assert_are_colnames(DT, c(id, sortBy))

  if (is.null(splitBy)) {
    splitBy <- id
  } else {
    splitBy <- c(id, splitBy)
  }

  assert_col_inherits(DT, id, c('numeric', 'character', 'integer'))
  assert_col_inherits(DT, splitBy, c('numeric', 'character', 'integer', 'IDate'))
  assert_col_inherits(DT, sortBy, 'POSIXct')

  assert_not_null(coords)
  assert_length(coords, 2L)
  assert_are_colnames(DT, coords)
  assert_col_inherits(DT, coords, 'numeric')

  dropRows <- DT[, .(dropped = .N < 2), by = c(splitBy)]

  if (dropRows[(dropped), .N] > 0) {
    warning('some rows dropped, cannot build lines with less than two points')
  }

  wo_drop <- DT[dropRows, on = splitBy][!(dropped)]

  data.table::setorderv(wo_drop, sortBy)

  lines <- sf::st_as_sf(
    wo_drop[, .(geometry = sf::st_sfc(sf::st_linestring(as.matrix(.SD)))),
       by = c(splitBy), .SDcols = coords],
    crs = sf::st_crs(crs)
  )
}

# Test build_lines
context('test build_lines')

DT <- fread('../testdata/DT.csv')

utm <- 'EPSG:32736'

DT[, datetime := as.POSIXct(datetime)]

test_that('DT is required', {
  expect_error(
    build_lines(
      DT = NULL,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'DT must be provided'
  )
})

test_that('coords, id, crs, sortBy must be provided, proper format', {
  expect_error(
    build_lines(
      DT = DT,
      id = NULL,
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'id must be provided'
  )

  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = NULL,
      sortBy = 'datetime'
    ),
    'crs must be provided'
  )

  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = NULL,
      crs = utm,
      sortBy = 'datetime'
    ),
    'coords must be provided'
  )

  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('ID', 'ID'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'coords must be of class numeric'
  )
  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'coords must be length 2', fixed = FALSE
  )

  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = NULL
    ),
    'sortBy must be provided'
  )
  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'ID'
    ),
    'sortBy must be', fixed = FALSE
  )

})


test_that('column names must exist in DT', {
  expect_error(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('potatoX', 'potatoY'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    build_lines(
      DT = DT,
      id = 'potato',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'not present in input',
    fixed = FALSE
  )

  expect_error(
    build_lines(
      DT = DT,
      id = 'potato',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'potato'
    ),
    'not present in input',
    fixed = FALSE
  )
})

test_that('returns same number of lines as unique IDs/splitBy provided', {
  # without splitBy
  expect_equal(nrow(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    )
  ),
  DT[, uniqueN(ID)])

  # with splitBy
  DT[, jul := data.table::yday(as.POSIXct(datetime))]
  splitBy <- c('ID', 'jul')
  DT[, count := .N, by = splitBy]
  subDT <- DT[count >= 2]

  expect_equal(nrow(
    build_lines(
      DT = subDT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      splitBy = 'jul',
      sortBy = 'datetime'
    )
  ),
  nrow(unique(subDT[, .SD, .SDcols = splitBy])))
})


test_that("build lines warns if < 2 locs per ID/byField", {
  # for ID (one row's ID is "potato")
  copyDT <- copy(DT)[1, ID := 'potato']

  expect_warning(
    build_lines(
      DT = copyDT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'some rows dropped, cannot build lines with less than two points'
  )

  copyDT <- copy(DT)[1, ID := 'potato']
  expect_warning(
    build_lines(
      DT = copyDT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      splitBy = 'population',
      sortBy = 'datetime'
    ),
    'some rows dropped, cannot build',
    fixed = FALSE
  )
})

test_that('splitBy and id provided are not correct format', {
  copyDT <- copy(DT)[, datetime := as.POSIXct(datetime)]
  expect_error(
    build_lines(
      DT = copyDT,
      id = 'datetime',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'id must be of class numeric/character/integer',
    fixed = FALSE
  )

  expect_error(
    build_lines(
      DT = copyDT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      splitBy = 'datetime',
      sortBy = 'datetime'
    ),
    'splitBy must be of class numeric/character/integer',
    fixed = FALSE
  )

  # with factor IDs
  copyDT <- copy(DT)[, ID := as.factor(ID)]
  expect_error(
    build_lines(
      DT = copyDT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'id must be of class numeric/character/integer',
    fixed = FALSE
  )
})

test_that('build_lines returns an sf object with LINESTRINGs', {
  expect_s3_class(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ),
    'sf'
  )

  expect_in(
    'LINESTRING',
    sf::st_geometry_type(build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      sortBy = 'datetime'
    ), by_geometry = FALSE)
  )
})

test_that('build_lines builds ordered lines', {
  base_lines <- build_lines(
    DT = DT,
    id = 'ID',
    coords = c('X', 'Y'),
    crs = utm,
    sortBy = 'datetime'
  )

  random_lines <- build_lines(
    DT = DT[sample(.N)],
    id = 'ID',
    coords = c('X', 'Y'),
    crs = utm,
    sortBy = 'datetime'
  )

  expect_equal(
    base_lines[order(base_lines$ID),],
    random_lines[order(random_lines$ID),]
  )
})




test_that('splitBy argument doesnt use splitBy column', {
  copyDT <- copy(DT)

  copyDT[, splitBy := as.IDate(datetime)]

  utm <- 'EPSG:32736'

  expect_equal(
    nrow(build_lines(
      DT = copyDT,
      id = 'ID',
      coords = c('X', 'Y'),
      crs = utm,
      splitBy = 'splitBy',
      sortBy = 'datetime'
    )),
    copyDT[, uniqueN(splitBy) * uniqueN(ID)]
  )
})


test_that('projection arg is deprecated', {
  expect_warning(
    build_lines(
      DT = DT,
      id = 'ID',
      coords = c('X', 'Y'),
      sortBy = 'datetime',
      projection = utm
    ),
    'projection argument is deprecated'
  )
})

#' Distance to group leader
#'
#' `distance_to_leader` calculates the distance to the leader of each
#' spatiotemporal group. The function expects a `data.table` with relocation
#' data appended with a `rank_position_group_direction` column indicating the
#' ranked position along the group direction generated with
#' `leader_direction_group(return_rank = TRUE)`. Relocation data should be in
#' two columns representing the X and Y coordinates, or in a geometry column
#' prepared by the helper function [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' This function expects a `rank_position_group_direction` column generated with
#' `leader_direction_group(return_rank = TRUE)`, a `group` column generated with
#' the `group_pts` function. The `group` argument expect the names of the column
#' in `DT` which corresponds to the group column.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Distance function" for details on underlying distance function used.
#'
#' @inheritSection distance_to_centroid Interface
#' @inheritSection distance_to_centroid Distance function
#'
#' @param DT input data.table with 'rank_position_group_direction' column
#'   generated by `leader_direction_group` and group column generated by
#'   `group_pts`
#' @inheritParams leader_direction_group
#' @inheritParams group_pts
#'
#' @return `distance_to_leader` returns the input `DT` appended with a
#'   `distance_leader` column indicating the distance to the group leader.
#'
#'   A message is returned when the `distance_leader` column already exist in
#'   the input `DT` because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @family Distance functions
#' @seealso [direction_to_leader], [leader_direction_group], [group_pts],
#' [sf::st_distance()]
#' @references
#'
#' See examples of using distance to leader and position within group:
#'  * \doi{doi:10.1111/jfb.15315}
#'  * \doi{doi:10.1098/rspb.2017.2629}
#'  * \doi{doi:10.1016/j.anbehav.2023.09.009}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 50, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate direction at each step
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Calculate group centroid
#' centroid_group(DT, coords = c('X', 'Y'))
#'
#' # Calculate group direction
#' direction_group(DT)
#'
#' # Calculate leader in terms of position along group direction
#' leader_direction_group(
#'   DT,
#'   coords = c('X', 'Y'),
#'   crs = 32736,
#'   return_rank = TRUE
#' )
#'
#' # Calculate distance to leader
#' distance_to_leader(DT, coords = c('X', 'Y'), crs = 32736)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' direction_step(
#'   DT = DT,
#'   id = 'ID'
#' )
#' centroid_group(DT)
#' direction_group(DT)
#' leader_direction_group(
#'   DT
#' )
#' distance_to_leader(DT)
distance_to_leader <- function(
    DT = NULL,
    coords = NULL,
    group = 'group',
    crs = NULL,
    geometry = 'geometry') {

  # Due to NSE notes
  geo <- lead <- x <- y <- x_leader <- y_leader <- . <-
    rank_position_group_direction <- has_leader <- distance_leader <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(group)
  assert_are_colnames(DT, group)

  leader_col <- 'rank_position_group_direction'
  assert_are_colnames(
    DT, leader_col,
    ', did you run leader_direction_group(return_rank = TRUE)?'
  )
  assert_col_inherits(DT, leader_col, 'numeric')

  check_leaderless <- DT[, .(
    has_leader = any(rank_position_group_direction == 1, na.rm = TRUE)),
    by = group,
    env = list(group = group)][!(has_leader)]

  out_col <- 'distance_leader'

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')

    zzz_geometry_leader <- 'zzz_geometry_leader'
    DT[, c(zzz_geometry_leader) :=
         sf::st_sf(rep(geo[which(rank_position_group_direction == 1)], .N)),
       env = list(geo = geometry, group = group),
       by = group]

    if (check_leaderless[, .N > 0]) {
      warning(
        'groups found missing leader (rank_position_group_direction == 1): \n',
        check_leaderless[, paste(group, collapse = ', ')]
      )
    }

    if (out_col %in% colnames(DT)) {
      message(
        paste0(out_col, ' column will be overwritten by this function')
      )
      data.table::set(DT, j = out_col, value = NULL)
    }

    crs <- sf::st_crs(DT[[geometry]])
    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    DT[!group %in% check_leaderless$group, c(out_col) := calc_distance(
      geometry_a = geo,
      geometry_b = lead,
      use_dist = use_dist
    ),
    env = list(
      geo = geometry, lead = zzz_geometry_leader
    )]

    data.table::set(DT, j = zzz_geometry_leader, value = NULL)

  } else {
    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')

    if (is.null(crs)) {
      crs <- sf::NA_crs_
    }

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)
    pre <- 'zzz_leader_'
    zzz_xcol_leader <- paste0(pre, xcol)
    zzz_ycol_leader <- paste0(pre, ycol)
    zzz_coords_leader  <- c(zzz_xcol_leader, zzz_ycol_leader)

    DT[, c(zzz_coords_leader) :=
         .SD[which(rank_position_group_direction == 1)],
       .SDcols = c(coords),
       env = list(group = group),
       by = group]

    if (check_leaderless[, .N > 0]) {
      warning(
        'groups found missing leader (rank_position_group_direction == 1): \n',
        check_leaderless[, paste(group, collapse = ', ')]
      )
    }

    if (out_col %in% colnames(DT)) {
      message(
        paste0(out_col, ' column will be overwritten by this function')
      )
      data.table::set(DT, j = out_col, value = NULL)
    }

    use_dist <- isFALSE(sf::st_is_longlat(crs)) || identical(crs, sf::NA_crs_)

    DT[!group %in% check_leaderless$group,
      c(out_col) := calc_distance(
        x_a = x,
        y_a = y,
        x_b = x_leader,
        y_b = y_leader,
        crs = crs,
        use_dist = use_dist
      ),
      env = list(
        x = xcol, y = ycol, x_leader = zzz_xcol_leader, y_leader = zzz_ycol_leader
      )
    ]

    data.table::set(DT, j = zzz_coords_leader, value = NULL)

  }

  return(DT[])
}

#' Directional correlation delay based edge-lists
#'
#' `edge_delay` returns edge-lists defined by the directional correlation
#' delay between individuals. The function expects a distance based edge-list
#' generated by `edge_dist` or `edge_nn`, a `data.table` with
#' relocation data, individual identifiers and a window argument. The window
#' argument is used to specify the temporal window within which to measure the
#' directional correlation delay. Relocation data should be in two columns
#' representing the X and Y coordinates.
#'
#' The `edges` and `DT` must be `data.table`s. If your data is a
#' `data.frame`, you can convert it by reference using
#' [data.table::setDT()].
#'
#' The `edges` argument expects a distance based edge-list generated with
#' `edge_nn` or `edge_dist`. The `DT` argument expects relocation
#' data with a timegroup column generated with `group_times`.
#'
#' The rows in `edges` and `DT` are internally matched in
#' `edge_delay` using the columns `timegroup` (from
#' `group_times`) and `ID1` and `ID2` (in `edges`, from
#' `dyad_id`) with `id` (in `DT`). This function expects a
#' `fusionID` present, generated with the `fusion_id` function, and a
#' `dyadID` present, generated with the `dyad_id` function. The
#' `id`, and `direction` arguments expect the names of a column in
#' `DT` which correspond to the id, and direction columns.
#'
#' @inheritParams centroid_fusion
#' @inheritParams direction_group
#' @param window temporal window in unit of timegroup column generated with
#'   `group_times`, eg. `window = 4` corresponds to the 4 timegroups
#'   before and after the focal observation
#'
#' @return `edge_delay` returns the input `edges` appended with a
#'   'direction_delay' column indicating the temporal delay (in units of
#'   timegroups) at which ID1's direction of movement is most similar to ID2's
#'   direction of movement, within the temporal window defined, and a
#'   'direction_diff' column indicating the absolute difference in direction.
#'   For example, if focal individual 'A' moves in a 45 degree direction at time
#'   2 and individual 'B' moves in a most similar direction within the window at
#'   time 5, the directional correlation delay between A and B is 3. Positive
#'   values of directional correlation delay indicate a directed leadership edge
#'   from ID1 to ID2.
#'
#'  Note: due to the merge required within this function, the output needs to be
#'  reassigned unlike some other `spatsoc` functions like `dyad_id`.
#'  See details in [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#'
#' @references
#'
#' The directional correlation delay is defined in Nagy et al. 2010
#' (\doi{doi:10.1038/nature08891}).
#'
#' See examples of measuring the directional correlation delay:
#'  * \doi{doi:10.1016/j.anbehav.2013.07.005}
#'  * \doi{doi:10.1073/pnas.1305552110}
#'  * \doi{doi:10.1111/jfb.15315}
#'  * \doi{doi:10.1371/journal.pcbi.1003446}
#'
#' @family Edge-list generation
#' @family Direction functions
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Select only individuals A, B, C for this example
#' DT <- DT[ID %in% c('A', 'B', 'C')]
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Calculate direction
#' direction_step(
#'   DT = DT,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   crs = 32736
#' )
#'
#' # Distance based edge-list generation
#' edges <- edge_dist(
#'   DT,
#'   threshold = 100,
#'   id = 'ID',
#'   coords = c('X', 'Y'),
#'   timegroup = 'timegroup',
#'   returnDist = TRUE,
#'   fillNA = FALSE
#' )
#'
#' # Generate dyad id
#' dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
#'
#' # Generate fusion id
#' fusion_id(edges, threshold = 100)
#'
#' # Directional correlation delay
#' delay <- edge_delay(
#'   edges = edges,
#'   DT = DT,
#'   window = 3,
#'   id = 'ID'
#' )
#'
#' delay[, mean(direction_delay, na.rm = TRUE), by = .(ID1, ID2)][V1 > 0]
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' direction_step(DT, id = 'ID')
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
#' fusion_id(edges, threshold = 100)
#' delay <- edge_delay(
#'   edges = edges,
#'   DT = DT,
#'   window = 3,
#'   id = 'ID'
#' )
edge_delay <- function(
    edges,
    DT,
    window = NULL,
    id = NULL,
    direction = 'direction') {
  # due to NSE notes in R CMD check
  . <- timegroup <- fusionID <- timegroup_min <- timegroup_max <-
    timegroup_delay <- ID1  <- ID2 <- direction_delay <- direction_diff <-
    dyadID <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)
  assert_not_null(edges)
  assert_not_null(id)

  check_cols_edges <- c('ID1', 'ID2', 'timegroup')
  assert_are_colnames(edges, check_cols_edges)

  check_cols_DT <- c(id, 'timegroup', direction)
  assert_are_colnames(DT, check_cols_DT)

  assert_not_null(window)
  assert_inherits(window, 'numeric')

  assert_col_inherits(DT, 'timegroup', 'integer')
  assert_col_inherits(edges, 'timegroup', 'integer')

  assert_are_colnames(edges, 'fusionID', ', did you run fusion_id?')
  assert_are_colnames(edges, 'dyadID', ', did you run dyad_id?')

  drop_nas <- data.table::copy(edges)[
    !(is.na(fusionID) | is.na(ID1) | is.na(ID2) | is.na(dyadID))]

  # "Forward": all edges ID1 -> ID2
  forward <- drop_nas[ID1 == tstrsplit(dyadID, '-')[[1L]]]

  forward[, timegroup_min :=
            data.table::fifelse(timegroup - window < min(timegroup),
                                min(timegroup),
                                timegroup - window),
          by = fusionID,
          env = list(window = window)]

  forward[, timegroup_max :=
            data.table::fifelse(timegroup + window > max(timegroup),
                                max(timegroup),
                                timegroup + window),
          by = fusionID,
          env = list(window = window)]

  forward[, c('timegroup_delay', 'direction_diff') := {
    focal_direction <- DT[timegroup == .BY$timegroup &
                            id == ID1, direction]
    sub <- DT[between(timegroup, timegroup_min, timegroup_max) & id == ID2,
              .(timegroup, diff = diff_rad(focal_direction, direction))]
    sub[which.min(diff)]
  },
  by = c('timegroup', 'dyadID'),
  env = list(id = id, direction = direction)]

  forward[, direction_delay := timegroup_delay - timegroup]

  data.table::set(forward,
                  j = c('timegroup_min', 'timegroup_max','timegroup_delay'),
                  value = NULL)

  # "Reverse": replicate forward but reverse direction ID1 <- ID2
  reverse <- data.table::copy(forward)
  data.table::setnames(reverse, c('ID1', 'ID2'), c('ID2', 'ID1'))
  reverse[, direction_delay := - direction_delay]

  out <- data.table::rbindlist(list(
    forward,
    reverse
  ), use.names = TRUE)

  data.table::setorder(out, timegroup)
  data.table::setcolorder(
    out,
    c('timegroup', 'ID1', 'ID2', 'dyadID', 'fusionID')
  )

  return(out)
}

#' Fission-fusion id
#'
#' `fusion_id` identifies fusion events in distance based edge-lists.
#' The function expects a distance based edge-list generated by
#' `edge_dist`, a threshold argument and arguments controlling how fusion
#' events are defined.
#'
#' The `edges` must be a `data.table` returned by the `edge_dist`
#' function. In addition, `fusion_id` requires a dyad ID set on the edge
#' list generated by `dyad_id`. If your data is a `data.frame`, you
#' can convert it by reference using
#' [data.table::setDT()].
#'
#' The `threshold` must be provided in the units of the coordinates. The
#' `threshold` must be larger than 0. Eg. in the case of UTM, a `threshold` = 50
#' would indicate a 50 m distance threshold.
#'
#' The `n_min_length` argument defines the minimum number of successive
#' fixes that are required to establish a fusion event. The `n_max_missing`
#' argument defines the the maximum number of allowable missing observations for
#' the dyad within a fusion event. The `allow_split` argument defines if a
#' single observation can be greater than the threshold distance without
#' initiating fission event.
#'
#' @return `fusion_id` returns the input `edges` appended with a
#'   `fusionID` column.
#'
#'   This column represents the fusion event id. As with `spatsoc`'s
#'   grouping functions, the actual value of `fusionID` is arbitrary and
#'   represents the identity of a given fusion event. If the data was reordered,
#'   the `fusionID` may change, but the membership of each fusion event
#'   would not.
#'
#'   A message is returned when a column named `fusionID` already exists in
#'   the input `edges`, because it will be overwritten.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @param edges distance based edge-list generated by `edge_dist` function,
#'   with 'dyadID' column generated by `dyad_ID`
#' @param threshold spatial distance threshold in the units of the crs
#' @param n_min_length minimum length of fusion events
#' @param n_max_missing maximum number of missing observations within a fusion
#'   event
#' @param allow_split logical defining if a single observation can be greater
#'   than the threshold distance without initiating fission event
#' @export
#' @seealso `edge_dist`
#' @references
#' See examples of identifying fission-fusion events with spatiotemporal data:
#'  * \doi{doi:10.1111/ele.12457}
#'  * \doi{doi:10.1016/j.anbehav.2018.03.014}
#'  * \doi{doi:10.1890/08-0345.1}
#' @examples
#'
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#'
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#'
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Edge-list generation
#' edges <- edge_dist(
#'     DT,
#'     threshold = 100,
#'     id = 'ID',
#'     coords = c('X', 'Y'),
#'     timegroup = 'timegroup',
#'     returnDist = TRUE,
#'     fillNA = TRUE
#'   )
#'
#' dyad_id(edges, 'ID1', 'ID2')
#'
#' fusion_id(
#'   edges = edges,
#'   threshold = 100,
#'   n_min_length = 1,
#'   n_max_missing = 0,
#'   allow_split = FALSE
#'   )
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
#' dyad_id(edges, id = 'ID1', id2 = 'ID2')
#' fusion_id(edges, threshold = 100)
fusion_id <- function(
    edges = NULL,
    threshold = NULL,
    n_min_length = 0,
    n_max_missing = 0,
    allow_split = FALSE)  {

  # due to NSE notes  in R CMD check
  . <- both_run <- distance <- dyadID <- fusionID <- tg_diff <- timegroup <-
    within_dist_run <- is_lead <- within_dist <- NULL

  assert_not_null(edges)

  check_colnames <- c('dyadID', 'timegroup', 'distance')
  assert_are_colnames(edges, check_colnames)

  assert_not_null(threshold)
  assert_inherits(threshold, 'numeric')
  assert_relation(threshold, `>`, 0)

  assert_inherits(n_min_length, 'numeric')
  assert_inherits(n_max_missing, 'numeric')
  assert_inherits(allow_split, 'logical')

  unique_edges <- unique(edges[, .(dyadID, timegroup, distance)])

  data.table::setorder(unique_edges, 'timegroup')

  # Define timegroup threshold
  tg_threshold <- 1 + n_max_missing

  # Check if edge distance less than threshold
  unique_edges[, within_dist := distance < threshold]

  # If allow split, check if previously within threshold, and
  #   timegroup difference between before and after is only 1
  if (allow_split) {
    unique_edges[, within_dist := data.table::fifelse(
      within_dist | timegroup == min(timegroup) | timegroup == max(timegroup),
      within_dist,
      data.table::shift(within_dist, type = 'lag') &
        data.table::shift(within_dist, type = 'lead') &
        abs(timegroup - data.table::shift(timegroup, type = 'lag')) <= tg_threshold &
        abs(timegroup - data.table::shift(timegroup, type = 'lead')) <= tg_threshold
    ), by = dyadID]
  }

  # Runs of within
  unique_edges[, within_dist_run := data.table::rleid(within_dist), by = dyadID]
  unique_edges[!(within_dist), within_dist_run := NA_integer_]

  # Check timegroup difference, unless first obs for dyad
  unique_edges[, tg_diff := data.table::fifelse(
    within_dist,
    timegroup - data.table::shift(timegroup, type = 'lag') <= 1 |
      timegroup == min(timegroup),
    NA
  ), by = dyadID]

  # If missing obs allowed, adjust timegroup difference to allow as long as
  #   previously within threshold distance
  if (n_max_missing > 0) {
    unique_edges[, tg_diff := data.table::fifelse(
      tg_diff,
      tg_diff,
      data.table::shift(within_dist, type = 'lag') &
        (timegroup - data.table::shift(timegroup, type = 'lag')) <=
        (tg_threshold)
    ), by = dyadID]
  }

  # Get runs on within and timegroup difference.
  unique_edges[(within_dist), both_run := data.table::rleid(within_dist_run, tg_diff),
               by = dyadID]

  # Adjust if runs of isolated observations together
  #  (eg. within T, T but timegroup diff F, F)
  unique_edges[(within_dist) & !(tg_diff),
               both_run := (both_run + seq.int(.N)) * -1,
               by = dyadID]

  # Check if shift lead is fusion and within spatial+temporal threshold
  unique_edges[, is_lead :=
    both_run < 0 &
    abs(timegroup - data.table::shift(timegroup, type = 'lead')) <=
      tg_threshold &
      within_dist &
      data.table::shift(within_dist, type = 'lead'),
  by = dyadID
  ]

  # If shift lead is fusion then shift id to lead id
  unique_edges[, both_run := fifelse(
    is_lead,
    data.table::shift(both_run, type = 'lead'),
    both_run
  ),
  by = dyadID]

  # If n minimum length > 0, check nrows and return NA if less than min
  if (n_min_length > 0) {
    unique_edges[!is.na(both_run), both_run := data.table::fifelse(
      .N >= n_min_length,
      both_run,
      NA_integer_),
      by = .(dyadID, both_run)]
  }

  # Set fusion id on runs and dyad id
  unique_edges[!is.na(both_run), fusionID := .GRP, by = .(dyadID, both_run)]

  # Merge fusion id onto input edges
  if ('fusionID' %in% colnames(edges)) {
    message('fusionID column will be overwritten by this function')
    data.table::set(edges, j = 'fusionID', value = NULL)
  }
  edges[unique_edges, fusionID := fusionID, on = .(timegroup, dyadID)]
  return(edges)
}

#' Direction to group centroid
#'
#' `direction_to_centroid` calculates the direction of each relocation to the
#' centroid of the spatiotemporal group identified by [group_pts()]. The
#' function expects a `data.table` with relocation data appended with a `group`
#' column from [group_pts()] and centroid columns from [centroid_group()].
#' Relocation data should be in two columns representing the X and Y
#' coordinates, or in a geometry column prepared by the helper function
#' [get_geometry()].
#'
#' The `DT` must be a `data.table`. If your data is a `data.frame`, you can
#' convert it by reference using [data.table::setDT()] or by reassigning using
#' [data.table::data.table()].
#'
#' This function expects a `group` column present generated with the
#' [group_pts()] function and centroid coordinates generated with the
#' [centroid_group()] function. The `group` argument expects the name of the
#' column in `DT` which correspond to the group column.
#'
#' See below under "Interface" for details on providing coordinates and under
#' "Direction function" for details on the underlying direction function used.
#'
#' @inheritSection direction_step Interface
#' @inheritSection direction_step Direction function
#'
#' @inheritParams direction_step
#' @inheritParams group_pts
#'
#' @return `direction_to_centroid` returns the input `DT` appended with a
#'   `direction_centroid` column indicating the direction to the group centroid
#'   in radians. A value of NaN is returned when the coordinates of the focal
#'   individual equal the coordinates of the centroid.
#'
#'   A message is returned when `direction_centroid` column already exist in the
#'   input `DT`, because they will be overwritten.
#'
#'   Missing values in coordinates / geometry are ignored and NA is returned.
#'
#'   See details for appending outputs using modify-by-reference in the
#'   [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).
#'
#' @export
#' @family Direction functions
#' @family Centroid functions
#' @seealso [centroid_group], [group_pts], [lwgeom::st_geod_azimuth()]
#' @references
#' See example of using direction to group centroid:
#'  * \doi{doi:10.1016/j.cub.2017.08.004}
#'
#' @examples
#' # Load data.table
#' library(data.table)
#' \dontshow{data.table::setDTthreads(1)}
#' # Read example data
#' DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
#'
#' # Cast the character column to POSIXct
#' DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
#' \dontshow{
#' # (Subset example data to reduce runtime on CRAN)
#' if (isFALSE(rlang::is_interactive())) {
#'   DT <- DT[as.Date(datetime) == '2017-01-17']
#' }
#' }
#' # Temporal grouping
#' group_times(DT, datetime = 'datetime', threshold = '20 minutes')
#'
#' # Spatial grouping with timegroup
#' group_pts(DT, threshold = 5, id = 'ID',
#'           coords = c('X', 'Y'), timegroup = 'timegroup')
#'
#' # Calculate group centroid
#' centroid_group(DT, coords = c('X', 'Y'), group = 'group')
#'
#' # Calculate direction to group centroid
#' direction_to_centroid(DT, coords = c('X', 'Y'), crs = 32736)
#'
#' # Or, using the new geometry interface
#' get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
#' group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
#' centroid_group(DT)
#' direction_to_centroid(DT)
direction_to_centroid <- function(
    DT = NULL,
    coords = NULL,
    crs = NULL,
    geometry = 'geometry') {

  # Due to NSE notes in R CMD check
  geo <- cent <- x <- y <- x_centroid <- y_centroid <- NULL

  assert_not_null(DT)
  assert_is_data_table(DT)

  out_colname <- 'direction_centroid'

  if (is.null(coords)) {
    if (!is.null(crs)) {
      message('crs argument is ignored when coords are null, using geometry')
    }

    assert_are_colnames(DT, geometry, ', did you run get_geometry()?')
    assert_col_inherits(DT, geometry, 'sfc_POINT')
    centroid_col <- 'centroid'
    assert_are_colnames(DT, centroid_col, ', did you run centroid_group?')
    assert_col_inherits(DT, centroid_col, 'sfc_POINT')

    if (out_colname %in% colnames(DT)) {
      message(out_colname, ' column will be overwritten by this function')
      data.table::set(DT, j = out_colname, value = NULL)
    }

    crs <- sf::st_crs(DT[[geometry]])
    use_transform <- !sf::st_is_longlat(crs)

    if (is.na(use_transform)) {
      rlang::abort(paste0('sf::st_is_longlat(crs) is ', use_transform,
                          ', ensure crs is provided for direction functions'))
    }

    DT[!sf::st_is_empty(geo) & !sf::st_is_empty(cent),
      c(out_colname) := calc_direction(
        geometry_a = geo,
        geometry_b = cent,
        use_transform = use_transform
      ),
      env = list(geo = geometry, cent = centroid_col)
    ]

  } else {
    assert_are_colnames(DT, coords)
    assert_length(coords, 2)
    assert_col_inherits(DT, coords, 'numeric')
    assert_not_null(crs)

    xcol <- data.table::first(coords)
    ycol <- data.table::last(coords)
    pre <- 'centroid_'
    xcol_centroid <- paste0(pre, xcol)
    ycol_centroid <- paste0(pre, ycol)
    coords_centroid  <- c(xcol_centroid, ycol_centroid)

    assert_are_colnames(DT, coords_centroid, ', did you run centroid_group?')
    assert_col_inherits(DT, coords_centroid, 'numeric')

    if (out_colname %in% colnames(DT)) {
      message(out_colname, ' column will be overwritten by this function')
      data.table::set(DT, j = out_colname, value = NULL)
    }

    use_transform <- !sf::st_is_longlat(crs)

    if (is.na(use_transform)) {
      rlang::abort(paste0('sf::st_is_longlat(crs) is ', use_transform,
                          ', ensure crs is provided for direction functions'))
    }

    DT[!is.na(x) & !is.na(y) & !is.na(x_centroid) & !is.na(y_centroid),
      c(out_colname) := calc_direction(
        x_a = x, y_a = y,
        x_b = x_centroid, y_b = y_centroid,
        crs = crs,
        use_transform = use_transform
      ),
      env = list(
        x = xcol, y = ycol,
        x_centroid = xcol_centroid, y_centroid = ycol_centroid
      )
    ]
  }

  return(DT[])
}

---
title: "Introduction to spatsoc"
author: "Alec L. Robitaille, Quinn Webber and Eric Vander Wal"
output: 
  rmarkdown::html_vignette:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Introduction to spatsoc}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r knitropts, include = FALSE}
knitr::opts_chunk$set(message = TRUE, 
                      warning = FALSE,
                      eval = FALSE, 
                      echo = FALSE)
```

The `spatsoc` package provides functionality for analyzing animal relocation data in time and space to identify potential interactions among individuals and build gambit-of-the-group data for constructing social networks. 

The package contains grouping and edge-list generating functions that are used for identifying spatially and temporally explicit groups from input data. In addition, we provide social network analysis functions for randomizing individual identifiers within groups, designed to test whether social networks generated from animal relocation data were based on non-random social proximity among individuals and for generating group by individual matrices.

The functions were developed for application across animal relocation data, for example, proximity based social network analyses and spatial and temporal clustering of points.

---

See the other vignettes for further information:

- [Introduction to spatsoc](https://docs.ropensci.org/spatsoc/articles/intro.html)
  - temporal grouping
  - spatiotemporal grouping with `group_pts`, `group_lines`, `group_polys`
  - distance based edge-list generation with `edge_dist`
- [Frequently asked questions about spatsoc](https://docs.ropensci.org/spatsoc/articles/faq.html)
  - install
  - function details for `group_times`, `group_pts`, `group_lines`, `group_polys`,
  `edge_dist`, `edge_nn`, and `randomizations`
  - package design including modify-by-reference, data.table column allocation
  - calculating summary information
- [Using spatsoc in social network analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html)
  - generating gambit-of-the-group data
  - generating observed networks
  - data stream randomization, randomized networks
  - network metrics
- [Using distance based edge-lists generating functions, dyad_id, and fusion_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html)
  - generate distance based edge-lists with `edge_dist` and `edge_nn`
  - generate dyad identifiers for edge-lists with `dyad_id`
  - identify fusion events with `fusion_id`
- [Geometry interface](https://docs.ropensci.org/spatsoc/articles/geometry-interface-and-spatial-measures.html)
  - using `get_geometry` to setup a geometry column and use the geometry interface
  - details of underlying distance, direction and centroid spatial measures
  - converting to and from related packages
- [Interspecific interactions](https://docs.ropensci.org/spatsoc/articles/interspecific-interactions.html)
  - combine two movement datasets
  - identify interspecific interactions

## Data preparation

```{r, eval = TRUE, echo = TRUE}
# Load packages
library(spatsoc)
library(data.table)
```

```{r, echo = FALSE, eval = TRUE}
data.table::setDTthreads(1)
```

```{r, eval = TRUE, echo = TRUE}
# Read in spatsoc's example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Use subset of individuals
DT <- DT[ID %in% c('H', 'I', 'J')]

# Cast character column 'datetime' as POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
DT <- DT[ID %chin% c('H', 'I', 'J')]
```

```{r, eval = TRUE, echo = FALSE}
knitr::kable(DT[, .SD[1:2], ID][order(ID)])
```


`spatsoc` expects a `data.table` for all of its functions. If you have a
`data.frame`, you can use `data.table::setDT()` to convert it by reference. If
your data is a CSV, you can use `data.table::fread()` to import it as a
`data.table`.

The data consist of relocations of `r DT[, uniqueN(ID)]` individuals over `r
DT[, max(yday(datetime)) - min(yday(datetime))]` days. Using these data, we can
compare the various grouping methods available in `spatsoc`. Note: these
examples will use a subset of the data, only individuals H, I and J.

## Temporal grouping
The `group_times` function is used to group relocations temporally. It is
flexible to a threshold provided in units of minutes, hours or days. Since GPS
fixes taken at regular intervals have some level of variability, we will provide
a time threshold (`threshold`), to consider all fixes within this threshold
taken at the same time. Alternatively, we may want to understand different
scales of grouping, perhaps daily movement trajectories or seasonal home range
overlap.

```{r groupmins, echo = TRUE}
group_times(DT, datetime = 'datetime', threshold = '5 minutes')
```

```{r tableSetUp, eval = TRUE}
nRows <- 9
```

```{r tabgroupmins, eval = TRUE}
knitr::kable(
  group_times(DT, threshold = '5 minutes', datetime = 'datetime')[, 
    .(ID, X, Y, datetime, minutes, timegroup)][
      order(datetime)][
        1:nRows])
```


A message is returned when `group_times` is run again on the same `DT`, as the columns already exist in the input `DT` and will be overwritten. 

```{r grouphours, echo = TRUE}
group_times(DT, datetime = 'datetime', threshold = '2 hours')
```

```{r tabgrouphours, eval = TRUE}
knitr::kable(
  group_times(DT, threshold = '2 hours', datetime = 'datetime')[, 
    .(ID, X, Y, datetime, hours, timegroup)][
      order(datetime)][
        1:nRows])
```

```{r groupdays, echo = TRUE}
group_times(DT, datetime = 'datetime', threshold = '5 days')
```

```{r tabgroupdays, eval = TRUE}
knitr::kable(
  group_times(
    DT, 
    threshold = '5 days', 
    datetime = 'datetime')[,
      .SD[sample(.N, 3)], by = .(timegroup, block)][order(datetime)][
        1:nRows, .(ID, X, Y, datetime, block, timegroup)]
)
```

## Spatial grouping
The `group_pts` function compares the relocations of all individuals in each
timegroup and groups individuals based on a distance threshold provided by the
user. The `group_pts` function uses the "chain rule" where three or more
individuals that are all within the defined threshold distance of at least one
other individual are considered in the same group. For point based spatial
grouping with a distance threshold that does not use the chain rule, see
`edge_dist` below.

```{r grouppts, echo = TRUE}
group_times(DT = DT, datetime = 'datetime', threshold = '15 minutes')
group_pts(DT, threshold = 50, id = 'ID', coords = c('X', 'Y'), timegroup = 'timegroup')
```

```{r fakegrouppts, eval = TRUE}
DT <- group_times(DT = DT, datetime = 'datetime', 
                     threshold = '15 minutes')
DT <- group_pts(
    DT = DT,
    threshold = 50, id = 'ID', coords = c('X', 'Y'),
    timegroup = 'timegroup')

knitr::kable(
  DT[
      between(group,  771, 774)][order(timegroup)][
        1:nRows, .(ID, X, Y, timegroup, group)]
)
```


The `group_lines` function groups individuals whose trajectories intersect in a
specified time interval. This represents a coarser grouping method than
`group_pts` which can help understand shared space at daily, weekly or other
temporal resolutions.

```{r fakegrouplines, echo = TRUE}
utm <- 32736

group_times(DT = DT, datetime = 'datetime', threshold = '1 day')
group_lines(DT, threshold = 50, crs = utm, 
            id = 'ID', coords = c('X', 'Y'),
            timegroup = 'timegroup', sortBy = 'datetime')
```

```{r grouplines, eval = TRUE}
utm <- 32736

DT <- group_times(DT = DT, datetime = 'datetime', 
                threshold = '1 day')
DT <- group_lines(DT,
                  threshold = 50, crs = utm, 
                  id = 'ID', coords = c('X', 'Y'), 
                  sortBy = 'datetime', timegroup = 'timegroup')
knitr::kable(
  unique(DT[, .(ID, timegroup, group)])[, .SD[1:3], ID][order(timegroup)]
)
```


The `group_polys` function groups individuals whose home ranges intersect. This
represents the coarsest grouping method, to provide a measure of overlap across
seasons, years or all available relocations. It can either return the proportion
of home range area overlapping between individuals or simple groups. Home ranges
are calculated using `adehabitatHR::kernelUD` or `adehabitatHR::mcp`.
Alternatively, a simple features POLYGON or MULTIPOLYGON object can be provided
to the `sfPolys` argument along with an id column.

```{r fakegrouppolys, echo = TRUE}
utm <- 32736
group_times(DT = DT, datetime = 'datetime', threshold = '8 days')
group_polys(DT = DT, area = TRUE, hrType = 'mcp',
           hrParams = list('percent' = 95),
           crs = utm,
           coords = c('X', 'Y'), id = 'ID')
```

```{r grouppolys, eval = TRUE}
utm <- 32736
DT <- group_times(DT = DT, datetime = 'datetime', threshold = '8 days')
knitr::kable(
  data.frame(group_polys(
    DT, 
    area = TRUE, hrType = 'mcp',
           hrParams = list('percent' = 95),
           crs = utm,
           coords = c('X', 'Y'), id = 'ID')[
             , .(ID1, ID2, area, proportion)])
)
```

## Edge-list generation
The `edge_dist` function calculates the geographic distance between between
individuals within each timegroup and returns all paired relocations within the
spatial threshold. `edge_dist` uses a distance matrix like group_pts, but, in
contrast, does not use the chain rule to group relocations.

```{r edgedist, echo = TRUE}
group_times(DT = DT, datetime = 'datetime', threshold = '15 minutes')
edge_dist(DT, threshold = 50, id = 'ID', 
          coords = c('X', 'Y'), timegroup = 'timegroup', fillNA = TRUE)
```

```{r fakeedgedist, eval = TRUE}
DT <- group_times(DT = DT, datetime = 'datetime', 
                     threshold = '15 minutes')
edges <- edge_dist(DT, threshold = 50, id = 'ID', 
                   coords = c('X', 'Y'), timegroup = 'timegroup', fillNA = TRUE)

knitr::kable(
  edges[between(timegroup, 158, 160)]
)
```

The `edge_nn` function calculates the nearest neighbour to each individual
within each time group. If the optional distance threshold is provided, it is
used to limit the maximum distance between neighbours. `edge_nn` returns an edge
list of each individual and their nearest neighbour.

```{r edgenn, echo = TRUE}
group_times(DT = DT, datetime = 'datetime', threshold = '15 minutes')
edge_nn(DT, id = 'ID', coords = c('X', 'Y'), timegroup = 'timegroup')
```

```{r fakeedgenn, eval = TRUE}
DT <- group_times(DT = DT, datetime = 'datetime', 
                     threshold = '15 minutes')
edges <- edge_nn(DT, id = 'ID', coords = c('X', 'Y'), timegroup = 'timegroup')

knitr::kable(
  edges[1:6]
)
```

---
title: "Using spatsoc in social network analysis"
author: "Alec L. Robitaille, Quinn Webber and Eric Vander Wal"
output:
  rmarkdown::html_vignette:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Using spatsoc in social network analysis}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r knitropts, include = FALSE}
knitr::opts_chunk$set(message = TRUE, 
                      warning = FALSE,
                      eval = FALSE, 
                      echo = TRUE)
```

`spatsoc` can be used in social network analysis to generate gambit of the group
format data from GPS relocation data, perform data stream randomization and
generate group by individual matrices.

Gambit of the group format data is generated using the grouping functions:

* `group_times`
* `group_pts`
* `group_lines`
* `group_polys`

Data stream randomization is performed using the `randomizations` function. Group
by individual matrices are generated using the `get_gbi` function. 

--- 

See the other vignettes for further information:

- [Introduction to spatsoc](https://docs.ropensci.org/spatsoc/articles/intro.html)
  - temporal grouping
  - spatiotemporal grouping with `group_pts`, `group_lines`, `group_polys`
  - distance based edge-list generation with `edge_dist`
- [Frequently asked questions about spatsoc](https://docs.ropensci.org/spatsoc/articles/faq.html)
  - install
  - function details for `group_times`, `group_pts`, `group_lines`, `group_polys`,
  `edge_dist`, `edge_nn`, and `randomizations`
  - package design including modify-by-reference, data.table column allocation
  - calculating summary information
- [Using spatsoc in social network analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html)
  - generating gambit-of-the-group data
  - generating observed networks
  - data stream randomization, randomized networks
  - network metrics
- [Using distance based edge-lists generating functions, dyad_id, and fusion_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html)
  - generate distance based edge-lists with `edge_dist` and `edge_nn`
  - generate dyad identifiers for edge-lists with `dyad_id`
  - identify fusion events with `fusion_id`
- [Geometry interface](https://docs.ropensci.org/spatsoc/articles/geometry-interface-and-spatial-measures.html)
  - using `get_geometry` to setup a geometry column and use the geometry interface
  - details of underlying distance, direction and centroid spatial measures
  - converting to and from related packages
- [Interspecific interactions](https://docs.ropensci.org/spatsoc/articles/interspecific-interactions.html)
  - combine two movement datasets
  - identify interspecific interactions

# Generate gambit of the group data
spatsoc provides users with one temporal (`group_times`) and three spatial
(`group_pts`, `group_lines`, `group_polys`) functions to generate gambit of the
group data from GPS relocations. Users can consider spatial grouping at three
different scales combined with an appropriate temporal grouping threshold. The
gambit of the group data is then used to generate a group by individual matrix
and build the network.


## 1. Load packages and prepare data
`spatsoc` expects a `data.table` for all `DT` arguments and date time columns to
be formatted `POSIXct`.

```{r}
## Load packages
library(spatsoc)
library(data.table)
library(asnipe)
library(igraph)
```

```{r, echo = FALSE, eval = TRUE}
data.table::setDTthreads(1)
```

```{r}
## Read data as a data.table
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

## Cast datetime column to POSIXct
DT[, datetime := as.POSIXct(datetime)]

## Calculate the year of the relocation 
DT[, yr := year(datetime)]
```


Next, we will group relocations temporally with `group_times` and spatially with
one of `group_pts`, `group_lines`, `group_polys`. Note: these are mutually
exclusive, only select one spatial grouping function at a time.

## 2. a) `group_pts` 

Point based grouping by calculating distance between relocations in each
timegroup. Depending on species and study system, relevant temporal and spatial
grouping thresholds are used. In this case, relocations within 5 minutes and 50
meters are grouped together.

```{r}
## Temporal groups
group_times(DT, datetime = 'datetime', threshold = '5 minutes')

## Spatial groups
group_pts(
  DT,
  threshold = 50,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup'
)

```

## 2. b) `group_lines`

Line based grouping by measuring intersection of, optionally buffered,
trajectories for each individual in each timegroup. Longer temporal thresholds
are used to measure, for example, intersecting daily trajectories.

```{r, eval = FALSE}
# EPSG code for relocations
utm <- 32736

## Group relocations by julian day
group_times(DT, datetime = 'datetime', threshold = '1 day')

## Group lines for each individual and julian day
group_lines(
  DT,
  threshold = 50,
  crs = utm,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  sortBy = 'datetime'
)
```


## 2. c) `group_polys`

Polygon based grouping by generating home ranges using `adehabitatHR` and
measuring intersection or proportional overlap. Longer temporal thresholds are
used to create seasonal, monthly, yearly home ranges.

```{r, eval = FALSE}
# EPSG code for relocations
utm <- 32736

## Option 1: area = FALSE and home range intersection 'group' column added to DT 
group_polys(
  DT,
  area = FALSE,
  hrType = 'mcp',
  hrParams = list(percent = 95),
  crs = utm,
  id = 'ID',
  coords = c('X', 'Y')
)

## Option 2: area = TRUE 
#  results must be assigned to a new variable 
#  data.table returned has ID1, ID2 and proportion and area overlap
areaDT <- group_polys(
  DT,
  area = TRUE,
  hrType = 'mcp',
  hrParams = list(percent = 95),
  crs = utm,
  id = 'ID',
  coords = c('X', 'Y')
)

```

# Build observed network 
Once we've generated groups using `group_times` and one of the spatial grouping
functions, we can generate a group by individual matrix.

The following code chunk showing `get_gbi` can be used for outputs from any of
`group_pts`, `group_lines` or `group_polys(area = FALSE)`. For the purpose of
this vignette however, we will consider the outputs from `group_pts` ([2.
a)](#a-group_pts)) for the following code chunk.

Note: we show this example creating the group by individual matrix and network
for only 2016 to illustrate how `spatsoc` can be used for simpler data with no
splitting of temporal or spatial subgroups (e.g.: yearly, population). See the
random network section for how to use `spatsoc` in social network analysis for
multi-year or other complex data.

## 3. `get_gbi`
```{r}
## Subset DT to only year 2016
subDT <- DT[yr == 2016]

## Generate group by individual matrix
# group column generated by spatsoc::group_pts
gbiMtrx <- get_gbi(DT = subDT, group = 'group', id = 'ID')
```

Note: `spatsoc::get_gbi` is identical in function to
`asnipe::get_group_by_individual`, but is more efficient (some benchmarks
measuring >10x improvements) thanks to `data.table::dcast`.


## 4. `asnipe::get_network`
Next, we can use `asnipe::get_network` to build the observed social network.
Ensure that the argument "data_format" is "GBI". Use other arguments that are
relevant to your analysis, here we calculate a Simple ratio index.

```{r}
## Generate observed network
net <- get_network(gbiMtrx,
                   data_format = "GBI",
                   association_index = "SRI")
```


# Data stream randomization
Three types of data stream randomization are provided by `spatsoc`'s
`randomizations` function:

* step: randomizes identities of relocations between individuals within each time step.
* daily: randomizes identities of relocations between individuals within each day.
* trajectory: randomizes daily trajectories within individuals 
([Spiegel et al. 2016](https://doi.org/10.1111/2041-210X.12553)).

The results of `randomizations` must be assigned. The function returns the `id`
and `datetime` columns provided (and anything provided to `splitBy`). In
addition, columns 'observed' and 'iteration' are returned indicating observed
rows and which iteration rows correspond to (where 0 is the observed).

As with spatial grouping functions, these methods are mutually exclusive. Pick
one `type` and rebuild the network after randomization.

Note: the `coords` argument is only required for trajectory type randomization,
since after randomizing with this method, the 'coords' are needed to redo
spatial grouping (with `group_pts`, `group_lines` or `group_polys`).


## 5. a) `type = 'step'`
`'step'` randomizes identities of relocations between individuals within each
time step. The `datetime` argument expects an integer group generated by
`group_times`. The `group` argument expects the column name of the group
generated from the spatial grouping functions.

Four columns are returned when `type = 'step'` along with `id`, `datetime` and
`splitBy` columns:

* 'randomID' - randomly selected ID from IDs within each time step
* 'observed' - observed rows (TRUE/FALSE)
* 'iteration' - which iteration rows correspond to (0 is observed)

```{r}
# Calculate year column to ensure randomization only occurs within years
#   since data spans multiple years
DT[, yr := year(datetime)]

## Step type randomizations
#  providing 'timegroup' (from group_times) as datetime
#  splitBy = 'yr' to force randomization only within year
randStep <- randomizations(
   DT,
   type = 'step',
   id = 'ID',
   group = 'group',
   coords = NULL,
   datetime = 'timegroup',
   iterations = 3,
   splitBy = 'yr'
)
```


## 5. b) `type = 'daily'`
`'daily'` randomizes identities of relocations between individuals within each
day. The `datetime` argument expects a datetime `POSIXct` format column.

Four columns are returned when `type = 'daily'` along with `id`, `datetime` and
`splitBy` columns:

* 'randomID' - randomly selected ID for each day
* 'jul' - julian day
* 'observed' - observed rows (TRUE/FALSE)
* 'iteration' - which iteration rows correspond to (0 is observed)

```{r}
# Calculate year column to ensure randomization only occurs within years 
#   since data spans multiple years
DT[, yr := year(datetime)]

## Daily type randomizations
# splitBy = 'yr' to force randomization only within year
randDaily <- randomizations(
   DT,
   type = 'daily',
   id = 'ID',
   group = 'group',
   coords = NULL,
   datetime = 'datetime',
   splitBy = 'yr',
   iterations = 20
)
```

## 5. c) `type = 'trajectory'`
`'trajectory'` randomizes daily trajectories within individuals 
([Spiegel et al. 2016](https://doi.org/10.1111/2041-210X.12553)). 
The `datetime` argument expects
a datetime `POSIXct` format column.

Five columns are returned when `type = 'trajectory'` along with `id`, `datetime`
and `splitBy` columns:

* random date time ("random" prefixed to *datetime* argument)
* 'jul' - observed julian day
* 'observed' - observed rows (TRUE/FALSE)
* 'iteration' - which iteration rows correspond to (0 is observed)
* 'randomJul' - random julian day relocations are swapped to from observed 
julian day

```{r}
# Calculate year column to ensure randomization only occurs within years 
#   since data spans multiple years
DT[, yr := year(datetime)]

## Trajectory type randomization
randTraj <- randomizations(
   DT,
   type = 'trajectory',
   id = 'ID',
   group = NULL,
   coords = c('X', 'Y'),
   datetime = 'datetime',
   splitBy = 'yr',
   iterations = 20
)
```


# Build random network 
Once we've randomized the data stream with `randomizations`, we can build the
random network.

We will use the `get_gbi` function directly when `type` is either 'step' or
'daily'. For `type = 'trajectory'`, we will recalculate spatial groups with one
of `group_pts`, `group_lines`, `group_polys` for the randomized data. In this
case, the example shows `group_pts`.

Since we want to create a group by individual matrix for each random iteration
(and in this case, each year), we will use `mapply` to work on subsets of the
randomized data.

Note: building the random networks depends on the `type` used and therefore, the
following chunks are mutually exclusive. Use the one that corresponds to the
randomization type you used above.

## 6. a) `type = 'step'`
`randomizations` with `type = 'step'` returns a 'randomID' which should be used
instead of the observed 'ID' to generate the group by individual matrix.

After `get_gbi`, we use `asnipe::get_network` to build the random network. 

```{r}
## Create a data.table of unique combinations of iteration and year, excluding observed rows
iterYearLs <- unique(randStep[!(observed), .(iteration, yr)])

## Generate group by individual matrix 
# for each combination of iteration number and year
# 'group' generated by spatsoc::group_pts
# 'randomID' used instead of observed ID (type = 'step')
gbiLs <- mapply(FUN = function(i, y) {
  get_gbi(randStep[iteration == i & yr == y],
          'group', 'randomID')
  },
  i = iterYearLs$iter,
  y = iterYearLs$yr,
  SIMPLIFY = FALSE
)

## Generate a list of random networks
netLs <- lapply(gbiLs, FUN = get_network,
                data_format = "GBI", association_index = "SRI")

```


## 6. b) `type = 'daily'`
`randomizations` with `type = 'step'` returns a 'randomID' which should be used
instead of the observed 'ID' to generate the group by individual matrix.

After `get_gbi`, we use `asnipe::get_network` to build the random network. 

In this case, we will generate a fake column representing a "population" to show
how we can translate the `mapply` chunk above to three (or more variables).

```{r}
## Generate fake population
randDaily[, population := sample(1:2, .N, replace = TRUE)]

## Create a data.table of unique combinations of iteration, year, and population, 
#    excluding observed rows
iterYearLs <- unique(randStep[!(observed), .(iteration, yr, population)])

## Generate group by individual matrix 
# for each combination of iteration number and year
# 'group' generated by spatsoc::group_pts
# 'randomID' used instead of observed ID (type = 'step')
gbiLs <- mapply(FUN = function(i, y, p) {
  get_gbi(randDaily[iteration == i & 
                      yr == y & population == p],
          'group', 'randomID')
  },
  i = iterYearLs$iter,
  y = iterYearLs$yr,
  p = iterYearLs$population,
  SIMPLIFY = FALSE
)

## Generate a list of random networks
netLs <- lapply(gbiLs, FUN = get_network,
                data_format = "GBI", association_index = "SRI")

```



## 6. c) `type = 'trajectory'`
`randomizations` with `type = 'trajectory'` returns a random date time which
should be used instead of the observed date time to generate random gambit of
the group data.

First, we pass the randomized data to `group_times` using the random date time
for `datetime`.

After `get_gbi`, we use `asnipe::get_network` to build the random network. 

```{r}
## Randomized temporal groups
# 'datetime' is the randomdatetime produced by randomizations(type = 'trajectory')
group_times(randTraj, datetime = 'randomdatetime', threshold = '5 minutes')

## Randomized spatial groups
# 'iteration' used in splitBy to ensure only points within each iteration are grouped
group_pts(randTraj, threshold = 50, id = 'ID', coords = c('X', 'Y'),
          timegroup = 'timegroup', splitBy = 'iteration')

## Create a data.table of unique combinations of iteration and year, excluding observed rows
iterYearLs <- unique(randStep[!(observed), .(iteration, yr)])

## Generate group by individual matrix 
# for each combination of iteration number and year
# 'group' generated by spatsoc::group_pts
# 'ID' used since datetimes were randomized within individuals
gbiLs <- mapply(FUN = function(i, y) {
  get_gbi(randTraj[iteration == i & yr == y],
          'group', 'ID')
  },
  i = iterYearLs$iter,
  y = iterYearLs$yr,
  SIMPLIFY = FALSE
)

## Generate a list of random networks
netLs <- lapply(gbiLs, FUN = get_network,
                data_format = "GBI", association_index = "SRI")

```


# Network metrics
Finally, we can calculate some network metrics. Please note that there are many
ways of interpreting, analyzing and measuring networks, so this will simply show
one option.


## 7. Calculate observed network metrics
To calculate observed network metrics, use the network (`net`) produced in
[4.](#asnipeget_network) from 2016 data.

```{r}
## Generate graph
g <- graph.adjacency(net, 'undirected', 
                     diag = FALSE, weighted = TRUE)

## Metrics for all individuals 
observed <- data.table(
  centrality = evcent(g)$vector,
  strength = graph.strength(g),
  degree = degree(g),
  ID = names(degree(g)),
  yr = subDT[, unique(yr)]
)
```


## 8. Calculate random network metrics
With the list of random networks from [6.](#build-random-network), we can
generate a list of graphs with `igraph::graph.adjacency` (for example) and
calculate random network metrics.

This example uses the `netLs` generated by [6. a)](#a-type-step-1) which was
split by year and iteration.

```{r}
## Generate graph and calculate network metrics
mets <- lapply(seq_along(netLs), function(n) {
  g <- graph.adjacency(netLs[[n]], 'undirected', 
                       diag = FALSE, weighted = TRUE)
  
  data.table(
    centrality = evcent(g)$vector,
    strength = graph.strength(g),
    degree = degree(g),
    ID = names(degree(g)),
    iteration = iterYearLs$iter[[n]],
    yr = iterYearLs$yr[[n]]
    )
})

## Metrics for all individuals across all iterations and years
random <- rbindlist(mets)

## Mean values for each individual and year
meanMets <- random[, lapply(.SD, mean), by = .(ID, yr),
                .SDcols = c('centrality', 'strength', 'degree')]
```

## 9. Compare observed and random metrics
Instead of calculating observed and random metrics separately (shown in
[7.](#calculate-observed-network-metrics) and
[8.](#calculate-random-network-metrics)), we can calculate metrics for both at
the same time and compare.

This chunk expects the outputs from [5. a)](#a-type-step), skipping steps 6.-8.

Note: by removing the `!(observed)` subset from `randStep` performed in [6.
a)](#a-type-step-1), we will include observed rows where `iteration == 0`. This
will return a `gbiLs` where the observed and random rows are included in the
same `data.table`.

```{r}
## Create a data.table of unique combinations of iteration and year, including observed and random rows
iterYearLs <- unique(randStep[, .(iteration, yr)])

## Generate group by individual matrix 
# for each combination of iteration and year
# 'group' generated by spatsoc::group_pts
# 'randomID' used instead of observed ID (type = 'step')
gbiLs <- mapply(FUN = function(i, y) {
  get_gbi(randStep[iteration == i & yr == y],
          'group', 'randomID')
  },
  i = iterYearLs$iter,
  y = iterYearLs$yr,
  SIMPLIFY = FALSE
)

## Generate a list of random networks
netLs <- lapply(gbiLs, FUN = get_network,
                data_format = "GBI", association_index = "SRI")

## Generate graph and calculate network metrics
mets <- lapply(seq_along(netLs), function(n) {
  g <- graph.adjacency(netLs[[n]], 'undirected', 
                       diag = FALSE, weighted = TRUE)
  
  data.table(
    centrality = evcent(g)$vector,
    strength = graph.strength(g),
    ID = names(degree(g)),
    iteration = iterYearLs$iter[[n]],
    yr = iterYearLs$yr[[n]]
    )
})

## Observed and random for all individuals across all iterations and years
out <- rbindlist(mets)

## Split observed and random
out[, observed := ifelse(iteration == 0, TRUE, FALSE)]

## Mean values for each individual and year, by observed/random
meanMets <- out[, lapply(.SD, mean), by = .(ID, yr, observed),
                .SDcols = c('centrality', 'strength')]

```

---
title: "Using distance based edge-list generating functions, dyad_id and fusion_id"
author: "Alec L. Robitaille, Quinn Webber and Eric Vander Wal"
output: 
  rmarkdown::html_vignette:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Using distance based edge-list generating functions, dyad_id and fusion_id}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  eval = FALSE,
  echo = TRUE,
  comment = "#>"
)
```

`spatsoc` can be used in social network analysis to generate distance based
edge-lists from GPS relocation data using either the `edge_dist` or the
`edge_nn` function.

---

See the other vignettes for further information:

- [Introduction to spatsoc](https://docs.ropensci.org/spatsoc/articles/intro.html)
  - temporal grouping
  - spatiotemporal grouping with `group_pts`, `group_lines`, `group_polys`
  - distance based edge-list generation with `edge_dist`
- [Frequently asked questions about spatsoc](https://docs.ropensci.org/spatsoc/articles/faq.html)
  - install
  - function details for `group_times`, `group_pts`, `group_lines`, `group_polys`,
  `edge_dist`, `edge_nn`, and `randomizations`
  - package design including modify-by-reference, data.table column allocation
  - calculating summary information
- [Using spatsoc in social network analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html)
  - generating gambit-of-the-group data
  - generating observed networks
  - data stream randomization, randomized networks
  - network metrics
- [Using distance based edge-lists generating functions, dyad_id, and fusion_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html)
  - generate distance based edge-lists with `edge_dist` and `edge_nn`
  - generate dyad identifiers for edge-lists with `dyad_id`
  - identify fusion events with `fusion_id`
- [Geometry interface](https://docs.ropensci.org/spatsoc/articles/geometry-interface-and-spatial-measures.html)
  - using `get_geometry` to setup a geometry column and use the geometry interface
  - details of underlying distance, direction and centroid spatial measures
  - converting to and from related packages
- [Interspecific interactions](https://docs.ropensci.org/spatsoc/articles/interspecific-interactions.html)
  - combine two movement datasets
  - identify interspecific interactions


## Generate edge-lists

spatsoc provides users with one temporal (`group_times`) and two distance based
edge-list generating functions (`edge_dist`, `edge_nn`) to generate edge-lists
from GPS relocations. Users can consider edges defined by either the spatial
proximity between individuals (with `edge_dist`), by nearest neighbour (with
`edge_nn`) or by nearest neighbour with a maximum distance (with `edge_nn`). The
edge-lists can be used directly by the animal social network package `asnipe` to
generate networks.

### 1. Load packages and prepare data

`spatsoc` expects a `data.table` for all `DT` arguments and date time columns to
be formatted `POSIXct`.

```{r, eval = TRUE}
## Load packages
library(spatsoc)
library(data.table)
```

```{r, echo = FALSE, eval = TRUE}
data.table::setDTthreads(1)
```

```{r, eval = TRUE}
## Read data as a data.table
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

## Cast datetime column to POSIXct
DT[, datetime := as.POSIXct(datetime)]
```

Next, we will group relocations temporally with `group_times` and generate edges
lists with one of `edge_dist`, `edge_dist`. Note: these are mutually exclusive,
only select one edge-list generating function at a time.

### 2. a) `edge_dist` 

Distance based edge-lists where relocations in each timegroup are considered
edges if they are within the spatial distance defined by the user with the
`threshold` argument. Depending on species and study system, relevant temporal
and spatial distance thresholds are used. In this case, relocations within 5
minutes and 50 meters are considered edges.

This is the non-chain rule implementation similar to `group_pts`. Edges are
defined by the distance threshold and NAs are returned for individuals within
each timegroup if they are not within the threshold distance of any other
individual (if `fillNA` is TRUE).

Optionally, `edge_dist` can return the distances between individuals (less than
the threshold) in a column named 'distance' with argument `returnDist = TRUE`.

```{r, eval = TRUE}
# Temporal groups
group_times(DT, datetime = 'datetime', threshold = '5 minutes')

# Edge-list generation
edges <- edge_dist(
  DT,
  threshold = 100,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  returnDist = TRUE,
  fillNA = TRUE
)
```

### 2. b) `edge_nn`

Nearest neighbour based edge-lists where each individual is connected to their
nearest neighbour. `edge_nn` can be used to generate edge-lists defined either
by nearest neighbour or nearest neighbour with a maximum distance. As with
grouping functions and `edge_dist`, temporal and spatial threshold depend on
species and study system.

NAs are returned for nearest neighbour for an individual was alone in a
timegroup (and/or splitBy) or if the distance between an individual and its
nearest neighbour is greater than the threshold.

Optionally, `edge_nn` can return the distances between individuals (less than
the threshold) in a column named 'distance' with argument `returnDist = TRUE`.

```{r, eval = FALSE}
# Temporal groups
group_times(DT, datetime = 'datetime', threshold = '5 minutes')

# Edge-list generation
edges <- edge_nn(
  DT,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup'
)

# Edge-list generation using maximum distance threshold
edges <- edge_nn(
  DT, 
  id = 'ID', 
  coords = c('X', 'Y'),
  timegroup = 'timegroup', 
  threshold = 100
)

# Edge-list generation using maximum distance threshold, returning distances
edges <- edge_nn(
  DT, 
  id = 'ID', 
  coords = c('X', 'Y'),
  timegroup = 'timegroup', 
  threshold = 100,
  returnDist = TRUE
)

```


## Dyads

### 3. `dyad_id`

The function `dyad_id` can be used to generate a unique, undirected dyad
identifier for edge-lists.

```{r, eval = TRUE}
# In this case, using the edges generated in 2. a) edge_dist
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')
```


Once we have generated dyad ids, we can measure consecutive relocations, start
and end relocation, etc. **Note:** since the edges are duplicated A-B and B-A,
you will need to use the unique timegroup*dyadID or divide counts by 2.


## Fusion events

### 4. `fusion_id`

The function `fusion_id` can be used to identify fusion events in distance based
edge-lists. The "n_min_length" argument defines the minimum number of successive
fixes that are required to establish a fusion event. The "n_max_missing"
argument defines the the maximum number of allowable missing observations for
the dyad within a fusion event. The "allow_split" argument defines if a single
observation can be greater than the threshold distance without initiating
fission event.


```{r, eval = TRUE}
fusion_id(
  edges = edges,
  threshold = 100,
  n_min_length = 1,
  n_max_missing = 0,
  allow_split = FALSE
)

# Print first 10 fusion events
print(edges[fusionID <= 5])
```

---
title: "Geometry interface and spatial measures"
author: "Alec L. Robitaille, Quinn Webber and Eric Vander Wal"
output: 
  rmarkdown::html_vignette:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Geometry interface and spatial measures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  eval = TRUE,
  echo = TRUE,
  comment = "#>"
)
```

## Introduction

{spatsoc} v 0.2.13 provides a geometry interface and internal functions
that provide consistent spatial measures across all {spatsoc} functions.

The motivation for `get_geometry` is to allow users to supply a geometry
column instead of providing coordinates and optionally crs through the
"coords" and "crs" args. This interface is enabled by default on all
functions that accept coordinates / geometry, by leaving the coords null
and ensuring the geometry column has been set up using `get_geometry()`.

Details on spatial measures below under [Spatial measures].

------------------------------------------------------------------------

See the other vignettes for further information:

-   [Introduction to
    spatsoc](https://docs.ropensci.org/spatsoc/articles/intro.html)
    -   temporal grouping
    -   spatiotemporal grouping with `group_pts`, `group_lines`,
        `group_polys`
    -   distance based edge-list generation with `edge_dist`
-   [Frequently asked questions about
    spatsoc](https://docs.ropensci.org/spatsoc/articles/faq.html)
    -   install
    -   function details for `group_times`, `group_pts`, `group_lines`,
        `group_polys`, `edge_dist`, `edge_nn`, and `randomizations`
    -   package design including modify-by-reference, data.table column
        allocation
    -   calculating summary information
-   [Using spatsoc in social network
    analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html)
    -   generating gambit-of-the-group data
    -   generating observed networks
    -   data stream randomization, randomized networks
    -   network metrics
-   [Using distance based edge-lists generating functions, dyad_id, and
    fusion_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html)
    -   generate distance based edge-lists with `edge_dist` and
        `edge_nn`
    -   generate dyad identifiers for edge-lists with `dyad_id`
    -   identify fusion events with `fusion_id`
-   [Geometry
    interface](https://docs.ropensci.org/spatsoc/articles/geometry-interface-and-spatial-measures.html)
    -   using `get_geometry` to setup a geometry column and use the
        geometry interface
    -   details of underlying distance, direction and centroid spatial
        measures
    -   converting to and from related packages
-   [Interspecific
    interactions](https://docs.ropensci.org/spatsoc/articles/interspecific-interactions.html)
    -   combine two movement datasets
    -   identify interspecific interactions

## Setup

Load packages

```{r}
library(spatsoc)
library(data.table)
```

Example data and variables

```{r}
# Read example data
DT <- fread(system.file('extdata', 'DT.csv', package = 'spatsoc'))

coords <- c('X', 'Y')
crs <- 32736
datetime <- 'datetime'
id <- 'ID'
```

## `get_geometry`

`get_geometry` can be used to setup a `DT` with a simple feature
geometry list column. Optionally, the input coordinates can be
transformed using the argument "output_crs". The default output geometry
column name can be changed using the argument "geometry_colname".

### Usage

```{r}
# Get geometry
get_geometry(DT, coords = coords, crs = crs)

print(DT)
```

## Spatial measures

### Distance

The underlying distance function used depends on the crs of the
coordinates or geometry provided.

-   If the crs is longlat degrees (as determined by
    `sf::st_is_longlat()`), the distance function is `sf::st_distance()`
    which passes to `s2::s2_distance()` if `sf::sf_use_s2()` is TRUE and
    `lwgeom::st_geod_distance()` if `sf::sf_use_s2()` is FALSE. The
    distance returned has units set according to the crs.

-   If the crs is not longlat degrees (eg. NULL, `NA_crs_`, or
    projected), the distance function used is `stats::dist()`,
    maintaining expected behaviour from previous versions. The distance
    returned does not have units set.

Note: in both cases, if the coordinates are NA then the result will be
NA.

### Direction

The underlying distance function used depends on the crs of the
coordinates or geometry provided.

-   If the crs is provided and longlat degrees (as determined by
    `sf::st_is_longlat()`), the distance function is
    `lwgeom::st_geod_azimuth()`.

-   If the crs is provided and not longlat degrees (eg. a projected
    UTM), the coordinates or geometry are transformed to
    `sf::st_crs(4326)` before the distance is measured using
    `lwgeom::st_geod_azimuth()`.

-   If the crs is NULL or `NA_crs_`, the distance function cannot be
    used and an error is returned.

The crs for direction functions (eg. `direction_step`,
`direction_to_centroid`) should be either longlat or a crs that can be
transformed to `sf::st_crs(4326)`. A user can check their crs by using
the function `sf::st_can_transform(src = crs, dst = 4326)`.

If a user provides a crs that is not longlat and cannot be transformed
to 4326, they will receive an internal error with the following
structure:

`"In CPL_transform(x, crs, aoi, pipeline, reverse, desired_accuracy, : GDAL Error 6: Cannot find coordinate operations from..."`

### Centroid

The underlying centroid function used depends on the crs of the
coordinates or geometry provided.

-   If the crs is longlat degrees (as determined by
    `sf::st_is_longlat()`) and `sf::sf_use_s2()` is TRUE, the distance
    function is `sf::st_centroid()` which passes to `s2::s2_centroid()`.

-   If the crs is longlat degrees but `sf::sf_use_s2()` is FALSE, the
    centroid calculated will be incorrect. See `sf::st_centroid()`.

-   If the crs is not longlat degrees (eg. NULL, NA_crs\_, or
    projected), the centroid function used is mean.

Note: if the input is length 1, the input is returned directly without
any processing.

The `distance_to_centroid`, `direction_to_centroid` and
`leader_direction_group` expect a centroid column named 'centroid' when
the geometry interface is used. Ensure the `centroid_group` function has
been used with the geometry interface before running these functions.

The following warning may be returned when centroid functions are used
with longlat coordinates / geometry and the {s2} package is not
available or `sf::sf_use_s2()` is `FALSE`. Please set
`sf::sf_use_s2(TRUE)` and try again.

## Converting from other packages

Instead of using `get_geometry()`, coming from other packages, we can
convert the objects to data.tables for processing with {spatsoc}.
Caution: {spatsoc}'s expectation is that the "geometry" column is
`sfc_POINT` representing each relocation.

### Converting from {sftrack}

```{r}
library(sftrack)
data("raccoon")
raccoon$timestamp <- as.POSIXct(raccoon$timestamp, "EST")
burstz <- list(id = raccoon$animal_id, month = as.POSIXlt(raccoon$timestamp)$mon)
my_track <- as_sftrack(raccoon,
  group = burstz, time = "timestamp",
  error = NA, coords = c("longitude", "latitude")
)

DT_sftrack <- as.data.table(my_track)

print(DT_sftrack)
```

### Converting from {move2}

```{r}
library(move2)
fishers <- mt_read(mt_example(file = "fishers.csv.gz"))
DT_move2 <- as.data.table(fishers)

print(DT_move2)
```

## Converting to other packages

After processing with {spatsoc}, optionally convert output to work with
other packages. Caution: {spatsoc}'s expectation is that users take
their input data.table through any required functions in {spatsoc}
before converting to other packages or using eg. {dplyr}. Intermixing
packages between {spatsoc} functions can lead to issues with
{data.table} class or column allocation. See more details under Package
Design in [FAQ](https://docs.ropensci.org/spatsoc/articles/faq.html).

An example of a potential issue with intermixing packages between
spatsoc functions is that attributes such as the sf column's bbox aren't
automatically updated when subsetting with data.table. This impacts the
bbox returned, and also the bbox used for plotting. The recommendation
is to run all required {spatsoc} functions then convert to standard
spatial formats as described below for further processing and
visualization.

```{r}
library(sf)
DT[, st_bbox(geometry)]
DT[1:10, st_bbox(geometry)]

DT[, plot(geometry)]
DT[1:10, plot(geometry)]

# Note the bbox can also be manually recomputed
DT[1:10, st_bbox(st_sfc(geometry, recompute_bbox = TRUE))]
DT[1:10, plot(st_sfc(geometry, recompute_bbox = TRUE))]
```

### Converting to {sf}

```{r}
DT_sf <- st_as_sf(DT, sf_column_name = 'geometry')
print(DT_sf)
plot(DT_sf)
```

### Converting to {sftrack}

```{r}
library(sftrack)

# First convert to sf as above
DT_sf <- st_as_sf(DT, sf_column_name = 'geometry')
# Then convert to sftrack
DT_sftrack <- as_sftrack(DT_sf, group = c(id = id), time = datetime)

head(DT_sftrack)
plot(DT_sftrack)
```

### Converting to {move2}

```{r}
library(move2)

# Convert to a {move2} object using mt_as_move2
DT_move2 <- mt_as_move2(DT,
  time_column = datetime,
  track_id_column = id,
  sf_column_name = 'geometry'
)

print(DT_move2)
plot(DT_move2)
```

---
title: "Frequently asked questions about spatsoc"
author: "Alec L. Robitaille, Quinn Webber and Eric Vander Wal"
output: 
  rmarkdown::html_vignette:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Frequently asked questions about spatsoc}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

```{r knitropts, include = FALSE}
knitr::opts_chunk$set(message = FALSE, 
                      warning = FALSE,
                      eval = FALSE, 
                      echo = TRUE)
```

spatsoc is an R package for detecting spatial and temporal groups in GPS
relocations. It can be used to build proximity-based social networks using
gambit-of-the-group format and edge-lists. In addition, the randomization
function provides data-stream randomization methods suitable for GPS data.


---

See the other vignettes for further information:

- [Introduction to spatsoc](https://docs.ropensci.org/spatsoc/articles/intro.html)
  - temporal grouping
  - spatiotemporal grouping with `group_pts`, `group_lines`, `group_polys`
  - distance based edge-list generation with `edge_dist`
- [Frequently asked questions about spatsoc](https://docs.ropensci.org/spatsoc/articles/faq.html)
  - install
  - function details for `group_times`, `group_pts`, `group_lines`, `group_polys`,
  `edge_dist`, `edge_nn`, and `randomizations`
  - package design including modify-by-reference, data.table column allocation
  - calculating summary information
- [Using spatsoc in social network analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html)
  - generating gambit-of-the-group data
  - generating observed networks
  - data stream randomization, randomized networks
  - network metrics
- [Using distance based edge-lists generating functions, dyad_id, and fusion_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html)
  - generate distance based edge-lists with `edge_dist` and `edge_nn`
  - generate dyad identifiers for edge-lists with `dyad_id`
  - identify fusion events with `fusion_id`
- [Geometry interface](https://docs.ropensci.org/spatsoc/articles/geometry-interface-and-spatial-measures.html)
  - using `get_geometry` to setup a geometry column and use the geometry interface
  - details of underlying distance, direction and centroid spatial measures
  - converting to and from related packages
- [Interspecific interactions](https://docs.ropensci.org/spatsoc/articles/interspecific-interactions.html)
  - combine two movement datasets
  - identify interspecific interactions
  
  
# Usage

`spatsoc` leverages `data.table` to modify by reference and iteratively work on
subsets of the input data. The first input for all functions in `spatsoc` is
`DT`, an input `data.table`. If your data is a `data.frame`, you can convert it
by reference using `setDT(DF)`.

## Spatial and temporal grouping

`spatsoc` is designed to work in two steps: temporal followed by either spatial
grouping or edge-list generating. Considering your specific study species and
system, determine a relevant temporal and spatial grouping threshold. This may
be 5 minutes and 50 meters or 2 days and 100 meters or any other thresholds -
the functions provided by `spatsoc` are flexible to user input. In some cases,
the spatial grouping function selected is only relevant with certain temporal
grouping thresholds. For example, we wouldn't expect a threshold of 5 minutes
with `group_polys`.

```{r, eval = TRUE}
# Load packages
library(spatsoc)
library(data.table)
```

```{r, echo = FALSE, eval = TRUE}
data.table::setDTthreads(1)
```

```{r, eval = TRUE}
# Read data as a data.table
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast datetime column to POSIXct
DT[, datetime := as.POSIXct(datetime)]

# Temporal groups
group_times(DT, datetime = 'datetime', threshold = '5 minutes')

# Spatial groups
group_pts(
  DT,
  threshold = 50,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup'
)
```


# Installation

## Package dependencies

```
adehabitatHR (>= 0.4.21),
data.table (>= 1.15.0),
igraph,
sf,
lwgeom (>= 0.2.15),
CircStats,
stats,
units (>= 0.8-6),
rlang
```

See [here](https://github.com/r-spatial/sf/?tab=readme-ov-file#installing) for
help installing `sf`, see [here](https://github.com/r-quantities/units/?tab=readme-ov-file#installation) 
for help installing `units` and [here](https://github.com/r-spatial/lwgeom?tab=readme-ov-file#installing) for
help installing `lwgeom`.


# Functions

## group_times


`group_times(DT, datetime, threshold)`

* `DT`: input `data.table`
* `datetime`: date time column name in input data.table
* `threshold`: threshold for grouping 

### DT

A `data.table` with a date time formatted column. The input `DT` will be
returned with columns appended. The `timegroup` column corresponds to the
temporal group assigned to each row. Please note that the actual value of the
time group is meaningless. Reordered data will return a different time group.
What is meaningful, however,  is the contents of each group. Each group will
contain all rows nearest to the threshold provided.


### datetime format

The `group_times` function expects either one column (`POSIXct`) or two columns
(`IDate` and `ITime`).

Given a character column representing the date time, convert it to `POSIXct` or
`IDate` and `ITime`:

```{r posixct}
DT[, datetime := as.POSIXct(datetime)]
DT[, c('idate', 'itime') := IDateTime(datetime)]
```


These are then provided to the function using the names of the column in the
input data.

`group_times(DT, datetime = 'datetime', threshold = '5 minutes')`

or

`group_times(DT, datetime = c('idate', 'itime'), threshold = '5 minutes')`


### threshold recommendations

The `threshold` provided to `group_times` should be related to the fix rate of
the input dataset and to the specific study system and species. If relocations
are recorded every two hours, a `threshold = '2 hours'` will group all rows to
the nearest two hour group (10am, 12pm, 2pm, 4pm, ...). This, however, means
that the relocations can be up to one hour apart from each other. Picking a
smaller threshold, e.g.: `threshold = '15 minutes'` may be more relevant in some
cases. The flexibility of `spatsoc`'s threshold argument means the user must
carefully consider what threshold is reasonable to their specific system.

### Limitations of threshold

The `threshold` of `group_times` is considered only within the scope of 24 hours
and this poses limitations on it:

* `threshold` must evenly divide into 60 minutes or 24 hours
* multi-day blocks are consistent **across years** and timegroups from these
are **by year**. 
* number of minutes cannot exceed 60
* `threshold` cannot be fractional

### Columns returned by group_times

The main column returned by `group_times` is "timegroup". It represents the
temporal group of each row, where those nearest (either above or below) within
the threshold are grouped. Its actual value does not have any meaning, but the
contents of each group do. That means if the data is reordered, a row may have a
different time group, but the other rows in that group should not change.

The extra columns are provided to help the user investigate, troubleshoot and
interpret the timegroup.

| threshold unit | column(s) added |
|----------------|-----------------|
| minute         | "minutes" column added identifying the nearest minute group for each row.      |
| hour           | "hours" column added identifying the nearest hour group for each row.         |
| day            | "block" columns added identifying the multiday block for each row. |


### Warnings and messages

* "columns found in input DT and will be overwritten by this function"

This message is returned to the user when a column matching those returned by
`group_times` is found in the input DT. This is commonly the case when
`group_times` is run multiple times consecutively.


* "no threshold provided, using the time field directly to group"

This message is returned to the user when the `threshold` is NULL. This is the
default setting of `threshold` and, at times, may be suitable. In this case, the
date times in the `datetime` column will be grouped exactly. Usually, a
threshold should be provided.

* "the minimum and maximum days in DT are not evenly divisible by the provided block length"

This warning is returned to the user when the `threshold` with unit days does
not divide evenly into the range of days in DT. For example, if DT had data
covering 30 days, and a threshold of '7 days' was used, this warning would be
returned. Note, this warning is returned for the range of days for the entire
data set and not by year.

## group_pts


`group_pts(DT, threshold, id, coords, timegroup, splitBy)`
  
* `DT`: input `data.table`
* `threshold`: threshold for grouping 
* `id`: column name of IDs in `DT`
* `coords`: column names of x and y coordinates in `DT`
* `timegroup`: column name of time group
* `splitBy`: (optional) column names of extra variables to group on

### <a name="group_pts DT"></a>DT

The input `data.table`. It will returned with a column named group appended,
which represents the spatial (and temporal if `timegroup` is provided) group.

### threshold

The threshold must be in the units of the coordinates. 

## group_lines

`group_lines(DT, threshold, crs, id, coords, timegroup, sortBy, splitBy, spLines)`
  
* `DT`: input `data.table`
* `threshold`: threshold for grouping 
* `crs`: crs of coordinates in `DT`
* `id`: column name of IDs in `DT`
* `coords`: column names of x and y coordinates in `DT`
* `timegroup`: (optional) column name of time group
* `sortBy`: column name of date time to sort rows for building lines
* `splitBy`: (optional) column names of extra variables to group on
* `sfLines`: alternatively, provide a sf LINESTRING object and id column name

### DT

See [3.2.1](#dt-1). 

### threshold

The `threshold` argument represents a buffer area around each line. When
`threshold = 0`, the lines are grouped by spatial overlap. If the threshold is
greater than 0, the lines buffered, then grouped by spatial overlap.

### crs

The `crs` argument expects a character string or numeric 
defining the coordinate reference system to be passed to `sf::st_crs`. 
For example, for UTM zone 36S (EPSG 32736), the crs
argument is `crs = "EPSG:32736"` or `crs = 32736`. 
See https://spatialreference.org for a list of EPSG codes. 
Please note, R spatial has followed updates to GDAL 
and PROJ for handling coordinate reference systems, see more at 
https://r-spatial.org/r/2020/03/17/wkt.html. 


### sortBy

The `sortBy` argument expects a date time formatted column name, which is used
to order the rows for each individual (and `splitBy`).

## group_polys

`group_polys(DT, area, hrType, hrParams, crs, id, coords, splitBy, spLines)`
  
* `DT`: input `data.table`
* `area`: logical argument if proportional area should be returned
* `hrType`: type of home range created 
* `hrParams`: parameters relevant to the type of home range created
* `crs`: crs of coordinates in `DT`
* `id`: column name of IDs in `DT`
* `coords`: column names of x and y coordinates in `DT`
* `splitBy`: (optional) column names of extra variables to group on
* `sfPolys`: alternatively, provide a simple features POLGON or 
MULTIPOLYGON object and an id column


### DT and area

If `area = FALSE`, see [3.2.1](#dt-1). If `area = TRUE`, the DT will not be
appended with a group column instead a `data.table` with IDs and proportional
area overlap will be returned.

The default unit for area overlap is square meters. 

<!-- direction of proportion -->

### crs

The `crs` argument expects a character string or numeric 
defining the coordinate reference system to be passed to `sf::st_crs`. 
For example, for UTM zone 36S (EPSG 32736), the crs
argument is `crs = "EPSG:32736"` or `crs = 32736`. 
See https://spatialreference.org for a list of EPSG codes. 
Please note, R spatial has followed updates to GDAL 
and PROJ for handling coordinate reference systems, see more at
https://r-spatial.org/r/2020/03/17/wkt.html. 


### hrType and hrParams

Currently, `spatsoc` offers two types of home ranges provided by the
`adehabitatHR` package: 'mcp' (`mcp`) and 'kernel' (`kernelUD` and
`getverticeshr`). The parameters must match the arguments of those functions.

Internally, we match arguments to the functions allowing the user to provide,
for example, both the percent (provided to `getverticeshr`) and grid arguments
(provided to `mcp`).

```{r}
group_polys(
  DT,
  area = FALSE,
  crs = utm,
  hrType = 'mcp',
  hrParams = list(grid = 60, percent = 95),
  id = 'ID',
  coords = c('X', 'Y')
)
```

## edge_dist

`edge_dist(DT = NULL, threshold = NULL, id = NULL, coords = NULL, timegroup = NULL, splitBy = NULL, fillNA = TRUE)`

* `DT`: input `data.table`
* `threshold`: threshold for grouping 
* `id`: column name of IDs in `DT`
* `coords`: column names of x and y coordinates in `DT`
* `timegroup`: column name of time group
* `splitBy`: (optional) column names of extra variables to group on
* `fillNA`: logical indicating if NAs should be returned for individuals that
were not within the threshold distance of any other. If TRUE, NAs are returned.
If FALSE, only edges between individuals within the threshold distance are 
returned.

This is the non-chain rule implementation similar to `group_pts`. Edges are
defined by the distance threshold and NAs are returned for individuals within
each timegroup if they are not within the threshold distance of any other
individual (if `fillNA` is TRUE).

**See the vignette [Using edge-list generating functions and dyad_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html) for details about the `edge_dist` function.**

## edge_nn

`edge_nn(DT = NULL, id = NULL, coords = NULL, timegroup = NULL, splitBy = NULL, threshold = NULL)`

* `DT`: input `data.table`
* `id`: column name of IDs in `DT`
* `coords`: column names of x and y coordinates in `DT`
* `timegroup`: column name of time group
* `splitBy`: (optional) column names of extra variables to group on
* `threshold`: (optional) spatial distance threshold to set maximum distance 
between an individual and their neighbour.

This function can be used to generate edge-lists defined either by nearest
neighbour or nearest neighbour with a maximum distance. NAs are returned for
nearest neighbour for an individual was alone in a timegroup (and/or splitBy) or
if the distance between an individual and it's nearest neighbour is greater than
the threshold.


**See the vignette [Using edge-list generating functions and dyad_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html) for details about the `edge_nn` function.**

## randomizations

`randomizations(DT, type, id, datetime, splitBy, iterations)`
  
* `DT`: input `data.table`
* `type`: one of 'daily', 'step' or 'trajectory' 
* `id`: Character string of ID column name
* `datetime`: field used for providing date time or time group - see details
* `splitBy`: List of fields in DT to split the randomization process by
* `iterations`: The number of iterations to randomize


**See the vignette [Using spatsoc in social network analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html) for details about the `randomizations` function (specifically the section 'Data stream randomization')**


# Package design

## Don't I need to reassign to save the output?

Many functions in `spatsoc` use data.table's modify-by-reference to reduce
recopying large datasets and improve performance. Included in this list are:

- `group_times`
- `group_pts`
- `group_lines`
- `group_polys(area = FALSE)`
- `centroid_group`
- `direction_to_leader`
- `direction_to_centroid`
- `direction_step`
- `direction_group`
- `direction_polarization`
- `distance_to_centroid`
- `distance_to_leader`
- `fusion_id`
- `dyad_id`
- `leader_direction_group`



Functions that require their outputs be reassigned include:

- `group_polys(area = TRUE)`
- `edge_dist`
- `edge_nn`
- `edge_delay`
- `edge_direction`
- `edge_alignment`
- `centroid_fusion`
- `centroid_dyad`
- `leader_edge_delay`
- `randomizations`


## Why does a function print the result, but columns aren't added to my DT?

Check that your `data.table` has columns allocated (with
`data.table::truelength`) and if not, use `data.table::setDT` or
`data.table::alloc.col`. This can happen if you are reading your data from `RDS`
or `RData` files.  [See
here.](https://cran.r-project.org/package=data.table/vignettes/datatable-faq.html#reading-data.table-from-rds-or-rdata-file)

```{r setdt}
if (truelength(DT) == 0) {
  setDT(DT)
}
# then go to spatsoc
group_times(DT, datetime = 'datetime', threshold = '5 minutes')
```

or simply:

```{r alloc}
DT <- readRDS('path/to/data.Rds')
alloc.col(DT)
```




# Summary information

Here are some useful code chunks for understanding the spatial and temporal
extent of your data and the outputs of `spatsoc` functions.

## Number of individuals
```{r}
# Number of unique individuals
DT[, uniqueN(ID)]

# Number of unique individuals by timegroup
DT[, uniqueN(ID), by = timegroup]
```

## Temporal range

```{r}
# Min, max datetime
DT[, range(datetime)]

# Difference between relocations in hours
DT[order(datetime), 
   .(difHours = as.numeric(difftime(datetime, shift(datetime), units = 'hours'))), 
   by = ID]

# Difference between relocations in hours
DT[order(datetime), 
   .(difMins = as.numeric(difftime(datetime, shift(datetime), units = 'mins'))), 
   by = ID]
```

## Spatial extent

Simple spatial extents can be calculated for all individuals or by individual.

```{r}
# All individuals
DT[, .(minX = min(X),
       maxX = max(X),
       minY = min(Y),
       maxY = max(Y),)]

# By individual
DT[, .(minX = min(X),
       maxX = max(X),
       minY = min(Y),
       maxY = max(Y),),
   by = ID]
```

## `spatsoc` outputs

After using the grouping functions, we can determine the number of individuals 
in a temporal or spatial group. 

```{r}
# Number of unique individuals by timegroup
DT[, uniqueN(ID), by = timegroup]

# Number of unique individuals by group
DT[, uniqueN(ID), by = group]
```

---
title: "Detecting interspecific interactions"
author: "Alec L. Robitaille, Quinn Webber and Eric Vander Wal"
output: 
  rmarkdown::html_vignette:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Detecting interspecific interactions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, echo = FALSE, eval = TRUE}
data.table::setDTthreads(1)
```

## Interspecific interactions

Interspecific data can be used with {spatsoc} to estimate interspecific
interactions, eg. predator-prey dynamics. 

---

See the other vignettes for further information:

- [Introduction to spatsoc](https://docs.ropensci.org/spatsoc/articles/intro.html)
  - temporal grouping
  - spatiotemporal grouping with `group_pts`, `group_lines`, `group_polys`
  - distance based edge-list generation with `edge_dist`
- [Frequently asked questions about spatsoc](https://docs.ropensci.org/spatsoc/articles/faq.html)
  - install
  - function details for `group_times`, `group_pts`, `group_lines`, `group_polys`,
  `edge_dist`, `edge_nn`, and `randomizations`
  - package design including modify-by-reference, data.table column allocation
  - calculating summary information
- [Using spatsoc in social network analysis](https://docs.ropensci.org/spatsoc/articles/using-in-sna.html)
  - generating gambit-of-the-group data
  - generating observed networks
  - data stream randomization, randomized networks
  - network metrics
- [Using distance based edge-lists generating functions, dyad_id, and fusion_id](https://docs.ropensci.org/spatsoc/articles/using-edge-and-dyad.html)
  - generate distance based edge-lists with `edge_dist` and `edge_nn`
  - generate dyad identifiers for edge-lists with `dyad_id`
  - identify fusion events with `fusion_id`
- [Geometry interface](https://docs.ropensci.org/spatsoc/articles/geometry-interface-and-spatial-measures.html)
  - using `get_geometry` to setup a geometry column and use the geometry interface
  - details of underlying distance, direction and centroid spatial measures
  - converting to and from related packages
- [Interspecific interactions](https://docs.ropensci.org/spatsoc/articles/interspecific-interactions.html)
  - combine two movement datasets
  - identify interspecific interactions
  

## Interspecific interactions

Given two movement datasets, simply bind them together and use 
the `group_*` functions as usual. 

```{r}
#| fig.alt: >
#|   Predator prey interactions

# Load packages
library(spatsoc)
library(data.table)

# Load data
predator <- fread(system.file("extdata", "DT_predator.csv", package = "spatsoc"))
prey <- fread(system.file("extdata", "DT_prey.csv", package = "spatsoc"))

# Combine data
DT <- rbindlist(list(predator, prey))

# Set the datetime as a POSIxct
DT[, datetime := as.POSIXct(datetime)]


# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '10 minutes')

# Spatial grouping
group_pts(DT, threshold = 50, id = 'ID', 
          coords = c('X', 'Y'), timegroup = 'timegroup')


# Calculate the number of types within each group
DT[, n_type := uniqueN(type), by = group]
DT[, interact := n_type > 1]

# Prey's perspective
sub_prey <- DT[type == 'prey']
sub_prey[, mean(interact)]


# Plot --------------------------------------------------------------------
# If we subset only where there are interactions
sub_interact <- DT[(interact)]

plot(sub_prey$X, sub_prey$Y, col = 'grey', pch = 21)
points(sub_interact$X, sub_interact$Y, col = factor(sub_interact$type))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{DT}
\alias{DT}
\title{Movement of 10 "Newfoundland Bog Cows"}
\format{
A data.table with 14297 rows and 5 variables: \describe{
\item{ID}{individual identifier} \item{X}{X coordinate of the relocation
(UTM 36N)} \item{Y}{Y coordinate of the relocation (UTM 36N)}
\item{datetime}{character string representing the date time}
\item{population}{sub population within the individuals} }
}
\description{
A dataset containing the GPS relocations of 10 individuals in winter
2016-2017.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_alignment.R
\name{edge_alignment}
\alias{edge_alignment}
\title{Directional alignment based edge-lists}
\usage{
edge_alignment(
  DT = NULL,
  id = NULL,
  direction = "direction",
  timegroup = "timegroup",
  group = NULL,
  splitBy = NULL,
  signed = FALSE
)
}
\arguments{
\item{DT}{input data.table}

\item{id}{character string of ID column name}

\item{direction}{character string of direction column name, default
"direction", expects that the unit of the direction column is radians.}

\item{timegroup}{character string of timegroup column name, default "timegroup"}

\item{group}{(optional) character string of group column name, used to restrict
the calculation of directional alignment to within spatiotemporal groups}

\item{splitBy}{(optional) vector of column names indicating subgroups within
which the direction alignment will be calculated}

\item{signed}{logical if signed difference should be returned, default FALSE}
}
\value{
\code{edge_alignment} returns a \code{data.table} with columns ID1, ID2,
timegroup, and a 'direction_diff' column indicating the difference in direction
between ID1 and ID2 in radians, along with any columns provided in splitBy.

Note: unlike many other functions (eg. \code{group_pts}) in \code{spatsoc},
\code{edge_alignment} needs to be reassigned. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{edge_alignment} returns edge-lists defined by directional alignment
(difference in movement direction) between individuals. The function expects
a \code{data.table} with relocation data and individual identifiers, a
direction column (generated by \code{direction_step}) and timegroup column
(generated by \code{group_times}).
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{id}, \code{direction}, \code{timegroup}, and optional \code{group}
and \code{splitBy} arguments expect the names of a column in \code{DT} which
correspond to the individual identifier, direction (generated by
\code{direction_step}), timegroup (generated by \code{group_times}), group
(generated by \code{group_pts}) and additional grouping columns.

There are two approaches to spatially restricting the calculation of
directional alignment. The \code{group} argument can be used to pass the
output group column from \code{group_pts} to calculate direction alignment
within spatiotemporal groups. Alternatively, the output of
\code{edge_alignment} can be merged with the output of \code{edge_dist} to
compare the difference in direction to the distance between individuals.

The \code{splitBy} argument offers further control over the calculation of
directional alignment.  If within your \code{DT}, you have multiple
populations, subgroups or other distinct parts, you can provide the name of
the column which identifies them to \code{splitBy}. \code{edge_alignment}
will only consider rows within each \code{splitBy} subgroup.
}
\examples{
# Load data.table
library(data.table)
\dontshow{
data.table::setDTthreads(1)
}

# Read example data
DT <- fread(system.file('extdata', 'DT.csv', package = 'spatsoc'))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Calculate direction
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Calculate directional alignment edge-list
align <- edge_alignment(
  DT,
  id = 'ID',
  signed = FALSE
)

# Or, calculate directional alignment within spatiotemporal groups
group_pts(DT, threshold = 5, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

align_group <- edge_alignment(
  DT,
  id = 'ID',
  group = 'group',
  signed = FALSE
)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
direction_step(DT, id = 'ID')
align_group <- edge_alignment(
  DT,
  id = 'ID',
  group = 'group',
  signed = FALSE
)
}
\references{
See examples of using directional alignment:
\itemize{
\item \doi{doi:10.1098/rsif.2013.0529}
\item \doi{doi:10.1016/j.beproc.2021.104473}
\item \doi{doi:10.1126/science.aaa5099}
}
}
\seealso{
Other Edge-list generation: 
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_dist}()},
\code{\link{edge_nn}()}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
\concept{Edge-list generation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_pts.R
\name{group_pts}
\alias{group_pts}
\title{Group points}
\usage{
group_pts(
  DT = NULL,
  threshold = NULL,
  id = NULL,
  coords = NULL,
  timegroup,
  crs = NULL,
  splitBy = NULL,
  geometry = "geometry"
)
}
\arguments{
\item{DT}{input data.table}

\item{threshold}{distance for grouping points, either numeric or units, in
the units of the crs / the coordinates or geometry}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{timegroup}{timegroup field in the DT within which the output will be
calculated}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}
}
\value{
\code{group_pts} returns the input \code{DT} appended with a
\code{group} column.

This column represents the spatialtemporal group. As with the other
grouping functions,  the actual value of \code{group} is arbitrary and
represents the identity of a given group where 1 or more individuals are
assigned to a group. If the data was reordered, the \code{group} may
change, but the contents of each group would not.

A message is returned when a column named \code{group} already exists in
the input \code{DT}, because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{group_pts} groups rows into spatial groups. The function expects a
\code{data.table} with relocation data, individual identifiers and a
threshold argument. The threshold argument is used to specify the criteria
for distance between points which defines a group. Relocation data should be
in two columns representing the X and Y coordinates, or in a
geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{id}, \code{timegroup}  (and optional \code{splitBy})
arguments expect the names of columns in \code{DT} which correspond to the
individual identifier, and timegroup (generated by
\code{group_times}) and additional grouping columns.

The \code{threshold} provided should match the units of the coordinates. The
\code{threshold} can be provided with units specified using the units package (eg.
\code{threshold = units::set_units(10, m)}) which will be checked against the
units of the coordinates using the \code{crs}. If units are not specified, the
\code{threshold} is assumed to be in the units of the coordinates.

The \code{timegroup} argument is required to define the temporal groups
within which spatial groups are calculated. The intended framework is to
group rows temporally with \code{group_times} then spatially with
\code{group_pts} (or \code{group_lines}, \code{group_polys}).
If you have already calculated temporal groups without
\code{group_times}, you can pass this column to the \code{timegroup}
argument. Note that the expectation is that each individual will be observed
only once per timegroup. Caution that accidentally including huge numbers of
rows within timegroups can overload your machine since all pairwise distances
are calculated within each timegroup.

The \code{splitBy} argument offers further control over grouping. If within
your \code{DT}, you have multiple populations, subgroups or other distinct
parts, you can provide the name of the column which identifies them to
\code{splitBy}. The grouping performed by \code{group_pts} will only consider
rows within each \code{splitBy} subgroup.

See below under "Interface" for details on providing coordinates and under
"Distance function" for details on underlying distance function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Distance function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is \code{\link[sf:geos_measures]{sf::st_distance()}} which
passes to \code{\link[s2:s2_is_collection]{s2::s2_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE and
\code{\link[lwgeom:geod]{lwgeom::st_geod_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE. The distance
returned has units set according to the crs.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
distance function used is \code{\link[stats:dist]{stats::dist()}}, maintaining expected behaviour
from previous versions. The distance returned does not have units set.
}

Note: in both cases, if the coordinates are NA then the result will be NA.
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Select only individuals A, B, C for this example
DT <- DT[ID \%in\% c('A', 'B', 'C')]

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 5, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Spatial grouping with timegroup and splitBy on population
group_pts(DT, threshold = 5, id = 'ID', coords = c('X', 'Y'),
         timegroup = 'timegroup', splitBy = 'population')

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 50, id = 'ID', timegroup = 'timegroup')
}
\seealso{
\code{group_times}

Other Spatial grouping: 
\code{\link{group_lines}()},
\code{\link{group_polys}()}
}
\concept{Spatial grouping}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_times.R
\name{group_times}
\alias{group_times}
\title{Group times}
\usage{
group_times(DT = NULL, datetime = NULL, threshold = NULL)
}
\arguments{
\item{DT}{input data.table}

\item{datetime}{name of date time column(s). either 1 POSIXct or 2 IDate and
ITime. e.g.: 'datetime' or c('idate', 'itime')}

\item{threshold}{threshold for grouping times. e.g.: '2 hours', '10 minutes',
etc. if not provided, times will be matched exactly. Note that provided
threshold must be in the expected format: '## unit'}
}
\value{
\code{group_times} returns the input \code{DT} appended with a
\code{timegroup} column and additional temporal grouping columns to help
investigate, troubleshoot and interpret the timegroup.

The actual value of \code{timegroup} is arbitrary and represents the
identity of a given \code{timegroup} which 1 or more individuals are
assigned to. If the data was reordered, the group may change, but the
contents of each group would not.

The temporal grouping columns added depend on the \code{threshold}
provided:

\itemize{ \item \code{threshold} with unit minutes: "minutes" column added
identifying the nearest minute group for each row. \item \code{threshold}
with unit hours: "hours" column added identifying the nearest hour group
for each row. \item \code{threshold} with unit days: "block" columns added
identifying the multiday block for each row. }

A message is returned when any of these columns already exist in the input
\code{DT}, because they will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{group_times} groups rows into time groups. The function expects date
time formatted data and a threshold argument. The threshold argument is used
to specify a time window within which rows are grouped.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{datetime} argument expects the name of a column in \code{DT} which
is of type \code{POSIXct} or the name of two columns in \code{DT} which are
of type \code{IDate} and \code{ITime}.

\code{threshold} must be provided in units of minutes, hours or days. The
character string should start with an integer followed by a unit, separated
by a space. It is interpreted in terms of 24 hours which poses the following
limitations:

\itemize{ \item minutes, hours and days cannot be fractional \item minutes
must divide evenly into 60 \item minutes must not exceed 60 \item minutes,
hours which are nearer to the next day, are grouped as such \item hours must
divide evenly into 24 \item multi-day blocks should divide into the range of
days, else the blocks may not be the same length }

In addition, the \code{threshold} is considered a fixed window throughout the
time series and the rows are grouped to the nearest interval.

If \code{threshold} is NULL, rows are grouped using the \code{datetime}
column directly.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

group_times(DT, datetime = 'datetime', threshold = '5 minutes')

group_times(DT, datetime = 'datetime', threshold = '2 hours')

group_times(DT, datetime = 'datetime', threshold = '10 days')

}
\seealso{
\code{group_pts} \code{group_lines}
\code{group_polys}
}
\concept{Temporal grouping}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{calc_centroid}
\alias{calc_centroid}
\title{Calculate centroid}
\usage{
calc_centroid(geometry, x, y, crs, use_mean = FALSE)
}
\arguments{
\item{geometry}{sfc (simple feature geometry list column) from
\code{\link[=get_geometry]{get_geometry()}}}

\item{x}{X coordinate column, numeric}

\item{y}{Y coordinate column, numeric}

\item{crs}{crs for x, y coordinates, ignored for geometry argument}

\item{use_mean}{boolean predetermine if centroid calculated via mean}
}
\value{
The underlying centroid function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}) and \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE, the distance function
is \code{\link[sf:geos_unary]{sf::st_centroid()}} which passes to \code{\link[s2:s2_boundary]{s2::s2_centroid()}}.
\item If the crs is longlat degrees but \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE, the centroid
calculated will be incorrect. See \code{\link[sf:geos_unary]{sf::st_centroid()}}.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
centroid function used is mean.
}

Note: if the input is length 1, the input is returned.
}
\description{
\strong{Internal function} - not developed to be used outside of spatsoc functions
}
\details{
Calculate centroid for one of:
\itemize{
\item geometry
\item the points in x, y
\item the pairwise points in geometry_a and geometry_b
\item the pairwise points in x_a, y_a and x_b, y_b
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{
data.table::setDTthreads(1)
}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

DT[, spatsoc:::calc_centroid(x = X, y = Y, crs = 32736)]

get_geometry(DT, coords = c('X', 'Y'), crs = 32736)

# Calculating centroids requires recomputing the bbox when:
#  1- by = , 2- providing geometry
DT[, centroid := spatsoc:::calc_centroid(geometry), by = ID]
DT[, centroid := sf::st_sfc(centroid, recompute_bbox = TRUE)]
plot(DT$centroid)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/centroid_dyad.R
\name{centroid_dyad}
\alias{centroid_dyad}
\title{Dyad centroid}
\usage{
centroid_dyad(
  edges = NULL,
  DT = NULL,
  id = NULL,
  coords = NULL,
  crs = NULL,
  timegroup = "timegroup",
  geometry = "geometry"
)
}
\arguments{
\item{edges}{edge-list generated generated by \code{edge_dist} or
\code{edge_nn}, with dyad ID column generated by \code{dyad_id}}

\item{DT}{input data.table with timegroup column generated with
\code{group_times} matching the input data.table used to generate the edge
list with \code{edge_nn} or \code{edge_dist}}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{timegroup}{character string of timegroup column name, default "timegroup"}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}
}
\value{
\code{centroid_dyad} returns the input \code{edges} appended with centroid
column(s) for each timestep and dyad id.

If \code{coords} are provided, the centroid columns will be named by prefixing
the coordinate column names with "centroid_" (eg. "X" = "centroid_X"). If
\code{geometry} is used, the centroid column will be named "centroid".

Note: due to the merge required within this function, the output needs to be
reassigned unlike some other \code{spatsoc} functions like \code{dyad_id} and
\code{group_pts}. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.

A message is returned when the centroid column(s) already exist in the input
because they will be overwritten.
}
\description{
\code{centroid_dyad} calculates the centroid (mean location) of a dyad in each
observation identified by \code{edge_nn} or \code{edge_dist}. The function expects an
edge-list generated by \code{edge_nn} or \code{edge_dist} and a \code{data.table} with
relocation data appended with a \code{timegroup} column from \code{group_times}.
Relocation data should be in two columns representing the X and Y
coordinates, or in a geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{edges} and \code{DT} must be \code{data.table}s. If your data is a \code{data.frame},
you can convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning
using \code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{edges} and \code{DT} are internally merged in this function using the columns
\code{id}, \code{dyadID} and \code{timegroup}. This function expects a \code{dyadID} present,
generated with the \code{dyad_id} function. The \code{id} and \code{timegroup} arguments
expect the names of a column in \code{DT} which correspond to the id and timegroup
columns.

See below under "Interface" for details on providing coordinates and under
"Centroid function" for details on the underlying centroid function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and optionally \code{crs}. The \code{coords} argument expects the
names of the X and Y coordinate columns. The \code{crs} argument expects a
character string or numeric defining the coordinate reference system to be
passed to \link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs
argument is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See
\url{https://spatialreference.org} for a list of EPSG codes. For centroid
calculations, if \code{crs} is NULL, it will be internally set to \code{NA_crs_}.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Centroid function}{


The underlying centroid function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}) and \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE, the distance function
is \code{\link[sf:geos_unary]{sf::st_centroid()}} which passes to \code{\link[s2:s2_boundary]{s2::s2_centroid()}}.
\item If the crs is longlat degrees but \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE, the centroid
calculated will be incorrect. See \code{\link[sf:geos_unary]{sf::st_centroid()}}.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
centroid function used is mean.
}

Note: if the input is length 1, the input is returned.
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
\dontshow{
# (Subset example data to reduce runtime on CRAN)
if (isFALSE(rlang::is_interactive())) {
  DT <- DT[as.Date(datetime) == '2017-01-17']
}
}
# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge-list generation
edges <- edge_dist(
    DT,
    threshold = 100,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup',
    returnDist = TRUE,
    fillNA = FALSE
  )

# Generate dyad id
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

# Calculate dyad centroid
centroids <- centroid_dyad(
  edges,
  DT,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup'
)

print(centroids)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup')
dyad_id(edges, id = 'ID1', id2 = 'ID2')
centroids <- centroid_dyad(
  edges,
  DT,
  id = 'ID',
  timegroup = 'timegroup'
)
print(centroids)
}
\seealso{
\link{dyad_id} \link{edge_dist} \link{edge_nn} \link{group_pts}

Other Centroid functions: 
\code{\link{centroid_fusion}()},
\code{\link{centroid_group}()},
\code{\link{direction_to_centroid}()},
\code{\link{distance_to_centroid}()}
}
\concept{Centroid functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/centroid_group.R
\name{centroid_group}
\alias{centroid_group}
\title{Group centroid}
\usage{
centroid_group(
  DT = NULL,
  coords = NULL,
  crs = NULL,
  group = "group",
  geometry = "geometry"
)
}
\arguments{
\item{DT}{input data.table with group column generated with \code{group_pts}}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{group}{Character string of group column}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}
}
\value{
\code{centroid_group} returns the input \code{DT} appended with
centroid column(s) for each group.

If the \code{crs} for \code{coords} or \code{st_crs(geometry)} for \code{geometry} is long lat
(see \code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), centroids will be calculated using
\code{\link[s2:s2_boundary]{s2::s2_centroid()}} through \code{\link[sf:geos_unary]{sf::st_centroid()}}. If the \code{crs} for \code{coords}
or \code{st_crs(geometry)} for \code{geometry} is projected or NA, the centroids will
be calculated using a mean on the coordinates.

If \code{coords} are provided, the centroid columns will be named by prefixing
the coordinate column names with "centroid_" (eg. "X" = "centroid_X"). If
\code{geometry} is used, the centroid column will be named "centroid".

A message is returned when the centroid column(s) already exist in the
input because they will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{centroid_group} calculates the centroid of all individuals in each
spatiotemporal group identified by \code{group_pts}. The function expects a
\code{data.table} with relocation data appended with a \code{group} column from
\code{group_pts}. Relocation data should be in two columns representing the X and
Y coordinates, or in a geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{group} argument expects the name of a column in
\code{DT} which correspond to the group column.

See below under "Interface" for details on providing coordinates and under
"Centroid function" for details on the underlying centroid function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and optionally \code{crs}. The \code{coords} argument expects the
names of the X and Y coordinate columns. The \code{crs} argument expects a
character string or numeric defining the coordinate reference system to be
passed to \link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs
argument is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See
\url{https://spatialreference.org} for a list of EPSG codes. For centroid
calculations, if \code{crs} is NULL, it will be internally set to \code{NA_crs_}.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Centroid function}{


The underlying centroid function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}) and \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE, the distance function
is \code{\link[sf:geos_unary]{sf::st_centroid()}} which passes to \code{\link[s2:s2_boundary]{s2::s2_centroid()}}.
\item If the crs is longlat degrees but \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE, the centroid
calculated will be incorrect. See \code{\link[sf:geos_unary]{sf::st_centroid()}}.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
centroid function used is mean.
}

Note: if the input is length 1, the input is returned.
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
\dontshow{
# (Subset example data to reduce runtime on CRAN)
if (isFALSE(rlang::is_interactive())) {
  DT <- DT[as.Date(datetime) == '2017-01-17']
}
}
# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 5, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate group centroid
centroid_group(DT, coords = c('X', 'Y'), group = 'group')

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
centroid_group(DT)
}
\seealso{
\code{group_pts}

Other Centroid functions: 
\code{\link{centroid_dyad}()},
\code{\link{centroid_fusion}()},
\code{\link{direction_to_centroid}()},
\code{\link{distance_to_centroid}()}
}
\concept{Centroid functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/direction_to_leader.R
\name{direction_to_leader}
\alias{direction_to_leader}
\title{Direction to group leader}
\usage{
direction_to_leader(
  DT = NULL,
  coords = NULL,
  group = "group",
  crs = NULL,
  geometry = "geometry"
)
}
\arguments{
\item{DT}{input data.table}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{group}{group column name, generated by \code{group_pts}, default
'group'}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}
}
\value{
\code{direction_to_leader} returns the input \code{DT} appended with a
\code{direction_leader} column indicating the direction to the group leader in
radians. A value of NaN is returned when the coordinates of the focal
individual equal the coordinates of the leader.

Missing values in coordinates / geometry are ignored and NA is returned.

A message is returned when the \code{direction_leader} column already exist in
the input \code{DT} because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{direction_to_leader} calculates the direction to the leader of each
spatiotemporal group. The function expects a \code{data.table} with relocation
data appended with a \code{rank_position_group_direction} column indicating the
ranked position along the group direction generated with
\code{leader_direction_group(return_rank = TRUE)}. Relocation data should be in
two columns representing the X and Y coordinates, or in a geometry column
prepared by the helper function \code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

This function expects a \code{rank_position_group_direction} column generated with
\code{leader_direction_group(return_rank = TRUE)}, a \code{group} column generated with
the \code{group_pts} function. The \code{group} argument expects the name of the column
in \code{DT} which correspond to the group column.

See below under "Interface" for details on providing coordinates and under
"Direction function" for details on the underlying direction function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Direction function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is provided and longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is provided and not longlat degrees (eg. a projected UTM),
the coordinates or geometry are transformed to \code{sf::st_crs(4326)} before the
distance is measured using \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is NULL or NA_crs_, the distance function cannot be used
and an error is returned.
}
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
\dontshow{
# (Subset example data to reduce runtime on CRAN)
if (isFALSE(rlang::is_interactive())) {
  DT <- DT[as.Date(datetime) == '2017-01-17']
}
}
# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 50, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate direction at each step
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Calculate group centroid
centroid_group(DT, coords = c('X', 'Y'))

# Calculate group direction
direction_group(DT)

# Calculate leader in terms of position along group direction
leader_direction_group(
  DT,
  coords = c('X', 'Y'),
  crs = 32736
)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
direction_step(
  DT = DT,
  id = 'ID'
)
centroid_group(DT)
direction_group(DT)
leader_direction_group(
  DT
)
direction_to_leader(DT)
}
\references{
See examples of using direction to leader and position within group:
\itemize{
\item \doi{doi:10.1016/j.anbehav.2023.09.009}
\item \doi{doi:10.1016/j.beproc.2013.10.007}
\item \doi{doi:10.1371/journal.pone.0036567}
}
}
\seealso{
\link{distance_to_leader}, \link{leader_direction_group}, \link{group_pts},
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}

Other Leadership functions: 
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
\concept{Leadership functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_gbi.R
\name{get_gbi}
\alias{get_gbi}
\title{Group by individual matrix}
\usage{
get_gbi(DT = NULL, group = "group", id = NULL)
}
\arguments{
\item{DT}{input data.table}

\item{group}{Character string of group column (generated from one of
spatsoc's spatial grouping functions)}

\item{id}{character string of ID column name}
}
\value{
\code{get_gbi} returns a group by individual matrix (columns
represent individuals and rows represent groups).

Note that \code{get_gbi} is identical in function for turning the outputs
of \code{spatsoc} into social networks as
\code{\link[asnipe:get_group_by_individual]{asnipe::get_group_by_individual()}}
but is more efficient thanks to
\code{\link[data.table:dcast.data.table]{data.table::dcast()}}.
}
\description{
\code{get_gbi} generates a group by individual matrix. The function expects a
\code{data.table} with individual identifiers and a group column. The group
by individual matrix can then be used to build a network using
\code{\link[asnipe:get_network]{asnipe::get_network()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{group} argument expects the name of a column which corresponds to
an integer group identifier (generated by \code{spatsoc}'s grouping
functions).

The \code{id} argument expects the name of a column which corresponds to the
individual identifier.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
DT[, yr := year(datetime)]

# EPSG code for example data
utm <- 'EPSG:32736'

group_polys(DT, area = FALSE, hrType = 'mcp',
            hrParams = list(percent = 95),
            crs = utm, id = 'ID', coords = c('X', 'Y'),
            splitBy = 'yr')

gbiMtrx <- get_gbi(DT = DT, group = 'group', id = 'ID')

}
\seealso{
\code{group_pts} \code{group_lines}
\code{group_polys}

Other Social network tools: 
\code{\link{randomizations}()}
}
\concept{Social network tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_polys.R
\name{group_polys}
\alias{group_polys}
\title{Group polygons}
\usage{
group_polys(
  DT = NULL,
  area = NULL,
  hrType = NULL,
  hrParams = NULL,
  crs = NULL,
  id = NULL,
  coords = NULL,
  splitBy = NULL,
  sfPolys = NULL,
  projection = NULL
)
}
\arguments{
\item{DT}{input data.table}

\item{area}{logical indicating either overlap group (when \code{FALSE}) or
area and proportion of overlap (when \code{TRUE})}

\item{hrType}{type of HR estimation, either 'mcp' or 'kernel'}

\item{hrParams}{a named list of parameters for \code{adehabitatHR} functions}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{sfPolys}{Alternatively, provide solely a simple features object with
POLYGONs or MULTIPOLYGONs. If sfPolys are provided, id is required and
splitBy cannot be used.}

\item{projection}{(deprecated) use crs argument instead}
}
\value{
When \code{area} is \code{FALSE}, \code{group_polys} returns the input \code{DT} appended
with a \code{group} column. As with the other grouping functions,  the actual
value of \code{group} is arbitrary and represents the identity of a given group
where 1 or more individuals are assigned to a group. If the data was
reordered, the \code{group} may change, but the contents of each group would
not. When \code{area} is \code{TRUE}, \code{group_polys} returns a proportional area
overlap \code{data.table}. In this case, ID refers to the focal individual of
which the total area is compared against the overlapping area of ID2.

If \code{area} is \code{FALSE}, a message is returned when a column named \code{group}
already exists in the input \code{DT}, because it will be overwritten.

Along with changes to follow the R-spatial evolution, \code{group_polys} also
now returns area and proportion of overlap with units explicitly specified
through the \code{units} package.

Note: if area is TRUE, the output of \code{group_polys} needs to be
reassigned. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{group_polys} groups rows into spatial groups by overlapping polygons (home
ranges). The function expects a \code{data.table} with relocation data, individual
identifiers and an \code{area} argument.  The relocation data is transformed into
home range POLYGONs using \code{\link[=build_polys]{build_polys()}} with \link[adehabitatHR:mcp]{adehabitatHR::mcp} or
\link[adehabitatHR:kernelUD]{adehabitatHR::kernelUD}. If the \code{area} argument is \code{FALSE}, \code{group_polys}
returns grouping calculated by spatial overlap. If the \code{area} argument is
\code{TRUE}, \code{group_polys} returns the area area and proportion of overlap.
Relocation data should be in two columns representing the X and Y
coordinates.
}
\details{
\subsection{R-spatial evolution}{

Please note, spatsoc has followed updates from R spatial, GDAL and PROJ for
handling coordinate reference systems, see more below and  details at
\url{https://r-spatial.org/r/2020/03/17/wkt.html}.

In addition, \code{group_polys} previously used rgeos::gIntersection,
rgeos::gIntersects and rgeos::gArea but has been
updated to use \link[sf:geos_binary_pred]{sf::st_intersects}, \link[sf:geos_binary_ops]{sf::st_intersection} and \link[sf:geos_measures]{sf::st_area}
according to the R-spatial evolution, see more
at \url{https://r-spatial.org/r/2022/04/12/evolution.html}.
}

\subsection{Notes on arguments}{

The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{id}, \code{coords} (and optional \code{splitBy}) arguments expect
the names of respective columns in \code{DT} which correspond to the
individual identifier, X and Y coordinates, and additional grouping columns.

The \code{crs} argument expects a character string or numeric defining the
coordinate reference system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, for
UTM zone 36S (EPSG 32736), the crs argument is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org} for a list of EPSG codes.

The \code{hrType} must be either one of "kernel" or "mcp". The
\code{hrParams} must be a named list of arguments matching those of
\code{\link[adehabitatHR:kernelUD]{adehabitatHR::kernelUD()}} or \code{\link[adehabitatHR:mcp]{adehabitatHR::mcp()}}.

The \code{splitBy} argument offers further control over grouping. If within
your \code{DT}, you have multiple populations, subgroups or other distinct
parts, you can provide the name of the column which identifies them to
\code{splitBy}. The grouping performed by \code{group_polys} will only
consider rows within each \code{splitBy} subgroup.
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# EPSG code for example data
utm <- 32736

group_polys(DT, area = FALSE, hrType = 'mcp',
            hrParams = list(percent = 95), crs = utm,
            id = 'ID', coords = c('X', 'Y'))

areaDT <- group_polys(DT, area = TRUE, hrType = 'mcp',
                      hrParams = list(percent = 95), crs = utm,
                      id = 'ID', coords = c('X', 'Y'))
print(areaDT)
}
\seealso{
\code{\link[=build_polys]{build_polys()}} \code{\link[=group_times]{group_times()}}

Other Spatial grouping: 
\code{\link{group_lines}()},
\code{\link{group_pts}()}
}
\concept{Spatial grouping}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_nn.R
\name{edge_nn}
\alias{edge_nn}
\title{Nearest neighbour based edge-lists}
\usage{
edge_nn(
  DT = NULL,
  id = NULL,
  coords = NULL,
  timegroup,
  crs = NULL,
  splitBy = NULL,
  threshold = NULL,
  geometry = "geometry",
  returnDist = FALSE
)
}
\arguments{
\item{DT}{input data.table}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{timegroup}{timegroup field in the DT within which the output will be
calculated}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{threshold}{(optional) spatial distance threshold to set maximum
distance between an individual and their neighbour.}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}

\item{returnDist}{logical indicating if the distance between individuals
should be returned. If FALSE (default), only individual columns (and
timegroup, splitBy columns if provided) are returned. If TRUE, a column
"distance" is also returned indicating the distance between individuals in
the units of the \code{crs}, or if \code{crs = NULL} no units are set.}
}
\value{
\code{edge_nn} returns a \code{data.table}  with three columns: timegroup, ID
and NN. If 'returnDist' is TRUE, a column 'distance' is returned indicating
the distance between ID and NN.  The ID and NN columns represent the edges
defined by the nearest neighbours (and temporal thresholds with
\code{group_times}).

If an individual was alone in a timegroup or splitBy, or did not have any
neighbours within the threshold distance, they are assigned NA for nearest
neighbour.

Note: unlike many other functions (eg. \code{group_pts}) in \code{spatsoc}, \code{edge_nn}
needs to be reassigned. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{edge_nn} returns edge-lists defined by the nearest neighbour. The function
expects a \code{data.table} with relocation data, individual identifiers and a
threshold argument. The threshold argument is used to specify the criteria
for distance between points which defines a group. Relocation data should be
in two columns representing the X and Y coordinates, or in a geometry column
prepared by the helper function \code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{id}, \code{timegroup} (and optional \code{splitBy}) arguments expect the names of
columns in \code{DT} which correspond to the individual identifier, and timegroup
(generated by \code{group_times}) and additional grouping columns.

If a \code{threshold} is provided, it should match the units of the coordinates.
The \code{threshold} can be provided with units specified using the units package
(eg. \code{threshold = units::set_units(10, m)}) which will be checked against the
units of the coordinates using the \code{crs}. If units are not specified, the
\code{threshold} is assumed to be in the units of the coordinates.

The \code{timegroup} argument is required to define the temporal groups within
which edge nearest neighbours are calculated. The intended framework is to
group rows temporally with \code{group_times} then spatially with \code{edge_nn}. If
you have already calculated temporal groups without \code{group_times}, you can
pass this column to the \code{timegroup} argument. Note that the expectation is
that each individual will be observed only once per timegroup. Caution that
accidentally including huge numbers of rows within timegroups can overload
your machine since all pairwise distances are calculated within each
timegroup.

The \code{splitBy} argument offers further control over grouping. If within your
\code{DT}, you have multiple populations, subgroups or other distinct parts, you
can provide the name of the column which identifies them to \code{splitBy}.
\code{edge_nn} will only consider rows within each \code{splitBy} subgroup.

See below under "Interface" for details on providing coordinates and under
"Distance function" for details on underlying distance function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Distance function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is \code{\link[sf:geos_measures]{sf::st_distance()}} which
passes to \code{\link[s2:s2_is_collection]{s2::s2_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE and
\code{\link[lwgeom:geod]{lwgeom::st_geod_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE. The distance
returned has units set according to the crs.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
distance function used is \code{\link[stats:dist]{stats::dist()}}, maintaining expected behaviour
from previous versions. The distance returned does not have units set.
}

Note: in both cases, if the coordinates are NA then the result will be NA.
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Select only individuals A, B, C for this example
DT <- DT[ID \%in\% c('A', 'B', 'C')]

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge-list generation
edges <- edge_nn(DT, id = 'ID', coords = c('X', 'Y'),
        timegroup = 'timegroup')

# Edge-list generation using maximum distance threshold
edges <- edge_nn(DT, id = 'ID', coords = c('X', 'Y'),
        timegroup = 'timegroup', threshold = 100)

# Edge-list generation, returning distance between nearest neighbours
edge_nn(DT, id = 'ID', coords = c('X', 'Y'),
        timegroup = 'timegroup', threshold = 100,
        returnDist = TRUE)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
edge_nn(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
}
\seealso{
Other Edge-list generation: 
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_dist}()}

Other Distance functions: 
\code{\link{distance_to_centroid}()},
\code{\link{distance_to_leader}()},
\code{\link{edge_dist}()},
\code{\link{edge_zones}()}
}
\concept{Distance functions}
\concept{Edge-list generation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fusion_id.R
\name{fusion_id}
\alias{fusion_id}
\title{Fission-fusion id}
\usage{
fusion_id(
  edges = NULL,
  threshold = NULL,
  n_min_length = 0,
  n_max_missing = 0,
  allow_split = FALSE
)
}
\arguments{
\item{edges}{distance based edge-list generated by \code{edge_dist} function,
with 'dyadID' column generated by \code{dyad_ID}}

\item{threshold}{spatial distance threshold in the units of the crs}

\item{n_min_length}{minimum length of fusion events}

\item{n_max_missing}{maximum number of missing observations within a fusion
event}

\item{allow_split}{logical defining if a single observation can be greater
than the threshold distance without initiating fission event}
}
\value{
\code{fusion_id} returns the input \code{edges} appended with a
\code{fusionID} column.

This column represents the fusion event id. As with \code{spatsoc}'s
grouping functions, the actual value of \code{fusionID} is arbitrary and
represents the identity of a given fusion event. If the data was reordered,
the \code{fusionID} may change, but the membership of each fusion event
would not.

A message is returned when a column named \code{fusionID} already exists in
the input \code{edges}, because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{fusion_id} identifies fusion events in distance based edge-lists.
The function expects a distance based edge-list generated by
\code{edge_dist}, a threshold argument and arguments controlling how fusion
events are defined.
}
\details{
The \code{edges} must be a \code{data.table} returned by the \code{edge_dist}
function. In addition, \code{fusion_id} requires a dyad ID set on the edge
list generated by \code{dyad_id}. If your data is a \code{data.frame}, you
can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{threshold} must be provided in the units of the coordinates. The
\code{threshold} must be larger than 0. Eg. in the case of UTM, a \code{threshold} = 50
would indicate a 50 m distance threshold.

The \code{n_min_length} argument defines the minimum number of successive
fixes that are required to establish a fusion event. The \code{n_max_missing}
argument defines the the maximum number of allowable missing observations for
the dyad within a fusion event. The \code{allow_split} argument defines if a
single observation can be greater than the threshold distance without
initiating fission event.
}
\examples{

# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge-list generation
edges <- edge_dist(
    DT,
    threshold = 100,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup',
    returnDist = TRUE,
    fillNA = TRUE
  )

dyad_id(edges, 'ID1', 'ID2')

fusion_id(
  edges = edges,
  threshold = 100,
  n_min_length = 1,
  n_max_missing = 0,
  allow_split = FALSE
  )

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
dyad_id(edges, id = 'ID1', id2 = 'ID2')
fusion_id(edges, threshold = 100)
}
\references{
See examples of identifying fission-fusion events with spatiotemporal data:
\itemize{
\item \doi{doi:10.1111/ele.12457}
\item \doi{doi:10.1016/j.anbehav.2018.03.014}
\item \doi{doi:10.1890/08-0345.1}
}
}
\seealso{
\code{edge_dist}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance_to_centroid.R
\name{distance_to_centroid}
\alias{distance_to_centroid}
\title{Distance to group centroid}
\usage{
distance_to_centroid(
  DT = NULL,
  coords = NULL,
  group = "group",
  crs = NULL,
  return_rank = TRUE,
  ties.method = NULL,
  geometry = "geometry"
)
}
\arguments{
\item{DT}{input data.table with centroid columns generated by eg.
\code{centroid_group}}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{group}{group column name, generated by \code{group_pts}, default
'group'}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{return_rank}{logical if rank distance should also be returned, default
TRUE}

\item{ties.method}{see \code{\link[data.table:frank]{?data.table::frank()}}}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}
}
\value{
\code{distance_to_centroid} returns the input \code{DT} appended with a
\code{distance_centroid} column indicating the distance to the group centroid
and, optionally, a \code{rank_distance_centroid} column indicating the within
group rank distance to the group centroid (if \code{return_rank = TRUE}).

A message is returned when \code{distance_centroid} and optional
\code{rank_distance_centroid} columns already exist in the input \code{DT}, because
they will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{distance_to_centroid} calculates the distance of each relocation to the
centroid of the spatiotemporal group identified by \code{group_pts}. The function
expects a \code{data.table} with relocation data appended with a \code{group} column
from \code{group_pts} and centroid columns from \code{centroid_group}. Relocation data
should be provided in two columns representing the X and Y coordinates, or in
a geometry column prepared by the helper function \code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

This function expects a \code{group} column present generated with the \code{group_pts}
function and centroid coordinate column(s) generated with the
\code{centroid_group} function. The \code{group} arguments expect the names of columns
in \code{DT} which correspond to the group column. The \code{return_rank} argument
controls if the rank of each individual's distance to the group centroid is
also returned. The \code{ties.method} argument is passed to \code{data.table::frank},
see details at \code{\link[data.table:frank]{?data.table::frank()}}.

See below under "Interface" for details on providing coordinates and under
"Distance function" for details on underlying distance function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Distance function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is \code{\link[sf:geos_measures]{sf::st_distance()}} which
passes to \code{\link[s2:s2_is_collection]{s2::s2_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE and
\code{\link[lwgeom:geod]{lwgeom::st_geod_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE. The distance
returned has units set according to the crs.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
distance function used is \code{\link[stats:dist]{stats::dist()}}, maintaining expected behaviour
from previous versions. The distance returned does not have units set.
}

Note: in both cases, if the coordinates are NA then the result will be NA.
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))
\dontshow{
# (Subset example data to reduce runtime on CRAN)
if (isFALSE(rlang::is_interactive())) {
  DT <- DT[as.Date(datetime) == '2017-01-17']
}
}
# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 5, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate group centroid
centroid_group(DT, coords = c('X', 'Y'), group = 'group')

# Calculate distance to group centroid
distance_to_centroid(
  DT,
  coords = c('X', 'Y'),
  group = 'group',
)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
centroid_group(DT)
direction_to_centroid(DT)
}
\references{
See examples of using distance to group centroid:
\itemize{
\item \doi{doi:10.1016/j.anbehav.2021.08.004}
\item \doi{doi:10.1111/eth.12336}
\item \doi{doi:10.1007/s13364-018-0400-2}
}
}
\seealso{
\link{centroid_group}, \link{group_pts}, \code{\link[sf:geos_measures]{sf::st_distance()}}

Other Distance functions: 
\code{\link{distance_to_leader}()},
\code{\link{edge_dist}()},
\code{\link{edge_nn}()},
\code{\link{edge_zones}()}

Other Centroid functions: 
\code{\link{centroid_dyad}()},
\code{\link{centroid_fusion}()},
\code{\link{centroid_group}()},
\code{\link{direction_to_centroid}()}
}
\concept{Centroid functions}
\concept{Distance functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_delay.R
\name{edge_delay}
\alias{edge_delay}
\title{Directional correlation delay based edge-lists}
\usage{
edge_delay(edges, DT, window = NULL, id = NULL, direction = "direction")
}
\arguments{
\item{edges}{edge-list generated generated by \code{\link[=edge_dist]{edge_dist()}} or
\code{\link[=edge_nn]{edge_nn()}}, with fusionID column generated by \code{\link[=fusion_id]{fusion_id()}}}

\item{DT}{input data.table with timegroup column generated with
\code{\link[=group_times]{group_times()}} matching the input data.table used to generate the edge
list with \code{\link[=edge_nn]{edge_nn()}} or \code{\link[=edge_dist]{edge_dist()}}}

\item{window}{temporal window in unit of timegroup column generated with
\code{group_times}, eg. \code{window = 4} corresponds to the 4 timegroups
before and after the focal observation}

\item{id}{character string of ID column name}

\item{direction}{character string of direction column name, default
"direction", expects that the unit of the direction column is radians.}
}
\value{
\code{edge_delay} returns the input \code{edges} appended with a
'direction_delay' column indicating the temporal delay (in units of
timegroups) at which ID1's direction of movement is most similar to ID2's
direction of movement, within the temporal window defined, and a
'direction_diff' column indicating the absolute difference in direction.
For example, if focal individual 'A' moves in a 45 degree direction at time
2 and individual 'B' moves in a most similar direction within the window at
time 5, the directional correlation delay between A and B is 3. Positive
values of directional correlation delay indicate a directed leadership edge
from ID1 to ID2.

Note: due to the merge required within this function, the output needs to be
reassigned unlike some other \code{spatsoc} functions like \code{dyad_id}.
See details in \href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{edge_delay} returns edge-lists defined by the directional correlation
delay between individuals. The function expects a distance based edge-list
generated by \code{edge_dist} or \code{edge_nn}, a \code{data.table} with
relocation data, individual identifiers and a window argument. The window
argument is used to specify the temporal window within which to measure the
directional correlation delay. Relocation data should be in two columns
representing the X and Y coordinates.
}
\details{
The \code{edges} and \code{DT} must be \code{data.table}s. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{edges} argument expects a distance based edge-list generated with
\code{edge_nn} or \code{edge_dist}. The \code{DT} argument expects relocation
data with a timegroup column generated with \code{group_times}.

The rows in \code{edges} and \code{DT} are internally matched in
\code{edge_delay} using the columns \code{timegroup} (from
\code{group_times}) and \code{ID1} and \code{ID2} (in \code{edges}, from
\code{dyad_id}) with \code{id} (in \code{DT}). This function expects a
\code{fusionID} present, generated with the \code{fusion_id} function, and a
\code{dyadID} present, generated with the \code{dyad_id} function. The
\code{id}, and \code{direction} arguments expect the names of a column in
\code{DT} which correspond to the id, and direction columns.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Select only individuals A, B, C for this example
DT <- DT[ID \%in\% c('A', 'B', 'C')]

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Calculate direction
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Distance based edge-list generation
edges <- edge_dist(
  DT,
  threshold = 100,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  returnDist = TRUE,
  fillNA = FALSE
)

# Generate dyad id
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

# Generate fusion id
fusion_id(edges, threshold = 100)

# Directional correlation delay
delay <- edge_delay(
  edges = edges,
  DT = DT,
  window = 3,
  id = 'ID'
)

delay[, mean(direction_delay, na.rm = TRUE), by = .(ID1, ID2)][V1 > 0]

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
direction_step(DT, id = 'ID')
edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
dyad_id(edges, id = 'ID1', id2 = 'ID2')
fusion_id(edges, threshold = 100)
delay <- edge_delay(
  edges = edges,
  DT = DT,
  window = 3,
  id = 'ID'
)
}
\references{
The directional correlation delay is defined in Nagy et al. 2010
(\doi{doi:10.1038/nature08891}).

See examples of measuring the directional correlation delay:
\itemize{
\item \doi{doi:10.1016/j.anbehav.2013.07.005}
\item \doi{doi:10.1073/pnas.1305552110}
\item \doi{doi:10.1111/jfb.15315}
\item \doi{doi:10.1371/journal.pcbi.1003446}
}
}
\seealso{
Other Edge-list generation: 
\code{\link{edge_alignment}()},
\code{\link{edge_direction}()},
\code{\link{edge_dist}()},
\code{\link{edge_nn}()}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
\concept{Edge-list generation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_lines.R
\name{group_lines}
\alias{group_lines}
\title{Group lines}
\usage{
group_lines(
  DT = NULL,
  threshold = NULL,
  crs = NULL,
  id = NULL,
  coords = NULL,
  timegroup = NULL,
  sortBy = NULL,
  splitBy = NULL,
  sfLines = NULL,
  projection = NULL
)
}
\arguments{
\item{DT}{input data.table}

\item{threshold}{The width of the buffer around the lines in the units of the
crs. Use \code{threshold = 0} to compare intersection without buffering.}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{timegroup}{timegroup field in the DT within which the output will be
calculated}

\item{sortBy}{Character string of date time column(s) to sort rows by. Must
be a POSIXct.}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{sfLines}{Alternatively to providing a DT, provide a simple feature
LINESTRING object generated with the sf package. The id argument is
required to provide the identifier matching each LINESTRING.
If an sfLines object is provided, groups cannot be calculated by timegroup
or splitBy.}

\item{projection}{(deprecated) use crs argument instead}
}
\value{
\code{group_lines} returns the input \code{DT} appended with a "group"
column.

This column represents the spatial (and if \code{timegroup} was provided -
spatiotemporal) group calculated by intersecting lines. As with the other
grouping functions,  the actual value of group is arbitrary and represents
the identity of a given group where 1 or more individuals are assigned to a
group. If the data was reordered, the group may change, but the contents of
each group would not.

A message is returned when a column named "group" already exists in the
input \code{DT}, because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{group_lines} groups rows into spatial groups by generating LINESTRINGs and
grouping based on spatial intersection. The function expects a \code{data.table}
with relocation data, individual identifiers and a distance threshold. The
relocation data is transformed into sf LINESTRINGs using \link{build_lines} and
intersecting LINESTRINGs are grouped. The threshold argument is used to
specify the distance criteria for grouping. Relocation data should be in two
columns representing the X and Y coordinates.
}
\details{
\subsection{R-spatial evolution}{

Please note, spatsoc has followed updates from R spatial, GDAL and PROJ for
handling coordinate reference systems, see more at
\url{https://r-spatial.org/r/2020/03/17/wkt.html}.

In addition, \code{group_lines} (and \link{build_lines}) previously used
\link[sp:SpatialLines]{sp::SpatialLines}, rgeos::gIntersects, rgeos::gBuffer but have been
updated to use \link[sf:st_as_sf]{sf::st_as_sf}, \link[sf:st]{sf::st_linestring}, \link[sf:geos_binary_pred]{sf::st_intersects}, and
\link[sf:geos_unary]{sf::st_buffer} according to the R-spatial evolution, see more at
\url{https://r-spatial.org/r/2022/04/12/evolution.html}.
}

\subsection{Notes on arguments}{

The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using \link[data.table:setDT]{data.table::setDT}.

The \code{id}, \code{coords}, \code{sortBy} (and optional \code{timegroup}
and \code{splitBy}) arguments expect the names of respective columns in
\code{DT} which correspond to the individual identifier, X and Y coordinates,
sorting, timegroup (generated by \link{group_times}) and additional grouping
columns.

The \code{crs} argument expects a numeric or character defining the coordinate
reference system. For example, for UTM zone 36N (EPSG 32736), the crs
argument is either \code{crs = 'EPSG:32736'} or \code{crs = 32736}. See details in
\code{\link[sf:st_crs]{sf::st_crs()}} and \url{https://spatialreference.org} for a list of EPSG codes.

The \code{sortBy} argument is used to order the input \code{DT} when creating sf
LINESTRINGs. It must a column in the input \code{DT} of type POSIXct to ensure the
rows are sorted by date time.

The \code{threshold} must be provided in the units of the coordinates. The
\code{threshold} can be equal to 0 if strict overlap is intended, otherwise it
should be some value greater than 0, in the units of the coordinates. Eg. in
the case of UTM, a \code{threshold = 50} would indicate a 50m distance threshold.

The \code{timegroup} argument is optional, but recommended to pair with
\link{group_times}. The intended framework is to group rows temporally with
\link{group_times} then spatially with \link{group_lines} (or \link{group_pts},
\link{group_polys}). With \link{group_lines}, pick a relevant \link{group_times} threshold
such as \code{'1 day'} or \code{'7 days'} which is informed by your study species,
system or question.

The \code{splitBy} argument offers further control building LINESTRINGs. If in
your input \code{DT}, you have multiple temporal groups (e.g.: years) for example,
you can provide the name of the column which identifies them and build
LINESTRINGs for each individual in each year. The grouping performed by
\link{group_lines} will only consider rows within each \code{splitBy} subgroup.
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Subset only individuals A, B, and C
DT <- DT[ID \%in\% c('A', 'B', 'C')]

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# EPSG code for example data
utm <- 32736

group_lines(DT, threshold = 50, crs = utm, sortBy = 'datetime',
            id = 'ID', coords = c('X', 'Y'))

## Daily movement tracks
# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '1 day')

# Subset only first 50 days
DT <- DT[timegroup < 25]

# Spatial grouping
group_lines(DT, threshold = 50, crs = utm,
            id = 'ID', coords = c('X', 'Y'),
            timegroup = 'timegroup', sortBy = 'datetime')

## Daily movement tracks by population
group_lines(DT, threshold = 50, crs = utm,
            id = 'ID', coords = c('X', 'Y'),
            timegroup = 'timegroup', sortBy = 'datetime',
            splitBy = 'population')
}
\seealso{
\link{build_lines} \link{group_times}

Other Spatial grouping: 
\code{\link{group_polys}()},
\code{\link{group_pts}()}
}
\concept{Spatial grouping}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/direction_to_centroid.R
\name{direction_to_centroid}
\alias{direction_to_centroid}
\title{Direction to group centroid}
\usage{
direction_to_centroid(
  DT = NULL,
  coords = NULL,
  crs = NULL,
  geometry = "geometry"
)
}
\arguments{
\item{DT}{input data.table}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}
}
\value{
\code{direction_to_centroid} returns the input \code{DT} appended with a
\code{direction_centroid} column indicating the direction to the group centroid
in radians. A value of NaN is returned when the coordinates of the focal
individual equal the coordinates of the centroid.

A message is returned when \code{direction_centroid} column already exist in the
input \code{DT}, because they will be overwritten.

Missing values in coordinates / geometry are ignored and NA is returned.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{direction_to_centroid} calculates the direction of each relocation to the
centroid of the spatiotemporal group identified by \code{\link[=group_pts]{group_pts()}}. The
function expects a \code{data.table} with relocation data appended with a \code{group}
column from \code{\link[=group_pts]{group_pts()}} and centroid columns from \code{\link[=centroid_group]{centroid_group()}}.
Relocation data should be in two columns representing the X and Y
coordinates, or in a geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

This function expects a \code{group} column present generated with the
\code{\link[=group_pts]{group_pts()}} function and centroid coordinates generated with the
\code{\link[=centroid_group]{centroid_group()}} function. The \code{group} argument expects the name of the
column in \code{DT} which correspond to the group column.

See below under "Interface" for details on providing coordinates and under
"Direction function" for details on the underlying direction function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Direction function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is provided and longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is provided and not longlat degrees (eg. a projected UTM),
the coordinates or geometry are transformed to \code{sf::st_crs(4326)} before the
distance is measured using \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is NULL or NA_crs_, the distance function cannot be used
and an error is returned.
}
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
\dontshow{
# (Subset example data to reduce runtime on CRAN)
if (isFALSE(rlang::is_interactive())) {
  DT <- DT[as.Date(datetime) == '2017-01-17']
}
}
# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 5, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate group centroid
centroid_group(DT, coords = c('X', 'Y'), group = 'group')

# Calculate direction to group centroid
direction_to_centroid(DT, coords = c('X', 'Y'), crs = 32736)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
centroid_group(DT)
direction_to_centroid(DT)
}
\references{
See example of using direction to group centroid:
\itemize{
\item \doi{doi:10.1016/j.cub.2017.08.004}
}
}
\seealso{
\link{centroid_group}, \link{group_pts}, \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}

Other Centroid functions: 
\code{\link{centroid_dyad}()},
\code{\link{centroid_fusion}()},
\code{\link{centroid_group}()},
\code{\link{distance_to_centroid}()}
}
\concept{Centroid functions}
\concept{Direction functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_geometry.R
\name{get_geometry}
\alias{get_geometry}
\title{Get geometry}
\usage{
get_geometry(
  DT = NULL,
  coords = NULL,
  crs = NULL,
  output_crs = NULL,
  geometry_colname = "geometry"
)
}
\arguments{
\item{DT}{input data.table}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{crs}{numeric or character defining the coordinate reference system to
be passed to \link[sf:st_crs]{sf::st_crs}. For example, \code{crs = "EPSG:32736"} or
\code{crs = 32736}.}

\item{output_crs}{default NULL, the output crs to transform the input
coordinates to with \link[sf:st_transform]{sf::st_transform}. If output_crs is NULL or
matching the crs argument, the coordinates will not be transformed}

\item{geometry_colname}{default "geometry", to optionally set output name
of simple feature geometry list column}
}
\value{
\code{get_geometry} returns the input \code{DT} appended with a
\code{geometry} column which represents the input coordinates
as a \code{sfc} (simple feature geometry list column). If the \code{output_crs}
was provided, the \code{geometry} will be transformed to the \code{output_crs}.

A message is returned when a column named \code{geometry} already exists in
the input \code{DT}, because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{get_geometry} sets up an input \code{DT} with a 'geometry' column for spatsoc's
geometry interface. The function expects a \code{data.table} with relocation data and
a coordinate reference system.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{coords} argument expects the names of columns in \code{DT} which correspond
to the X and Y coordinates.

The \code{output_crs} argument allows the user to set an output \code{crs} for their
geometry column. Note: some functions in spatsoc (eg. those that measure
directions like \code{edge_direction} and \code{direction_to_leader}) require
geographic coordinates and it is therefore simpler to leave the default
\code{output_crs = 4326}.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file('extdata', 'DT.csv', package = 'spatsoc'))

# Get geometry
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)

# Print
print(DT)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/randomizations.R
\name{randomizations}
\alias{randomizations}
\title{Data-stream randomizations}
\usage{
randomizations(
  DT = NULL,
  type = NULL,
  id = NULL,
  group = NULL,
  coords = NULL,
  datetime = NULL,
  splitBy = NULL,
  iterations = NULL
)
}
\arguments{
\item{DT}{input data.table}

\item{type}{one of 'daily', 'step' or 'trajectory' - see details}

\item{id}{character string of ID column name}

\item{group}{generated from spatial grouping functions - see details}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{datetime}{field used for providing date time or time group - see
details}

\item{splitBy}{List of fields in DT to split the randomization process by}

\item{iterations}{The number of iterations to randomize}
}
\value{
\code{randomizations} returns the random date time or random id along
with the original \code{DT}, depending on the randomization \code{type}.
The length of the returned \code{data.table} is the original number of rows
multiplied by the number of iterations + 1. For example, 3 iterations will
return 4x - one observed and three randomized.

Two columns are always returned: \itemize{ \item observed - if the rows
represent the observed (TRUE/FALSE) \item iteration - iteration of rows
(where 0 is the observed) }

In addition, depending on the randomization type, random ID or random date
time columns are returned:

\itemize{ \item step - \code{randomID} each time step \item daily -
\code{randomID} for each day and \code{jul} indicating julian day \item
trajectory - a random date time ("random" prefixed to \code{datetime}
argument), observed \code{jul} and \code{randomJul} indicating the random
day relocations are swapped to. }
}
\description{
\code{randomizations} performs data-stream social network randomization. The
function expects a \code{data.table} with relocation data, individual
identifiers and a randomization \code{type}. The \code{data.table} is
randomized either using \code{step} or \code{daily} between-individual
methods, or within-individual daily \code{trajectory} method described by
Spiegel et al. (2016).
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}}.

Three randomization \code{type}s are provided: \enumerate{ \item step -
randomizes identities of relocations between individuals within each time
step. \item daily - randomizes identities of relocations between individuals
within each day. \item trajectory - randomizes daily trajectories within
individuals (Spiegel et al. 2016). }

Depending on the \code{type}, the \code{datetime} must be a certain format:

\itemize{ \item step - datetime is integer group created by
\code{group_times} \item daily - datetime is \code{POSIXct} format \item
trajectory - datetime is \code{POSIXct} format }

The \code{id}, \code{datetime},  (and optional \code{splitBy}) arguments
expect the names of respective columns in \code{DT} which correspond to the
individual identifier, date time, and additional grouping columns. The
\code{coords} argument is only required when the \code{type} is "trajectory",
since the coordinates are required for recalculating spatial groups with
\code{group_pts}, \code{group_lines} or \code{group_polys}.

Please note that if the data extends over multiple years, a column indicating
the year should be provided to the \code{splitBy} argument. This will ensure
randomizations only occur within each year.

The \code{group} argument is expected only when \code{type} is 'step' or
'daily'.

For example, using \code{\link[data.table:IDateTime]{data.table::year()}}:

\preformatted{ DT[, yr := year(datetime)] randomizations(DT, type = 'step',
id = 'ID', datetime = 'timegroup', splitBy = 'yr') }

\code{iterations} is set to 1 if not provided. Take caution with a large
value for \code{iterations} with large input \code{DT}.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Select only individuals A, B, C for this example
DT <- DT[ID \%in\% c('A', 'B', 'C')]

# Date time columns
DT[, datetime := as.POSIXct(datetime)]
DT[, yr := year(datetime)]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '5 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 5, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Randomization: step
randStep <- randomizations(
    DT,
    type = 'step',
    id = 'ID',
    group = 'group',
    datetime = 'timegroup',
    splitBy = 'yr',
    iterations = 2
)

# Randomization: daily
randDaily <- randomizations(
    DT,
    type = 'daily',
    id = 'ID',
    group = 'group',
    datetime = 'datetime',
    splitBy = 'yr',
    iterations = 2
)

# Randomization: trajectory
randTraj <- randomizations(
    DT,
    type = 'trajectory',
    id = 'ID',
    group = NULL,
    coords = c('X', 'Y'),
    datetime = 'datetime',
    splitBy = 'yr',
    iterations = 2
)

}
\references{
\url{doi:10.1111/2041-210X.12553}
}
\seealso{
Other Social network tools: 
\code{\link{get_gbi}()}
}
\concept{Social network tools}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_polys.R
\name{build_polys}
\alias{build_polys}
\title{Build polygons}
\usage{
build_polys(
  DT = NULL,
  crs = NULL,
  hrType = NULL,
  hrParams = NULL,
  id = NULL,
  coords = NULL,
  splitBy = NULL,
  spPts = NULL,
  projection = NULL
)
}
\arguments{
\item{DT}{input data.table}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}.}

\item{hrType}{type of HR estimation, either 'mcp' or 'kernel'}

\item{hrParams}{a named list of parameters for \code{adehabitatHR} functions}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{spPts}{alternatively, provide solely a SpatialPointsDataFrame with one
column representing the ID of each point, as specified by \link[adehabitatHR:mcp]{adehabitatHR::mcp}
or \link[adehabitatHR:kernelUD]{adehabitatHR::kernelUD}}

\item{projection}{(deprecated) use crs argument instead}
}
\value{
\code{build_polys} returns a simple feature collection with POLYGONs
for each individual (and optionally \code{splitBy} combination).

An error is returned when \code{hrParams} do not match the arguments
of the respective \code{hrType} \code{adehabitatHR} function.
}
\description{
\code{build_polys} generates a simple feature collection with POLYGONs from a
\code{data.table}. The function expects a \code{data.table} with
relocation data, individual identifiers, a crs,
home range type and parameters. The relocation
data is transformed into POLYGONs using either \link[adehabitatHR:mcp]{adehabitatHR::mcp} or
\link[adehabitatHR:kernelUD]{adehabitatHR::kernelUD} for each individual and, optionally,
combination of columns listed in \code{splitBy}. Relocation data should be in two
columns representing the X and Y coordinates.
}
\details{
\link{group_polys} uses \code{build_polys} for grouping overlapping
polygons created from relocations.
\subsection{R-spatial evolution}{

Please note, spatsoc has followed updates from R spatial, GDAL and PROJ for
handling coordinate reference systems, see more below and  details at
\url{https://r-spatial.org/r/2020/03/17/wkt.html}.

In addition, \code{build_polys} previously used \link[sp:SpatialPoints]{sp::SpatialPoints} but has been
updated to use \link[sf:st_as_sf]{sf::st_as_sf} according to the R-spatial evolution, see more
at \url{https://r-spatial.org/r/2022/04/12/evolution.html}.
}

\subsection{Notes on arguments}{

The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \link[data.table:setDT]{data.table::setDT}.

The \code{id}, \code{coords} (and optional \code{splitBy}) arguments
expect the names of respective columns in \code{DT} which correspond
to the individual identifier, X and Y coordinates, and additional
grouping columns.

The \code{crs} argument expects a character string or numeric
defining the coordinate reference system to be passed to \link[sf:st_crs]{sf::st_crs}.
For example, for UTM zone 36S (EPSG 32736), the crs
argument is \code{crs = "EPSG:32736"} or \code{crs = 32736}.
See \url{https://spatialreference.org}
for a list of EPSG codes.

The \code{hrType} must be either one of "kernel" or "mcp". The
\code{hrParams} must be a named list of arguments matching those
of \link[adehabitatHR:kernelUD]{adehabitatHR::kernelUD} and \link[adehabitatHR:getverticeshr]{adehabitatHR::getverticeshr}
or \link[adehabitatHR:mcp]{adehabitatHR::mcp}.

The \code{splitBy} argument offers further control building
POLYGONs. If in your \code{DT}, you have multiple
temporal groups (e.g.: years) for example, you can provide the
name of the column which identifies them and build POLYGONs
for each individual in each year.
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# EPSG code for example data
utm <- 32736

# Build polygons for each individual using kernelUD and getverticeshr
build_polys(DT, crs = utm, hrType = 'kernel',
            hrParams = list(grid = 60, percent = 95),
            id = 'ID', coords = c('X', 'Y'))

# Build polygons for each individual by year
DT[, yr := year(datetime)]
build_polys(DT, crs = utm, hrType = 'mcp',
            hrParams = list(percent = 95),
            id = 'ID', coords = c('X', 'Y'), splitBy = 'yr')
}
\seealso{
\link{group_polys}

Other Build functions: 
\code{\link{build_lines}()}
}
\concept{Build functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{diff_rad}
\alias{diff_rad}
\title{Difference of two angles measured in radians}
\usage{
diff_rad(x, y, signed = FALSE, return_units = FALSE)
}
\arguments{
\item{x}{angle in radians}

\item{y}{angle in radians}

\item{signed}{logical if signed difference should be returned, default FALSE}

\item{return_units}{return difference with units = 'rad'}
}
\value{
Difference between x and y in radians. If signed is TRUE, the signed
difference is returned. If signed is FALSE, the absolute difference is
returned. Note: The difference is the smallest difference, eg. the
difference between 2 rad and -2.5 rad is 1.78.
}
\description{
\strong{Internal function} - not developed to be used outside of spatsoc functions
}
\examples{
# Load data.table, units
library(data.table)
library(units)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Set order using data.table::setorder
setorder(DT, datetime)

# Calculate direction
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Differences
spatsoc:::diff_rad(DT[1, direction], DT[2, direction])

# Note smallest difference returned
spatsoc:::diff_rad(as_units(2, 'rad'), as_units(-2.5, 'rad'))
}
\references{
adapted from https://stackoverflow.com/a/7869457
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_zones.R
\name{edge_zones}
\alias{edge_zones}
\title{Behavioural zones}
\usage{
edge_zones(
  edges = NULL,
  zone_thresholds = NULL,
  zone_labels = NULL,
  blind_volume = NULL
)
}
\arguments{
\item{edges}{distance based edge-list generated by \code{edge_dist},
optionally with direction columns from \code{direction_step} and
\code{edge_direction} (see Details)}

\item{zone_thresholds}{upper thresholds to define behavioural zones, eg.
c(10, 20, 30) defines behavioural zones (0-10], (10-20], (20-30]}

\item{zone_labels}{labels for zones defined by zone_thresholds, must match
zone_thresholds in length}

\item{blind_volume}{(optional) interindividual direction to define
symmetrical window outside of focal individual's perception, eg. 2 becomes
(-2, 2), see Details}
}
\value{
\code{edge_zones} returns the input \code{edges} appended with a zone
column indicating the behavioural zone, using the zone label provided.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.

A message is returned when a zone column already exists in the input
\code{edges}, because it will be overwritten.
}
\description{
\code{edge_zones} returns edge-lists defined by behavioural zones (Couzin
2002). The function expects a distance based edge-list generated by
\code{edge_dist} (optionally with directions measured by
\code{direction_step} and \code{edge_direction}), zone thresholds, labels
and (optionally) a blind volume.
}
\details{
\code{edge_zones} uses interindividual distances, and optionally directions,
to assign neighboring individuals to a focal individual's behavioural zones.
The user provides zone thresholds (eg. 25 m, 100 m, 250 m) along with zone
labels (eg. zone of repulsion, zone of orientation, zone of attraction),
according to their objectives, study species and system. The optional blind
volume can be provided to define a range of interindividual directions that
correspond to the limits of the focal individual's perception.

Two workflows for this function exist, depending on if the blind volume
argument is used:

a) If the blind volume is not provided, simply provide your distance based
edge-lists from \code{edge_dist} with zone thresholds and labels.

b) If the blind volume is provided, the following order of functions is
expected to ensure the relevant direction columns are available:
\enumerate{
\item direction_step(DT)
\item edges <- edge_dist(DT)
\item dyad_id(edges)
\item dyad_directions <- edge_direction(edges, DT)
\item edge_zones(dyad_directions)
}

Interindividual distances are converted into behavioural zones using
\code{cut}. The thresholds provided are used as cut points for a series of
intervals that are open on the left and closed on the right, starting at 0.
See details in \code{\link[base:cut]{base::cut()}}.

The (optional) blind volume defines the range of interindividual directions
between the focal individual (ID1) and the neighbour (ID2) that is outside
of the focal individual's perception. The interindividual direction (column
"direction_dyad" from \code{edge_direction}) is made relative to the focal
individual's movement direction (column "direction" from
\code{direction_step}). The argument blind_volume expects a single value to
define a symmetrical window behind the focal individual's movement direction
eg. where blind_volume = 2, the symmetrical window is from (-2, 2).

The \code{edges} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge list generation
edges <- edge_dist(
  DT,
  threshold = 100,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  returnDist = TRUE,
  fillNA = FALSE
)

# Calculate behavioural zones
edge_zones(
  edges,
  zone_thresholds = c(25, 50, 75),
  zone_labels = c('repulsion', 'orientation', 'attraction')
)

# Alternatively, if a user wants to specify a blind volume,
#   we need to measure directions
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Edge list generation
edges <- edge_dist(
  DT,
  threshold = 100,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  returnDist = TRUE,
  fillNA = FALSE
)

# Generate dyad id
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

# Interindividual directions
dyad_directions <- edge_direction(
  edges,
  DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736,
  timegroup = 'timegroup'
)

# Calculate behavioural zones
edge_zones(
  dyad_directions,
  zone_thresholds = c(25, 50, 75),
  zone_labels = c('repulsion', 'orientation', 'attraction'),
  blind_volume = 2
)
print(dyad_directions[, .SD[1:3], by = zone])

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
direction_step(DT, id = 'ID')
edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
dyad_id(edges, id = 'ID1', id2 = 'ID2')
dyad_directions <- edge_direction(
  edges,
  DT,
  id = 'ID',
  timegroup = 'timegroup'
)
edge_zones(
  dyad_directions,
  zone_thresholds = c(25, 50, 75),
  zone_labels = c('repulsion', 'orientation', 'attraction'),
  blind_volume = 2
)
print(dyad_directions[, .SD[1:3], by = zone])
}
\references{
The behavioural zones metric is defined in Couzin et al. 2002
(\doi{doi:10.1006/jtbi.2002.3065}).

See examples of measuring behavioural zones:
\itemize{
\item \doi{doi:10.1073/pnas.1001763107}
\item \doi{doi:10.1371/journal.pcbi.1008832}
\item \doi{doi:10.3389/fphy.2021.715996}
}
}
\seealso{
\link{edge_dist} \link{direction_step} \link{edge_direction}

Other Distance functions: 
\code{\link{distance_to_centroid}()},
\code{\link{distance_to_leader}()},
\code{\link{edge_dist}()},
\code{\link{edge_nn}()}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
\concept{Distance functions}
\concept{Edge-list generating functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leader_direction_group.R
\name{leader_direction_group}
\alias{leader_direction_group}
\title{Leadership along group direction}
\usage{
leader_direction_group(
  DT = NULL,
  group_direction = "group_direction",
  coords = NULL,
  group = "group",
  crs = NULL,
  geometry = "geometry",
  return_rank = TRUE,
  ties.method = NULL
)
}
\arguments{
\item{DT}{input data.table with group direction columns generated by
\code{direction_group} and centroid columns generated by
\code{centroid_group}}

\item{group_direction}{group_direction column name generated using
\code{direction_group}, default 'group_direction'}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{group}{group column name, generated by \code{group_pts}, default
'group'}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}

\item{return_rank}{logical if rank distance should also be returned, default
TRUE}

\item{ties.method}{see \code{\link[data.table:frank]{?data.table::frank()}}}
}
\value{
\code{leader_direction_group} returns the input \code{DT} appended with a
\code{position_group_direction} column indicating the position along the group
direction in the units of the crs and, optionally when \code{return_rank = TRUE}, a \code{rank_position_group_direction} column indicating the ranked
position along the group direction.

A message is returned when \code{position_group_direction} or
\code{rank_position_group_direction} columns already exist in the input \code{DT},
because they will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
Given the mean direction of a group of individuals, \code{leader_direction_group}
shifts the coordinate system to a new origin at the group centroid and
rotates the coordinate system by the mean direction to return each
individual's position along the mean direction, representing leadership in
terms of the front-back position in each group's mean direction. Relocation
data should be in two columns representing the X and Y coordinates, or in a
geometry column prepared by the helper function \code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The function expects a \code{data.table} with relocation data appended with a
\code{group_direction} column from \code{\link[=direction_group]{direction_group()}} and group centroid columns
from \code{\link[=centroid_group]{centroid_group()}}.

The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{group_direction} argument expects the names of columns in \code{DT} which
correspond to the mean group direction generated by \code{\link[=direction_group]{direction_group()}}. The
mean group direction column is expected in units of radians. The
\code{return_rank} argument controls if the rank of each individual's distance to
the group centroid is also returned. If \code{return_rank} is TRUE, the \code{group}
argument is required to specify the group column generated by \code{\link[=group_pts]{group_pts()}}.
The \code{ties.method} argument is passed to \code{\link[data.table:frank]{data.table::frank()}}, see details at
\code{\link[data.table:frank]{?data.table::frank()}}.

See below under "Interface" for details on providing coordinates.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}
# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]
\dontshow{
# (Subset example data to reduce runtime on CRAN)
if (isFALSE(rlang::is_interactive())) {
  DT <- DT[as.Date(datetime) == '2017-01-17']
}
}
# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 50, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate direction at each step
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Calculate group centroid
centroid_group(DT, coords = c('X', 'Y'))

# Calculate group direction
direction_group(DT)

# Calculate leader in terms of position along group direction
leader_direction_group(DT, coords = c('X', 'Y'), crs = 32736)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
direction_step(DT = DT, id = 'ID')
centroid_group(DT)
direction_group(DT)
leader_direction_group(DT)
}
\references{
See examples of measuring leadership along group direction (also called
forefront index):
\itemize{
\item \doi{doi:10.1371/journal.pone.0036567}
\item \doi{doi:10.1111/jfb.15315}
\item \doi{doi:10.1098/rspb.2021.0839}
}
}
\seealso{
\code{direction_group}, \code{centroid_group}

Other Leadership functions: 
\code{\link{direction_to_leader}()},
\code{\link{leader_edge_delay}()}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
\concept{Leadership functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{calc_direction}
\alias{calc_direction}
\title{Calculate direction}
\usage{
calc_direction(geometry_a, geometry_b, x_a, y_a, x_b, y_b, crs, use_transform)
}
\arguments{
\item{geometry_a, geometry_b}{sfc (simple feature geometry list column) from
\code{\link[=get_geometry]{get_geometry()}}}

\item{x_a, x_b}{X coordinate column, numeric}

\item{y_a, y_b}{Y coordinate column, numeric}

\item{crs}{crs for x_a, y_a (and if provided, x_b, y_b) coordinates,
ignored for geometry_a and geometry_b arguments}

\item{use_transform}{boolean predetermine if coordinates/geometry need transform}
}
\value{
Direction in units of radians, in range of pi, -pi where North = 0.
}
\description{
\strong{Internal function} - not developed to be used outside of spatsoc functions
}
\details{
Calculate direction using \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}} between one of:
\itemize{
\item the sequence of points in geometry_a
\item the sequence of points in x_a, y_a
\item the pairwise points in geometry_a and geometry_b
\item the pairwise points in x_a, y_a and x_b, y_b
}

Requirements:
\itemize{
\item matching length between a and b objects if b provided
\item crs is provided. If crs is not longlat (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the coordinates or geometry will be transformed to
\code{sf::st_crs(4326)}
\item no missing values in coordinates
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Example result for East, North, West, South directions
example <- data.table(
  X = c(0, 5, 5, 0, 0),
  Y = c(0, 0, 5, 5, 0)
)
# E, N, W, S
example[, spatsoc:::calc_direction(x_a = X, y_a = Y, crs = 4326, use_transform = FALSE)]
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{calc_distance}
\alias{calc_distance}
\title{Calculate distance}
\usage{
calc_distance(geometry_a, geometry_b, x_a, y_a, x_b, y_b, crs, use_dist)
}
\arguments{
\item{geometry_a, geometry_b}{sfc (simple feature geometry list column)
from \code{\link[=get_geometry]{get_geometry()}}}

\item{x_a, x_b}{X coordinate column, numeric}

\item{y_a, y_b}{Y coordinate column, numeric}

\item{crs}{crs for x_a, y_a (and if provided, x_b, y_b) coordinates,
ignored for geometry_a and geometry_b arguments}

\item{use_dist}{boolean predetermine if distance calculated via dist}
}
\value{
The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is \code{\link[sf:geos_measures]{sf::st_distance()}} which
passes to \code{\link[s2:s2_is_collection]{s2::s2_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE and
\code{\link[lwgeom:geod]{lwgeom::st_geod_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE. The distance
returned has units set according to the crs.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
distance function used is \code{\link[stats:dist]{stats::dist()}} (or Euclidean distance for
pairwise distances), maintaining expected behaviour from previous versions.
The distance returned does not have units set.
}

Note: in both cases, if the coordinates are NA then the result will be NA.
}
\description{
\strong{Internal function} - not developed to be used outside of spatsoc functions
}
\details{
Calculate distance for one of the following combinations:
\itemize{
\item the distance matrix of points in geometry_a
\item the distance matrix of points in x_a, y_a
\item the pairwise distance between points in geometry_a and geometry_b
\item the pairwise distance between points in x_a, y_a and x_b, y_b
}

Requirements:
\itemize{
\item matching length between a and b objects if b provided
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Example points
example <- data.table(
  X = c(0, 5, 5, 0, 0, NA_real_, 0,        NA_real_),
  Y = c(0, 0, 5, 5, 0, 0,        NA_real_, NA_real_)
)
# E, N, W, S
example[, spatsoc:::calc_distance(x_a = X, y_a = Y, crs = 4326, use_dist = FALSE)]
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/direction_step.R
\name{direction_step}
\alias{direction_step}
\title{Direction step}
\usage{
direction_step(
  DT = NULL,
  id = NULL,
  coords = NULL,
  crs = NULL,
  splitBy = NULL,
  geometry = "geometry",
  projection = NULL
)
}
\arguments{
\item{DT}{input data.table}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}

\item{projection}{(deprecated) use crs argument instead}
}
\value{
\code{direction_step} returns the input \code{DT} appended with a \code{direction}
column with units set to radians using the \code{units} package.

This column represents the azimuth between the sequence of points for each
individual computed using \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}. Note, the order of
points is not modified by this function and therefore it is crucial the
user sets the order of rows to their specific question before using
\code{direction_step}. In addition, the direction column will include an \code{NA}
value for the last point in each sequence of points since there is no
future point to calculate a direction to.

A message is returned when a direction column are already exists in the
input \code{DT}, because it will be overwritten.

An error is returned if there are any missing values in coordinates /
geometry as the underlying direction function (\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}})
does not accept missing values.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{direction_step} calculates the direction of movement steps in radians. The
function expects a \code{data.table} with relocation data and individual
identifiers. Relocation data should be in two columns representing the X and
Y coordinates, or in a geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}. Note the order of rows is not modified by this function and
therefore users must be cautious to set it explicitly. See example for one
approach to setting order of rows using a datetime field.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{id}, and optional \code{splitBy} arguments expect the names of a column in
\code{DT} which correspond to the individual identifier and additional grouping
columns.

The \code{splitBy} argument offers further control over grouping. If within your
\code{DT}, you have distinct sampling periods for each individual, you can provide
the column name(s) which identify them to \code{splitBy}. The direction
calculation by \code{\link[=direction_step]{direction_step()}} will only consider rows within each \code{id}
and \code{splitBy} subgroup.

See below under "Interface" for details on providing coordinates and under
"Direction function" for details on the underlying direction function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Direction function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is provided and longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is provided and not longlat degrees (eg. a projected UTM),
the coordinates or geometry are transformed to \code{sf::st_crs(4326)} before the
distance is measured using \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is NULL or NA_crs_, the distance function cannot be used
and an error is returned.
}
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Set order using data.table::setorder
setorder(DT, datetime)

# Calculate direction
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Or: geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
direction_step(DT, id = 'ID')

# Example result for East, North, West, South steps
example <- data.table(
  X = c(0, 5, 5, 0, 0),
  Y = c(0, 0, 5, 5, 0),
  step = c('E', 'N', 'W', 'S', NA),
  ID = 'A'
)

direction_step(example, 'ID', c('X', 'Y'), crs = 4326)
example[, .(step, direction, units::set_units(direction, 'degree'))]
}
\seealso{
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}, \code{\link[amt:steps]{amt::direction_abs()}},
\code{\link[geosphere:bearing]{geosphere::bearing()}}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_lines.R
\name{build_lines}
\alias{build_lines}
\title{Build lines}
\usage{
build_lines(
  DT = NULL,
  crs = NULL,
  id = NULL,
  coords = NULL,
  sortBy = NULL,
  splitBy = NULL,
  projection = NULL
)
}
\arguments{
\item{DT}{input data.table}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{sortBy}{Character string of date time column(s) to sort rows by. Must
be a POSIXct.}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{projection}{(deprecated) use crs argument instead}
}
\value{
\code{build_lines} returns an sf LINESTRING object with a line
for each individual (and optionally \code{splitBy} combination).

Individuals (or combinations of individuals and \code{splitBy}) with less than two
relocations are dropped since it requires at least two relocations to
build a line.
}
\description{
\code{build_lines} generates a simple feature collection with LINESTRINGs from a
\code{data.table}. The function expects a \code{data.table} with relocation data,
individual identifiers, a sorting column and a \code{crs}. The relocation
data is transformed into LINESTRINGs for each individual and, optionally,
combination of columns listed in \code{splitBy}. Relocation data should be in two
columns representing the X and Y coordinates.
}
\details{
\subsection{R-spatial evolution}{

Please note, spatsoc has followed updates from R spatial, GDAL and PROJ
for handling coordinate reference systems, see more at
\url{https://r-spatial.org/r/2020/03/17/wkt.html}.

In addition, \code{build_lines} previously used \link[sp:SpatialLines]{sp::SpatialLines} but has been
updated to use \link[sf:st_as_sf]{sf::st_as_sf} and \link[sf:st]{sf::st_linestring} according to the
R-spatial evolution, see more at
\url{https://r-spatial.org/r/2022/04/12/evolution.html}.
}

\subsection{Notes on arguments}{

The \code{crs} argument expects a numeric or character defining the
coordinate reference system.
For example, for UTM zone 36N (EPSG 32736), the crs argument is either
\code{crs = 'EPSG:32736'} or \code{crs = 32736}.
See details in \code{\link[sf:st_crs]{sf::st_crs()}} and \url{https://spatialreference.org}
for a list of EPSG codes.

The \code{sortBy} argument is used to order the input \code{DT} when creating
sf LINESTRINGs. It must a column in the input \code{DT} of type
POSIXct to ensure the rows are sorted by date time.

The \code{splitBy} argument offers further control building LINESTRINGs.
If in your input \code{DT}, you have multiple temporal groups (e.g.: years) for
example, you can provide the name of the column which identifies them and
build LINESTRINGs for each individual in each year.

\code{build_lines} is used by \code{group_lines} for grouping overlapping
lines generated from relocations.
}
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# EPSG code for example data
utm <- 32736

# Build lines for each individual
lines <- build_lines(DT, crs = utm, id = 'ID', coords = c('X', 'Y'),
            sortBy = 'datetime')

# Build lines for each individual by year
DT[, yr := year(datetime)]
lines <- build_lines(DT, crs = utm, id = 'ID', coords = c('X', 'Y'),
            sortBy = 'datetime', splitBy = 'yr')
}
\seealso{
\code{group_lines}

Other Build functions: 
\code{\link{build_polys}()}
}
\concept{Build functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leader_edge_delay.R
\name{leader_edge_delay}
\alias{leader_edge_delay}
\title{Leadership in directional correlation delay}
\usage{
leader_edge_delay(edges = NULL, threshold = NULL, splitBy = NULL)
}
\arguments{
\item{edges}{edge-list generated generated by \code{\link[=edge_dist]{edge_dist()}} or
\code{\link[=edge_nn]{edge_nn()}}, with fusionID column generated by \code{\link[=fusion_id]{fusion_id()}}}

\item{threshold}{(optional) threshold difference in direction used to subset
rows included in calculation of mean directional delay. eg.
\code{threshold = 0.5} corresponds to only rows where direction_diff is
less than 0.5. Expects that unit is radians, see \code{edge_delay}.}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the mean directional correlation delay will be
calculated}
}
\value{
\code{leader_edge_delay} returns the input \code{edges} aggregated
with a \code{mean_direction_delay_dyad} column indicating the mean
directional correlation delay between ID1 and ID2 and a
\code{mean_direction_delay} column indicating the mean directional
correlation delay for each individual in 'ID1' column.

Note: since \code{leader_edge_delay} returns an aggregation of the input
\code{edges}, the output needs to be reassigned unlike some other
\code{spatsoc} functions like \code{dyad_id}. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
Given the directional correlation delay, \code{leader_edge_delay}
calculates the mean directional correlation delay for individuals in a
group to identify leadership patterns.
}
\details{
The function expects an edge-list from \code{edge_delay} with columns
'direction_delay' indicating the directional correlation delay between
individuals and 'direction_diff' indicating the unsigned difference in
movement directions at the temporal delay, columns 'ID1' and 'ID2' indicating
individuals and column 'dyadID' indicating the dyad.

The \code{edge} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Select only individuals A, B, C for this example
DT <- DT[ID \%in\% c('A', 'B', 'C')]

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Calculate direction
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Distance based edge-list generation
edges <- edge_dist(
  DT,
  threshold = 100,
  id = 'ID',
  coords = c('X', 'Y'),
  timegroup = 'timegroup',
  returnDist = TRUE,
  fillNA = FALSE
)

# Generate dyad id
dyad_id(edges, id1 = 'ID1', id2 = 'ID2')

# Generate fusion id
fusion_id(edges, threshold = 100)

# Directional correlation delay
delay <- edge_delay(
  edges = edges,
  DT = DT,
  window = 3,
  id = 'ID'
)

# Leadership from directional correlation delay
leadership <- leader_edge_delay(
  delay,
  threshold = 0.5
)
print(leadership)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
direction_step(DT, id = 'ID')
edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
dyad_id(edges, id = 'ID1', id2 = 'ID2')
fusion_id(edges, threshold = 100)
delay <- edge_delay(
  edges = edges,
  DT = DT,
  window = 3,
  id = 'ID'
)
leadership <- leader_edge_delay(
  delay,
  threshold = 0.5
)
print(leadership)
}
\references{
See examples of measuring leadership using the directional correlation
delay:
\itemize{
\item \doi{doi:10.1016/j.anbehav.2013.07.005}
\item \doi{doi:10.1073/pnas.1305552110}
\item \doi{doi:10.1126/science.aap7781}
\item \doi{doi:10.1111/jfb.15315}
\item \doi{doi:10.1371/journal.pcbi.1003446}
}
}
\seealso{
\code{edge_delay}

Other Leadership functions: 
\code{\link{direction_to_leader}()},
\code{\link{leader_direction_group}()}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()}
}
\concept{Direction functions}
\concept{Leadership functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dyad_id.R
\name{dyad_id}
\alias{dyad_id}
\title{Dyad id}
\usage{
dyad_id(DT = NULL, id1 = NULL, id2 = NULL)
}
\arguments{
\item{DT}{input data.table with columns id1 and id2, as generated by
\code{edge_dist} or \code{edge_nn}}

\item{id1}{ID1 column name generated by \code{edge_dist} or \code{edge_nn}}

\item{id2}{ID2 column name generated by \code{edge_dist} or \code{edge_nn}}
}
\value{
\code{dyad_id} returns the input \code{data.table} with appended "dyadID"
column.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
Generate a dyad ID for edge-list generated by \code{edge_nn} or
\code{edge_dist}.
}
\details{
An undirected edge identifier between, for example individuals A and B will
be A-B (and reverse B and A will be A-B). Internally sorts and pastes id
columns.

More details in the edge and dyad vignette (in progress).
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge-list generation
edges <- edge_dist(
    DT,
    threshold = 100,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup',
    returnDist = TRUE,
    fillNA = TRUE
  )

# Generate dyad IDs
dyad_id(edges, 'ID1', 'ID2')

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
edges <- edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup')
dyad_id(edges, id = 'ID1', id2 = 'ID2')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_direction.R
\name{edge_direction}
\alias{edge_direction}
\title{Direction based edge-lists}
\usage{
edge_direction(
  edges = NULL,
  DT = NULL,
  id = NULL,
  coords = NULL,
  crs = NULL,
  timegroup = "timegroup",
  geometry = "geometry",
  projection = NULL
)
}
\arguments{
\item{edges}{edge-list generated generated by \code{edge_dist} or
\code{edge_nn}, with dyad ID column generated by \code{dyad_id}}

\item{DT}{input data.table with timegroup column generated with
\code{group_times} matching the input data.table used to generate the edge
list with \code{edge_nn} or \code{edge_dist}}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{timegroup}{character string of timegroup column name, default "timegroup"}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}

\item{projection}{(deprecated) use crs argument instead}
}
\value{
\code{\link[=edge_direction]{edge_direction()}} returns the input \code{edges} appended with a
"direction_dyad" column representing the direction between ID1 and ID2 in
radians. A value of NaN is returned when the coordinates of ID1 equal the
coordinates of ID2.

If the "direction" column is found in input DT, it will be retained for ID1
in the output for use in downstream functions (eg. \code{\link[=edge_zones]{edge_zones()}}).

Missing values in coordinates / geometry are ignored and NA is returned.

Note: due to the merge required within this function, the output needs to
be reassigned unlike some other \code{spatsoc} functions like \code{\link[=dyad_id]{dyad_id()}} and
\code{\link[=group_pts]{group_pts()}}. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{\link[=edge_direction]{edge_direction()}} returns edge lists defined by the direction between
individuals. The function expects a distance based edge-list generated by
\code{edge_nn} or \code{\link[=edge_dist]{edge_dist()}} and a \code{data.table} with relocation data appended
with a \code{timegroup} column from \code{\link[=group_times]{group_times()}}. It is required to use the
argument \code{fillNA = FALSE} for \code{\link[=edge_dist]{edge_dist()}} to ensure there are no NAs in the
coordinate columns. Relocation data should be in two columns representing the
X and Y coordinates, or in a geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{edges} and \code{DT} must be \code{data.table}s. If your data is a \code{data.frame},
you can convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning
using \code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{edges} and \code{DT} are internally merged in this function using the columns
\code{id} / \code{ID1} and \code{ID2}, \code{timegroup}. The \code{id}, and \code{timegroup} arguments
expect the names of columns which correspond to the ID, and timegroup
columns in \code{DT}.

See below under "Interface" for details on providing coordinates and under
"Direction function" for details on the underlying direction function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Direction function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is provided and longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is provided and not longlat degrees (eg. a projected UTM),
the coordinates or geometry are transformed to \code{sf::st_crs(4326)} before the
distance is measured using \code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}.
\item If the crs is NULL or NA_crs_, the distance function cannot be used
and an error is returned.
}
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge list generation
edges <- edge_dist(
    DT,
    threshold = 100,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup',
    returnDist = TRUE,
    fillNA = FALSE
  )

# Direction based edge-lists
dyad_directions <- edge_direction(
  edges,
  DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736,
  timegroup = 'timegroup'
)

print(dyad_directions)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
dyad_directions <- edge_direction(
  edges,
  DT,
  id = 'ID',
  timegroup = 'timegroup'
)
}
\references{
See examples of measuring the direction between individuals:
\itemize{
\item \doi{doi:10.1098/rsif.2013.0529}
\item \doi{doi:10.1098/rstb.2019.0380}
\item \doi{doi:10.1111/jfb.15315}
}
}
\seealso{
\link{dyad_id}, \link{edge_dist}, \link{edge_nn}, \link{group_times},
\code{\link[lwgeom:st_geod_azimuth]{lwgeom::st_geod_azimuth()}}

Other Edge-list generation: 
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_dist}()},
\code{\link{edge_nn}()}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
\concept{Edge-list generation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/direction_polarization.R
\name{direction_polarization}
\alias{direction_polarization}
\title{Polarization}
\usage{
direction_polarization(DT, direction = "direction", group = "group")
}
\arguments{
\item{DT}{input data.table with direction column generated by
\code{\link[=direction_step]{direction_step()}} and group column generated with \code{\link[=group_pts]{group_pts()}}}

\item{direction}{character string of direction column name, default
"direction", expects that the unit of the direction column is radians.}

\item{group}{character string of group column name, default "group"}
}
\value{
\code{direction_polarization} returns the input \code{DT} appended
with a \code{polarization} column representing the direction polarization
of all individuals in each spatiotemporal group.

The direction polarization is calculated using \code{\link[CircStats:r.test]{CircStats::r.test()}}
which expects units of radians.

A message is returned when the \code{polarization} columns already
exists in the input \code{DT}, because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{direction_polarization} calculates the polarization of individual
directions in each spatiotemporal group identified by \code{group_pts}. The
function expects a \code{data.table} with relocation data appended with a
\code{direction} column from \code{direction_step} and a \code{group} column
from \code{group_pts}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a
\code{data.frame}, you can convert it by reference using
\code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{direction} and \code{group} arguments expect the names of columns
in \code{DT} which correspond to the direction and group columns. The
direction column is expected in units of radians and the polarization is
calculated with \code{\link[CircStats:r.test]{CircStats::r.test()}}.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 50, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate direction at each step
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Calculate polarization
direction_polarization(DT)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
direction_step(DT, id = 'ID')
direction_polarization(DT)
}
\references{
See examples of using polarization:
\itemize{
\item \doi{doi:10.1016/j.cub.2017.08.004}
\item \doi{doi:10.1371/journal.pcbi.1009437}
\item \doi{doi:10.7554/eLife.19505}
}
}
\seealso{
\code{direction_step}, \code{group_pts},
\code{\link[CircStats:r.test]{CircStats::r.test()}}

Other Direction functions: 
\code{\link{direction_group}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatsoc.R
\docType{package}
\name{spatsoc}
\alias{spatsoc}
\alias{spatsoc-package}
\title{spatsoc}
\description{
spatsoc is an R package for detecting spatial and temporal groups in GPS
relocations and measuring intragroup social dynamics. It can be used to convert GPS relocations to gambit-of-the-group
format to build proximity-based social networks, identify nearest neighbours distances and direction, and measure individual's position with respect to the group centroid or group leader. See all available functions below, in the manual or on the acommpanying documentation website at https://docs.ropensci.org/spatsoc/.
}
\details{
Temporal grouping:
\itemize{
\item \code{\link[=group_times]{group_times()}}
}

Spatial grouping:
\itemize{
\item \code{\link[=group_lines]{group_lines()}}
\item \code{\link[=group_polys]{group_polys()}}
\item \code{\link[=group_pts]{group_pts()}}
}

Edge-list generation:
\itemize{
\item \code{\link[=edge_dist]{edge_dist()}}
\item \code{\link[=edge_nn]{edge_nn()}}
\item \code{\link[=edge_delay]{edge_delay()}}
\item \code{\link[=edge_alignment]{edge_alignment()}}
\item \code{\link[=edge_direction]{edge_direction()}}
\item \code{\link[=edge_zones]{edge_zones()}}
}

Social network tools:
\itemize{
\item \code{\link[=randomizations]{randomizations()}}
\item \code{\link[=get_gbi]{get_gbi()}}
}

Dyad functions
\itemize{
\item \code{\link[=dyad_id]{dyad_id()}}
\item \code{\link[=fusion_id]{fusion_id()}}
}

Centroid functions
\itemize{
\item \code{\link[=centroid_group]{centroid_group()}}
\item \code{\link[=centroid_dyad]{centroid_dyad()}}
\item \code{\link[=centroid_fusion]{centroid_fusion()}}
}

Direction functions
\itemize{
\item \code{\link[=direction_step]{direction_step()}}
\item \code{\link[=direction_to_centroid]{direction_to_centroid()}}
\item \code{\link[=direction_to_leader]{direction_to_leader()}}
\item \code{\link[=direction_group]{direction_group()}}
\item \code{\link[=direction_polarization]{direction_polarization()}}
}

Distance functions
\itemize{
\item \code{\link[=distance_to_centroid]{distance_to_centroid()}}
\item \code{\link[=distance_to_leader]{distance_to_leader()}}
}

Leadership functions
\itemize{
\item \code{\link[=leader_direction_group]{leader_direction_group()}}
\item \code{\link[=leader_edge_delay]{leader_edge_delay()}}
}

Geometry interface functions
\itemize{
\item \code{\link[=get_geometry]{get_geometry()}}
}

Build functions
\itemize{
\item \code{\link[=build_lines]{build_lines()}}
\item \code{\link[=build_polys]{build_polys()}}
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/spatsoc/}
  \item \url{https://github.com/ropensci/spatsoc}
  \item Report bugs at \url{https://github.com/ropensci/spatsoc/issues}
}

}
\author{
\strong{Maintainer}: Alec L. Robitaille \email{robit.alec@gmail.com} (\href{https://orcid.org/0000-0002-4706-1762}{ORCID})

Authors:
\itemize{
  \item Quinn Webber (\href{https://orcid.org/0000-0002-0434-9360}{ORCID})
  \item Eric Vander Wal (\href{https://orcid.org/0000-0002-8534-4317}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_dist.R
\name{edge_dist}
\alias{edge_dist}
\title{Distance based edge-lists}
\usage{
edge_dist(
  DT = NULL,
  threshold,
  id = NULL,
  coords = NULL,
  timegroup,
  crs = NULL,
  splitBy = NULL,
  geometry = "geometry",
  returnDist = FALSE,
  fillNA = TRUE
)
}
\arguments{
\item{DT}{input data.table}

\item{threshold}{distance for grouping points, either numeric or units, in
the units of the crs / the coordinates or geometry}

\item{id}{character string of ID column name}

\item{coords}{character vector of X coordinate and Y coordinate column names.
Note: the order is assumed X followed by Y column names}

\item{timegroup}{timegroup field in the DT within which the output will be
calculated}

\item{crs}{numeric or character defining the coordinate reference
system to be passed to \link[sf:st_crs]{sf::st_crs}. For example, either
\code{crs = "EPSG:32736"} or \code{crs = 32736}. Used only if coords are provided,
see details under Interface}

\item{splitBy}{(optional) character string or vector of grouping column
name(s) upon which the output will be calculated}

\item{geometry}{simple feature geometry list column name, generated by
\code{\link[=get_geometry]{get_geometry()}}. Default 'geometry', see details under Interface}

\item{returnDist}{logical indicating if the distance between individuals
should be returned. If FALSE (default), only individual columns (and
timegroup, splitBy columns if provided) are returned. If TRUE, a column
"distance" is also returned indicating the distance between individuals in
the units of the \code{crs}, or if \code{crs = NULL} no units are set.}

\item{fillNA}{logical indicating if NAs should be returned for individuals
that were not within the threshold distance of any other. If TRUE, NAs are
returned. If FALSE, only edges between individuals within the threshold
distance are returned.}
}
\value{
\code{edge_dist} returns a \code{data.table} with columns ID1, ID2, timegroup
(if supplied) and any columns provided in splitBy. If 'returnDist' is TRUE,
column 'distance' is returned indicating the distance between ID1 and ID2.

The ID1 and ID2 columns represent the edges defined by the spatial (and
temporal with \code{group_times}) thresholds.

Note: unlike many other functions (eg. \code{group_pts}) in \code{spatsoc},
\code{edge_dist} needs to be reassigned. See details in
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{edge_dist} returns edge-lists defined by a spatial distance within the user
defined threshold. The function expects a \code{data.table} with relocation data,
individual identifiers and a threshold argument. The threshold argument is
used to specify the criteria for distance between points which defines a
group. Relocation data should be in two columns representing the X and Y
coordinates, or in a geometry column prepared by the helper function
\code{\link[=get_geometry]{get_geometry()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}}.

The \code{id}, \code{timegroup} (and optional \code{splitBy}) arguments expect the names of
columns in \code{DT} which correspond to the individual identifier, and timegroup
(generated by \code{group_times}) and additional grouping columns.

The \code{threshold} provided should match the units of the coordinates. The
\code{threshold} can be provided with units specified using the units package (eg.
\code{threshold = units::set_units(10, m)}) which will be checked against the
units of the coordinates using the \code{crs}. If units are not specified, the
\code{threshold} is assumed to be in the units of the coordinates.

The \code{timegroup} argument is required to define the temporal groups within
which edges are calculated. The intended framework is to group rows
temporally with \code{group_times} then spatially with \code{edge_dist}. If you have
already calculated temporal groups without \code{group_times}, you can pass this
column to the \code{timegroup} argument. Note that the expectation is that each
individual will be observed only once per timegroup. Caution that
accidentally including huge numbers of rows within timegroups can overload
your machine since all pairwise distances are calculated within each
timegroup.

The \code{splitBy} argument offers further control over grouping. If within your
\code{DT}, you have multiple populations, subgroups or other distinct parts, you
can provide the name of the column which identifies them to \code{splitBy}.
\code{edge_dist} will only consider rows within each \code{splitBy} subgroup.

See below under "Interface" for details on providing coordinates and under
"Distance function" for details on underlying distance function used.
}
\section{Interface}{

Two interfaces are available for providing coordinates:
\enumerate{
\item Provide \code{coords} and \code{crs}. The \code{coords} argument expects the names of
the X and Y coordinate columns. The \code{crs} argument expects a character
string or numeric defining the coordinate reference system to be passed to
\link[sf:st_crs]{sf::st_crs}. For example, for UTM zone 36S (EPSG 32736), the crs argument
is \code{crs = "EPSG:32736"} or \code{crs = 32736}. See \url{https://spatialreference.org}
for a list of EPSG codes.
\item (New!) Provide \code{geometry}. The \code{geometry} argument allows the user to
supply a \code{geometry} column that represents the coordinates as a simple
feature geometry list column. This interface expects the user to prepare
their input DT with \code{\link[=get_geometry]{get_geometry()}}. To use this interface, leave the
\code{coords} and \code{crs} arguments \code{NULL}, and the default argument for \code{geometry}
('geometry') will be used directly.
}
}

\section{Distance function}{


The underlying distance function used depends on the crs of the coordinates
or geometry provided.
\itemize{
\item If the crs is longlat degrees (as determined by
\code{\link[sf:st_is_longlat]{sf::st_is_longlat()}}), the distance function is \code{\link[sf:geos_measures]{sf::st_distance()}} which
passes to \code{\link[s2:s2_is_collection]{s2::s2_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is TRUE and
\code{\link[lwgeom:geod]{lwgeom::st_geod_distance()}} if \code{\link[sf:s2]{sf::sf_use_s2()}} is FALSE. The distance
returned has units set according to the crs.
\item If the crs is not longlat degrees (eg. NULL, NA_crs_, or projected), the
distance function used is \code{\link[stats:dist]{stats::dist()}}, maintaining expected behaviour
from previous versions. The distance returned does not have units set.
}

Note: in both cases, if the coordinates are NA then the result will be NA.
}

\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Edge-list generation
edges <- edge_dist(
    DT,
    threshold = 100,
    id = 'ID',
    coords = c('X', 'Y'),
    timegroup = 'timegroup',
    crs = 32736,
    returnDist = TRUE,
    fillNA = TRUE
  )

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
edge_dist(DT, threshold = 100, id = 'ID', timegroup = 'timegroup', returnDist = TRUE)
}
\seealso{
\code{\link[sf:geos_measures]{sf::st_distance()}}

Other Edge-list generation: 
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_nn}()}

Other Distance functions: 
\code{\link{distance_to_centroid}()},
\code{\link{distance_to_leader}()},
\code{\link{edge_nn}()},
\code{\link{edge_zones}()}
}
\concept{Distance functions}
\concept{Edge-list generation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/direction_group.R
\name{direction_group}
\alias{direction_group}
\title{Group mean direction}
\usage{
direction_group(DT, direction = "direction", group = "group")
}
\arguments{
\item{DT}{input data.table with direction column generated by
\code{\link[=direction_step]{direction_step()}} and group column generated with \code{\link[=group_pts]{group_pts()}}}

\item{direction}{character string of direction column name, default
"direction", expects that the unit of the direction column is radians.}

\item{group}{character string of group column name, default "group"}
}
\value{
\code{direction_group} returns the input \code{DT} appended with a
\code{group_direction} column representing the mean direction of all individuals
in each spatiotemporal group.

The mean direction is calculated using \code{\link[CircStats:circ.mean]{CircStats::circ.mean()}} which
expects units of radians.

A message is returned when the \code{group_direction} columns already exists in
the input \code{DT}, because it will be overwritten.

See details for appending outputs using modify-by-reference in the
\href{https://docs.ropensci.org/spatsoc/articles/faq.html}{FAQ}.
}
\description{
\code{direction_group} calculates the mean direction of all individuals in each
spatiotemporal group identified by \code{\link[=group_pts]{group_pts()}}. The function expects a
\code{data.table} with relocation data appended with a \code{direction} column from
\code{\link[=direction_step]{direction_step()}} and a \code{group} column from \code{\link[=group_pts]{group_pts()}}.
}
\details{
The \code{DT} must be a \code{data.table}. If your data is a \code{data.frame}, you can
convert it by reference using \code{\link[data.table:setDT]{data.table::setDT()}} or by reassigning using
\code{\link[data.table:data.table]{data.table::data.table()}}.

The \code{direction} and \code{group} arguments expect the names of columns in \code{DT}
which correspond to the direction and group columns. The direction column is
expected in units of radians and the mean calculated with
\code{\link[CircStats:circ.mean]{CircStats::circ.mean()}}.
}
\examples{
# Load data.table
library(data.table)
\dontshow{data.table::setDTthreads(1)}

# Read example data
DT <- fread(system.file("extdata", "DT.csv", package = "spatsoc"))

# Cast the character column to POSIXct
DT[, datetime := as.POSIXct(datetime, tz = 'UTC')]

# Temporal grouping
group_times(DT, datetime = 'datetime', threshold = '20 minutes')

# Spatial grouping with timegroup
group_pts(DT, threshold = 50, id = 'ID',
          coords = c('X', 'Y'), timegroup = 'timegroup')

# Calculate direction at each step
direction_step(
  DT = DT,
  id = 'ID',
  coords = c('X', 'Y'),
  crs = 32736
)

# Calculate group direction
direction_group(DT)

# Or, using the new geometry interface
get_geometry(DT, coords = c('X', 'Y'), crs = 32736)
group_pts(DT, threshold = 5, id = 'ID', timegroup = 'timegroup')
direction_step(DT, id = 'ID')
direction_group(DT)
}
\references{
See examples of using mean group direction:
\itemize{
\item \doi{doi:10.1098/rsos.170148}
\item \doi{doi:10.1098/rsos.201128}
\item \doi{doi:10.1016/j.beproc.2018.01.013}
}
}
\seealso{
\code{\link[=direction_step]{direction_step()}}, \code{\link[=group_pts]{group_pts()}},
\code{\link[CircStats:circ.mean]{CircStats::circ.mean()}}

Other Direction functions: 
\code{\link{direction_polarization}()},
\code{\link{direction_step}()},
\code{\link{direction_to_centroid}()},
\code{\link{direction_to_leader}()},
\code{\link{edge_alignment}()},
\code{\link{edge_delay}()},
\code{\link{edge_direction}()},
\code{\link{edge_zones}()},
\code{\link{leader_direction_group}()},
\code{\link{leader_edge_delay}()}
}
\concept{Direction functions}
